{
*****************************************************************************
*  Name    : DHCP.BAS                                                       *
*  Date    : 30/08/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module DHCP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "UDP.bas"
Include "Tick.bas"

Const
   DHCP_CLIENT_PORT                     = (68),
   DHCP_SERVER_PORT                     = (67),

   BOOT_REQUEST                         = (1),
   BOOT_REPLY                           = (2),
   BOOT_HW_TYPE                         = (1),
   BOOT_LEN_OF_HW_TYPE                  = (6),

   DHCP_MESSAGE_TYPE                    = (53),
   DHCP_MESSAGE_TYPE_LEN                = (1),

   DHCP_UNKNOWN_MESSAGE                 = (0),

   DHCP_DISCOVER_MESSAGE                = (1),
   DHCP_OFFER_MESSAGE                   = (2),
   DHCP_REQUEST_MESSAGE                 = (3),
   DHCP_DECLINE_MESSAGE                 = (4),
   DHCP_ACK_MESSAGE                     = (5),
   DHCP_NAK_MESSAGE                     = (6),
   DHCP_RELEASE_MESSAGE                 = (7),

   DHCP_SERVER_IDENTIFIER               = (54),
   DHCP_SERVER_IDENTIFIER_LEN           = (4),

   DHCP_OPTION_ACK_MESSAGE			    = (53),
   DHCP_PARAM_REQUEST_LIST              = (55),
   DHCP_PARAM_REQUEST_LIST_LEN          = (4),
   DHCP_PARAM_REQUEST_IP_ADDRESS        = (50),
   DHCP_PARAM_REQUEST_IP_ADDRESS_LEN    = (4),
   DHCP_SUBNET_MASK                     = (1),
   DHCP_ROUTER                          = (3),
   DHCP_DNS						        = (6),
   DHCP_HOST_NAME					    = (12),
   DHCP_IP_LEASE_TIME                   = (51),
   DHCP_END_OPTION                      = (255)

Structure BOOTP_HEADER
   MessageType As Byte
   HardwareType As Byte
   HardwareLen As Byte
   Hops As Byte
   TransactionID As DWORD
   SecondsElapsed As Word
   BootpFlags As Word
   ClientIP As IP_ADDR
   YourIP As IP_ADDR
   NextServerIP As IP_ADDR
   RelayAgentIP As IP_ADDR
   ClientMAC As MAC_ADDR
End Structure

// enumerations...
Type SM_DHCP = Byte
Const
   SM_DHCP_DISABLED = 0,
   SM_DHCP_GET_SOCKET = 1,
   SM_DHCP_SEND_DISCOVERY = 2,
   SM_DHCP_GET_OFFER = 3,
   SM_DHCP_SEND_REQUEST = 4,
   SM_DHCP_GET_REQUEST_ACK = 5,
   SM_DHCP_BOUND = 6,
   SM_DHCP_SEND_RENEW = 7,
   SM_DHCP_GET_RENEW_ACK = 8,
   SM_DHCP_SEND_RENEW2 = 9,
   SM_DHCP_GET_RENEW_ACK2 = 10,
   SM_DHCP_SEND_RENEW3 = 11,
   SM_DHCP_GET_RENEW_ACK3 = 12

Structure TDHCPBits
   bIsBound As Boolean
   bOfferReceived As Boolean
   bDHCPServerDetected As Boolean
End Structure

Structure DHCP_CLIENT_FLAGS
   Val As Byte
   Bits As TDHCPBits Union
End Structure

Public Dim 
   DHCPFlags As DHCP_CLIENT_FLAGS,
   DHCPBindCount As Byte 
Dim   
   smDHCPState As SM_DHCP

(*
 *********************************************************************
 * Macro:           BOOL DHCPIsBound(void)
 * PreCondition:    None
 * Input:           None
 * Output:          TRUE if DHCP is bound to given configuration
 *                  FALSE if DHCP has yet to be bound.
 * Side Effects:    None
 * Overview:        None
 * Note:
 ********************************************************************/
*)
Const
   DHCP_TIMEOUT                  =  2 * TICK_SECOND

Public Function DHCPIsBound() As Boolean
   result = DHCPFlags.Bits.bIsBound
End Function

Dim
   DHCPSocket As UDP_SOCKET,
   DHCPServerID As DWORD_VAL,
   DHCPLeaseTime As DWORD_VAL,
   tempIPAddress As IP_ADDR,
   tempGateway As IP_ADDR,
   tempMask As IP_ADDR
   
#if defined(STACK_USE_DNS)
Dim
   tempDNS As IP_ADDR,
   tempDNS2 As IP_ADDR
#endif

//static BYTE tempHostName[16];

Structure TValidBits
   IPAddress As Boolean
   Gateway As Boolean
   Mask As Boolean
   DNS As Boolean
   DNS2 As Boolean
   HostName As Boolean
End Structure   

Structure TValidValues
   Val As Byte
   Bits As TValidBits Union
End Structure
Dim ValidValues As TValidValues
(*
 *********************************************************************
 * Function:        void DHCPReset(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Resets the DHCP client, giving up any lease, 
 *					knowledge of DHCP servers, etc.
 * Note:            None
 ********************************************************************/
*)
Public Sub DHCPReset()
   // Do nothing if DHCP is disabled
   If smDHCPState = SM_DHCP_DISABLED Then
      Exit
   EndIf

   If DHCPSocket <> INVALID_UDP_SOCKET Then
      smDHCPState = SM_DHCP_SEND_DISCOVERY
   Else
      smDHCPState = SM_DHCP_GET_SOCKET
   EndIf
      
   DHCPBindCount = 0
   DHCPFlags.Bits.bIsBound = false
End Sub
(*
 *********************************************************************
 * Function:        void DHCPDisable(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Puts DHCPTask into unhandled state "SM_DHCP_DISABLED"
 *                  and hence DHCP is effictively disabled.
 * Note:            If the board was configured by DHCP before 
 *					DHCPDisable() was called, then the configuration 
 *					will continue to be used, but the DHCP module 
 *					will no longer perform any lease renewals or 
 *					otherwise.  In other words, if the lease expires 
 *					the board will still be using the old IP address 
 *					and the DHCP server may decide to give the IP 
 *					address to someone else.  The application should 
 *					change the current IP address and other 
 *					configuration with static information provided by 
 *					the user following a call to DHCPDisable().
 ********************************************************************/
*)
Public Sub DHCPDisable()
   If DHCPSocket <> INVALID_UDP_SOCKET Then
      UDPClose(DHCPSocket)
  	  DHCPSocket = INVALID_UDP_SOCKET
   EndIf	  
   smDHCPState = SM_DHCP_DISABLED
   AppConfig.Flags.bIsDHCPEnabled = false
End Sub
(*
 *********************************************************************
 * Function:        void DHCPSend(byte MessageType)
 ********************************************************************/
*)
Sub _DHCPSend(MessageType As Byte)
   Dim i As Byte
   Dim MyIP As IP_ADDR

   UDPPut(BOOT_REQUEST)                       // op
   UDPPut(BOOT_HW_TYPE)                       // htype
   UDPPut(BOOT_LEN_OF_HW_TYPE)                // hlen
   UDPPut(0)                                  // hops
   UDPPut($12)                                // xid[0]
   UDPPut($23)                                // xid[1]
   UDPPut($34)                                // xid[2]
   UDPPut($56)                                // xid[3]
   UDPPut(0)                                  // secs[0]
   UDPPut(0)                                  // secs[1]
   UDPPut($80)                                // flags[0] with BF set
   UDPPut(0)                                  // flags[1]

   // If this is DHCP REQUEST message, use previously allocated IP address.
'   #if 0
'   If MessageType = DHCP_REQUEST_MESSAGE Then
'      UDPPutArray(@tempIPAddress, SizeOf(tempIPAddress))
'   Else
'   #endif
   UDPPut($00)
   UDPPut($00)
   UDPPut($00)
   UDPPut($00)
'EndIf

   // Set yiaddr, siaddr, giaddr as zeros,
   i = 0
   While i < 12
      UDPPut($00)
      Inc(i)
   Wend   

   // Load chaddr - Client hardware address.
   UDPPutArray(@AppConfig.MyMACAddr, SizeOf(AppConfig.MyMACAddr))

   // Set chaddr[6..15], sname and file as zeros.
   i = 0
   While i < 202
      UDPPut(0)
      Inc(i)
   Wend   

   // Load magic cookie as per RFC 1533.
   UDPPut(99)
   UDPPut(130)
   UDPPut(83)
   UDPPut(99)

   // Load message type.
   UDPPut(DHCP_MESSAGE_TYPE)
   UDPPut(DHCP_MESSAGE_TYPE_LEN)
   UDPPut(MessageType)

   If MessageType = DHCP_DISCOVER_MESSAGE Then
      // Reset offered flag so we know to act upon the next valid offer
	  DHCPFlags.Bits.bOfferReceived = false
   EndIf

   If (MessageType <> DHCP_DISCOVER_MESSAGE) And (tempIPAddress.Val <> $0000) Then
      // DHCP REQUEST message may include server identifier,
      // to identify the server we are talking to.
      // DHCP ACK may include it too.  To simplify logic,
      // we will include server identifier in DHCP ACK message too.
      // _DHCPReceive() would populate "serverID" when it
      // receives DHCP OFFER message. We will simply use that
      // when we are replying to server.
      // If this is a renwal request, do not include server id.
      UDPPut(DHCP_SERVER_IDENTIFIER)
      UDPPut(DHCP_SERVER_IDENTIFIER_LEN)
      UDPPut(DHCPServerID.v(3))
      UDPPut(DHCPServerID.v(2))
      UDPPut(DHCPServerID.v(1))
      UDPPut(DHCPServerID.v(0))
   EndIf

   // Load our interested parameters
   // This is hardcoded list.  If any new parameters are desired,
   // new lines must be added here.
   UDPPut(DHCP_PARAM_REQUEST_LIST)
   UDPPut(DHCP_PARAM_REQUEST_LIST_LEN)
   UDPPut(DHCP_SUBNET_MASK)
   UDPPut(DHCP_ROUTER)

   UDPPut(DHCP_DNS)
   UDPPut(DHCP_HOST_NAME)

   // Add requested IP address to DHCP Request Message
   If MessageType = DHCP_REQUEST_MESSAGE Then
      UDPPut(DHCP_PARAM_REQUEST_IP_ADDRESS)
      UDPPut(DHCP_PARAM_REQUEST_IP_ADDRESS_LEN)
      UDPPutArray(@tempIPAddress, DHCP_PARAM_REQUEST_IP_ADDRESS_LEN)
   EndIf

   // Add any new paramter request here.

   // End of Options.
   UDPPut(DHCP_END_OPTION)


   // Add zero padding to ensure compatibility with old BOOTP relays that discard 
   // small packets (<300 UDP octets)
   While UDPTxCount < 300
      UDPPut(0)    
   Wend
   
   // Make sure we advirtise a 0.0.0.0 IP address so all DHCP servers will respond.  
   // If we have a static IP outside the DHCP server's scope, it may simply ignore 
   // discover messages.
   MyIP.Val = AppConfig.MyIPAddr.Val
   AppConfig.MyIPAddr.Val = $00000000
   UDPFlush()
   AppConfig.MyIPAddr.Val = MyIP.Val
End Sub
(*
 *********************************************************************
        DHCP PACKET FORMAT As PER RFC 1541

   0                   1                   2                   3
   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     op (1)    |   htype (1)   |   hlen (1)    |   Hops (1)    |
   +---------------+---------------+---------------+---------------+
   |                            xid (4)                            |
   +-------------------------------+-------------------------------+
   |           secs (2)            |           Flags (2)           |
   +-------------------------------+-------------------------------+
   |                          ciaddr  (4)                          |
   +---------------------------------------------------------------+
   |                          yiaddr  (4)                          |
   +---------------------------------------------------------------+
   |                          siaddr  (4)                          |
   +---------------------------------------------------------------+
   |                          giaddr  (4)                          |
   +---------------------------------------------------------------+
   |                                                               |
   |                          chaddr  (16)                         |
   |                                                               |
   |                                                               |
   +---------------------------------------------------------------+
   |                                                               |
   |                          sname   (64)                         |
   +---------------------------------------------------------------+
   |                                                               |
   |                          file    (128)                        |
   +---------------------------------------------------------------+
   |                                                               |
   |                          options (312)                        |
   +---------------------------------------------------------------+

 ********************************************************************/
*)
Function _DHCPReceive() As Byte
   Dim v As Byte
   Dim i, j As Byte
   Dim _Type As Byte
   Dim  lbDone As Boolean
   Dim tempServerID As DWORD_VAL


   // Assume unknown message until proven otherwise.
   _Type = DHCP_UNKNOWN_MESSAGE

   UDPGet(v)                             // op
    
   // Make sure this is BOOT_REPLY.
   If v = BOOT_REPLY Then
      // Discard htype, hlen, hops, xid, secs, flags, ciaddr.
      i = 0
      While i < 15
         UDPGet(v)
         Inc(i)
      Wend   

      // Check to see if this is the first offer
	  If DHCPFlags.Bits.bOfferReceived Then
         // Discard offered IP address, we already have an offer
         i = 0
	     While i < 4
            UDPGet(v)
            Inc(i)
         Wend
      Else
         // Save offered IP address until we know for sure that we have it.
         UDPGetArray(@tempIPAddress, SizeOf(tempIPAddress))
		 ValidValues.Bits.IPAddress = true
	  EndIf

      // Ignore siaddr, giaddr
      i = 0
      While i < 8
         UDPGet(v)
         Inc(i)
      Wend   

      // Check to see if chaddr (Client Hardware Address) belongs to us.
      i = 0
      While i < 6
         UDPGet(v)
         If v <> AppConfig.MyMACAddr.v(i) Then
            GoTo UDPInvalid
         EndIf   
         Inc(i)
      Wend

      // Ignore part of chaddr, sname, file, magic cookie.
      i = 0
      While i < 206
         UDPGet(v)
         Inc(i)
      Wend

      lbDone = false
      Repeat
         // Get the Option number
		 // Break out eventually in case if this is a malformed 
		 // DHCP message, ie: missing DHCP_END_OPTION marker
		 If Not UDPGet(v) Then
		    lbDone = true
			Break
         EndIf

         Select v
            Case DHCP_MESSAGE_TYPE:
               UDPGet(v)                         // Skip len
                
               // Len must be 1.
               If  v = 1 Then
                  UDPGet(_Type)                  // Get type

                  // Throw away the packet if we know we don't need it (ie: another offer when we already have one)
	              If DHCPFlags.Bits.bOfferReceived And (_Type = DHCP_OFFER_MESSAGE) Then
			         GoTo UDPInvalid
			      EndIf
               Else
                  GoTo UDPInvalid
               EndIf

            Case DHCP_SUBNET_MASK:
               UDPGet(v)                     // Skip len
               
               // Len must be 4.
               If v = 4 Then
			      // Check to see if this is the first offer
				   If DHCPFlags.Bits.bOfferReceived Then
		               
                       // Discard offered IP mask, we already have an offer
				       i = 0
                       While i < 4
		                  UDPGet(v)
		                  Inc(i)
		               Wend
				   Else
					  UDPGetArray(@tempMask, SizeOf(tempMask))
					  ValidValues.Bits.Mask = true
                   EndIf
               Else
                  GoTo UDPInvalid
               EndIf

            Case DHCP_ROUTER:
                UDPGet(j)
                
                // Len must be >= 4.
                If j >= 4 Then
   				   // Check to see if this is the first offer
				   If DHCPFlags.Bits.bOfferReceived Then
		              // Discard offered Gateway address, we already have an offer
				      i = 0
				      While i < 4
		                 UDPGet(v)
		                 Inc(i)
		              Wend
				   Else
				      UDPGetArray(@tempGateway, SizeOf(tempGateway))
					  ValidValues.Bits.Gateway = true
				   EndIf
                Else
                   GoTo UDPInvalid
                EndIf    

                // Discard any other router addresses.
                j = j - 4
                While j <> 0 
                    UDPGet(v)
                    Dec(j)
                Wend

            #if defined(STACK_USE_DNS)
            Case DHCP_DNS:
                UDPGet(j)
                
                // Len must be >= 4.
                If j < 4 Then
                   GoTo UDPInvalid
                EndIf   

				// NOTE : for some routers, the Gateway can also be
				// the primary DNS
                // Check to see if this is the first offer
				If Not DHCPFlags.Bits.bOfferReceived Then
					UDPGetArray(@tempDNS, SizeOf(tempDNS))
					ValidValues.Bits.DNS = true
					j = j - 4
                EndIf

				// Len must be >= 4 for a secondary DNS server address
				If j >= 4 Then
					// Check to see if this is the first offer
					If Not DHCPFlags.Bits.bOfferReceived Then
					   UDPGetArray(@tempDNS2, SizeOf(tempDNS2))
					   ValidValues.Bits.DNS2 = true
					   j = j - 4
					EndIf
				EndIf

                // Discard any other DNS server addresses
                While j <> 0
                    UDPGet(v)
                    Dec(j)
                Wend
            #endif
(*
            case DHCP_HOST_NAME:
                UDPGet(j)
                
                // Len must be >= 4.
                if j < 1 then
                   goto UDPInvalid
                endif   

				// Check to see if this is the first offer
				if DHCPFlags.bits.bOfferReceived then
		           // Discard offered host name, we already have an offer
	               while j <> 0
                      UDPGet(v)
                      dec(j)
                   wend   
				else
				   i = 0
				   while i < sizeof(tempHostName)
				      UDPGet(tempHostName(i))
				      inc(i)
				      dec(j)
				   wend
       			   while j <> 0
                      UDPGet(v)
                      dec(j)
                   wend
				   ValidValues.bits.HostName = true
				endif
*)
            Case DHCP_SERVER_IDENTIFIER:
                UDPGet(v)                         // Get len
                
                // Len must be 4.
                If v = 4 Then
                   UDPGet(tempServerID.v(3))   // Get the id
                   UDPGet(tempServerID.v(2))
                   UDPGet(tempServerID.v(1))
                   UDPGet(tempServerID.v(0))
                Else
                   GoTo UDPInvalid
                EndIf

            Case DHCP_END_OPTION:
                lbDone = true

            Case DHCP_IP_LEASE_TIME:
                UDPGet(v)                         // Get len
                
                // Len must be 4.
                If v = 4 Then
					// Check to see if this is the first offer
					If DHCPFlags.Bits.bOfferReceived Then
				        // Discard offered lease time, we already have an offer
				        i = 0
				        While i < 4
		                   UDPGet(v)
		                   Inc(i)
		                Wend   
					Else
                       UDPGet(DHCPLeaseTime.v(3))
	                   UDPGet(DHCPLeaseTime.v(2))
	                   UDPGet(DHCPLeaseTime.v(1))
	                   UDPGet(DHCPLeaseTime.v(0))

	                   // In case if our clock is not as accurate as the remote 
					   // DHCP server's clock, let's treat the lease time as only 
					   // 96.875% of the value given
					   DHCPLeaseTime.Val = DHCPLeaseTime.Val - (DHCPLeaseTime.Val >> 5)
					EndIf
                Else
                    GoTo UDPInvalid
                EndIf

            Else // select case
               // Ignore all unsupport tags.
               UDPGet(j)                     // Get option len
               While j <> 0                     // Ignore option values
                  UDPGet(v)
                  Dec(j)
               Wend
         End Select
      Until lbDone
   EndIf

   // If this is an OFFER message, remember current server id.
   If _Type = DHCP_OFFER_MESSAGE Then
      DHCPServerID.Val = tempServerID.Val
	  DHCPFlags.Bits.bOfferReceived = true
   Else
      // For other types of messages, make sure that received
      // server id matches with our previous one.
      If DHCPServerID.Val <> tempServerID.Val Then
         _Type = DHCP_UNKNOWN_MESSAGE
      EndIf
   EndIf

   UDPDiscard()                             // We are done with this packet
   result = _Type
   Exit
   
UDPInvalid:
   UDPDiscard()
   result = DHCP_UNKNOWN_MESSAGE
End Function

(*
 *********************************************************************
 * Function:        void DHCPEnable(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Puts DHCPTask into the original initialization 
 *					state, if currently disabled.  Does nothing if 
 *					already enabled.
 * Note:            None
 ********************************************************************/
*)
Public Sub DHCPEnable()
   If smDHCPState = SM_DHCP_DISABLED Then
      AppConfig.Flags.bIsDHCPEnabled = 
	  smDHCPState = SM_DHCP_GET_SOCKET
	  DHCPBindCount = 0
	  DHCPFlags.Bits.bIsBound = false
   EndIf
End Sub
(*
 *********************************************************************
 * Function:        void DHCPTask(void)
 * PreCondition:    DHCPInit() is already called
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Gets And generates DHCP messages To obtain a 
 *					lease And keep it from expiring
 * Note:            None
 ********************************************************************/
*)
Dim eventTime As TICK // DHCPTask() static...
Public Sub DHCPTask()
   
   If smDHCPState = SM_DHCP_GET_SOCKET Then
      // Open a socket to send and receive broadcast messages on
      DHCPSocket = UDPOpen(DHCP_CLIENT_PORT, 0, DHCP_SERVER_PORT)
	  If DHCPSocket = INVALID_UDP_SOCKET Then
	     Exit
	  EndIf
      smDHCPState = SM_DHCP_SEND_DISCOVERY
   EndIf
      
   If smDHCPState =	SM_DHCP_SEND_DISCOVERY Then
      If UDPIsPutReady(DHCPSocket) < 300 Then
         Exit
      EndIf

      // Ensure that we transmit to the broadcast IP and MAC addresses
	  // The UDP Socket remembers who it was last talking to
	  // *** assumes remote node at structure offset 0 ***
      memset(@UDPSocketInfo(DHCPSocket), $FF, SizeOf(NODE_INFO))

	  // Assume default IP Lease time of 60 seconds.
	  // This should be minimum possible to make sure that if
	  // server did not specify lease time, we try again after this minimum time.
	  DHCPLeaseTime.Val = 60
	  ValidValues.Val = $00
	  DHCPBindCount = 0
	  DHCPFlags.Bits.bIsBound = false	
	  DHCPFlags.Bits.bOfferReceived = false
			
      // Send the DHCP Discover broadcast
	  _DHCPSend(DHCP_DISCOVER_MESSAGE)
			
      // Start a timer and begin looking for a response
	  eventTime = TickGet() + DHCP_TIMEOUT
	  smDHCPState = SM_DHCP_GET_OFFER
	  Exit
   EndIf
   
   If smDHCPState = SM_DHCP_GET_OFFER Then

      // Check to see if a packet has arrived
	  If UDPIsGetReady(DHCPSocket) < 250 Then
	  
         // Go back and transmit a new discovery if we didn't get an offer after 2 seconds
		 If LongInt(eventTime - TickGet) <= 0 Then
            smDHCPState = SM_DHCP_SEND_DISCOVERY
         EndIf      
         Exit
      EndIf

      // Let the DHCP server module know that there is a DHCP server 
	  // on this network
	  DHCPFlags.Bits.bDHCPServerDetected = true
      
	  // Check to see if we received an offer
	  If _DHCPReceive() <> DHCP_OFFER_MESSAGE Then
	     Exit
	  EndIf
	  smDHCPState = SM_DHCP_SEND_REQUEST

   EndIf

   If smDHCPState = SM_DHCP_SEND_REQUEST Then
      If UDPIsPutReady(DHCPSocket) < 258 Then
         Exit
      EndIf

	  // Send the DHCP request message
	  _DHCPSend(DHCP_REQUEST_MESSAGE)
			
      // Start a timer and begin looking for a response
	  eventTime = TickGet() + DHCP_TIMEOUT
      smDHCPState = SM_DHCP_GET_REQUEST_ACK
      Exit
   EndIf

   If smDHCPState = SM_DHCP_GET_REQUEST_ACK Then
 
      // Check to see if a packet has arrived
	  If UDPIsGetReady(DHCPSocket) < 250 Then
         // Go back and transmit a new discovery if we didn't get an ACK after 2 seconds
		 If LongInt(eventTime - TickGet()) <= 0 Then
		    smDHCPState = SM_DHCP_SEND_DISCOVERY
	     EndIf
	     Exit
      EndIf
      
      // Check to see if we received an offer
	  Select _DHCPReceive()
	     Case DHCP_ACK_MESSAGE:
		    eventTime = TickGet() + TICK_SECOND
		    smDHCPState = SM_DHCP_BOUND

            DHCPFlags.Bits.bIsBound = true	
	        DHCPBindCount = DHCPBindCount + 1
		    If ValidValues.Bits.IPAddress Then
               AppConfig.MyIPAddr = tempIPAddress
            EndIf   
            If ValidValues.Bits.Mask Then
               AppConfig.MyMask = tempMask
            EndIf   
            If ValidValues.Bits.Gateway Then
			   AppConfig.MyGateway = tempGateway
			EndIf   
            #if defined(STACK_USE_DNS)
			If ValidValues.Bits.DNS Then
			   AppConfig.PrimaryDNSServer = tempDNS
			EndIf   
			If ValidValues.Bits.DNS2 Then
      		   AppConfig.SecondaryDNSServer = tempDNS2
			EndIf	
     	    #endif
//			if ValidValues.bits.HostName then
//			   AppConfig.NetBIOSName = tempHostName
//          endif
         Case DHCP_NAK_MESSAGE:
		    smDHCPState = SM_DHCP_SEND_DISCOVERY
      End Select
      Exit
   EndIf

   If smDHCPState = SM_DHCP_BOUND Then
      If LongInt(eventTime - TickGet) >= 0 Then
	     Exit
      EndIf

      // Check to see if our lease has nearly expired and we must renew
	  eventTime = eventTime + TICK_SECOND
	  DHCPLeaseTime.Val = DHCPLeaseTime.Val - 1
	  If DHCPLeaseTime.Val <> 0 Then
	     Exit
	  EndIf
      smDHCPState = SM_DHCP_SEND_RENEW
   EndIf

   If (smDHCPState = SM_DHCP_SEND_RENEW) Or (smDHCPState = SM_DHCP_SEND_RENEW2) Or (smDHCPState = SM_DHCP_SEND_RENEW3) Then
      If UDPIsPutReady(DHCPSocket) < 258 Then
         Exit
      EndIf

      // Send the DHCP request message
	  _DHCPSend(DHCP_REQUEST_MESSAGE)
	  DHCPFlags.Bits.bOfferReceived = false
			
      // Start a timer and begin looking for a response
	  eventTime = TickGet() + DHCP_TIMEOUT
	  smDHCPState = smDHCPState + 1
      Exit
   EndIf
   
   If (smDHCPState = SM_DHCP_GET_RENEW_ACK) Or (smDHCPState = SM_DHCP_GET_RENEW_ACK2) Or (smDHCPState = SM_DHCP_GET_RENEW_ACK3) Then
      // Check to see if a packet has arrived
	  If UDPIsGetReady(DHCPSocket) < 250 Then
         // Go back and transmit a new discovery if we didn't get an ACK after 2 seconds
		 If LongInt(eventTime - TickGet()) <= 0 Then
            smDHCPState = smDHCPState + 1
			If smDHCPState > SM_DHCP_GET_RENEW_ACK3 Then
  			   smDHCPState = SM_DHCP_SEND_DISCOVERY
			EndIf
         EndIf
         Exit
      EndIf   

      // Check to see if we received an offer
	  Select _DHCPReceive()
	     Case DHCP_ACK_MESSAGE:
		    eventTime = TickGet() + TICK_SECOND
			DHCPBindCount = DHCPBindCount + 1
            smDHCPState = SM_DHCP_BOUND
         Case DHCP_NAK_MESSAGE:
		    smDHCPState = SM_DHCP_SEND_DISCOVERY
      End Select
   EndIf
End Sub

// initialise
#if defined(STACK_USE_DHCP_CLIENT)
smDHCPState = SM_DHCP_GET_SOCKET
DHCPFlags.Val = $00
DHCPBindCount = 0
DHCPSocket = INVALID_UDP_SOCKET
#endif
