{
*****************************************************************************
*  Name    : ENC28J60.BAS                                                   *
*  Date    : 28/08/2007                                                     *
*  Version : 1.1 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
*      v1.1: move SPI functions to encspi.bas                               *
*          : add support for polling ENC_INT pin (optional)                 *
*          : replace Lo/Hi functions with macros lowbyte/highbyte           *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module ENC28J60

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"

// ENC spi interface
include "encspi.bas"
   
// Since the ENC28J60 doesn't support auto-negotiation, full-duplex mode is
// not compatible with most switches/routers.  If a dedicated network is used
// where the duplex of the remote node can be manually configured, you may
// change this configuration.  Otherwise, half duplex should always be used
// note: this was originally done as a '#define'
#option ETH_DUPLEX = HALF_DUPLEX
#if (ETH_DUPLEX = HALF_DUPLEX)
  // default setting
#elseif (ETH_DUPLEX = FULL_DUPLEX)
  #warning "ETH FULL_DUPLEX may cause problems"
#else
  #error "invalid ETH_DUPLEX setting. must be FULL_DUPLEX or HALF_DUPLEX"
#endif
   
// A generic structure representing the Ethernet header starting all Ethernet 
// frames
Public Structure MAC_ETHER_HEADER
   DestMACAddr As MAC_ADDR
   SourceMACAddr As MAC_ADDR
   _Type As WORD_VAL
End Structure

// Rest are Receive Buffers
Public Const
   MAC_IP      = ($00),
   MAC_ARP     = ($06),
   MAC_UNKNOWN = ($FF) 

// bring defines into program...
Public Const
   MAC_TX_BUFFER_SIZE = _MAC_TX_BUFFER_SIZE,
   RAMSIZE = _RAMSIZE,
   TXSTART = _TXSTART,
   RXSTART = _RXSTART,
   RXSTOP = _RXSTOP,
   RXSIZE = _RXSIZE,
   BASE_TX_ADDR = _BASE_TX_ADDR,
   BASE_TCB_ADDR = _BASE_TCB_ADDR,
   BASE_HTTPB_ADDR = _BASE_HTTPB_ADDR
   
// Structures
Structure TEIE 
   Val As Byte
   RXERIE As Val.0
   TXERIE As Val.1
   WOLIE As Val.2
   TXIE As Val.3
   LINKIE As Val.4
   DMAIE As Val.5
   PKTIE As Val.6
   INTIE As Val.7
End Structure

Structure TEIR
   Val As Byte
   RXERIF As Val.0
   TXERIF As Val.1
   WOLIF As Val.2
   TXIF As Val.3
   LINKIF As Val.4
   DMAIF As Val.5
   PKTIF As Val.6
End Structure

Structure TESTAT
   Val As Byte
   CLKRDY As Val.Booleans(0)
   TXABRT As Val.1
   RXBUSY As Val.booleans(2)
   LATECOL As Val.4
   Int As Val.7
End Structure

Structure TECON2
   Val As Byte
   VRPS As Val.3
   VRTP As Val.4
   PWRSV As Val.5
   PKTDEC As Val.6
   AUTOINC As Val.7
End Structure
        
Structure TECON1
   Val As Byte
   BSEL0 As Val.0
   BSEL1 As Val.1
   RXEN As Val.2
   TXRTS As Val.Booleans(3)
   CSUMEN As Val.4
   DMAST As Val.Booleans(5)
   RXRST As Val.6
   TXRST As Val.7
End Structure
        
Structure TERXFCON
   Val As Byte
   BCEN As Val.0
   MCEN As Val.1
   HTEN As Val.2
   MPEN As Val.3
   PMEN As Val.4
   CRCEN As Val.5
   ANDOR As Val.6
   UCEN As Val.7
End Structure
        
Structure TMACON1
   Val As Byte
   MARXEN As Val.0
   PASSALL As Val.1
   RXPAUS As Val.2
   TXPAUS As Val.3
   LOOPBK As Val.4
End Structure
        
Structure TMACON3
   Val As Byte
   FULDPX As Val.0
   FRMLNEN As Val.1
   HFRMEN As Val.2
   PHDREN As Val.3
   TXCRCEN As Val.4
   PADCFG0 As Val.5
   PADCFG1 As Val.6
   PADCFG2 As Val.7
   
   // alt...
   _PADCFG As Val     // 5..7 (3 Bits)
End Structure

Structure TMACON4
   Val As Byte
   PUREPRE As Val.0
   LONGPRE As Val.1
   NOBKOFF As Val.4
   BPEN As Val.5
   DEFER As Val.6
End Structure

Structure TMICMD
   Val As Byte
   MIIRD As Val.0
   MIISCAN As Val.1
End Structure

Structure TEBSTCON
   Val As Byte
   BISTST As Val.0
   TME As Val.1
   TMSEL0 As Val.2
   TMSEL1 As Val.3
   PSEL As Val.4
   PSV0 As Val.5
   PSV1 As Val.6
   PSV2 As Val.7

   // alt...
   _TMSEL As Val    // 2..3 (2 bits)
   _PSV As Val      // 5..7 (3 bits)
End Structure

Structure TMISTAT
   Val As Byte
   BUSY As Val.Booleans(0)
   SCAN As Val.1
   NVALID As Val.2
End Structure

Structure TECOCON
   Val As Byte
   COCON0 As Val.0
   COCON1 As Val.1
   COCON2 As Val.2

   // alt...
   _COCON As Val   // 0..2 (3 bits)
End Structure
        
Structure TEFLOCON
   Val As Byte
   FCEN0 As Val.0
   FCEN1 As Val.1
   FULDPXS As Val.2
   
   // alt...
   _FCEN As Val    // 0..1 (2 bits)
End Structure

Public Structure REG
   Val As Byte
   EIEbits As TEIE Union
   EIRbits As TEIR Union
   ESTATbits As TESTAT Union
   ECON2bits As TECON2 Union
   ECON1bits As TECON1 Union
   ERXFCONbits As TERXFCON Union
   MACON1bits As TMACON1 Union
   MACON3bits As TMACON3 Union
   MACON4bits As TMACON4 Union
   MICMDbits As TMICMD Union
   EBSTCONbits As TEBSTCON Union
   MISTATbits As TMISTAT Union
   ECOCONbits As TECOCON Union
   EFLOCONbits As TEFLOCON Union
End Structure

Structure TPHCON1
   Val As Word
   PDPXMD As Val.8
   PPWRSV As Val.11
   PLOOPBK As Val.14
   PRST As Val.15
End Structure

Structure TPHSTAT1
   Val As Word
   JBSTAT As Val.1
   LLSTAT As Val.Booleans(2)
   PHDPX As Val.11
   PFDPX As Val.12
End Structure

Structure TPHID2
   Val As Word
   PREV0 As Val.0
   PREV1 As Val.1
   PREV2 As Val.2
   PREV3 As Val.3
   PPN0 As Val.4
   PPN1 As Val.5
   PPN2 As Val.6
   PPN3 As Val.7
   PPN4 As Val.8
   PPN5 As Val.9
   PID19 As Val.10
   PID20 As Val.11
   PID21 As Val.12
   PID22 As Val.13
   PID23 As Val.14
   PID24 As Val.15
   
   // alt
   PREV As Val         // 0..3   (bits 4)
   PPNL As Val         // 4..7   (bits 4)
   PPNH As Val.Byte1   // 8..9   (bits 2)
   PID As Val.Byte1    // 10..15 (bits 6)
End Structure

Structure TPHCON2
   Val As Word
   HDLDIS As Val.8
   JABBER As Val.10
   TXDIS As Val.13
   FRCLNK As Val.14
End Structure

Structure TPHSTAT2
   Val As Word
   PLRITY As Val.5
   DPXSTAT As Val.9
   LSTAT As Val.10
   COLSTAT As Val.11
   RXSTAT As Val.12
   TXSTAT As Val.13
End Structure

Structure TPHIE
   Val As Word
   PGEIE As Val.1
   PLNKIE As Val.4
End Structure

Structure TPHIR
   Val As Word
   PGIF As Val.2
   PLNKIF As Val.4
End Structure

Structure TPHLCON
   Val As Word
   STRCH As Val.1
   LFRQ0 As Val.2
   LFRQ1 As Val.3
   LBCFG0 As Val.4
   LBCFG1 As Val.5
   LBCFG2 As Val.6
   LBCFG3 As Val.7
   LACFG0 As Val.8
   LACFG1 As Val.9
   LACFG2 As Val.10
   LACFG3 As Val.11

   // alt...
   LFRQ As Val         // 2..3  (bits 2)
   LBCFG As Val        // 4..7  (bits 4)
   LACFG As Val.Byte1  // 8..11 (bits 4)
End Structure

Public Structure PHYREG
   Val As WORD_VAL
   PHCON1bits As TPHCON1 Union
   PHSTAT1bits As TPHSTAT1 Union
   PHID2bits As TPHID2 Union
   PHCON2bits As TPHCON2 Union
   PHSTAT2bits As TPHSTAT2 Union
   PHIEbits As TPHIE Union
   PHIRbits As TPHIR Union
   PHLCONbits As TPHLCON Union
End Structure

// PH Register Locations
Public Const
   PHCON1       = $00,
   PHSTAT1      = $01,
   PHID1        = $02,
   PHID2        = $03,
   PHCON2       = $10,
   PHSTAT2      = $11,
   PHIE         = $12,
   PHIR         = $13,
   PHLCON       = $14

// TXSTATUS bits...
Structure TXBits
   v(7) As Byte
   ByteCount As v(0).AsWord                  // 16
   _CollisionCount As v(2)                   // 4
   CRCError As v(2).4                        // 1
   LengthCheckError As v(2).5                // 1
   LengthOutOfRange As v(2).6                // 1
   Done As v(2).7                            // 1
   Multicast As v(3).0                       // 1
   Broadcast As v(3).1                       // 1
   PacketDefer As v(3).2                     // 1
   ExcessiveDefer As v(3).3                  // 1
   MaximumCollisions As v(3).4               // 1
   LateCollision As v(3).5                   // 1
   Giant As v(3).6                           // 1
   Underrun As v(3).7                        // 1
   BytesTransmittedOnWire As v(4).AsWord     // 16
   ControlFrame As v(6).0                    // 1
   PauseControlFrame As v(6).1               // 1
   BackpressureApplied As v(6).2             // 1
   VLANTaggedFrame As v(6).3                 // 1
                                             // 4 zeros
End Structure   

// TXSTATUS...
Structure TXSTATUS
   v(7) As Byte
   Bits As TXBits Union
End Structure

// RXSTATUS bits...
Structure RXBits
   v(4) As Byte
   ByteCount As v(0).AsWord           // 16
   PreviouslyIgnored As v(2).0        // 1
   RXDCPreviouslySeen As v(2).1       // 1
   CarrierPreviouslySeen As v(2).2    // 1
   CodeViolation As v(2).3            // 1
   CRCError As v(2).4                 // 1
   LengthCheckError As v(2).5         // 1
   LengthOutOfRange As v(2).6         // 1
   ReceiveOk As v(2).Booleans(7)      // 1
   Multicast As v(3).0                // 1
   Broadcast As v(3).1                // 1
   DribbleNibble As v(3).2            // 1
   ControlFrame As v(3).3             // 1
   PauseControlFrame As v(3).4        // 1
   UnsupportedOpcode As v(3).5        // 1 
   VLANType As v(3).6                 // 1
   zero As v(3).7                    // 1 zero
End Structure   

// RXSTATUS
Structure RXSTATUS
   v(4) As Byte
   Bits As RXBits Union
End Structure
{
*******************************************************************************
* Register locations
*******************************************************************************
}
// Bank 0 registers --------
Const
   ERDPTL       = $00,
   ERDPTH       = $01,
   EWRPTL       = $02,
   EWRPTH       = $03,
   ETXSTL       = $04,
   ETXSTH       = $05,
   ETXNDL       = $06,
   ETXNDH       = $07,
   ERXSTL       = $08,
   ERXSTH       = $09,
   ERXNDL       = $0A,
   ERXNDH       = $0B,
   ERXRDPTL     = $0C,
   ERXRDPTH     = $0D,
   ERXWRPTL     = $0E,
   ERXWRPTH     = $0F,
   EDMASTL      = $10,
   EDMASTH      = $11,
   EDMANDL      = $12,
   EDMANDH      = $13,
   EDMADSTL     = $14,
   EDMADSTH     = $15,
   EDMACSL      = $16,
   EDMACSH      = $17,
   EIE          = $1B,
   EIR          = $1C,
   ESTAT        = $1D,
   ECON2        = $1E,
   ECON1        = $1F

// Bank 1 registers -----
Const
   EHT0         = $100,
   EHT1         = $101,
   EHT2         = $102,
   EHT3         = $103,
   EHT4         = $104,
   EHT5         = $105,
   EHT6         = $106,
   EHT7         = $107,
   EPMM0        = $108,
   EPMM1        = $109,
   EPMM2        = $10A,
   EPMM3        = $10B,
   EPMM4        = $10C,
   EPMM5        = $10D,
   EPMM6        = $10E,
   EPMM7        = $10F,
   EPMCSL       = $110,
   EPMCSH       = $111,
   EPMOL        = $114,
   EPMOH        = $115,
   ERXFCON      = $118,
   EPKTCNT      = $119

// Bank 2 registers -----
Const
   MACON1       = $200,
   MACON3       = $202,
   MACON4       = $203,
   MABBIPG      = $204,
   MAIPGL       = $206,
   MAIPGH       = $207,
   MACLCON1     = $208,
   MACLCON2     = $209,
   MAMXFLL      = $20A,
   MAMXFLH      = $20B,
   MICMD        = $212,
   MIREGADR     = $214,
   MIWRL        = $216,
   MIWRH        = $217,
   MIRDL        = $218,
   MIRDH        = $219

// Bank 3 registers -----
Const
   MAADR5       = $300,
   MAADR6       = $301,
   MAADR3       = $302,
   MAADR4       = $303,
   MAADR1       = $304,
   MAADR2       = $305,
   EBSTSD       = $306,
   EBSTCON      = $307,
   EBSTCSL      = $308,
   EBSTCSH      = $309,
   MISTAT       = $30A,
   EREVID       = $312,
   ECOCON       = $315,
   EFLOCON      = $317,
   EPAUSL       = $318,
   EPAUSH       = $319

// Structures
// .. see NETTypes

{
*******************************************************************************
* Individual Register Bits
*******************************************************************************
}
// ETH/MAC/MII bits
Const
   // EIE bits ----------
   EIE_INTIE       = (1<<7),
   EIE_PKTIE       = (1<<6),
   EIE_DMAIE       = (1<<5),
   EIE_LINKIE      = (1<<4),
   EIE_TXIE        = (1<<3),
   EIE_WOLIE       = (1<<2),
   EIE_TXERIE      = (1<<1),
   EIE_RXERIE      = (1),

   // EIR bits ----------
   EIR_PKTIF       = (1<<6),
   EIR_DMAIF       = (1<<5),
   EIR_LINKIF      = (1<<4),
   EIR_TXIF        = (1<<3),
   EIR_WOLIF       = (1<<2),
   EIR_TXERIF      = (1<<1),
   EIR_RXERIF      = (1),
    
   // ESTAT bits ---------
   ESTAT_INT       = (1<<7),
   ESTAT_LATECOL   = (1<<4),
   ESTAT_RXBUSY    = (1<<2),
   ESTAT_TXABRT    = (1<<1),
   ESTAT_CLKRDY    = (1),
    
   // ECON2 bits --------
   ECON2_AUTOINC   = (1<<7),
   ECON2_PKTDEC    = (1<<6),
   ECON2_PWRSV     = (1<<5),
   ECON2_VRPS      = (1<<3),
    
   // ECON1 bits --------
   ECON1_TXRST     = (1<<7),
   ECON1_RXRST     = (1<<6),
   ECON1_DMAST     = (1<<5),
   ECON1_CSUMEN    = (1<<4),
   ECON1_TXRTS     = (1<<3),
   ECON1_RXEN      = (1<<2),
   ECON1_BSEL1     = (1<<1),
   ECON1_BSEL0     = (1),
    
   // ERXFCON bits ------
   ERXFCON_UCEN    = (1<<7),
   ERXFCON_ANDOR   = (1<<6),
   ERXFCON_CRCEN   = (1<<5),
   ERXFCON_PMEN    = (1<<4),
   ERXFCON_MPEN    = (1<<3),
   ERXFCON_HTEN    = (1<<2),
   ERXFCON_MCEN    = (1<<1),
   ERXFCON_BCEN    = (1),
    
   // MACON1 bits --------
   MACON1_LOOPBK   = (1<<4),
   MACON1_TXPAUS   = (1<<3),
   MACON1_RXPAUS   = (1<<2),
   MACON1_PASSALL  = (1<<1),
   MACON1_MARXEN   = (1),
    
   // MACON3 bits --------
   MACON3_PADCFG2  = (1<<7),
   MACON3_PADCFG1  = (1<<6),
   MACON3_PADCFG0  = (1<<5),
   MACON3_TXCRCEN  = (1<<4),
   MACON3_PHDREN   = (1<<3),
   MACON3_HFRMEN   = (1<<2),
   MACON3_FRMLNEN  = (1<<1),
   MACON3_FULDPX   = (1),
    
   // MACON4 bits --------
   MACON4_DEFER    = (1<<6),
   MACON4_BPEN     = (1<<5),
   MACON4_NOBKOFF  = (1<<4),
   MACON4_LONGPRE  = (1<<1),
   MACON4_PUREPRE  = (1),
    
   // MICMD bits ---------
   MICMD_MIISCAN   = (1<<1),
   MICMD_MIIRD     = (1),

   // EBSTCON bits -----
   EBSTCON_PSV2    = (1<<7),
   EBSTCON_PSV1    = (1<<6),
   EBSTCON_PSV0    = (1<<5),
   EBSTCON_PSEL    = (1<<4),
   EBSTCON_TMSEL1  = (1<<3),
   EBSTCON_TMSEL0  = (1<<2),
   EBSTCON_TME     = (1<<1),
   EBSTCON_BISTST  = (1),

   // MISTAT bits --------
   MISTAT_NVALID   = (1<<2),
   MISTAT_SCAN     = (1<<1),
   MISTAT_BUSY     = (1),
    
   // ECOCON bits -------
   ECOCON_COCON2   = (1<<2),
   ECOCON_COCON1   = (1<<1),
   ECOCON_COCON0   = (1),
    
   // EFLOCON bits -----
   EFLOCON_FULDPXS = (1<<2),
   EFLOCON_FCEN1   = (1<<1),
   EFLOCON_FCEN0   = (1)

// PHY bits
Const

   // PHCON1 bits ----------
   PHCON1_PRST As Word      = (1<<15),
   PHCON1_PLOOPBK As Word   = (1<<14),
   PHCON1_PPWRSV As Word    = (1<<11),
   PHCON1_PDPXMD As Word    = (1<<8),

   // PHSTAT1 bits --------
   PHSTAT1_PFDPX As Word    = (1<<12),
   PHSTAT1_PHDPX As Word    = (1<<11),
   PHSTAT1_LLSTAT As Word   = (1<<2),
   PHSTAT1_JBSTAT As Word   = (1<<1),

   // PHID2 bits --------
   PHID2_PID24 As Word      = (1<<15),
   PHID2_PID23 As Word      = (1<<14),
   PHID2_PID22 As Word      = (1<<13),
   PHID2_PID21 As Word      = (1<<12),
   PHID2_PID20 As Word      = (1<<11),
   PHID2_PID19 As Word      = (1<<10),
   PHID2_PPN5 As Word       = (1<<9),
   PHID2_PPN4 As Word       = (1<<8),
   PHID2_PPN3 As Word       = (1<<7),
   PHID2_PPN2 As Word       = (1<<6),
   PHID2_PPN1 As Word       = (1<<5),
   PHID2_PPN0 As Word       = (1<<4),
   PHID2_PREV3 As Word      = (1<<3),
   PHID2_PREV2 As Word      = (1<<2),
   PHID2_PREV1 As Word      = (1<<1),
   PHID2_PREV0 As Word      = (1),

   // PHCON2 bits ----------
   PHCON2_FRCLNK As Word    = (1<<14),
   PHCON2_TXDIS As Word     = (1<<13),
   PHCON2_JABBER As Word    = (1<<10),
   PHCON2_HDLDIS As Word    = (1<<8),

   // PHSTAT2 bits --------
   PHSTAT2_TXSTAT As Word   = (1<<13),
   PHSTAT2_RXSTAT As Word   = (1<<12),
   PHSTAT2_COLSTAT As Word  = (1<<11),
   PHSTAT2_LSTAT As Word    = (1<<10),
   PHSTAT2_DPXSTAT As Word  = (1<<9),
   PHSTAT2_PLRITY As Word   = (1<<5),

   // PHIE bits -----------
   PHIE_PLNKIE As Word      = (1<<4),
   PHIE_PGEIE As Word       = (1<<1),

   // PHIR bits -----------
   PHIR_PLNKIF As Word      = (1<<4),
   PHIR_PGIF As Word        = (1<<2),

   // PHLCON bits -------
   PHLCON_LACFG3 As Word    = (1<<11),
   PHLCON_LACFG2 As Word    = (1<<10),
   PHLCON_LACFG1 As Word    = (1<<9),
   PHLCON_LACFG0 As Word    = (1<<8),
   PHLCON_LBCFG3 As Word    = (1<<7),
   PHLCON_LBCFG2 As Word    = (1<<6),
   PHLCON_LBCFG1 As Word    = (1<<5),
   PHLCON_LBCFG0 As Word    = (1<<4),
   PHLCON_LFRQ1 As Word     = (1<<3),
   PHLCON_LFRQ0 As Word     = (1<<2),
   PHLCON_STRCH As Word     = (1<<1)
 
// v1.1 lowbyte/highbyte macros (was Lo/High Pseudo Functions)
macro lowbyte(val)
    byte(val)
end macro
macro highbyte(val)
    byte((val)>>8)
end macro

// ENC28J60 Opcodes (to be ORed with a 5 bit address)
Const
   WCR          = ($2<<5),          // Write Control Register command
   BFS          = ($4<<5),          // Bit Field Set command
   BFC          = ($5<<5),          // Bit Field Clear command
   RCR          = ($0<<5),          // Read Control Register command
   RBM          = (($1<<5) Or $1A), // Read Buffer Memory command
   WBM          = (($3<<5) Or $1A), // Write Buffer Memory command
   SR           = (($7<<5) Or $1F)  // System Reset command does not use an address.  
                                    //   It requires 0x1F, however.
Const
   ETHER_IP     = ($00),
   ETHER_ARP    = ($06)

// A header appended at the start of all RX frames by the hardware
Structure ENC_PREAMBLE
   NextPacketPointer As Word
   StatusVector As RXSTATUS
   DestMACAddr As MAC_ADDR
   SourceMACAddr As MAC_ADDR
   _Type As WORD_VAL
End Structure

// Internal MAC level variables and flags.
Dim
   NextPacketLocation As WORD_VAL,
   CurrentPacketLocation As WORD_VAL,
   WasDiscarded As Boolean,
   ENCRevID As Byte

// *****************************************************************************
// NOTE: All code in this module expect Bank 0 to be currently selected.  
// If code ever changes the bank, it must restore it to Bank 0 before returning
// *****************************************************************************
{
 ******************************************************************************
 * Function:        void WriteReg(Byte Address, Byte Data)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  Bank Select Bits must be set corresponding To the register 
 *                  To modify.
 * Input:           5 Bit address of the ETH, MAC, Or MII register To modify.  
 *                  The top 3 Bits must be 0.  
 *                  Byte To be written into the register.
 * Output:          None
 * Side Effects:    None
 * Overview:        WriteReg sends the 8 Bit WCR opcode/Address Byte over the 
 *                  SPI And Then sends the data To write in the Next 8 SPI 
 *                  clocks.
 * Note:            This routine is almost identical To the BFCReg() And 
 *                  BFSReg() functions.  It is seperate To maximize speed.  
 *                  Unlike the ReadETHReg/ReadMACReg functions, WriteReg() 
 *                  can write To any ETH Or MAC register.  Writing To PHY 
 *                  registers must be accomplished With WritePHYReg().
 *****************************************************************************/
}
Sub WriteReg(Address As Byte, Data As Byte)
    enc.assertCS()
    enc.transfer(WCR or Address)    // send the opcode + address
    enc.transfer(data)              // send the data
    enc.deassertCS()
End Sub

{
 ******************************************************************************
 * Function:        void BFCReg(Byte Address, Byte Data)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  Bank Select Bits must be set corresponding To the register 
 *                  To modify.
 * Input:           5 Bit address of the register To modify.  The top 3 Bits 
 *                  must be 0.  
 *                  Byte To be used With the Bit Field Clear operation.
 * Output:          None
 * Side Effects:    None
 * Overview:        BFCReg sends the 8 Bit BFC opcode/Address Byte over the 
 *                  SPI And Then sends the data in the Next 8 SPI clocks.
 * Note:            This routine is almost identical To the WriteReg() And 
 *                  BFSReg() functions.  It is separate To maximize speed.  
 *                  BFCReg() must only be used on ETH registers.
 *****************************************************************************/
}
Sub BFCReg(Address As Byte, Data As Byte)
    enc.assertCS()
    enc.transfer(BFC or Address)    // send the opcode + address
    enc.transfer(data)              // send the data
    enc.deassertCS()
End Sub

{
 ******************************************************************************
 * Function:        void BFSReg(Byte Address, Byte Data)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  Bank Select Bits must be set corresponding To the register 
 *                  To modify.
 * Input:           5 Bit address of the register To modify.  The top 3 Bits 
 *                  must be 0.  
 *                  Byte To be used With the Bit Field Set operation.
 * Output:          None
 * Side Effects:    None
 * Overview:        BFSReg sends the 8 Bit BFC opcode/Address Byte over the 
 *                  SPI And Then sends the data in the Next 8 SPI clocks.
 * Note:            This routine is almost identical To the WriteReg() And 
 *                  BFCReg() functions.  It is separate To maximize speed.
 *                  BFSReg() must only be used on ETH registers.
 *****************************************************************************/
}
Sub BFSReg(Address As Byte, Data As Byte)
    enc.assertCS()
    enc.transfer(BFS or Address)    // send the opcode + address
    enc.transfer(data)              // send the data
    enc.deassertCS()
End Sub

{
 ******************************************************************************
 * Function:        BankSel
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           Register address With the High Byte containing the 2 bank 
 *                  Select 2 Bits.
 * Output:          None
 * Side Effects:    None
 * Overview:        BankSel takes the High Byte of a register address And 
 *                  changes the bank Select Bits in ETHCON1 To match.
 * Note:            None
 *****************************************************************************/
}
Sub BankSel(Register As Word) 
   BFCReg(ECON1, ECON1_BSEL1 Or ECON1_BSEL0)
   BFSReg(ECON1, Register.Byte1)
End Sub

{
 ******************************************************************************
 * Function:        REG ReadETHReg(Byte Address)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  Bank Select Bits must be set corresponding To the register 
 *                  To read from.
 * Input:           5 Bit address of the ETH control register To read from.
 *                  The top 3 Bits must be 0.
 * Output:          Byte read from the Ethernet controller's ETH register.
 * Side Effects:    None
 * Overview:        ReadETHReg sends the 8 Bit RCR opcode/Address Byte over 
 *                  the SPI And Then retrives the register contents in the 
 *                  Next 8 SPI clocks.
 * Note:            This routine cannot be used To Access MAC/MII Or PHY 
 *                  registers.  Use ReadMACReg() Or ReadPHYReg() For that 
 *                  purpose.  
 *****************************************************************************/
}
Function ReadETHReg(Address As Byte) As REG
    // Select the chip and send the Read Control Register opcode/address
    enc.assertCS()
    enc.transfer(RCR or Address)    // send the opcode + address
    Result.Val = enc.transfer()              // read result
    enc.deassertCS()
End Function

{
 ******************************************************************************
 * Function:        REG ReadMACReg(Byte Address)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  Bank Select Bits must be set corresponding To the register 
 *                  To read from.
 * Input:           5 Bit address of the MAC Or MII register To read from.
 *                  The top 3 Bits must be 0.
 * Output:          Byte read from the Ethernet controller's MAC/MII register.
 * Side Effects:    None
 * Overview:        ReadMACReg sends the 8 Bit RCR opcode/Address Byte As well 
 *                  As a dummy Byte over the SPI And Then retrives the 
 *                  register contents in the last 8 SPI clocks.
 * Note:            This routine cannot be used To Access ETH Or PHY 
 *                  registers.  Use ReadETHReg() Or ReadPHYReg() For that 
 *                  purpose.  
 *****************************************************************************/
}
Function ReadMACReg(Address As Byte) As REG
    // Select the chip and send the Read Control Register opcode/address
    enc.assertCS()
    enc.transfer(RCR or Address)    // send the opcode + address
    enc.transfer()                  // send a dummy byte
    Result.Val = enc.transfer()     // send another dummy byte to read result
    enc.deassertCS()
End Function

{
 ******************************************************************************
 * Function:        ReadPHYReg
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           Address of the PHY register To read from.
 * Output:          16 Bits of data read from the PHY register.
 * Side Effects:    None
 * Overview:        ReadPHYReg performs an MII read operation.  While in 
 *                  progress, it simply polls the MII BUSY Bit wasting time 
 *                  (10.24us).
 * Note:            None
 *****************************************************************************/
}
Function ReadPHYReg(Register As Byte) As PHYREG
   Dim RegVal As REG
   
   // Set the right address and start the register read operation
   BankSel(MIREGADR)
   WriteReg(Byte(MIREGADR), Register)
   WriteReg(Byte(MICMD), MICMD_MIIRD)   

   // Loop to wait until the PHY register has been read through the MII
   // This requires 10.24us
   BankSel(MISTAT)
   RegVal = ReadMACReg(Byte(MISTAT))
   While RegVal.MISTATbits.BUSY
      RegVal = ReadMACReg(Byte(MISTAT))
   Wend

   // Stop reading
   BankSel(MIREGADR)
   WriteReg(Byte(MICMD), $00)   
    
   // Obtain results and return
   Result.Val.v(0) = ReadMACReg(Byte(MIRDL))
   Result.Val.v(1) = ReadMACReg(Byte(MIRDH))

   BankSel(ERDPTL)  // Return to Bank 0
End Function

{
 ******************************************************************************
 * Function:        WritePHYReg
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           Address of the PHY register To write To.
 *                  16 Bits of data To write To PHY register.
 * Output:          None
 * Side Effects:    Alters bank Bits To point To Bank 3
 * Overview:        WritePHYReg performs an MII write operation.  While in 
 *                  progress, it simply polls the MII BUSY Bit wasting time.
 * Note:            None
 *****************************************************************************/
}
Sub WritePHYReg(Register As Byte, Data As Word)
   Dim RegVal As REG
   
   // Write the register address
   BankSel(MIREGADR)
   WriteReg(Byte(MIREGADR), Register)
    
   // Write the data
   // Order is important: write low byte first, high byte last
   WriteReg(Byte(MIWRL), Data.Byte0)  // 1st byte
   WriteReg(Byte(MIWRH), Data.Byte1)  // 2nd byte

   // Wait until the PHY register has been written
   BankSel(MISTAT)
   RegVal = ReadMACReg(Byte(MISTAT))
   While RegVal.MISTATbits.BUSY
      RegVal = ReadMACReg(Byte(MISTAT))
   Wend
   BankSel(ERDPTL)  // Return to Bank 0
End Sub

{
 ******************************************************************************
 * Macro:           void SetLEDConfig(Word NewConfig)
 * PreCondition:    SPI bus must be initialized (done in MACInit()).
 * Input:           NewConfig - xxx0: Pulse stretching disabled
 *                              xxx2: Pulse stretch To 40ms (default)
 *                              xxx6: Pulse stretch To 73ms
 *                              xxxA: Pulse stretch To 139ms
 *                              
 *                              xx1x: LEDB - TX
 *                              xx2x: LEDB - RX (default)
 *                              xx3x: LEDB - collisions
 *                              xx4x: LEDB - link
 *                              xx5x: LEDB - duplex
 *                              xx7x: LEDB - TX And RX
 *                              xx8x: LEDB - on
 *                              xx9x: LEDB - off
 *                              xxAx: LEDB - blink fast
 *                              xxBx: LEDB - blink slow
 *                              xxCx: LEDB - link And RX
 *                              xxDx: LEDB - link And TX And RX
 *                              xxEx: LEDB - duplex And collisions
 *
 *                              x1xx: LEDA - TX
 *                              x2xx: LEDA - RX
 *                              x3xx: LEDA - collisions
 *                              x4xx: LEDA - link (default)
 *                              x5xx: LEDA - duplex
 *                              x7xx: LEDA - TX And RX
 *                              x8xx: LEDA - on
 *                              x9xx: LEDA - off
 *                              xAxx: LEDA - blink fast
 *                              xBxx: LEDA - blink slow
 *                              xCxx: LEDA - link And RX
 *                              xDxx: LEDA - link And TX And RX
 *                              xExx: LEDA - duplex And collisions
 * Output:          None
 * Side Effects:    None
 * Overview:        Writes the value of NewConfig into the PHLCON PHY register.  
 *                  The LED pins will beginning outputting the new 
 *                  configuration immediately.
 * Note:            
 *****************************************************************************/
}
Public Sub SetLEDConfig(NewConfig As Word)
   WritePHYReg(PHLCON, NewConfig)
End Sub

{
 ******************************************************************************
 * Macro:           Word GetLEDConfig(void)
 * PreCondition:    SPI bus must be initialized (done in MACInit()).
 * Input:           None
 * Output:          Word -  xxx0: Pulse stretching disabled
 *                          xxx2: Pulse stretch To 40ms (default)
 *                          xxx6: Pulse stretch To 73ms
 *                          xxxA: Pulse stretch To 139ms
 *                              
 *                          xx1x: LEDB - TX
 *                          xx2x: LEDB - RX (default)
 *                          xx3x: LEDB - collisions
 *                          xx4x: LEDB - link
 *                          xx5x: LEDB - duplex
 *                          xx7x: LEDB - TX And RX
 *                          xx8x: LEDB - on
 *                          xx9x: LEDB - off
 *                          xxAx: LEDB - blink fast
 *                          xxBx: LEDB - blink slow
 *                          xxCx: LEDB - link And RX
 *                          xxDx: LEDB - link And TX And RX
 *                          xxEx: LEDB - duplex And collisions
 *
 *                          x1xx: LEDA - TX
 *                          x2xx: LEDA - RX
 *                          x3xx: LEDA - collisions
 *                          x4xx: LEDA - link (default)
 *                          x5xx: LEDA - duplex
 *                          x7xx: LEDA - TX And RX
 *                          x8xx: LEDA - on
 *                          x9xx: LEDA - off
 *                          xAxx: LEDA - blink fast
 *                          xBxx: LEDA - blink slow
 *                          xCxx: LEDA - link And RX
 *                          xDxx: LEDA - link And TX And RX
 *                          xExx: LEDA - duplex And collisions
 * Side Effects:    None
 * Overview:        Returns the current value of the PHLCON register.
 * Note:            None
 *****************************************************************************/
}
Public Function GetLEDConfig() As PHYREG
   result = ReadPHYReg(PHLCON)
End Function

{
 ******************************************************************************
 * Function:        static void SendSystemReset(void)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        SendSystemReset sends the System Reset SPI command To 
 *                  the Ethernet controller.  It resets all register contents 
 *                  (except For ECOCON) And returns the Device To the power 
 *                  on default state.
 * Note:            None
 *****************************************************************************/
}
Sub SendSystemReset()
    enc.assertCS()
    enc.transfer(SR)
    enc.deassertCS()
End Sub

{
 ******************************************************************************
 * Function:        void MACPut(Byte Val)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  EWRPT must point To the location To begin writing.
 * Input:           Byte To write into the ENC28J60 buffer memory
 * Output:          None
 * Side Effects:    None
 * Overview:        MACPut outputs the Write Buffer Memory opcode/constant 
 *                  (8 Bits) And data To write (8 Bits) over the SPI.  
 *                  EWRPT is incremented after the write.
 * Note:            None
 *****************************************************************************/
}
Public Sub MACPut(Val As Byte)
    enc.assertCS()
    enc.transfer(WBM)    // send the opcode
    enc.transfer(Val)
    enc.deassertCS()
End Sub

{
 ******************************************************************************
 * Function:        Byte MACGet()
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  ERDPT must point To the place To read from.
 * Input:           None
 * Output:          Byte read from the ENC28J60's RAM
 * Side Effects:    None
 * Overview:        MACGet returns the Byte pointed To by ERDPT And 
 *                  increments ERDPT so MACGet() can be called again.  The 
 *                  increment will follow the receive buffer wrapping boundary.
 * Note:            None
 *****************************************************************************/
}
Public Function MACGet() As Byte
    enc.assertCS()
    enc.transfer(RBM)    // send the opcode
    result = enc.transfer()
    enc.deassertCS()
End Function

{
 ******************************************************************************
 * Function:        void MACInit(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        MACInit sets up the PIC's SPI module and all the registers
 *                  in the ENC28J60 so that normal operation can begin
 * Note:            None
 *****************************************************************************/
}
Public Sub MACInit()
    Dim i As Byte

    // v1.1 Set up the SPI module on the PIC for communications with the ENC28J60
    enc.Initialize()

    // Wait for CLKRDY to become set.
    // Bit 3 in ESTAT is an unimplemented bit.  If it reads out as '1' that
    // means the part is in RESET or there is something wrong with the SPI 
    // connection.  This loop makes sure that we can communicate with the 
    // ENC28J60 before proceeding.
    i = ReadETHReg(ESTAT)
    While (i And $08) <> 0 Or (Not i And ESTAT_CLKRDY) <> 0
        i = ReadETHReg(ESTAT)
    Wend

    // RESET the entire ENC28J60, clearing all registers
    SendSystemReset()    
    DelayMS(1)

    // Start up in Bank 0 and configure the receive buffer boundary pointers 
    // and the buffer write protect pointer (receive buffer read pointer)
    WasDiscarded = true
    NextPacketLocation.Val = RXSTART

    WriteReg(ERXSTL, lowbyte(RXSTART))
    WriteReg(ERXSTH, highbyte(RXSTART))   
   
    WriteReg(ERXRDPTL, lowbyte(RXSTOP))   // Write low byte first
    WriteReg(ERXRDPTH, highbyte(RXSTOP))   // Write high byte last
    WriteReg(ERXNDL, lowbyte(RXSTOP))
    WriteReg(ERXNDH, highbyte(RXSTOP))
    WriteReg(ETXSTL, lowbyte(TXSTART))
    WriteReg(ETXSTH, highbyte(TXSTART))

    // Write a permanant per packet control byte of 0x00
    WriteReg(EWRPTL, lowbyte(TXSTART))
    WriteReg(EWRPTH, highbyte(TXSTART))
    MACPut($00)

    // Enter Bank 1 and configure Receive Filters 
    // (No need to reconfigure - Unicast OR Broadcast with CRC checking is acceptable)
    // Write ERXFCON_CRCEN only to ERXFCON to enter promiscuous mode

    // Promiscious mode example:
    //BankSel(ERXFCON)
    //WriteReg(byte(ERXFCON), ERXFCON_CRCEN)
    
    // Enter Bank 2 and configure the MAC
    BankSel(MACON1)

    // Enable the receive portion of the MAC
    WriteReg(Byte(MACON1), MACON1_TXPAUS Or MACON1_RXPAUS Or MACON1_MARXEN)
    
    // Pad packets to 60 bytes, add CRC, and check Type/Length field.
    // Set the back-to-back inter-packet gap time to IEEE specified requirements.
    // The meaning of the MABBIPG value changes with the duplex state, so it must
    // be updated in this function.
    // In full duplex, 0x15 represents 9.6us; 0x12 is 9.6us in half duplex
  #if (ETH_DUPLEX = FULL_DUPLEX)
    WriteReg(Byte(MACON3), MACON3_PADCFG0 Or MACON3_TXCRCEN Or MACON3_FRMLNEN Or MACON3_FULDPX)
    WriteReg(Byte(MABBIPG), $15) 
  #else    // ETH_DUPLEX = HALF_DUPLEX
    WriteReg(Byte(MACON3), MACON3_PADCFG0 Or MACON3_TXCRCEN Or MACON3_FRMLNEN)
    WriteReg(Byte(MABBIPG), $12) 
  #endif    // ETH_DUPLEX

    // Allow infinite deferals if the medium is continuously busy 
    // (do not time out a transmission if the half duplex medium is 
    // completely saturated with other people's data)
    WriteReg(Byte(MACON4), MACON4_DEFER)

    // Late collisions occur beyond 63+8 bytes (8 bytes for preamble/start of frame delimiter)
    // 55 is all that is needed for IEEE 802.3, but ENC28J60 B5 errata for improper link pulse 
    // collisions will occur less often with a larger number.
    WriteReg(Byte(MACLCON2), 63)
    
    // Set non-back-to-back inter-packet gap to 9.6us.  The back-to-back 
    // inter-packet gap (MABBIPG) is set by MACSetDuplex() which is called later
    WriteReg(Byte(MAIPGL), $12)
    WriteReg(Byte(MAIPGH), $0C)

    // Set the maximum packet size which the controller will accept
    WriteReg(Byte(MAMXFLL), lowbyte(6 + 6 + 2 + 1500 + 4))   // 1518 is the IEEE 802.3 specified limit
    WriteReg(Byte(MAMXFLH), highbyte(6 + 6 + 2 + 1500 + 4))   // 1518 is the IEEE 802.3 specified limit
      
    // Enter Bank 3 and initialize physical MAC address registers
    BankSel(MAADR1)
    WriteReg(Byte(MAADR1), AppConfig.MyMACAddr.v(0))
    WriteReg(Byte(MAADR2), AppConfig.MyMACAddr.v(1))
    WriteReg(Byte(MAADR3), AppConfig.MyMACAddr.v(2))
    WriteReg(Byte(MAADR4), AppConfig.MyMACAddr.v(3))
    WriteReg(Byte(MAADR5), AppConfig.MyMACAddr.v(4))
    WriteReg(Byte(MAADR6), AppConfig.MyMACAddr.v(5))

    // Disable the CLKOUT output to reduce EMI generation
    WriteReg(Byte(ECOCON), $00)  // Output off (0V)
    //WriteReg(BYTE(ECOCON), $01)    // 25.000MHz
    //WriteReg(BYTE(ECOCON), $03)    // 8.3333MHz ( *4 with PLL is 33.3333MHz)

    // Get the Rev ID so that we can implement the correct errata workarounds
    ENCRevID = ReadETHReg(Byte(EREVID))

    // Disable half duplex loopback in PHY. Bank bits changed to Bank 2 as a side effect
    WritePHYReg(PHCON2, PHCON2_HDLDIS)

    // Configure LEDA to display LINK status, LEDB to display TX/RX activity
    SetLEDConfig($3472)
    
    // enable INT output for receive PKTIE and LINKIE
    WriteReg(EIE, EIE_INTIE or EIE_LINKIE or EIE_PKTIE)

    // Set the PHY duplex to match the MAC duplex state
  #if (ETH_DUPLEX = FULL_DUPLEX)
    WritePHYReg(PHCON1, PHCON1_PDPXMD)
  #else    // ETH_DUPLEX = HALF_DUPLEX
    WritePHYReg(PHCON1, $0000)
  #endif
    BankSel(ERDPTL)      // Return to default Bank 0

    // Enable packet reception
    BFSReg(ECON1, ECON1_RXEN)
End Sub

//
//---------------------------------------------------------------------------
//  Function:        bool MACIsIntrAsserted()
//  Overview:        polls the intr pin and returns true/false if intr is asserted
//---------------------------------------------------------------------------
//
public dim MACIsIntrAsserted as enc.getINT

{
 ******************************************************************************
 * Function:        BOOL MACIsLinked(void)
 * Output:          TRUE: If the PHY reports that a link partner is present 
 *                        And the link has been up continuously since the last 
 *                        call To MACIsLinked()
 *                  FALSE: If the PHY reports no link partner, Or the link went 
 *                         down momentarily since the last call To MACIsLinked()
 * Side Effects:    None
 * Overview:        Returns the PHSTAT1.LLSTAT Bit.
 * Note:            None
 *****************************************************************************/
}
Public Function MACIsLinked() As Boolean
   Dim PHYRegister As PHYREG
   // LLSTAT is a latching low link status bit.  Therefore, if the link 
   // goes down and comes back up before a higher level stack program calls
   // MACIsLinked(), MACIsLinked() will still return FALSE.  The next 
   // call to MACIsLinked() will return TRUE (unless the link goes down 
   // again).
   PHYRegister = ReadPHYReg(PHSTAT1)
   result = PHYRegister.PHSTAT1bits.LLSTAT
End Function

{
 ******************************************************************************
 * Function:        BOOL MACIsTxReady(void)
 * PreCondition:    None
 * Input:           None
 * Output:          TRUE: If no Ethernet transmission is in progress
 *                  FALSE: If a previous transmission was started, And it has 
 *                         Not completed yet.  While FALSE, the data in the 
 *                         transmit buffer And the TXST/TXND pointers must Not
 *                         be changed.
 * Side Effects:    None
 * Overview:        Returns the ECON1.TXRTS Bit
 * Note:            None
 *****************************************************************************/
}
Public Function MACIsTxReady() As Boolean
   Dim Register As REG
   Register = ReadETHReg(ECON1)
   result = Not Register.ECON1bits.TXRTS
End Function

{
 ******************************************************************************
 * Function:        void MACDiscardRx(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Marks the last received packet (obtained using 
 *                  MACGetHeader())As being processed And frees the buffer 
 *                  memory associated With it
 * Note:            None
 *****************************************************************************/
}
Public Sub MACDiscardRx()
   Dim NewRXRDLocation As WORD_VAL

   // Make sure the current packet was not already discarded
   If WasDiscarded Then
      Exit
   EndIf
   WasDiscarded = true
    
   // Decrement the next packet pointer before writing it into 
   // the ERXRDPT registers.  This is a silicon errata workaround.
   // RX buffer wrapping must be taken into account if the 
   // NextPacketLocation is precisely RXSTART.
   NewRXRDLocation.Val = NextPacketLocation.Val - 1
   If NewRXRDLocation.Val > RXSTOP Then
      NewRXRDLocation.Val = RXSTOP
   EndIf

   // Decrement the RX packet counter register, EPKTCNT
   BFSReg(ECON2, ECON2_PKTDEC)

   // Move the receive read pointer to unwrite-protect the memory used by the 
   // last packet.  The writing order is important: set the low byte first, 
   // high byte last.
   WriteReg(ERXRDPTL, NewRXRDLocation.v(0))
   WriteReg(ERXRDPTH, NewRXRDLocation.v(1))
End Sub

{
 ******************************************************************************
 * Function:        Word MACGetFreeRxSize(void)
 * PreCondition:    None
 * Input:           None
 * Output:          A Word estimate of how much RX buffer space is free at 
 *                  the present time.
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 *****************************************************************************/
}
Public Function MACGetFreeRxSize() As Word
   Dim ReadPT, WritePT As WORD_VAL
   Dim Register As REG
   
   // Read the Ethernet hardware buffer write pointer.  Because packets can be 
   // received at any time, it can change between reading the low and high 
   // bytes.  A loop is necessary to make certain a proper low/high byte pair
   // is read.
   BankSel(EPKTCNT)
   Repeat
      // Save EPKTCNT in a temporary location
      ReadPT.v(0) = ReadETHReg(Byte(EPKTCNT))

      BankSel(ERXWRPTL)
      WritePT.v(0) = ReadETHReg(ERXWRPTL)
      WritePT.v(1) = ReadETHReg(ERXWRPTH)
    
      BankSel(EPKTCNT)
      Register = ReadETHReg(Byte(EPKTCNT))
   Until Register.Val = ReadPT.v(0)
    
   // Determine where the write protection pointer is
   BankSel(ERXRDPTL)
   ReadPT.v(0) = ReadETHReg(ERXRDPTL)
   ReadPT.v(1) = ReadETHReg(ERXRDPTH)
    
   // Calculate the difference between the pointers, taking care to account 
   // for buffer wrapping conditions
   If WritePT.Val > ReadPT.Val Then
      result = (RXSTOP - RXSTART) - (WritePT.Val - ReadPT.Val)
   ElseIf WritePT.Val = ReadPT.Val Then
      result = RXSIZE - 1
   Else
     result = ReadPT.Val - WritePT.Val - 1
   EndIf
End Function 

{
 ******************************************************************************
 * Function:        Word MACGetArray(Byte *Val, Word len)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  ERDPT must point To the place To read from.
 * Input:           *Val: Pointer To storage location
 *                  len:  Number of Bytes To read from the data buffer.
 * Output:          Byte(S) of data read from the data buffer.
 * Side Effects:    None
 * Overview:        Burst reads several sequential Bytes from the data buffer 
 *                  And places them into local memory.  With SPI burst support, 
 *                  it performs much faster than multiple MACGet() calls.
 *                  ERDPT is incremented after each Byte, following the same 
 *                  rules As MACGet().
 * Note:            None
 *****************************************************************************/
}
Public Function MACGetArray(dest As Word, len As Word) As Word 
    result = len
    
    enc.assertCS()
    enc.transfer(RBM)   // Send the Read Buffer Memory opcode
    enc.burst_read(dest, len)
    enc.deassertCS()    // Terminate the burst operation
End Function

{
 ******************************************************************************
 * Function:        void MACPutArray(Byte *Val, Word len)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  EWRPT must point To the location To begin writing.
 * Input:           *Val: Pointer To source of Bytes To copy.
 *                  len:  Number of Bytes To write To the data buffer.
 * Output:          None
 * Side Effects:    None
 * Overview:        MACPutArray writes several sequential Bytes To the 
 *                  ENC28J60 RAM.  It performs faster than multiple MACPut()
 *                  calls.  EWRPT is incremented by len.
 * Note:            None
 *****************************************************************************/
}
Public Sub MACPutArray(src As Word, len As Word)
    enc.assertCS()
    enc.transfer(WBM)     // Send the Write Buffer Memory opcode
    enc.burst_write(src, len)
    enc.deassertCS()    // Terminate the burst operation
End Sub

{
 ******************************************************************************
 * Function:        BOOL MACGetHeader(MAC_ADDR *remote, Byte* Type)
 * PreCondition:    None
 * Input:           *remote: Location To store the Source MAC address of the 
 *                           received frame.
 *                  *Type: Location of a Byte To store the constant 
 *                         MAC_UNKNOWN, ETHER_IP, Or ETHER_ARP, representing 
 *                         the contents of the Ethernet Type field.
 * Output:          TRUE: If a packet was waiting in the RX buffer.  The 
 *                        remote, And Type values are updated.
 *                  FALSE: If a packet was Not pending.  remote And Type are 
 *                         Not changed.
 * Side Effects:    Last packet is discarded If MACDiscardRx() hasn't already
 *                  been called.
 * Overview:        None
 * Note:            None
 *****************************************************************************/
}
Public Function MACGetHeader(ByRef remote As MAC_ADDR, ByRef _Type As Byte) As Boolean
   Dim header As ENC_PREAMBLE
   Dim PacketCount As Byte

    // set default return value (no packet available)
    result = false

    // check to see if INT pin is asserted
    if (MACIsIntrAsserted() = false) then
        exit
    endif

   // Test if at least one packet has been received and is waiting
   BankSel(EPKTCNT)
   PacketCount = ReadETHReg(Byte(EPKTCNT))
   BankSel(ERDPTL)
   If PacketCount = 0 Then
      Exit
   EndIf

   // Make absolutely certain that any previous packet was discarded
   If WasDiscarded = false Then
      MACDiscardRx()
      Exit
   EndIf

   // Set the SPI read pointer to the beginning of the next unprocessed packet
   CurrentPacketLocation.Val = NextPacketLocation.Val
   WriteReg(ERDPTL, CurrentPacketLocation.v(0))
   WriteReg(ERDPTH, CurrentPacketLocation.v(1))

   // Obtain the MAC header from the Ethernet buffer
   MACGetArray(@header, SizeOf(header))

   // The EtherType field, like most items transmitted on the Ethernet medium
   // are in big endian.
   header._Type.Val = NETUtils.swaps(header._Type.Val)
   

   // Validate the data returned from the ENC28J60.  Random data corruption, 
   // such as if a single SPI bit error occurs while communicating or a 
   // momentary power glitch could cause this to occur in rare circumstances.

   If header.NextPacketPointer > RXSTOP Or 
      header.NextPacketPointer.0 <> 0 Or 
      header.StatusVector.Bits.zero <> 0 Or
      header.StatusVector.Bits.CRCError <> 0 Or
      header.StatusVector.Bits.ByteCount > 1518 Or
      Not header.StatusVector.Bits.ReceiveOk Then
      Reset()
   EndIf

   // Save the location where the hardware will write the next packet to
   NextPacketLocation.Val = header.NextPacketPointer

   // Return the Ethernet frame's Source MAC address field to the caller
   // This parameter is useful for replying to requests without requiring an 
   // ARP cycle.
   // memcpy((void}remote->v, (void}header.SourceMACAddr.v, SizeOf( *remote));
   remote.v = header.SourceMACAddr.v

    // Return a simplified version of the EtherType field to the caller
    _Type = MAC_UNKNOWN
    If  header._Type.v(1) = $08 And (header._Type.v(0) = ETHER_IP Or header._Type.v(0) = ETHER_ARP) Then
       _Type = header._Type.v(0)
    EndIf

    // Mark this packet as discardable
    WasDiscarded = false
    result = true
End Function

{
 ******************************************************************************
 * Function:        void MACPutHeader(MAC_ADDR *remote, Byte Type, Word dataLen)
 * PreCondition:    MACIsTxReady() must return TRUE.
 * Input:           *remote: Pointer To memory which contains the destination
 *                           MAC address (6 Bytes)
 *                  Type: The constant ETHER_ARP Or ETHER_IP, defining which 
 *                        value To write into the Ethernet header's type field.
 *                  dataLen: Length of the Ethernet data payload
 * Output:          None
 * Side Effects:    None
 * Overview:        None
 * Note:            Because of the dataLen parameter, it is probably 
 *                  advantagous To call this Function immediately before 
 *                  transmitting a packet rather than initially when the 
 *                  packet is first created.  The order in which the packet
 *                  is constructed (header first Or data first) is Not 
 *                  important.
 *****************************************************************************/
}
Public Sub MACPutHeader(ByRef remote As MAC_ADDR, _Type As Byte, dataLen As Word)
   // Set the SPI write pointer to the beginning of the transmit buffer (post per packet control byte)
   WriteReg(EWRPTL, lowbyte(TXSTART+1))
   WriteReg(EWRPTH, highbyte(TXSTART+1))

   // Calculate where to put the TXND pointer
   dataLen = dataLen + SizeOf(MAC_ETHER_HEADER) + TXSTART

   // Write the TXND pointer into the registers, given the dataLen given
   WriteReg(ETXNDL, dataLen.Byte0)
   WriteReg(ETXNDH, dataLen.Byte1)

   // Set the per-packet control byte and write the Ethernet destination 
   // address
   MACPutArray(@remote, SizeOf(MAC_ADDR))

   // Write our MAC address in the Ethernet source field
   MACPutArray(@AppConfig.MyMACAddr, SizeOf(AppConfig.MyMACAddr))

   // Write the appropriate Ethernet Type WORD for the protocol being used
   MACPut($08)
   If _Type = MAC_IP Then
      MACPut(ETHER_IP)
   Else
      MACPut(ETHER_ARP)
   EndIf
End Sub

{
 ******************************************************************************
 * Function:        void MACFlush(void)
 * PreCondition:    A packet has been created by calling MACPut() And 
 *                  MACPutHeader().
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        MACFlush causes the current TX packet To be sent out on 
 *                  the Ethernet medium.  The Hardware MAC will take control
 *                  And handle CRC generation, collision retransmission And 
 *                  other details.
 * Note:            After transmission completes (MACIsTxReady() returns TRUE), 
 *                  the packet can be modified And transmitted again by calling 
 *                  MACFlush() again.  Until MACPutHeader() Or MACPut() is 
 *                  called (in the TX data area), the data in the TX buffer 
 *                  will Not be corrupted.
 *****************************************************************************/
}
Public Sub MACFlush()
   Dim AttemptCounter As Word
   Dim Register As REG
   Dim ReadPtrSave As WORD_VAL
   Dim TXEnd As WORD_VAL
   Dim TXSTAT As TXSTATUS
   Dim i As Byte        
      
   // Reset transmit logic if a TX Error has previously occured
   // This is a silicon errata workaround
   BFSReg(ECON1, ECON1_TXRST)
   BFCReg(ECON1, ECON1_TXRST)
   BFCReg(EIR, EIR_TXERIF Or EIR_TXIF)

   // Start the transmission
   // After transmission completes (MACIsTxReady() returns TRUE), the packet 
   // can be modified and transmitted again by calling MACFlush() again.
   // Until MACPutHeader() is called, the data in the TX buffer will not be 
   // corrupted.
   BFSReg(ECON1, ECON1_TXRTS)     
   
   // Revision B5 and B7 silicon errata workaround
   If (ENCRevID = $05) Or (ENCRevID = $06) Then

      AttemptCounter = $0000
      Register = ReadETHReg(EIR)

      While Not (Register.Val And (EIR_TXERIF Or EIR_TXIF)) <> 0 And (AttemptCounter < 1000)
         Inc(AttemptCounter)
         Register = ReadETHReg(EIR)
         If Register.EIRbits.TXERIF <> 0 Or (AttemptCounter >= 1000) Then

           // Cancel the previous transmission if it has become stuck set
           BFCReg(ECON1, ECON1_TXRTS)

           // Save the current read pointer (controlled by application)
           ReadPtrSave.v(0) = ReadETHReg(ERDPTL)
           ReadPtrSave.v(1) = ReadETHReg(ERDPTH)

            // Get the location of the transmit status vector
            TXEnd.v(0) = ReadETHReg(ETXNDL)
            TXEnd.v(1) = ReadETHReg(ETXNDH)
            Inc(TXEnd.Val)
            
            // Read the transmit status vector
            WriteReg(ERDPTL, TXEnd.v(0))
            WriteReg(ERDPTH, TXEnd.v(1))
            MACGetArray(@TXSTAT, SizeOf(TXSTAT))

            // Implement retransmission if a late collision occured (this can 
            // happen on B5 when certain link pulses arrive at the same time 
            // as the transmission)
            i = 0
            While i < 16
               Register = ReadETHReg(EIR)
               If Register.EIRbits.TXERIF <> 0 And TXSTAT.Bits.LateCollision <> 0 Then
                   
                  // Reset the TX logic
                  BFSReg(ECON1, ECON1_TXRST)
                  BFCReg(ECON1, ECON1_TXRST)
                  BFCReg(EIR, EIR_TXERIF Or EIR_TXIF)

                  // Transmit the packet again
                  BFSReg(ECON1, ECON1_TXRTS)
                  Register = ReadETHReg(EIR) 
                  While (Register.Val And (EIR_TXERIF Or EIR_TXIF)) = 0
                     Register = ReadETHReg(EIR)
                  Wend

                  // Cancel the previous transmission if it has become stuck set
                  BFCReg(ECON1, ECON1_TXRTS)

                  // Read transmit status vector
                  WriteReg(ERDPTL, TXEnd.v(0))
                  WriteReg(ERDPTH, TXEnd.v(1))
                  MACGetArray(@TXSTAT, SizeOf(TXSTAT))
               Else
                  Break
               EndIf
               Inc(i)
            Wend

            // Restore the current read pointer
            WriteReg(ERDPTL, ReadPtrSave.v(0))
            WriteReg(ERDPTH, ReadPtrSave.v(1))
         EndIf
         
         // again
         Register = ReadETHReg(EIR)
      Wend
   EndIf
End Sub

{
 ******************************************************************************
 * Function:        void MACSetReadPtrInRx(Word offset)
 * PreCondition:    A packet has been obtained by calling MACGetHeader() And 
 *                  getting a TRUE result.
 * Input:           offset: Word specifying how many Bytes beyond the Ethernet 
 *                          header's type field to relocate the SPI read 
 *                          pointer.
 * Output:          None
 * Side Effects:    None
 * Overview:        SPI read pointer are updated.  All calls To 
 *                  MACGet() And MACGetArray() will use these new values.
 * Note:            RXSTOP must be statically defined As being > RXSTART For 
 *                  this Function To work correctly.  In other Words, do Not 
 *                  define an RX buffer which spans the 0x1FFF->0x0000 memory
 *                  boundary.
 *****************************************************************************/
}
Public Sub MACSetReadPtrInRx(offset As Word)
   Dim ReadPT As WORD_VAL

   // Determine the address of the beginning of the entire packet
   // and adjust the address to the desired location
   ReadPT.Val = CurrentPacketLocation.Val + SizeOf(ENC_PREAMBLE) + offset
    
   // Since the receive buffer is circular, adjust if a wraparound is needed
   If ReadPT.Val > RXSTOP Then
      ReadPT.Val = ReadPT.Val - RXSIZE
   EndIf   
    
   // Set the SPI read pointer to the new calculated value
   WriteReg(ERDPTL, ReadPT.v(0))
   WriteReg(ERDPTH, ReadPT.v(1))
End Sub

{
 ******************************************************************************
 * Function:        Word MACSetWritePtr(Word Address)
 * PreCondition:    None
 * Input:           Address: Address To seek To
 * Output:          Word: Old EWRPT location
 * Side Effects:    None
 * Overview:        SPI write pointer is updated.  All calls To 
 *                  MACPut() And MACPutArray() will use this new value.
 * Note:            None
 *****************************************************************************/
}
Public Function MACSetWritePtr(Address As Word) As Word
   Dim oldVal As WORD_VAL

   oldVal.v(0) = ReadETHReg(EWRPTL)
   oldVal.v(1) = ReadETHReg(EWRPTH)

   // Set the SPI write pointer to the new calculated value
   WriteReg(EWRPTL, Address.Byte0)
   WriteReg(EWRPTH, Address.Byte1)

   result = oldVal.Val
End Function

{
 ******************************************************************************
 * Function:        Word MACSetReadPtr(Word Address)
 * PreCondition:    None
 * Input:           Address: Address To seek To
 * Output:          Word: Old ERDPT value
 * Side Effects:    None
 * Overview:        SPI write pointer is updated.  All calls To 
 *                  MACPut() And MACPutArray() will use this new value.
 * Note:            None
 *****************************************************************************/
}
Public Function MACSetReadPtr(Address As Word) As Word
   Dim oldVal As WORD_VAL

   oldVal.v(0) = ReadETHReg(ERDPTL)
   oldVal.v(1) = ReadETHReg(ERDPTH)

   // Set the SPI write pointer to the new calculated value
   WriteReg(ERDPTL, Address.Byte0)
   WriteReg(ERDPTH, Address.Byte1)
   result = oldVal.Val
End Function

{
 ******************************************************************************
 * Function:        Word CalcIPBufferChecksum(Word len)
 * PreCondition:    Read buffer pointer set To starting of checksum data
 * Input:           len: Total number of Bytes To calculate the checksum over. 
 *                       The first Byte included in the checksum is the Byte 
 *                       pointed To by ERDPT, which is updated by calls To 
 *                       MACSetReadPtr(), MACGet(), MACGetArray(), 
 *                       MACGetHeader(), etc.
 * Output:          16-Bit checksum As defined by RFC 793
 * Side Effects:    None
 * Overview:        This Function performs a checksum calculation in the MAC
 *                  buffer itself.  The ENC28J60 has a Hardware DMA Module 
 *                  which can calculate the checksum faster than software, so 
 *                  this Function replaces the CaclIPBufferChecksum() Function 
 *                  defined in the Helpers.c file.  Through the use of 
 *                  preprocessor defines, this replacement is automatic.
 * Note:            This Function works either in the RX buffer area Or the TX
 *                  buffer area.  No validation is Done on the len parameter.
 *****************************************************************************/
}
Public Function CalcIPBufferChecksum(Len As Word) As Word
   Dim Start As WORD_VAL
   Dim Checksum As DWORD_VAL 
   Dim ChunkLen, Temp As Word
   Dim DataBuffer(20) As Byte   // Must be an even size
   Dim DataPtr As FSR0
   
   Checksum.Val = $00000000
   
   // Save the SPI read pointer starting address
   Start.v(0) = ReadETHReg(ERDPTL)
   Start.v(1) = ReadETHReg(ERDPTH)

   While Len <> 0
      // Obtain a chunk of data (less SPI overhead compared 
      // to requesting one byte at a time)
      If Len > SizeOf(DataBuffer) Then
         ChunkLen = SizeOf(DataBuffer)
      Else
         ChunkLen = Len
      EndIf
         
      MACGetArray(@DataBuffer, ChunkLen)

      Len = Len - ChunkLen

      // Take care of a last odd numbered data byte
      If ChunkLen.0 <> 0 Then
         DataBuffer(ChunkLen) = $00
         Inc(ChunkLen)
      EndIf

      // Calculate the checksum over this chunk
      DataPtr = @DataBuffer
      While ChunkLen <> 0
         Temp.Byte0 = POSTINC0
         Temp.Byte1 = POSTINC0
         Checksum.Val = Checksum.Val + Temp
         Dec(ChunkLen, 2)
      Wend
   Wend
    
   // Restore old read pointer location
   WriteReg(ERDPTL, Start.v(0))
   WriteReg(ERDPTH, Start.v(1))
    
   // Do an end-around carry (one's complement arrithmatic)
   Checksum.Val = Checksum.w(0) + Checksum.w(1)

   // Do another end-around carry in case if the prior add 
   // caused a carry out
   Checksum.w(0) = Checksum.w(0) + Checksum.w(1)

   // Return the resulting checksum
   result = Not Checksum.w(0)
End Function

{
 ******************************************************************************
 * Function:        Word MACCalcRxChecksum(Word offset, Word len)
 * PreCondition:    None
 * Input:           offset  - Number of Bytes beyond the beginning of the 
 *                          Ethernet data (first Byte after the Type field) 
 *                          where the checksum should begin
 *                  len     - Total number of Bytes To Include in the checksum
 * Output:          16-Bit checksum As defined by RFC 793.
 * Side Effects:    None
 * Overview:        This Function performs a checksum calculation in the MAC
 *                  buffer itself
 * Note:            None
 *****************************************************************************/
}
Public Function MACCalcRxChecksum(offset As Word, Len As Word) As Word
   Dim temp As WORD_VAL
   Dim RDSave As WORD_VAL

   // Add the offset requested by firmware plus the Ethernet header
   temp.Val = CurrentPacketLocation.Val + SizeOf(ENC_PREAMBLE) + offset
   If temp.Val > RXSTOP Then        // Adjust value if a wrap is needed
      temp.Val = temp.Val - RXSIZE
   EndIf
   
   RDSave.v(0) = ReadETHReg(ERDPTL)
   RDSave.v(1) = ReadETHReg(ERDPTH)

   WriteReg(ERDPTL, temp.v(0))
   WriteReg(ERDPTH, temp.v(1))
    
   temp.Val = CalcIPBufferChecksum(Len)

   WriteReg(ERDPTL, RDSave.v(0))
   WriteReg(ERDPTH, RDSave.v(1))
   result = temp.Val
End Function

{
 ******************************************************************************
 * Function:        void MACMemCopyAsync(Word destAddr, Word sourceAddr, Word len)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           destAddr:   Destination address in the Ethernet memory To 
 *                              copy To.  If the MSb is set, the current EWRPT 
 *                              value will be used instead.
 *                  sourceAddr: Source address To read from.  If the MSb is 
 *                              set, the current ERDPT value will be used 
 *                              instead.
 *                  len:        Number of Bytes To copy
 * Output:          Byte read from the ENC28J60's RAM
 * Side Effects:    None
 * Overview:        Bytes are asynchrnously transfered within the buffer.  Call 
 *                  MACIsMemCopyDone() To see when the transfer is complete.  
 * Note:            If a prior transfer is already in progress prior To 
 *                  calling this Function, this Function will block Until it 
 *                  can start this transfer.
 *
 *                  If a negative value is used For the sourceAddr Or destAddr 
 *                  parameters, Then that pointer will get updated With the 
 *                  Next address after the read Or write.
 *****************************************************************************/
}
Public Sub MACMemCopyAsync(destAddr As Word, sourceAddr As Word, Len As Word)
   Dim ReadSave, WriteSave As WORD_VAL
   Dim UpdateWritePointer As Boolean
   Dim UpdateReadPointer As Boolean 
   Dim Register As REG

   UpdateWritePointer = false
   UpdateReadPointer = false
 
   If destAddr.15 <> 0 Then
      UpdateWritePointer = true
      destAddr.Byte0 = ReadETHReg(EWRPTL)
      destAddr.Byte1 = ReadETHReg(EWRPTH)
   EndIf   

   If sourceAddr.15 <> 0 Then
      UpdateReadPointer = true
      sourceAddr.Byte0 = ReadETHReg(ERDPTL)
      sourceAddr.Byte1 = ReadETHReg(ERDPTH)
   EndIf
      
   // Handle special conditions where len == 0 or len == 1
   // The DMA module is not capable of handling those corner cases
   If Len <= 1 Then
      If Not UpdateReadPointer Then
         ReadSave.v(0) = ReadETHReg(ERDPTL)
         ReadSave.v(1) = ReadETHReg(ERDPTH)
      EndIf
      
      If Not UpdateWritePointer Then
         WriteSave.v(0) = ReadETHReg(EWRPTL)
         WriteSave.v(1) = ReadETHReg(EWRPTH)
      EndIf

      WriteReg(ERDPTL, sourceAddr.Byte0)
      WriteReg(ERDPTH, sourceAddr.Byte1)
      WriteReg(EWRPTL, destAddr.Byte0)
      WriteReg(EWRPTH, destAddr.Byte1)
      While Len <> 0
         MACPut(MACGet())
         Dec(Len)
      Wend   
      If Not UpdateReadPointer Then
         WriteReg(ERDPTL, ReadSave.v(0))
         WriteReg(ERDPTH, ReadSave.v(1))
      EndIf   
      
      If Not UpdateWritePointer Then
         WriteReg(EWRPTL, WriteSave.v(0))
         WriteReg(EWRPTH, WriteSave.v(1))
      EndIf
   Else
      If UpdateWritePointer Then
         WriteSave.Val = destAddr + Len
         WriteReg(EWRPTL, WriteSave.v(0))
         WriteReg(EWRPTH, WriteSave.v(1))
      EndIf
      Len = Len + (sourceAddr - 1)
      Register = ReadETHReg(ECON1)
      While Register.ECON1bits.DMAST
         Register = ReadETHReg(ECON1)
      Wend
      
      WriteReg(EDMASTL, sourceAddr.Byte0)
      WriteReg(EDMASTH, sourceAddr.Byte1)
      WriteReg(EDMADSTL, destAddr.Byte0)
      WriteReg(EDMADSTH, destAddr.Byte1)
    
      If (sourceAddr <= RXSTOP) And (Len > RXSTOP) Then //&& (sourceAddr >= RXSTART))
         Len = Len - RXSIZE
      EndIf
         
      WriteReg(EDMANDL, Len.Byte0)
      WriteReg(EDMANDH, Len.Byte1)
      
      BFCReg(ECON1, ECON1_CSUMEN)
      BFSReg(ECON1, ECON1_DMAST)
      If UpdateReadPointer Then
         Inc(Len)
         If (sourceAddr <= RXSTOP) And (Len > RXSTOP) Then //&& (sourceAddr >= RXSTART))
            Len = Len - RXSIZE
         EndIf   
         WriteReg(ERDPTL, Len.Byte0)
         WriteReg(ERDPTH, Len.Byte1)
      EndIf
   EndIf
End Sub

Public Function MACIsMemCopyDone() As Boolean
   Dim Register As REG
   Register = ReadETHReg(ECON1)
   result = Not Register.ECON1bits.DMAST
End Function

{
 ******************************************************************************
 * Function:        void MACPutROMArray(ROM Byte *Val, Word len)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 *                  EWRPT must point To the location To begin writing.
 * Input:           *Val: Pointer To source of Bytes To copy.
 *                  len:  Number of Bytes To write To the data buffer.
 * Output:          None
 * Side Effects:    None
 * Overview:        MACPutArray writes several sequential Bytes To the 
 *                  ENC28J60 RAM.  It performs faster than multiple MACPut()
 *                  calls.  EWRPT is incremented by len.
 * Note:            None
 *****************************************************************************/
}
Public Sub MACPutROMArray(src As TROMAddress, Len As Word)
    enc.assertCS()
    enc.transfer(WBM)     // Send the Write Buffer Memory opcode
    enc.burst_write_from_rom(src, len)
    enc.deassertCS()    // Terminate the burst operation
End Sub

{
 ******************************************************************************
 * Function:        void MACPowerDown(void)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        MACPowerDown puts the ENC28J60 in Low power sleep mode. In
 *                  sleep mode, no packets can be transmitted Or received.  
 *                  All MAC And PHY registers should Not be accessed.
 * Note:            If a packet is being transmitted While this Function is 
 *                  called, this Function will block Until it is it complete.
 *                  If anything is being received, it will be completed.
 *****************************************************************************/
}
Sub MACPowerDown()
   Dim Register As REG
   
   // Disable packet reception
   BFCReg(ECON1, ECON1_RXEN)

   // Make sure any last packet which was in-progress when RXEN was cleared 
   // is completed
   Register = ReadETHReg(ESTAT)
   While Register.ESTATbits.RXBUSY
      Register = ReadETHReg(ESTAT)
   Wend

   // If a packet is being transmitted, wait for it to finish
   Register = ReadETHReg(ECON1)
   While Register.ECON1bits.TXRTS
      Register = ReadETHReg(ECON1)
   Wend 
   
   // Enter sleep mode
   BFSReg(ECON2, ECON2_PWRSV)
End Sub

{
 ******************************************************************************
 * Function:        void MACPowerUp(void)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        MACPowerUp returns the ENC28J60 back To normal operation
 *                  after a previous call To MACPowerDown().  Calling this 
 *                  Function when already powered up will have no effect.
 * Note:            If a link partner is present, it will take 10S of 
 *                  milliseconds before a new link will be established after
 *                  waking up.  While Not linked, packets which are 
 *                  transmitted will most likely be lost.  MACIsLinked() can 
 *                  be called To determine If a link is established.
 *****************************************************************************/
}
Sub MACPowerUp()
   Dim Register As REG
   
   // Leave power down mode
   BFCReg(ECON2, ECON2_PWRSV)

   // Wait for the 300us Oscillator Startup Timer (OST) to time out.  This 
   // delay is required for the PHY module to return to an operational state.
   Register = ReadETHReg(ESTAT)
   While Not Register.ESTATbits.CLKRDY
      Register = ReadETHReg(ESTAT)
   Wend
    
   // Enable packet reception
   BFSReg(ECON1, ECON1_RXEN)
End Sub

{
 ******************************************************************************
 * Function:        void SetCLKOUT(Byte NewConfig)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           NewConfig - 0x00: CLKOUT disabled (pin driven Low)
 *                              0x01: Divide by 1 (25 MHz)
 *                              0x02: Divide by 2 (12.5 MHz)
 *                              0x03: Divide by 3 (8.333333 MHz)
 *                              0x04: Divide by 4 (6.25 MHz, POR default)
 *                              0x05: Divide by 8 (3.125 MHz)
 * Output:          None
 * Side Effects:    None
 * Overview:        Writes the value of NewConfig into the ECOCON register.  
 *                  The CLKOUT pin will beginning outputting the new frequency
 *                  immediately.
 * Note:            
 *****************************************************************************/
}
Sub SetCLKOUT(NewConfig As Byte)
   BankSel(ECOCON)
   WriteReg(Byte(ECOCON), NewConfig)
   BankSel(ERDPTL)
End Sub

{
 ******************************************************************************
 * Function:        Byte GetCLKOUT(void)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           None
 * Output:          Byte - 0x00: CLKOUT disabled (pin driven Low)
 *                         0x01: Divide by 1 (25 MHz)
 *                         0x02: Divide by 2 (12.5 MHz)
 *                         0x03: Divide by 3 (8.333333 MHz)
 *                         0x04: Divide by 4 (6.25 MHz, POR default)
 *                         0x05: Divide by 8 (3.125 MHz)
 *                         0x06: Reserved
 *                         0x07: Reserved
 * Side Effects:    None
 * Overview:        Returns the current value of the ECOCON register.
 * Note:            None
 *****************************************************************************/
}
Function GetCLKOUT() As Byte
   BankSel(ECOCON)
   Result = ReadETHReg(Byte(ECOCON))
   BankSel(ERDPTL)
End Function

{
 ******************************************************************************
 * Function:        void SetRXHashTableEntry(MAC_ADDR DestMACAddr)
 * PreCondition:    SPI bus must be initialized (Done in MACInit()).
 * Input:           DestMACAddr: 6 Byte group destination MAC address To allow 
 *                               through the Hash Table Filter
 * Output:          Sets the appropriate Bit in the EHT* registers To allow 
 *                  packets sent To DestMACAddr To be received If the Hash 
 *                  Table receive filter is enabled
 * Side Effects:    None
 * Overview:        Calculates a CRC-32 using polynomial 0x4C11DB7 And Then, 
 *                  using Bits 28:23 of the CRC, sets the appropriate Bit in 
 *                  the EHT* registers
 * Note:            This code is commented out To Save code space on systems 
 *                  that do Not need this Function.  Change the "#if 0" line 
 *                  To "#if 1" To uncomment it.
 *****************************************************************************/
'void SetRXHashTableEntry(MAC_ADDR DestMACAddr)
'{
'   DWORD_VAL CRC = {0xFFFFFFFF};
'   Byte HTRegister;
'   Byte i, j;

'   // Calculate a CRC-32 over the 6 byte MAC address 
'   // using polynomial 0x4C11DB7
'   For(i = 0; i < SizeOf(MAC_ADDR); i++)
'   {
'       BYTE  crcnext;
'   
'       // shift in 8 bits
'       for(j = 0; j < 8; j++)
'       {
'           crcnext = 0;
'           if(((BYTE_VAL* )&(CRC.v[3]))->bits.b7)
'               crcnext = 1;
'           crcnext ^= (((BYTE_VAL* )&DestMACAddr.v[i])->bits.b0);
'   
'           CRC.Val <<= 1;
'           if(crcnext)
'               CRC.Val ^= 0x4C11DB7;
'           // next bit
'           DestMACAddr.v[i] >>= 1;
'       }
'   }
'   
'   // CRC-32 calculated, now extract bits 28:23
'   // Bits 25:23 define where within the Hash Table byte the bit needs to be set
'   // Bits 28:26 define which of the 8 Hash Table bytes that bits 25:23 apply to
'   i = CRC.v[3] & 0x1F;
'   HTRegister = (i >> 2) + (Byte)EHT0;
'   i = (i << 1) & 0x06;
'   ((BYTE_VAL* )&i)->Bits.b0 = ((BYTE_VAL* )&CRC.v[2])->Bits.b7;
'   
'   // Set the proper bit in the Hash Table
'   BankSel(EHT0);
'   BFSReg(HTRegister, 1<<i);
'
'   BankSel(ERDPTL);            // Return to Bank 0
'}

end module
