//
//---------------------------------------------------------------------------
//  Name    : ENCx24J600.BAS
//  Date    : 6/1/2013
//  Version : 2.3.6 - Based around Microchip TCP/IP stack version 5.41
//  Notes   : SPI mode interface only
//
// ver 1.0 (v5.10)
//
//---------------------------------------------------------------------------
// Copyright 2002-2007 Microchip Technology Inc.  All rights reserved.
//---------------------------------------------------------------------------
//
module ENCx24J600

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"

// ENC spi interface
include "encspi.bas"
   
// eth init eEthOpenFlags (speed, duplex, loopback, and pause settings)
include "ethphy.bas"

// select between latching (true) or real-time link status (false) bit
// for the MACIsLinked() function (the 28J60 bit is latching)
#define ENC_USE_LATCHING_LINKSTAT = false

public type ram_addr_t = word
public type rom_addr_t = word

// Binary constant identifiers for ReadMemoryWindow() and WriteMemoryWindow() functions
public const
    UDA_WINDOW_BIT  = 0,
    GP_WINDOW_BIT   = 1,
    RX_WINDOW_BIT   = 2,
    UDA_WINDOW      = (1<<UDA_WINDOW_BIT),      // $01
    GP_WINDOW       = (1<<GP_WINDOW_BIT),       // $02
    RX_WINDOW       = (1<<RX_WINDOW_BIT)        // $04

const
    ENC100_RAM_SIZE     = (24*1024)

//--------------------------------------------------
// ENC424J600/624J600 SPI Opcodes
//--------------------------------------------------
const
    RCR         = $00<<5,   // Read Control Register
    WCR         = $02<<5,   // Write Control Register
    RCRU        = $20,      // Read Control Register Unbanked
    WCRU        = $22,      // Write Control Register Unbanked
    BFS         = $04<<5,   // Bit Field Set
    BFSU        = $24,      // Bit Field Set Unbanked
    BFC         = $05<<5,   // Bit Field Clear
    BFCU        = $26,      // Bit Field Clear Unbanked
    RBMGP       = $28,      // Read Buffer Memory General Purpose
    WBMGP       = $2A,      // Write Buffer Memory General Purpose
    RBMRX       = $2C,      // Read Buffer Memory RX
    WBMRX       = $2E,      // Write Buffer Memory RX
    RBMUDA      = $30,      // Read Buffer Memory User Defined Area
    WBMUDA      = $32,      // Write Buffer Memory User Defined Area
    WGPRDPT     = $60,      // Write General Purpose Read Pointer
    RGPRDPT     = $62,      // Read General Purpose Read Pointer
    WRXRDPT     = $64,      // Write RX Read Pointer
    RRXRDPT     = $66,      // Read RX Read Pointer
    WUDARDPT    = $68,      // Write User Defined Area Read Pointer
    RUDARDPT    = $6A,      // Read User Defined Area Read Pointer
    WGPWRPT     = $6C,      // Write General Purpose Write Pointer
    RGPWRPT     = $6E,      // Read General Purpose Write Pointer
    WRXWRPT     = $70,      // Write RX Write Pointer
    RRXWRPT     = $72,      // Read RX Write Pointer
    WUDAWRPT    = $74,      // Write User Defined Area Write Pointer
    RUDAWRPT    = $76,      // Read User Defined Area Write Pointer
    B0SEL       = $C0,      // Bank 0 Select
    B1SEL       = $C2,      // Bank 1 Select
    B2SEL       = $C4,      // Bank 2 Select
    B3SEL       = $C6,      // Bank 3 Select
    RBSEL       = $C8,      // Read Bank Select
    SETETHRST   = $CA,      // Set ETHRST bit (perform system reset)
    FCDIS       = $E0,      // Flow Control Disable
    FCSINGLE    = $E2,      // Flow Control Single
    FCMULTIPLE  = $E4,      // Flow Control Multiple
    FCCLEAR     = $E6,      // Flow Control Clear
    SETPKTDEC   = $CC,      // Set PKTDEC bit (decrement RX packet pending counter)
    DMASTOP     = $D0,      // DMA Stop
    DMACKSUM    = $D8,      // DMA Start Checksum
    DMACKSUMS   = $DA,      // DMA Start Checksum with Seed
    DMACOPY     = $DC,      // DMA Start Copy
    DMACOPYS    = $DE,      // DMA Start Copy and Checksum with Seed
    SETTXRTS    = $D4,      // Set TXRTS bit (transmit a packet)
    ENABLERX    = $E8,      // Enable RX
    DISABLERX   = $EA,      // Disable RX
    SETEIE      = $EC,      // Set Ethernet Interrupt Enable (EIE)
    CLREIE      = $EE       // Clear Ethernet Interrupt Enable (EIE)

public const
    SET_OFFSET  = $0100,
    CLR_OFFSET  = $0180

// ENC424J600 register addresses
// note: these are actually word addresses with A8=0 ($7Exx)

// SPI Bank 0 registers --------
public const
    ETXST       = $00,
    ETXSTL      = $00,
    ETXSTH      = $01,
    ETXLEN      = $02,
    ETXLENL     = $02,
    ETXLENH     = $03,
    ERXST       = $04,
    ERXSTL      = $04,
    ERXSTH      = $05,
    ERXTAIL     = $06,
    ERXTAILL    = $06,
    ERXTAILH    = $07,
    ERXHEAD     = $08,
    ERXHEADL    = $08,
    ERXHEADH    = $09,
    EDMAST      = $0A,
    EDMASTL     = $0A,
    EDMASTH     = $0B,
    EDMALEN     = $0C,
    EDMALENL    = $0C,
    EDMALENH    = $0D,
    EDMADST     = $0E,
    EDMADSTL    = $0E,
    EDMADSTH    = $0F,
    EDMACS      = $10,
    EDMACSL     = $10,
    EDMACSH     = $11,
    ETXSTAT     = $12,
    ETXSTATL    = $12,
    ETXSTATH    = $13,
    ETXWIRE     = $14,
    ETXWIREL    = $14,
    ETXWIREH    = $15

// SPI all bank registers
public const
    EUDAST      = $16,
    EUDASTL     = $16,
    EUDASTH     = $17,
    EUDAND      = $18,
    EUDANDL     = $18,
    EUDANDH     = $19,
    ESTAT       = $1A,
    ESTATL      = $1A,
    ESTATH      = $1B,
    EIR         = $1C,
    EIRL        = $1C,
    EIRH        = $1D,
    ECON1       = $1E,
    ECON1L      = $1E,
    ECON1H      = $1F

// SPI Bank 1 registers -----
public const
    EHT1        = $20,
    EHT1L       = $20,
    EHT1H       = $21,
    EHT2        = $22,
    EHT2L       = $22,
    EHT2H       = $23,
    EHT3        = $24,
    EHT3L       = $24,
    EHT3H       = $25,
    EHT4        = $26,
    EHT4L       = $26,
    EHT4H       = $27,
    EPMM1       = $28,
    EPMM1L      = $28,
    EPMM1H      = $29,
    EPMM2       = $2A,
    EPMM2L      = $2A,
    EPMM2H      = $2B,
    EPMM3       = $2C,
    EPMM3L      = $2C,
    EPMM3H      = $2D,
    EPMM4       = $2E,
    EPMM4L      = $2E,
    EPMM4H      = $2F,
    EPMCS       = $30,
    EPMCSL      = $30,
    EPMCSH      = $31,
    EPMO        = $32,
    EPMOL       = $32,
    EPMOH       = $33,
    ERXFCON     = $34,
    ERXFCONL    = $34,
    ERXFCONH    = $35

// SPI all bank registers from 0x36 to 0x3F

// SPI Bank 2 registers -----
public const
    MACON1      = $40,
    MACON1L     = $40,
    MACON1H     = $41,
    MACON2      = $42,
    MACON2L     = $42,
    MACON2H     = $43,
    MABBIPG     = $44,
    MABBIPGL    = $44,
    MABBIPGH    = $45,
    MAIPG       = $46,
    MAIPGL      = $46,
    MAIPGH      = $47,
    MACLCON     = $48,
    MACLCONL    = $48,
    MACLCONH    = $49,
    MAMXFL      = $4A,
    MAMXFLL     = $4A,
    MAMXFLH     = $4B,
    MICMD       = $52,
    MICMDL      = $52,
    MICMDH      = $53,
    MIREGADR    = $54,
    MIREGADRL   = $54,
    MIREGADRH   = $55

// SPI all bank registers from 0x56 to 0x5F

// SPI Bank 3 registers -----
public const
    MAADR3      = $60,
    MAADR3L     = $60,
    MAADR3H     = $61,
    MAADR2      = $62,
    MAADR2L     = $62,
    MAADR2H     = $63,
    MAADR1      = $64,
    MAADR1L     = $64,
    MAADR1H     = $65,
    MIWR        = $66,
    MIWRL       = $66,
    MIWRH       = $67,
    MIRD        = $68,
    MIRDL       = $68,
    MIRDH       = $69,
    MISTAT      = $6A,
    MISTATL     = $6A,
    MISTATH     = $6B,
    EPAUS       = $6C,
    EPAUSL      = $6C,
    EPAUSH      = $6D,
    ECON2       = $6E,
    ECON2L      = $6E,
    ECON2H      = $6F,
    ERXWM       = $70,
    ERXWML      = $70,
    ERXWMH      = $71,
    EIE         = $72,
    EIEL        = $72,
    EIEH        = $73,
    EIDLED      = $74,
    EIDLEDL     = $74,
    EIDLEDH     = $75

// SPI all bank registers from 0x66 to 0x6F

// SPI Non-banked Special Function Registers
public const
    EGPDATA     = $80,
    EGPDATAL    = $80,
    // r        = $81
    ERXDATA     = $82,
    ERXDATAL    = $82,
    // r        = $83
    EUDADATA    = $84,
    EUDADATAL   = $84,
    // r        = $85
    EGPRDPT     = $86,
    EGPRDPTL    = $86,
    EGPRDPTH    = $87,
    EGPWRPT     = $88,
    EGPWRPTL    = $88,
    EGPWRPTH    = $89,
    ERXRDPT     = $8A,
    ERXRDPTL    = $8A,
    ERXRDPTH    = $8B,
    ERXWRPT     = $8C,
    ERXWRPTL    = $8C,
    ERXWRPTH    = $8D,
    EUDARDPT    = $8E,
    EUDARDPTL   = $8E,
    EUDARDPTH   = $8F,
    EUDAWRPT    = $90,
    EUDAWRPTL   = $90,
    EUDAWRPTH   = $91

//---------------------------------------------------------------------------
// ENC424J600/624J600 PHY Register Addresses
//---------------------------------------------------------------------------
public const
    PHCON1  = $00,
    PHSTAT1 = $01,
    PHANA   = $04,
    PHANLPA = $05,
    PHANE   = $06,
    PHCON2  = $11,
    PHSTAT2 = $1B,
    PHSTAT3 = $1F

//---------------------------------------------------------------------------
// ENC424J600/624J600 register bits               //
//---------------------------------------------------------------------------
// ESTAT bits ----------
public const
    ESTAT_INT       = (1<<15),
    ESTAT_FCIDLE    = (1<<14),
    ESTAT_RXBUSY    = (1<<13),
    ESTAT_CLKRDY    = (1<<12),
    ESTAT_RSTDONE   = (1<<11),      // undocumented in datasheet (reserved:ignore on read)
    ESTAT_PHYDPX    = (1<<10),
    ESTAT_PHYRDY    = (1<<9),       // undocumented in datasheet (reserved:ignore on read)
    ESTAT_PHYLNK    = (1<<8),
    ESTAT_PKTCNT7   = (1<<7),
    ESTAT_PKTCNT6   = (1<<6),
    ESTAT_PKTCNT5   = (1<<5),
    ESTAT_PKTCNT4   = (1<<4),
    ESTAT_PKTCNT3   = (1<<3),
    ESTAT_PKTCNT2   = (1<<2),
    ESTAT_PKTCNT1   = (1<<1),
    ESTAT_PKTCNT0   = (1),

    ESTAT_RXBUSY_BIT = 13,
    ESTAT_CLKRDY_BIT = 12,
    ESTAT_PHYDPX_BIT = 10,
    ESTAT_PHYLNK_BIT = 8

// EIR bits ------------
public const
    EIR_CRYPTEN     = (1<<15),
    EIR_MODEXIF     = (1<<14),
    EIR_HASHIF      = (1<<13),
    EIR_AESIF       = (1<<12),
    EIR_LINKIF      = (1<<11),
    EIR_PRDYIF      = (1<<10),
    EIR_r9          = (1<<9),
    EIR_r8          = (1<<8),
    EIR_r7          = (1<<7),
    EIR_PKTIF       = (1<<6),
    EIR_DMAIF       = (1<<5),
    EIR_r4          = (1<<4),
    EIR_TXIF        = (1<<3),
    EIR_TXABTIF     = (1<<2),
    EIR_RXABTIF     = (1<<1),
    EIR_PCFULIF     = (1),

    EIR_CRYPTEN_BIT = 15,
    EIR_LINKIF_BIT  = 11,
    EIR_PKTIF_BIT   = 6,
    EIR_TXIF_BIT    = 3,
    EIR_TXABTIF_BIT = 2,
    EIR_RXABTIF_BIT = 1,
    EIR_PCFULIF_BIT = 0

// ECON1 bits ----------
public const
    ECON1_MODEXST   = (1<<15),
    ECON1_HASHEN    = (1<<14),
    ECON1_HASHOP    = (1<<13),
    ECON1_HASHLST   = (1<<12),
    ECON1_AESST     = (1<<11),
    ECON1_AESOP1    = (1<<10),
    ECON1_AESOP0    = (1<<9),
    ECON1_PKTDEC    = (1<<8),
    ECON1_FCOP1     = (1<<7),
    ECON1_FCOP0     = (1<<6),
    ECON1_DMAST     = (1<<5),
    ECON1_DMACPY    = (1<<4),
    ECON1_DMACSSD   = (1<<3),
    ECON1_DMANOCS   = (1<<2),
    ECON1_TXRTS     = (1<<1),
    ECON1_RXEN      = (1),

    ECON1_DMAST_BIT = 5,
    ECON1_TXRTS_BIT = 1

// ETXSTAT bits --------
public const
    ETXSTAT_r12     = (1<<12),
    ETXSTAT_r11     = (1<<11),
    ETXSTAT_LATECOL = (1<<10),
    ETXSTAT_MAXCOL  = (1<<9),
    ETXSTAT_EXDEFER = (1<<8),
    ETXSTAT_DEFER   = (1<<7),
    ETXSTAT_r6      = (1<<6),
    ETXSTAT_r5      = (1<<5),
    ETXSTAT_CRCBAD  = (1<<4),
    ETXSTAT_COLCNT3 = (1<<3),
    ETXSTAT_COLCNT2 = (1<<2),
    ETXSTAT_COLCNT1 = (1<<1),
    ETXSTAT_COLCNT0 = (1)

// ERXFCON bits --------
public const
    ERXFCON_HTEN    = (1<<15),
    ERXFCON_MPEN    = (1<<14),
    ERXFCON_NOTPM   = (1<<12),
    ERXFCON_PMEN3   = (1<<11),
    ERXFCON_PMEN2   = (1<<10),
    ERXFCON_PMEN1   = (1<<9),
    ERXFCON_PMEN0   = (1<<8),
    ERXFCON_CRCEEN  = (1<<7),
    ERXFCON_CRCEN   = (1<<6),
    ERXFCON_RUNTEEN = (1<<5),
    ERXFCON_RUNTEN  = (1<<4),
    ERXFCON_UCEN    = (1<<3),
    ERXFCON_NOTMEEN = (1<<2),
    ERXFCON_MCEN    = (1<<1),
    ERXFCON_BCEN    = (1)

// MACON1 bits ---------
public const
    MACON1_r15      = (1<<15),
    MACON1_r14      = (1<<14),
    MACON1_r11      = (1<<11),
    MACON1_r10      = (1<<10),
    MACON1_r9       = (1<<9),
    MACON1_r8       = (1<<8),
    MACON1_LOOPBK   = (1<<4),
    MACON1_r3       = (1<<3),
    MACON1_RXPAUS   = (1<<2),
    MACON1_PASSALL  = (1<<1),
    MACON1_r0       = (1)

// MACON2 bits ---------
public const
    MACON2_DEFER    = (1<<14),
    MACON2_BPEN     = (1<<13),
    MACON2_NOBKOFF  = (1<<12),
    MACON2_r9       = (1<<9),
    MACON2_r8       = (1<<8),
    MACON2_PADCFG2  = (1<<7),
    MACON2_PADCFG1  = (1<<6),
    MACON2_PADCFG0  = (1<<5),
    MACON2_TXCRCEN  = (1<<4),
    MACON2_PHDREN   = (1<<3),
    MACON2_HFRMEN   = (1<<2),
    MACON2_r1       = (1<<1),
    MACON2_FULDPX   = (1),

    MACON2_FULDPX_BIT = 0

// MABBIPG bits --------
public const
    MABBIPG_BBIPG6  = (1<<6),
    MABBIPG_BBIPG5  = (1<<5),
    MABBIPG_BBIPG4  = (1<<4),
    MABBIPG_BBIPG3  = (1<<3),
    MABBIPG_BBIPG2  = (1<<2),
    MABBIPG_BBIPG1  = (1<<1),
    MABBIPG_BBIPG0  = (1)

// MAIPG bits ----------
public const
    MAIPG_r14       = (1<<14),
    MAIPG_r13       = (1<<13),
    MAIPG_r12       = (1<<12),
    MAIPG_r11       = (1<<11),
    MAIPG_r10       = (1<<10),
    MAIPG_r9        = (1<<9),
    MAIPG_r8        = (1<<8),
    MAIPG_IPG6      = (1<<6),
    MAIPG_IPG5      = (1<<5),
    MAIPG_IPG4      = (1<<4),
    MAIPG_IPG3      = (1<<3),
    MAIPG_IPG2      = (1<<2),
    MAIPG_IPG1      = (1<<1),
    MAIPG_IPG0      = (1)

// MACLCON bits --------
public const
    MACLCON_r13     = (1<<13),
    MACLCON_r12     = (1<<12),
    MACLCON_r11     = (1<<11),
    MACLCON_r10     = (1<<10),
    MACLCON_r9      = (1<<9),
    MACLCON_r8      = (1<<8),
    MACLCON_MAXRET3 = (1<<3),
    MACLCON_MAXRET2 = (1<<2),
    MACLCON_MAXRET1 = (1<<1),
    MACLCON_MAXRET0 = (1)

// MICMD bits ----------
public const
    MICMD_MIISCAN   = (1<<1),
    MICMD_MIIRD     = (1)

// MIREGADR bits -------
public const
    MIREGADR_r12    = (1<<12),
    MIREGADR_r11    = (1<<11),
    MIREGADR_r10    = (1<<10),
    MIREGADR_r9     = (1<<9),
    MIREGADR_r8     = (1<<8),
    MIREGADR_PHREG4 = (1<<4),
    MIREGADR_PHREG3 = (1<<3),
    MIREGADR_PHREG2 = (1<<2),
    MIREGADR_PHREG1 = (1<<1),
    MIREGADR_PHREG0 = (1)

// MISTAT bits ---------
public const
    MISTAT_r3       = (1<<3),
    MISTAT_NVALID   = (1<<2),
    MISTAT_SCAN     = (1<<1),
    MISTAT_BUSY     = (1),
    MISTAT_BUSY_BIT = 0

// ECON2 bits ----------
public const
    ECON2_ETHEN     = (1<<15),
    ECON2_STRCH     = (1<<14),
    ECON2_TXMAC     = (1<<13),
    ECON2_SHA1MD5   = (1<<12),
    ECON2_COCON3    = (1<<11),
    ECON2_COCON2    = (1<<10),
    ECON2_COCON1    = (1<<9),
    ECON2_COCON0    = (1<<8),
    ECON2_AUTOFC    = (1<<7),
    ECON2_TXRST     = (1<<6),
    ECON2_RXRST     = (1<<5),
    ECON2_ETHRST    = (1<<4),
    ECON2_MODLEN1   = (1<<3),
    ECON2_MODLEN0   = (1<<2),
    ECON2_AESLEN1   = (1<<1),
    ECON2_AESLEN0   = (1)

// ERXWM bits ----------
public const
    ERXWM_RXFWM7    = (1<<15),
    ERXWM_RXFWM6    = (1<<14),
    ERXWM_RXFWM5    = (1<<13),
    ERXWM_RXFWM4    = (1<<12),
    ERXWM_RXFWM3    = (1<<11),
    ERXWM_RXFWM2    = (1<<10),
    ERXWM_RXFWM1    = (1<<9),
    ERXWM_RXFWM0    = (1<<8),
    ERXWM_RXEWM7    = (1<<7),
    ERXWM_RXEWM6    = (1<<6),
    ERXWM_RXEWM5    = (1<<5),
    ERXWM_RXEWM4    = (1<<4),
    ERXWM_RXEWM3    = (1<<3),
    ERXWM_RXEWM2    = (1<<2),
    ERXWM_RXEWM1    = (1<<1),
    ERXWM_RXEWM0    = (1)

// EIE bits ------------
public const
    EIE_INTIE       = (1<<15),
    EIE_MODEXIE     = (1<<14),
    EIE_HASHIE      = (1<<13),
    EIE_AESIE       = (1<<12),
    EIE_LINKIE      = (1<<11),
    EIE_PRDYIE      = (1<<10),
    EIE_r9          = (1<<9),
    EIE_r8          = (1<<8),
    EIE_r7          = (1<<7),
    EIE_PKTIE       = (1<<6),
    EIE_DMAIE       = (1<<5),
    EIE_r4          = (1<<4),
    EIE_TXIE        = (1<<3),
    EIE_TXABTIE     = (1<<2),
    EIE_RXABTIE     = (1<<1),
    EIE_PCFULIE     = (1),

    EIE_INTIE_BIT   = 15


// EIDLED bits ---------
public const
    EIDLED_LACFG3   = (1<<15),
    EIDLED_LACFG2   = (1<<14),
    EIDLED_LACFG1   = (1<<13),
    EIDLED_LACFG0   = (1<<12),
    EIDLED_LBCFG3   = (1<<11),
    EIDLED_LBCFG2   = (1<<10),
    EIDLED_LBCFG1   = (1<<9),
    EIDLED_LBCFG0   = (1<<8),
    EIDLED_DEVID2   = (1<<7),
    EIDLED_DEVID1   = (1<<6),
    EIDLED_DEVID0   = (1<<5),
    EIDLED_REVID4   = (1<<4),
    EIDLED_REVID3   = (1<<3),
    EIDLED_REVID2   = (1<<2),
    EIDLED_REVID1   = (1<<1),
    EIDLED_REVID0   = (1)

// PHCON1 bits ---------
public const
    PHCON1_PRST     = (1<<15),
    PHCON1_PLOOPBK  = (1<<14),
    PHCON1_SPD100   = (1<<13),
    PHCON1_ANEN     = (1<<12),
    PHCON1_PSLEEP   = (1<<11),
    PHCON1_r10      = (1<<10),
    PHCON1_RENEG    = (1<<9),
    PHCON1_PFULDPX  = (1<<8),
    PHCON1_r7       = (1<<7),
    PHCON1_r6       = (1<<6),
    PHCON1_r5       = (1<<5),
    PHCON1_r4       = (1<<4),
    PHCON1_r3       = (1<<3),
    PHCON1_r2       = (1<<2),
    PHCON1_r1       = (1<<1),
    PHCON1_r0       = (1),

    PHCON1_PRST_BIT     = 15,
    PHCON1_SPD100_BIT   = 13,
    PHCON1_ANEN_BIT     = 12,
    PHCON1_PSLEEP_BIT   = 11,
    PHCON1_PFULDPX_BIT  = 8

// PHSTAT1 bits --------
public const
    PHSTAT1_r15     = (1<<15),
    PHSTAT1_FULL100 = (1<<14),
    PHSTAT1_HALF100 = (1<<13),
    PHSTAT1_FULL10  = (1<<12),
    PHSTAT1_HALF10  = (1<<11),
    PHSTAT1_r10     = (1<<10),
    PHSTAT1_r9      = (1<<9),
    PHSTAT1_r8      = (1<<8),
    PHSTAT1_r7      = (1<<7),
    PHSTAT1_r6      = (1<<6),
    PHSTAT1_ANDONE  = (1<<5),
    PHSTAT1_LRFAULT = (1<<4),
    PHSTAT1_ANABLE  = (1<<3),
    PHSTAT1_LLSTAT  = (1<<2),
    PHSTAT1_r1      = (1<<1),
    PHSTAT1_EXTREGS = (1),

    PHSTAT1_ANDONE_BIT = 5,
    PHSTAT1_LLSTAT_BIT = 2

// PHANA bits ----------
public const
    PHANA_ADNP      = (1<<15),
    PHANA_r14       = (1<<14),
    PHANA_ADFAULT   = (1<<13),
    PHANA_r12       = (1<<12),
    PHANA_ADPAUS1   = (1<<11),
    PHANA_ADPAUS0   = (1<<10),
    PHANA_r9        = (1<<9),
    PHANA_AD100FD   = (1<<8),
    PHANA_AD100     = (1<<7),
    PHANA_AD10FD    = (1<<6),
    PHANA_AD10      = (1<<5),
    PHANA_ADIEEE4   = (1<<4),
    PHANA_ADIEEE3   = (1<<3),
    PHANA_ADIEEE2   = (1<<2),
    PHANA_ADIEEE1   = (1<<1),
    PHANA_ADIEEE0   = (1),

    PHANA_AD100FD_BIT   = 8,
    PHANA_AD100_BIT     = 7,
    PHANA_AD10FD_BIT    = 6,
    PHANA_AD10_BIT      = 5


// PHANLPA bits --------
public const
    PHANLPA_LPNP    = (1<<15),
    PHANLPA_LPACK   = (1<<14),
    PHANLPA_LPFAULT = (1<<13),
    PHANLPA_r12     = (1<<12),
    PHANLPA_LPPAUS1 = (1<<11),
    PHANLPA_LPPAUS0 = (1<<10),
    PHANLPA_LP100T4 = (1<<9),
    PHANLPA_LP100FD = (1<<8),
    PHANLPA_LP100   = (1<<7),
    PHANLPA_LP10FD  = (1<<6),
    PHANLPA_LP10    = (1<<5),
    PHANLPA_LPIEEE4 = (1<<4),
    PHANLPA_LPIEEE3 = (1<<3),
    PHANLPA_LPIEEE2 = (1<<2),
    PHANLPA_LPIEEE1 = (1<<1),
    PHANLPA_LPIEEE0 = (1)

// PHANE bits ----------
public const
    PHANE_r15       = (1<<15),
    PHANE_r14       = (1<<14),
    PHANE_r13       = (1<<13),
    PHANE_r12       = (1<<12),
    PHANE_r11       = (1<<11),
    PHANE_r10       = (1<<10),
    PHANE_r9        = (1<<9),
    PHANE_r8        = (1<<8),
    PHANE_r7        = (1<<7),
    PHANE_r6        = (1<<6),
    PHANE_r5        = (1<<5),
    PHANE_PDFLT     = (1<<4),
    PHANE_r3        = (1<<3),
    PHANE_r2        = (1<<2),
    PHANE_LPARCD    = (1<<1),
    PHANA_LPANABL   = (1)

// PHCON2 bits ---------
public const
    PHCON2_r15      = (1<<15),
    PHCON2_r14      = (1<<14),
    PHCON2_EDPWRDN  = (1<<13),
    PHCON2_r12      = (1<<12),
    PHCON2_EDTHRES  = (1<<11),
    PHCON2_r10      = (1<<10),
    PHCON2_r9       = (1<<9),
    PHCON2_r8       = (1<<8),
    PHCON2_r7       = (1<<7),
    PHCON2_r6       = (1<<6),
    PHCON2_r5       = (1<<5),
    PHCON2_r4       = (1<<4),
    PHCON2_r3       = (1<<3),
    PHCON2_FRCLNK   = (1<<2),
    PHCON2_EDSTAT   = (1<<1),
    PHCON2_r0       = (1)

// PHSTAT2 bits ---------
public const
    PHSTAT2_r15     = (1<<15),
    PHSTAT2_r14     = (1<<14),
    PHSTAT2_r13     = (1<<13),
    PHSTAT2_r12     = (1<<12),
    PHSTAT2_r11     = (1<<11),
    PHSTAT2_r10     = (1<<10),
    PHSTAT2_r9      = (1<<9),
    PHSTAT2_r8      = (1<<8),
    PHSTAT2_r7      = (1<<7),
    PHSTAT2_r6      = (1<<6),
    PHSTAT2_r5      = (1<<5),
    PHSTAT2_PLRITY  = (1<<4),
    PHSTAT2_r3      = (1<<3),
    PHSTAT2_r2      = (1<<2),
    PHSTAT2_r1      = (1<<1),
    PHSTAT2_r0      = (1)

// PHSTAT3 bits --------
public const
    PHSTAT3_r15     = (1<<15),
    PHSTAT3_r14     = (1<<14),
    PHSTAT3_r13     = (1<<13),
    PHSTAT3_r12     = (1<<12),
    PHSTAT3_r11     = (1<<11),
    PHSTAT3_r10     = (1<<10),
    PHSTAT3_r9      = (1<<9),
    PHSTAT3_r8      = (1<<8),
    PHSTAT3_r7      = (1<<7),
    PHSTAT3_r6      = (1<<6),
    PHSTAT3_r5      = (1<<5),
    PHSTAT3_SPDDPX2 = (1<<4),
    PHSTAT3_SPDDPX1 = (1<<3),
    PHSTAT3_SPDDPX0 = (1<<2),
    PHSTAT3_r1      = (1<<1),
    PHSTAT3_r0      = (1)

// ethertype field (16-bit value, ie $0800 or $0806)
const
    ETHER_TYPE  as byte = $08,      // high-byte
    ETHER_IP    as byte = $00,      // low-bytes
    ETHER_ARP   as byte = $06

public const
    MAC_IP      = ETHER_IP,
    MAC_ARP     = ETHER_ARP,
    MAC_UNKNOWN = ($FF)

// bring defines into program...
public const
    MAC_TX_BUFFER_SIZE  = _MAC_TX_BUFFER_SIZE,
    RAMSIZE             = _RAMSIZE,
    TXSTART             = _TXSTART,
    RXSTART             = _RXSTART,
    RXSTOP              = _RXSTOP,
    RXSIZE              = _RXSIZE,
    BASE_TX_ADDR        = _BASE_TX_ADDR,
    BASE_TCB_ADDR       = _BASE_TCB_ADDR,
    BASE_HTTPB_ADDR     = _BASE_HTTPB_ADDR

// RXSTATUS bits...
structure RXBits
    v(6) as byte

    ByteCount               as v(0).asWord          // 16

    PreviouslyIgnored       as v(2).0               // 1
    RXDCPreviouslySeen      as v(2).1               // 1
    CarrierPreviouslySeen   as v(2).2               // 1
    CodeViolation           as v(2).3               // 1
    CRCError                as v(2).4               // 1
    LengthCheckError        as v(2).5               // 1
    LengthOutOfRange        as v(2).6               // 1
    ReceiveOk               as v(2).booleans(7)     // 1

    Multicast               as v(3).0               // 1
    Broadcast               as v(3).1               // 1
    DribbleNibble           as v(3).2               // 1
    ControlFrame            as v(3).3               // 1
    PauseControlFrame       as v(3).4               // 1
    UnsupportedOpcode       as v(3).5               // 1
    VLANType                as v(3).6               // 1
    RuntMatch               as v(3).7               // 1

    filler                  as v(4).0               // 1
    HashMatch               as v(4).1               // 1
    MagicPacketMatch        as v(4).2               // 1
    PatternMatch            as v(4).3               // 1
    UnicastMatch            as v(4).4               // 1
    BroadcastMatch          as v(4).5               // 1
    MulticastMatch          as v(4).6               // 1
    ZeroH                   as v(4).7               // 1

    Zero                    as v(5)                 // 8
end structure

// Receive Status Vector bit fields
structure RXSTATUS
    v(6) as byte
    Bits as RXBits union
end structure

// A generic structure representing the Ethernet header starting all Ethernet frames
public structure MAC_ETHER_HEADER
    DestMACAddr      as MAC_ADDR
    SourceMACAddr    as MAC_ADDR
    htype            as WORD_VAL
end structure

// A header appended at the start of all RX frames by the hardware
public structure ENC100_PREAMBLE
    NextPacketPointer   as word
    StatusVector        as RXSTATUS
    DestMACAddr         as MAC_ADDR
    SourceMACAddr       as MAC_ADDR
    htype               as WORD_VAL
end structure

// Internal MAC level variables and flags
dim
    vCurrentBank            as byte,
    wCurrentPacketPointer   as word,
    wNextPacketPointer      as word

public structure ENC100Flags_t
    b as byte
    bWasDiscarded   as b.booleans(0)      // Status tracking bit indicating if the last received packet has been discarded via MACDiscardRx() or not.
    PoweredDown     as b.booleans(1)      // Local cached bit indicating CRYPTEN == ETHEN == 0 and PSLEEP == 1
    CryptoEnabled   as b.booleans(2)      // Local cached bit indicating CRYPTEN == 1
end structure

public dim
    ENC100Flags as ENC100Flags_t,
    ENCRevID    as byte

// pic18 asm instructions
inline sub reset()
    asm
        TRAP
        RESET
    end asm
end sub

inline sub nop()
    asm
        nop
    end asm
end sub

//
//------------------------------------------------------------------------
// SPI register interface functions
// PreCondition:    SPI bus must be initialized (done in MACInit()).
//------------------------------------------------------------------------
//
public function SelectBank(bAddr as byte) as boolean
    dim vBank as byte
    
    // See if we need to change register banks
    vBank = byte(bAddr) and $E0         // mask off upper three bits
    if (vBank.bits(7) = 0) then         // bit 7 = 0 (%000-%011)
        if (vBank <> vCurrentBank) then
            if (vBank = ($00<<5)) then
                bAddr = B0SEL
            elseif (vBank = ($01<<5)) then
                bAddr = B1SEL
            elseif (vBank = ($02<<5)) then
                bAddr = B2SEL
            elseif (vBank = ($03<<5)) then
                bAddr = B3SEL
            endif                    
            enc.assertCS()
            enc.transfer(bAddr)    // Send the command/opcode
            enc.deassertCS()
            vCurrentBank = vBank
        endif
        result = true
    else            // bank > 4 (%1xx)
        result = false
    endif
end function

//
//------------------------------------------------------------------------
// WORD ReadReg(byte bAddr)
//------------------------------------------------------------------------
//
public function ReadReg(bAddr as byte) as word
    if (SelectBank(bAddr)) then
        bAddr = bAddr and $1F       // combine opcode with reg addr
        bAddr = bAddr or RCR
        enc.assertCS()
        enc.transfer(bAddr)         // send the opcode/addr
        result.byte0 = enc.transfer()
        result.byte1 = enc.transfer()
        enc.deassertCS()
    else
        enc.assertCS()
        enc.transfer(RCRU)          // Send the command/opcode
        enc.transfer(bAddr)
        result.bytes(0) = enc.transfer()
        result.bytes(1) = enc.transfer()
        enc.deassertCS()
    endif
end function

//
//------------------------------------------------------------------------
// WriteReg(byte bAddr, WORD wValue)
//------------------------------------------------------------------------
//
public sub WriteReg(bAddr as byte, wValue as word)
    if (SelectBank(bAddr)) then
        bAddr = bAddr and $1F       // combine opcode with reg addr
        bAddr = bAddr or WCR
        enc.assertCS()
        enc.transfer(bAddr)         // send the opcode/addr
        enc.transfer(wValue.bytes(0))
        enc.transfer(wValue.bytes(1))
        enc.deassertCS()
    else
        enc.assertCS()
        enc.transfer(WCRU)          // Send the command/opcode
        enc.transfer(bAddr)
        enc.transfer(wValue.bytes(0))
        enc.transfer(wValue.bytes(1))
        enc.deassertCS()
    endif
end sub

//
//------------------------------------------------------------------------
// Function:        void BFSReg(WORD wAddress, WORD wBitMask)
//                  void BFCReg(WORD wAddress, WORD wBitMask)
// Overview:        Sets/clears bits in Ethernet SFR registers
// Note:            These functions cannot be used to access ESFR registers
//                  they set addr lines A8/A7 as appropriate
//------------------------------------------------------------------------
//
public sub BFSReg(bAddr as byte, wBitMask as word)
    SelectBank(bAddr)
    bAddr = bAddr and $1F       // combine opcode with reg addr
    bAddr = bAddr or BFS
    enc.assertCS()
    enc.transfer(bAddr)         // send the opcode/addr
    enc.transfer(wBitMask.bytes(0))
    enc.transfer(wBitMask.bytes(1))
    enc.deassertCS()
end sub

public sub BFCReg(bAddr as byte, wBitMask as word)
    SelectBank(bAddr)
    bAddr = bAddr and $1F       // combine opcode with reg addr
    bAddr = bAddr or BFC
    enc.assertCS()
    enc.transfer(bAddr)         // send the opcode/addr
    enc.transfer(wBitMask.bytes(0))
    enc.transfer(wBitMask.bytes(1))
    enc.deassertCS()
end sub

//
//---------------------------------------------------------------------------
// burst read and write functions
//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
// void ReadMemoryWindow(BYTE vWindow, BYTE * vData, WORD wLength)
// Input:
//      vWindow: UDA_WINDOW, GP_WINDOW, or RX_WINDOW corresponding
//               to the window register to read from
//      vData: Pointer to local PIC RAM which will be written with data from
//              the ENC624J600 Family
//      wLength: Number of bytes to copy from window to vData
// Overview:
//  Copies 0 or more bytes from the ENC624J600 RAM using one of the UDA,
//  TX, or RX read window pointers.  This pointer is incremented by the
//  number of bytes read
//---------------------------------------------------------------------------
//
public sub ReadMemoryWindow(vWindow as byte, dest as word, wLength as word)
    dim vOpcode as byte

    if (wLength = 0) then
        exit
    endif

    vOpcode = RBMUDA
    if ((vWindow and GP_WINDOW) <> 0) then
        vOpcode = RBMGP
    endif        
    if ((vWindow and RX_WINDOW) <> 0) then
        vOpcode = RBMRX
    endif
    enc.assertCS()
    enc.transfer(vOpcode)                   // Send the command/opcode
    FSR0 = dest
    while (wLength <> 0)
        POSTINC0 = enc.transfer()
        wLength = wLength - 1
    end while
    enc.deassertCS()
end sub

//
//---------------------------------------------------------------------------
// void WriteMemoryWindow(BYTE vWindow, BYTE * vData, WORD wLength)
// Input:
//      vWindow: UDA_WINDOW, GP_WINDOW, or RX_WINDOW corresponding
//               to the window register to write to
//      vData: Pointer to local PIC RAM which contains the source data
//      wLength: Number of bytes to copy from vData to window
// Overview:
//  Copies 0 or more bytes from CPU RAM to the ENCX24J600 RAM using one
//  of the UDA, TX, or RX write window pointers. This pointer is incremented
//  by the number of bytes written
//  this routine takes ~70us to transfer 128 bytes
//---------------------------------------------------------------------------
//
public sub WriteMemoryWindow(vWindow as byte, src as word, wLength as word)
    dim vOpcode as byte

    vOpcode = WBMUDA
    if ((vWindow and GP_WINDOW) <> 0) then
        vOpcode = WBMGP
    endif        
    if ((vWindow and RX_WINDOW) <> 0) then
        vOpcode = WBMRX
    endif
    enc.assertCS()
    enc.transfer(vOpcode)                   // Send the command/opcode
    FSR0 = src
    while (wLength <> 0)
        enc.transfer(POSTINC0)
        wLength = wLength - 1
    end while
    enc.deassertCS()
end sub

inline function TBLRD_POSTINC() as TABLAT
    asm
        TBLRD*+
    end asm
end function

public sub WriteROMMemoryWindow(vWindow as byte, src as rom_addr_t, wLength as word)
    dim vOpcode as byte

    vOpcode = WBMUDA
    if ((vWindow and GP_WINDOW) <> 0) then
        vOpcode = WBMGP
    endif        
    if ((vWindow and RX_WINDOW) <> 0) then
        vOpcode = WBMRX
    endif

    enc.assertCS()
    enc.transfer(vOpcode)    // Send the Write Buffer Memory opcode

    TABLEPTR = src
    while (wLength <> 0)
        enc.transfer(TBLRD_POSTINC())
        wLength = wLength - 1
    end while
    
    enc.deassertCS()
end sub

//
//---------------------------------------------------------------------------
// void SetCLKOUT(BYTE clk_config)
// byte GetCLKOUT()
// PreCondition:
//      SPI bus must be initialized
// Input: clk_config
//        0x00: CLKOUT disabled (pin driven low)
//        0x01: 33.33 MHz
//        0x02: 25.00 MHz
//        0x03: 20.00 MHz
//        0x04: 16.67 MHz
//        0x05: 12.50 MHz
//        0x06: 10.00 MHz
//        0x07:  8.33 MHz
//        0x08:  8.00 MHz (duty cycle is not 50%)
//        0x09:  6.25 MHz
//        0x0a:  5.00 MHz
//        0x0b:  4.00 MHz
//        0x0c:  3.125 MHz
//        0x0d: no output...pin driven low
//        0x0e: 100 kHz
//        0x0f:  50 kHz
// Overview:
//      Writes the value of clk_config into the ECOCON register. The CLKOUT pin
//      will beginning outputting the new frequency immediately
//---------------------------------------------------------------------------
//
public sub SetCLKOUT(clk_config as byte)
    dim w as word

    // mask lower nibble
    clk_config = clk_config and $0f
    // get setting and clear current
    w = ReadReg(ECON2)
    w.byte1 = w.byte1 and $f0
    // set new
    w.byte1 = w.byte1 or clk_config
    WriteReg(ECON2, w)
end sub

public function GetCLKOUT() as byte
    dim w as word

    w = ReadReg(ECON2)
    result = w.byte1 and $0f
end function

//
//---------------------------------------------------------------------------
//  Function:        Byte MACGetRevID()
//  Overview:        reads enc rev id reg
//---------------------------------------------------------------------------
//
public function MACGetRevID() as byte
    result = ReadReg(EIDLED)
end function

//
//---------------------------------------------------------------------------
// WORD ReadPHYReg(BYTE Register)
// PreCondition:
//      SPI bus must be initialized
// Input:
//      Address of the PHY register to read from
// Output:
//      16 bits of data read from the PHY register
// Overview:
//      ReadPHYReg performs an MII read operation.  While in progress, it simply
//      polls the MII BUSY bit wasting time (25.6us)
//---------------------------------------------------------------------------
//
public function ReadPHYReg(bRegister as byte) as word
    dim b as byte

    // set the register address, being sure to set the reserved bit 8
    WriteReg(MIREGADR, $0100 or bRegister)

    // start the register read operation
    WriteReg(MICMD, MICMD_MIIRD)

    // the datasheet says to wait 25.6us (the C code doesn't do this)
    delayus(25)

    // wait until the PHY register has been read through the MII (requires 25.6us)
    repeat
        b = ReadReg(MISTAT)
    until (b.bits(MISTAT_BUSY_BIT) = 0)

    // stop read operation
    WriteReg(MICMD, $0000)

    // get results and return
    result = ReadReg(MIRD)
end function

//
//---------------------------------------------------------------------------
// void WritePHYReg(BYTE Register, WORD Data)
// PreCondition:
//      SPI bus must be initialized
// Input:
//      Address of the PHY register to write to
//      16 bits of data to write to PHY register
// Overview:
//      WritePHYReg performs an MII write operation.  While in progress,
//      it simply polls the MII BUSY bit wasting time
//---------------------------------------------------------------------------
//
public sub WritePHYReg(bRegister as byte, wData as word)
    dim b as byte

    // set the register address, being sure to set the reserved bit 8
    WriteReg(MIREGADR, $0100 or bRegister)

    // write the data. data must be written low byte, then high byte
    WriteReg(MIWR, wData)

    // the datasheet says to wait 25.6us (the C code doesn't do this)
    delayus(25)

    // wait until the PHY register has been written (requires 25.6us)
    repeat
        b = ReadReg(MISTAT)
    until (b.bits(MISTAT_BUSY_BIT) = 0)
end sub

//
//---------------------------------------------------------------------------
// PTR_BASE MACSetUDAWritePtr(PTR_BASE address)
// Input:
//      Address: Address to seek to
// Overview:
//      SPI write pointer is updated.  All calls to MACPutUDA() and MACPutArrayUDA()
//      will use this new value
//---------------------------------------------------------------------------
//
public sub MACSetUDAWritePtr(address as word)
    // Set the TX Write Pointer to the new calculated value
    WriteReg(EUDAWRPT, address)
end sub

//
//---------------------------------------------------------------------------
// PTR_BASE MACSetUDAReadPtr(PTR_BASE Address)
// Input:
//      Address: Address to seek to
// Overview:
//      SPI read pointer is updated.  All calls to MACGetUDA() and MACGetArrayUDA()
//      will use these new values
//---------------------------------------------------------------------------
//
public sub MACSetUDAReadPtr(address as word)
    // Set the RX Read Pointer to the new calculated value
    WriteReg(EUDARDPT, address)
end sub

//
//---------------------------------------------------------------------------
// BYTE MACGetUDA()
// PreCondition:
//      SPI bus must be initialized
//      EUDARDPT must point to the place to read from
// Output:
//      Byte read from the ENCx24J600's RAM
// Overview:
//      MACGet returns the byte pointed to by EUDARDPT and increments EUDARDPT so MACGetUDA()
//      can be called again. The increment will follow the receive buffer wrapping boundary
//---------------------------------------------------------------------------
//
public function MACGetUDA() as byte
    ReadMemoryWindow(UDA_WINDOW, @result, 1)
end function

//
//---------------------------------------------------------------------------
// void MACPutUDA(BYTE val)
// PreCondition:
//      SPI bus must be initialized
//      EUDAWRPT must point to the location to begin writing
// Input:
//      Byte to write into the ENCx24J600 buffer memory
// Overview:
//      MACPutUDA outputs the Write Buffer Memory opcode/constant (8 bits) and
//      data to write (8 bits) over the SPI. EUDAWRPT is incremented after the write
//---------------------------------------------------------------------------
//
public sub MACPutUDA(val as byte)
    WriteMemoryWindow(UDA_WINDOW, @val, 1)
end sub

//
//---------------------------------------------------------------------------
// WORD MACGetArrayUDA(BYTE * val, WORD len)
// PreCondition:
//      SPI bus must be initialized
//      EUDARDPT must point to the place to read from
// Input:
//      val: Pointer to storage location
//      len:  Number of bytes to read from the data buffer
// Output:
//      Byte(s) of data read from the data buffer
// Overview:
//      Burst reads several sequential bytes from the data buffer and places them
//      into local memory. With SPI burst support, it performs much faster than
//      multiple MACGetUDA() calls. EUDARDPT is incremented after each byte, following
//      the same rules as MACGetUDA()
//---------------------------------------------------------------------------
//
public sub MACGetArrayUDA(dest as word, len as word)
    ReadMemoryWindow(UDA_WINDOW, dest, len)
end sub

//
//---------------------------------------------------------------------------
// void MACPutArrayUDA(BYTE * val, WORD len)
// PreCondition:
//      SPI bus must be initialized
//      EUDAWRPT must point to the location to begin writing
// Input:
//      val: Pointer to source of bytes to copy
//      len:  Number of bytes to write to the data buffer
// Overview:
//      MACPutArrayUDA writes several sequential bytes to the ENCx24J600 RAM.
//      It performs faster than multiple MACPutUDA() calls. EUDAWRPT is incremented by len
//---------------------------------------------------------------------------
//
public sub MACPutArrayUDA(src as word, len as word)
    WriteMemoryWindow(UDA_WINDOW, src, len)
end sub

//
//---------------------------------------------------------------------------
// static void ToggleCRYPTEN(void)
// Summary:
//    Toggles the CRYPTEN (EIR<15>) cryptographic enable bit in an errata safe way
// Description:
//    Due to an ENC424J600/624J600 Rev. A2 silicon errata, it is not possible to
//    set or clear the CRYPTEN bit (EIR<15>) by using the BFSReg() or BFCReg()
//    functions. Instead a register write to the volatile EIR interrupt flags
//    register is required.  This function reads EIR, toggles the CRYPTEN state,
//    and writes the value back to EIR in a manner that does not destroy or
//    glitch the interrupt states
//---------------------------------------------------------------------------
//
private sub ToggleCRYPTEN()
    dim
        wLinkStat as word,
        wInterruptEnable as word,
        w as word

    // About to do read-modify-write to interrupt flags - disable interrupts temporarily
    wInterruptEnable = ReadReg(EIE)
    if (wInterruptEnable.bits(EIE_INTIE_BIT) <> 0) then
        BFCReg(EIE, EIE_INTIE)
    endif

    // Get current link status so that we can regenerate LINKIF if it is missed
    wLinkStat = ReadReg(ESTAT)

    // Turn on/off crypto enable
    w = ReadReg(EIR)
    if (w.bits(EIR_CRYPTEN_BIT) = 1) then        // toggle CRYPTEN bit
        BFCReg(EIR, EIR_CRYPTEN)
    else
        BFSReg(EIR, EIR_CRYPTEN)
    endif
    
    // Generate link change interrupt if the status has changed while we were
    // doing the read-modify-write operation (unlikely but possible)
    if (((ReadReg(ESTAT) xor wLinkStat) and ESTAT_PHYLNK) <> 0) then
        BFSReg(EIR, EIR_LINKIF)
    endif

    // Restore interrupts
    if (wInterruptEnable.bits(EIE_INTIE_BIT) <> 0) then
        BFSReg(EIE, EIE_INTIE)
    endif
end sub

//
//---------------------------------------------------------------------------
// void MACWaitForAutoneg()
// Overview:
// wait for autonegotiation to complete if it's enabled.
//---------------------------------------------------------------------------
//
public sub MACWaitForAutoneg()
    dim st as word,
        tout as word

    st = ReadPHYReg(PHCON1)
    if (st.bits(PHCON1_ANEN_BIT) = 1) then      // autoneg enabled...
        tout = 500                              // wait a max of 500ms
        repeat
            st = ReadPHYReg(PHSTAT1)
            delayms(1)
            tout = tout - 1
            if (tout = 0) then
                break                           // we've waited too long...leave
            endif
        until (st.bits(PHSTAT1_ANDONE_BIT) = 1)    // autoneg is complete
    endif
end sub

//
//---------------------------------------------------------------------------
// void MACUpdateDuplexStatus()
// Overview:
// Check to see if the duplex status has changed
// This can change if the user unplugs the cable and plugs it into a different node
// Auto-negotiation will automatically set the duplex in the PHY, but we must also
// update the MAC inter-packet gap timing and duplex state to match
//---------------------------------------------------------------------------
//
public sub MACUpdateDuplexStatus()
    dim w as word,
        st as word

    // update MAC duplex settings to match PHY duplex setting
    w = ReadReg(MACON2)
    st = ReadReg(ESTAT)
    if (st.bits(ESTAT_PHYDPX_BIT) = 1) then
        // switching to full duplex
        WriteReg(MABBIPG, $15)
        w.bits(MACON2_FULDPX_BIT) = 1
    else
        // switching to half duplex
        WriteReg(MABBIPG, $12)
        w.bits(MACON2_FULDPX_BIT) = 0
    endif
    WriteReg(MACON2, w)
    
    // added... wait for autoneg to complete
    MACWaitForAutoneg()
end sub

//
//---------------------------------------------------------------------------
//  Function:        bool MACIsIntrAsserted()
//  Overview:        polls the intr pin and returns true/false if intr is asserted
//---------------------------------------------------------------------------
//
public dim MACIsIntrAsserted as enc.getINT

//
//---------------------------------------------------------------------------
//  Function:        void MACResetLinkChangeIntr(void)
//  Overview:        resets PHY PLNKIF
//---------------------------------------------------------------------------
//
public sub MACResetLinkChangeIntr()
    dim w as word

    w = ReadReg(EIR)
    if (w.bits(EIR_LINKIF_BIT) = 1) then
        // clear the LINKIF bit in EIR intr request register
        BFCReg(EIR, EIR_LINKIF)

        // check to see if we just established a link and need to setup duplex
        w = ReadReg(ESTAT)
        if (w.bits(ESTAT_PHYLNK_BIT) = 1) then
            MACUpdateDuplexStatus()
        endif

    endif
end sub

public sub MACSetLinkChangeIntr()
    // set the LINKIF bit in EIR intr request register
    BFSReg(EIR, EIR_LINKIF)
end sub

//
//---------------------------------------------------------------------------
// BOOL MACIsLinked(void)
// Output:
//      TRUE: An Ethernet link is established
//      FALSE: Ethernet link is not established
// Overview:
//      Returns the ESTAT<PHYLNK> bit
// Note: changed to read latching LLSTAT bit for compat with ENC28J60 function
//---------------------------------------------------------------------------
//
public function MACIsLinked() as boolean
    dim w as word

  #if (ENC_USE_LATCHING_LINKSTAT)
    // LLSTAT is a latching low link status bit.  Therefore, if the link goes down and
    // comes back up before a higher level stack program calls MACIsLinked(), MACIsLinked()
    // will still return FALSE.  The next call to MACIsLinked() will return TRUE (unless
    // the link goes down again)
    w = ReadPHYReg(PHSTAT1)
    if (w.bits(PHSTAT1_LLSTAT_BIT) = 0) then
        result = false
    else
        result = true
    endif
  #else
    w = ReadReg(ESTAT)
    if (w.bits(ESTAT_PHYLNK_BIT) = 0) then
        result = false
    else
        result = true
    endif
  #endif
end function

//
//---------------------------------------------------------------------------
// byte MACGetPhySpeed(void)
// returns 10 or 100
//---------------------------------------------------------------------------
//
public function MACGetPhySpeed() as byte
    dim b as byte

    b = ReadPHYReg(PHSTAT3)
    b = (b >> 2) and %111
    if ((b = %010) or (b = %110)) then
        result = 100
    elseif ((b = %001) or (b = %101)) then
        result = 10
    else
        result = 0      // undefined!!!
    endif
end function

//
//---------------------------------------------------------------------------
// byte MACGetPhyDuplex(void)
// returns 1 (half) or 2 (full)
//---------------------------------------------------------------------------
//
public function MACGetPhyDuplex() as byte
    dim b as byte

    b = ReadPHYReg(PHSTAT3)
    b = (b >> 2) and %111
    if ((b = %001) or (b = %010)) then
        result = 1      // half duplex
    elseif ((b = %101) or (b = %110)) then
        result = 2      // full duplex
    else
        result = 0      // undefined!!!
    endif
end function

//
//---------------------------------------------------------------------------
// BOOL MACIsTxReady(void)
// Output:
//      TRUE: If no Ethernet transmission is in progress
//      FALSE: If a previous transmission was started, and it has not completed yet
//          While FALSE, the data in the transmit buffer and the TXST/TXLEN pointers
//          must not be changed
// Overview:
//      Returns the ECON1<TXRTS> bit
//---------------------------------------------------------------------------
//
public function MACIsTxReady() as boolean
    dim b as byte

    result = true       // assume true
    b = ReadReg(ECON1)
    if (b.bits(ECON1_TXRTS_BIT) = 1) then
        result = false
    endif
end function

//
//---------------------------------------------------------------------------
// void MACDiscardRx(void)
// Overview:
//      Marks the last received packet (obtained using MACGetHeader())as being
//      processed and frees the buffer memory associated with it
// Note:
//      Is is safe to call this function multiple times between MACGetHeader() calls
//      Extra packets won't be thrown away until MACGetHeader() makes it available
//---------------------------------------------------------------------------
//
public sub MACDiscardRx()
    dim wNewRXTail as word

    // Make sure the current packet was not already discarded
    if (ENC100Flags.bWasDiscarded) then
        exit
    endif
    ENC100Flags.bWasDiscarded = true

    // Decrement the next packet pointer before writing it into the ERXRDPT registers
    // RX buffer wrapping must be taken into account if the NextPacketLocation is precisely RXSTART
    wNewRXTail = wNextPacketPointer - 2
    if (wNextPacketPointer = RXSTART) then
        wNewRXTail = ENC100_RAM_SIZE - 2
    endif

    // Decrement the RX packet counter register, EPKTCNT
    BFSReg(ECON1, ECON1_PKTDEC)

    // Move the receive read pointer to unwrite-protect the memory used by the
    // last packet.  The writing order is important: set the low byte first,
    // high byte last (handled automatically in WriteReg())
    WriteReg(ERXTAIL, wNewRXTail)
end sub

//
//---------------------------------------------------------------------------
// WORD MACGetFreeRxSize(void)
// Output:
//      A WORD estimate of how much RX buffer space is free at the present time
//---------------------------------------------------------------------------
//
public function MACGetFreeRxSize() as word
    dim wHeadPtr as word

    wHeadPtr = ReadReg(ERXHEAD)

    // Calculate the difference between the pointers, taking care to account
    // for buffer wrapping conditions
    if (wHeadPtr > wCurrentPacketPointer) then
        result = (RXSTOP - RXSTART) - (wHeadPtr - wCurrentPacketPointer)
        exit
    endif

    result = wCurrentPacketPointer - wHeadPtr - 2
end function

//
//---------------------------------------------------------------------------
// BOOL MACGetHeader(MAC_ADDR * remote, BYTE * type)
// Input:
//      remote: Location to store the Source MAC address of the received frame
//      type: Location of a BYTE to store the constant MAC_UNKNOWN, ETHER_IP,
//          or ETHER_ARP, representing the contents of the Ethernet type field
// Output:
//      TRUE: If a packet was waiting in the RX buffer. The remote and type
//          values are updated
//      FALSE: If a packet was not pending. remote and type are not changed
// Side Effects:
//      Last packet is discarded if MACDiscardRx() hasn't already been called
//---------------------------------------------------------------------------
//
#define ENABLE_MAC_RESET_TRACE_CODE

public function MACGetHeader(byref remote as MAC_ADDR, byref htype as byte) as boolean
    dim header as ENC100_PREAMBLE,
        b as byte,
        btmp as WREG

    // assume no packets waiting
    result = false

    // check to see if INT pin is asserted
    if (MACIsIntrAsserted() = false) then
        exit
    endif

    // Test if at least one packet has been received and is waiting
    b = ReadReg(EIR)
    if (b.bits(EIR_PKTIF_BIT) = 0) then
        exit
    endif

    // Discard the last packet, if the user application hasn't done so already
    if (ENC100Flags.bWasDiscarded = false) then
        MACDiscardRx()

        // Test again if at least one packet has been received and is waiting
        b = ReadReg(EIR)
        if (b.bits(EIR_PKTIF_BIT) = 0) then
            exit
        endif
    endif

    // Set the RX Read Pointer to the beginning of the next unprocessed packet
    wCurrentPacketPointer = wNextPacketPointer
    WriteReg(ERXRDPT, wCurrentPacketPointer)

    // Obtain the MAC header from the Ethernet buffer
    ReadMemoryWindow(RX_WINDOW, @header, sizeof(header))

    // The EtherType field, like most items transmitted on the Ethernet medium are in big endian
    // header.htype.Val = swaps(header.htype.Val)
    btmp = header.htype.v(0)
    header.htype.v(0) = header.htype.v(1)
    header.htype.v(1) = btmp

    // Validate the data returned from the ENC624J600 Family device.  Random data corruption,
    // such as if a single SPI/PSP bit error occurs while communicating or a momentary power
    // glitch could cause this to occur in rare circumstances.  Also, certain hardware bugs
    // such as violations of the absolute maximum electrical specs can cause this.  For example,
    // if an MCU with a high slew rate were to access the interface, parasitic inductance in the
    // traces could cause excessive voltage undershoot. If the voltage goes too far below ground,
    // the ENCx24J600's internal ESD structure may activate and disrupt the communication.  To prevent
    // this, ensure that you have a clean board layout and consider adding resistors in series
    // with the MCU output pins to limit the slew rate of signals going to the ENCx24J600. 100 Ohm
    // resistors is a good value to start testing with
  #if defined(ENABLE_MAC_RESET_TRACE_CODE)
    // enable this code if you need to find the reason for the reset
    b = 0
    if (header.NextPacketPointer > RXSTOP) then
        b = 1
    elseif (header.NextPacketPointer.bits(0) <> 0) then
        b = 2
    elseif (header.StatusVector.bits.Zero <> 0) then
        b = 3
    elseif (header.StatusVector.bits.ZeroH <> 0) then
        b = 4
    elseif (header.StatusVector.bits.CRCError <> 0) then
        b = 5
    elseif (header.StatusVector.bits.ByteCount > 1522) then
        b = 6
    endif
    if (b <> 0) then
        nop()
        nop()
        reset()
    endif
  #else
    if ((header.NextPacketPointer > RXSTOP) or
        (header.NextPacketPointer.bits(0) <> 0) or
        (header.StatusVector.bits.Zero <> 0) or
        (header.StatusVector.bits.ZeroH <> 0) or
        (header.StatusVector.bits.CRCError <> 0) or
        (header.StatusVector.bits.ByteCount > 1522)) then
        nop()
        reset()
    endif
  #endif

    // Save the location where the hardware will write the next packet to
    wNextPacketPointer = header.NextPacketPointer

    // Return the Ethernet frame's Source MAC address field to the caller
    // This parameter is useful for replying to requests without requiring an ARP cycle
    remote.v = header.SourceMACAddr.v

    // Return a simplified version of the EtherType field to the caller
    htype = MAC_UNKNOWN
    if ((header.htype.v(1) = ETHER_TYPE) and ((header.htype.v(0) = ETHER_IP) or (header.htype.v(0) = ETHER_ARP))) then
       htype = header.htype.v(0)
    endif

    // Mark this packet as discardable
    ENC100Flags.bWasDiscarded = false
    result = true
end function

//
//---------------------------------------------------------------------------
// void MACPutHeader(MAC_ADDR * remote, BYTE type, WORD dataLen)
// PreCondition:
//      MACIsTxReady() must return TRUE
// Input:
//      remote: Pointer to memory which contains the destination MAC address (6 bytes)
//      type: The constant ETHER_ARP or ETHER_IP, defining which value to write
//          into the Ethernet header's type field
//      dataLen: Length of the Ethernet data payload
// Note:
//      Because of the dataLen parameter, it is probably advantagous to call
//      this function immediately before transmitting a packet rather than
//      initially when the packet is first created. The order in which the packet
//      is constructed (header first or data first) is not important
//---------------------------------------------------------------------------
//
public sub MACPutHeader(byref remote as MAC_ADDR, htype as byte, dataLen as word)
    dim wEthernetType as word

    wEthernetType.byte0 = ETHER_TYPE
    if (htype = MAC_IP) then
        wEthernetType.byte1 = ETHER_IP
    else
        wEthernetType.byte1 = ETHER_ARP
    endif

    // set the Window Write Pointer to the beginning of the transmit buffer
    WriteReg(EGPWRPT, TXSTART)
    WriteReg(ETXLEN, dataLen + sizeof(MAC_ETHER_HEADER))

    // write the Ethernet destination MAC address, our source MAC address, and the Ethernet Type field
    WriteMemoryWindow(GP_WINDOW, addressof(remote), sizeof(MAC_ADDR))
    WriteMemoryWindow(GP_WINDOW, addressof(AppConfig.MyMACAddr), sizeof(AppConfig.MyMACAddr))
    WriteMemoryWindow(GP_WINDOW, addressof(wEthernetType), 2)
end sub

//
//---------------------------------------------------------------------------
// void MACFlush(void)
// PreCondition:
//      A packet has been created by calling MACPut() and MACPutHeader()
// Overview:
//      MACFlush causes the current TX packet to be sent out on the Ethernet medium
//      The hardware MAC will take control and handle CRC generation, collision
//      retransmission and other details
// Note:
//      After transmission completes (MACIsTxReady() returns TRUE), the packet
//      can be modified and transmitted again by calling MACFlush() again
//      Until MACPutHeader() or MACPut() is called (in the TX data area), the
//      data in the TX buffer will not be corrupted
//---------------------------------------------------------------------------
//
public sub MACFlush()
    dim w as word

    // original C code
    // Check to see if the duplex status has changed.  This can change if the user unplugs the cable 
    // and plugs it into a different node.  Auto-negotiation will automatically set the PHY duplex, 
    // but we must also update the MAC inter-packet gap timing and duplex state to match
    w = ReadReg(EIR)
    if (w.bits(EIR_LINKIF_BIT) = 1) then
        BFCReg(EIR, EIR_LINKIF)
        MACUpdateDuplexStatus()
    endif

    // start the transmission, but only if we are linked. Supressing transmissing when unlinked
    // is necessary to avoid stalling the TX engine if we are in PHY energy detect power down mode
    // and no link is present. A stalled TX engine won't do any harm in itself, but will cause the
    // MACIsTXReady() function to continuously return FALSE, which will ultimately stall the TCP/IP
    // stack since there is blocking code elsewhere in other files that expect the TX engine to
    // always self-free itself very quickly
    w = ReadReg(ESTAT)
    if (w.bits(ESTAT_PHYLNK_BIT) = 1) then
        BFSReg(ECON1, ECON1_TXRTS)
    endif
end sub

//
//---------------------------------------------------------------------------
// void MACSetReadPtrInRx(WORD offset)
// PreCondition:
//      A packet has been obtained by calling MACGetHeader() and getting a TRUE result
// Input:
//      offset: WORD specifying how many bytes beyond the Ethernet header's type
//          field to relocate the SPI read pointer
// Overview:
//      SPI read pointer is updated.  All calls to MACGet() and MACGetArray()
//      will use these new values
// Note:
//      RXSTOP must be statically defined as being > RXSTART for this function to
//      work correctly. In other words, do not define an RX buffer which spans
//      the 0x1FFF->0x0000 memory boundary
//---------------------------------------------------------------------------
//
public sub MACSetReadPtrInRx(offset as word)
    dim wReadPtr as word

    // Determine the address of the beginning of the entire packet
    // and adjust the address to the desired location
    wReadPtr = wCurrentPacketPointer + sizeof(ENC100_PREAMBLE) + offset

    // Since the receive buffer is circular, adjust if a wraparound is needed
    if (wReadPtr > RXSTOP) then
        wReadPtr = wReadPtr - RXSIZE
    endif

    // Set the RX Window Read pointer to the new calculated value
    WriteReg(ERXRDPT, wReadPtr)
end sub

//
//---------------------------------------------------------------------------
// PTR_BASE MACSetWritePtr(PTR_BASE address)
// Input:
//      Address: Address to seek to
// Output:
//      WORD: Old EWRPT location
// Overview:
//      SPI write pointer is updated.  All calls to MACPut() and MACPutArray()
//      will use this new value
//---------------------------------------------------------------------------
//
public function MACSetWritePtr(address as word) as word
    result = ReadReg(EGPWRPT)

    // Set the TX Write Pointer to the new calculated value
    WriteReg(EGPWRPT, address)
end function

//
//---------------------------------------------------------------------------
// PTR_BASE MACSetReadPtr(PTR_BASE Address)
// Input:
//      Address: Address to seek to
// Output:
//      WORD: Old ERDPT value
// Overview:
//      SPI read pointer is updated.  All calls to MACGet() and MACGetArray()
//      will use these new values
//---------------------------------------------------------------------------
//
public function MACSetReadPtr(address as word) as word
    result = ReadReg(ERXRDPT)

    // Set the RX Read Pointer to the new calculated value
    WriteReg(ERXRDPT, address)
end function

//
//---------------------------------------------------------------------------
// dma transfer helper function
//---------------------------------------------------------------------------
//
private sub wait_for_dma_done()
    dim b as byte

    repeat
        b = ReadReg(ECON1)
    until (b.bits(ECON1_DMAST_BIT) = 0)
end sub

//
//---------------------------------------------------------------------------
// checksum helper function
//---------------------------------------------------------------------------
//
private function MACCalcChecksum(wAddr as word, len as word) as word
    // If in power down, temporarily bring power up so we can use the DMA
    if (ENC100Flags.PoweredDown) then
        BFSReg(ECON2, ECON2_ETHEN)
    endif

    // Calculate the checksum using the ENCX24J600 DMA engine
    wait_for_dma_done()
    WriteReg(EDMAST, wAddr)
    WriteReg(EDMALEN, len)
    BFCReg(ECON1, ECON1_DMACPY or ECON1_DMANOCS or ECON1_DMACSSD)
    BFSReg(ECON1, ECON1_DMAST)
    wait_for_dma_done()

    // Restore power down state, if applicable
    if (ENC100Flags.PoweredDown) then
        BFCReg(ECON2, ECON2_ETHEN)
    endif

    result = ReadReg(EDMACS)
end function

//
//---------------------------------------------------------------------------
// WORD CalcIPBufferChecksum(WORD len)
// PreCondition:
//      Read buffer pointer set to starting of checksum data
// Input:
//      len: Total number of bytes to calculate the checksum over. The first
//          byte included in the checksum is the byte pointed to by ERDPT,
//          which is updated by calls to MACSetReadPtr(), MACGet(), MACGetArray(),
//          MACGetHeader(), etc
// Output:
//      16-bit checksum as defined by RFC 793
// Overview:
//      This function performs a checksum calculation in the MAC buffer itself
//      The ENCx24J600 has a hardware DMA module which can calculate the checksum
//      faster than software, so this function replaces the CaclIPBufferChecksum()
//      function defined in the helpers.c file.  Through the use of preprocessor
//      defines, this replacement is automatic
// Note:
//      This function works either in the RX buffer area or the TX buffer area
//      No validation is done on the len parameter
//---------------------------------------------------------------------------
//
public function CalcIPBufferChecksum(len as word) as word
    dim w as word

    w = ReadReg(ERXRDPT)
    result = MACCalcChecksum(w, len)
end function

//
//---------------------------------------------------------------------------
// WORD MACCalcRxChecksum(WORD offset, WORD len)
// Input:
//      offset  - Number of bytes beyond the beginning of the Ethernet data
//          (first byte after the type field) where the checksum should begin
//      len     - Total number of bytes to include in the checksum
// Output:
//      16-bit checksum as defined by RFC 793
// Overview:
//      This function performs a checksum calculation in the MAC buffer itself
//---------------------------------------------------------------------------
//
public function MACCalcRxChecksum(offset as word, len as word) as word
    dim wStartAddress as word

    // Add the offset requested by firmware plus the Ethernet header
    wStartAddress = wCurrentPacketPointer + sizeof(ENC100_PREAMBLE) + offset
    if (wStartAddress > RXSTOP) then
        wStartAddress = wStartAddress - RXSIZE
    endif

    result = MACCalcChecksum(wStartAddress, len)
end function

//
//---------------------------------------------------------------------------
// void MACMemCopyAsync(PTR_BASE destAddr, PTR_BASE sourceAddr, WORD len)
// Summary:
//      Asynchronously copies data from one address to another within the 24KB
//      Ethernet memory
// Description:
//      Asynchronously copies data from one address to another within the 24KB
//      Ethernet memory using the ENCX24J600 hardware DMA engine (very fast).
//      Overlapped memory regions are allowed (with restrictions)
//      The addresses do not have to be aligned
// Precondition:
//      SPI bus must be initialized
// Parameters:
//      destAddr - Destination address in the Ethernet memory to copy to.
//          If (PTR_BASE)-1 is specified, the current EGPWRPT value will be used
//          instead
//      sourceAddr - Source address to read from.  If (PTR_BASE)-1 is specified,
//          the current EGPRDPT value will be used instead.
//      len - Number of bytes to copy
// Remarks:
//      Call MACIsMemCopyDone() to see when the transfer is complete.
//      Copying to a destination region that overlaps with the source address
//      is supported only if the destination start address is at a lower memory
//      address (closer to 0x0000) than the source pointer.  However, if they do
//      overlap there must be at least 2 bytes of non-overlap to ensure correct
//      results due to hardware DMA requirements.  For example, destAddr = 0;
//      sourceAddr = 1; is illegal while destAddr = 0; sourceAddr = 2; is fine.
//      If a prior transfer is already in progress prior to calling this function,
//      this function will block until it can start this transfer.
//      If (PTR_BASE)-1 is used for the sourceAddr or destAddr parameters,
//      then that pointer will get updated with the next address after the read or write
//---------------------------------------------------------------------------
//
public sub MACMemCopyAsync(destAddr as word, sourceAddr as word, len as word)
    dim wNewReadPtr as word

    // Decode destination and source addresses
    if (destAddr = word(-1)) then
        destAddr = ReadReg(EGPWRPT)
        WriteReg(EGPWRPT, destAddr + len)
    endif

    if (sourceAddr = word(-1)) then
        sourceAddr = ReadReg(ERXRDPT)
        wNewReadPtr = sourceAddr + len
        if (wNewReadPtr > RXSTOP) then
            wNewReadPtr = wNewReadPtr - RXSIZE
        endif
        WriteReg(ERXRDPT, wNewReadPtr)
    endif

    // If in power down, temporarily bring power up so we can use the DMA
    if (ENC100Flags.PoweredDown) then
        BFSReg(ECON2, ECON2_ETHEN)
    endif

    // Start the copy using the DMA
    wait_for_dma_done()
    WriteReg(EDMAST, sourceAddr)
    WriteReg(EDMADST, destAddr)
    WriteReg(EDMALEN, len)
    BFSReg(ECON1, ECON1_DMAST or ECON1_DMACPY)
end sub

public function MACIsMemCopyDone() as boolean
    dim b as byte

    b = ReadReg(ECON1)
    if (b.bits(ECON1_DMAST_BIT) = 1) then
        result = false
    else
        result = true
        // Restore power down state, if applicable
        if (ENC100Flags.PoweredDown) then
            BFCReg(ECON2, ECON2_ETHEN)
        endif
    endif
end function

//
//---------------------------------------------------------------------------
// BYTE MACGet()
// PreCondition:
//      SPI bus must be initialized
//      ERDPT must point to the place to read from
// Output:
//      Byte read from the ENCx24J600's RAM
// Overview:
//      MACGet returns the byte pointed to by ERDPT and increments ERDPT so MACGet()
//      can be called again. The increment will follow the receive buffer wrapping boundary
//---------------------------------------------------------------------------
//
public function MACGet() as byte
    ReadMemoryWindow(RX_WINDOW, @result, 1)
end function

//
//---------------------------------------------------------------------------
// WORD MACGetArray(BYTE * val, WORD len)
// PreCondition:
//      SPI bus must be initialized
//      ERDPT must point to the place to read from
// Input:
//      val: Pointer to storage location
//      len:  Number of bytes to read from the data buffer
// Output:
//      Byte(s) of data read from the data buffer
// Overview:
//      Burst reads several sequential bytes from the data buffer and places them
//      into local memory. With SPI burst support, it performs much faster than
//      multiple MACGet() calls. ERDPT is incremented after each byte, following
//      the same rules as MACGet()
//---------------------------------------------------------------------------
//
public function MACGetArray(dest as word, len as word) as word
    dim wNewReadPtr as word

    if (dest <> 0) then
        ReadMemoryWindow(RX_WINDOW, dest, len)
    else
        wNewReadPtr = ReadReg(ERXRDPT) + len
        if (wNewReadPtr > RXSTOP) then
            wNewReadPtr = wNewReadPtr - RXSIZE
        endif
        WriteReg(ERXRDPT, wNewReadPtr)
    endif

    result = len
end function

//
//---------------------------------------------------------------------------
// void MACPut(BYTE val)
// PreCondition:
//      SPI bus must be initialized
//      EWRPT must point to the location to begin writing
// Input:
//      Byte to write into the ENCx24J600 buffer memory
// Overview:
//      MACPut outputs the Write Buffer Memory opcode/constant (8 bits) and
//      data to write (8 bits) over the SPI. EWRPT is incremented after the write
//---------------------------------------------------------------------------
//
public sub MACPut(val as byte)
    WriteMemoryWindow(GP_WINDOW, @val, 1)
end sub

//
//---------------------------------------------------------------------------
// void MACPutArray(BYTE * val, WORD len)
// PreCondition:
//      SPI bus must be initialized
//      EWRPT must point to the location to begin writing
// Input:
//      val: Pointer to source of bytes to copy
//      len:  Number of bytes to write to the data buffer
// Overview:
//      MACPutArray writes several sequential bytes to the ENCx24J600 RAM.
//      It performs faster than multiple MACPut() calls. EWRPT is incremented by len
//---------------------------------------------------------------------------
//
public sub MACPutArray(src as word, len as word)
    WriteMemoryWindow(GP_WINDOW, src, len)
end sub

//
//---------------------------------------------------------------------------
// void MACPutROMArray(ROM BYTE * val, WORD len)
// PreCondition:
//      SPI bus must be initialized
//      EWRPT must point to the location to begin writing
// Input:
//      val: Pointer to source of bytes to copy
//      len: Number of bytes to write to the data buffer
// Overview:
//      MACPutArray writes several sequential bytes to the ENCx24J600 RAM.
//      It performs faster than multiple MACPut() calls. EWRPT is incremented by len
//---------------------------------------------------------------------------
//
public sub MACPutROMArray(src as rom_addr_t, len as word)
    WriteROMMemoryWindow(GP_WINDOW, src, len)
end sub

//
//---------------------------------------------------------------------------
// void MACPowerDown(void)
// PreCondition:
//      SPI bus must be initialized
// Overview:
//      MACPowerDown puts the ENCx24J600 in low power sleep mode. In sleep mode,
//      no packets can be transmitted or received. All MAC and PHY registers
//      should not be accessed. To exit power down, call MACPowerUp()
// Note:
//      If a packet is being transmitted while this function is called, this
//      function will block until it is it complete. If anything is being received,
//      it will be completed
//---------------------------------------------------------------------------
//
public sub MACPowerDown()
    // Disable packet reception
    BFCReg(ECON1, ECON1_RXEN)

    // Make sure any last packet which was in-progress when RXEN was cleared
    // is completed
    while ((ReadReg(ESTAT) and ESTAT_RXBUSY) <> 0)
    end while

    // If a packet is being transmitted, the DMA is operating, the Modular
    // Exponentiation or AES engine is running, wait for it to finish
    while ((ReadReg(ECON1) and (ECON1_TXRTS or ECON1_DMAST or ECON1_MODEXST or ECON1_AESST)) <> 0)
    end while

    // Enter sleep mode
    ENC100Flags.PoweredDown = true
    ENC100Flags.CryptoEnabled = false
    if ((ReadReg(EIR) and EIR_CRYPTEN) <> 0) then              // Turn off ModdEx/AES clock
        ToggleCRYPTEN()
    endif
    WritePHYReg(PHCON1, PHCON1_PSLEEP)          // Turn off the PHY
    BFCReg(ECON2, ECON2_ETHEN or ECON2_STRCH)   // Turn off general internal clocks and LED stretching so they immediately turn off and don't get stuck on
end sub

//
//---------------------------------------------------------------------------
// void MACEDPowerDown(void)
// PreCondition:
//      SPI bus must be initialized
// Overview:
//      MACEDPowerDown puts the ENCx24J600 PHY in Energy Detect mode. In this mode,
//      the PHY will passively listen for link pulses and automatically link up,
//      if detected. This can be detected via the MACIsLinked() function.
//      This power state will save power only when an Ethernet cable is unattached.
//      To exit energy detect power down, call MACPowerUp().
// Note:
//      The ENCx24J600 is fully operational when in energy detect mode.
//      If a Ethernet link is detected or already established, full TX/RX activity
//      will work as normal and no power saving will occur
//---------------------------------------------------------------------------
//
public sub MACEDPowerDown()
    // Put the PHY into energy detect mode
    WritePHYReg(PHCON2, PHCON2_EDPWRDN)
end sub

//
//---------------------------------------------------------------------------
// void MACPowerUp(void)
// PreCondition:
//      SPI bus must be initialized
// Overview:
//      MACPowerUp returns the ENCx24J600 back to normal operation after a previous
//      call to MACPowerDown().  Calling this function when already powered up
//      will have no effect
// Note:
//      If a link partner is present, it will take 10s of msecs before a new link
//      will be established after waking up.  While not linked, packets which are
//      transmitted will most likely be lost.  MACIsLinked() can be called to
//      determine if a link is established
//---------------------------------------------------------------------------
//
public sub MACPowerUp()
    // Start up general clocks and reenable LED stretching
    ENC100Flags.PoweredDown = false
    BFSReg(ECON2, ECON2_ETHEN or ECON2_STRCH)

    // Power PHY back up and put in auto-negotation mode to reestablish a link
    if ((ReadPHYReg(PHCON1) and PHCON1_PSLEEP) <> 0) then
        WritePHYReg(PHCON1, PHCON1_ANEN)
    endif

    // Disable energy detect PHY power down
    WritePHYReg(PHCON2, $0000)

    // Ensure the PLL and PHY are ready
    while ((ReadReg(ESTAT) and (ESTAT_PHYRDY or ESTAT_CLKRDY)) = 0)
    end while

    // Enable packet reception
    BFSReg(ECON1, ECON1_RXEN)
end sub

//
//---------------------------------------------------------------------------
// static void SendSystemReset(void)
// PreCondition:
//      SPI or PSP bus must be initialized
// Overview:
//      SendSystemReset reliably resets the Ethernet controller. It resets all
//      register contents (except for COCON bits of ECON2) and returns the device
//      to the power on default state. This function should be called instead of
//      directly attempting to perform a reset via the ECON2<ETHRST> bit
// Note:
//      This function is a blocking function and will lock up the application
//      if a non-recoverable problem is present (original C code...no longer true)
// NOTE: the original C code for this function differs from that given in the datasheet
//      it uses UNDOCUMENTED bits in the ESTAT register(ESTAT_RSTDONE, ESTAT_PHYRDY)!!
//      also, a check of PRST in PHCON1 was added in v2.0, as it seems perhaps the
//      undocumented bits don't work very well
// Modified to add timeout and return status
//        0 = ok
//      < 0 = failure
//---------------------------------------------------------------------------
//
public function SendSystemReset() as byte
    dim w as word
{
    // Power cycle the ENCx24J600 device, if any sort of POR pin is defined
    #if defined(ENC100_POR_IO)
        ENC100_POR_IO = 0;
        ENC100_POR_TRIS = 0;
        DelayMs(2);

        // If the INT/SPISEL signal is connected, force it to the correct state 
        // for latching SPI or PSP mode.
        #if defined(ENC100_INT_TRIS)    
            #if ENC100_INTERFACE_MODE == 0  // SPI
                ENC100_INT_IO = 1;
            #else   // PSP
                ENC100_INT_IO = 0;
            #endif
            ENC100_INT_TRIS = 0;
        #endif

        // Turn on power and wait for interface latching to occur
        ENC100_POR_IO = 1;
        vCurrentBank = 0;
        Delay10us(40);
        
        // Tri-state interrupt GPIO so that we don't cause bus contention
        #if defined(ENC100_INT_TRIS)
            ENC100_INT_TRIS = 1;
        #endif
    #endif
}    

    // Perform a reset via the SPI/PSP interface
    repeat
        // Set and clear a few bits that clears themselves upon reset.  
        // If EUDAST cannot be written to and your code gets stuck in this 
        // loop, you have a hardware problem of some sort (SPI or PMP not 
        // initialized correctly, I/O pins aren't connected or are 
        // shorted to something, power isn't available, etc.)
        repeat
            WriteReg(EUDAST, $1234)
        until (ReadReg(EUDAST) = $1234)

        // Issue a reset and wait for it to complete
        BFSReg(ECON2, ECON2_ETHRST)
        // after setting ETHRST, a min delay of 25us is req'd before accessing the 424J600
        delayus(50)
        vCurrentBank = 0
        repeat
            w = ReadReg(ESTAT) and (ESTAT_CLKRDY or ESTAT_RSTDONE or ESTAT_PHYRDY)
        until (w = (ESTAT_CLKRDY or ESTAT_RSTDONE or ESTAT_PHYRDY))

        // Check to see if the reset operation was successful by 
        // checking if EUDAST went back to its reset default.  This test 
        // should always pass, but certain special conditions might make 
        // this test fail, such as a PSP pin shorted to logic high.
    until (ReadReg(EUDAST) = 0)
    
    // Really ensure reset is done and give some time for power to be stable
    // wait at least 256us for the PHY registers to become available
    delayms(1)

    // v2.0: added check to wait for the PHY reset to complete by checking
    // PRST in PHCON1. the phy sets PRST = 0 when reset is completed
    repeat
        w = ReadPHYReg(PHCON1)
    until (w.bits(PHCON1_PRST_BIT) = 0)    // phy reset completed

    result = 0
end function

//
//---------------------------------------------------------------------------
//  Function:        MACGetAddress()
//  Overview:        reads enc mac address from the chip into AppConfig
//---------------------------------------------------------------------------
//
public sub MACGetAddress(byref v() as byte)
    dim w as word

    // use ENCx24J600 preprogrammed MAC address
    w = ReadReg(MAADR1)
    v(0) = w.byte0
    v(1) = w.byte1

    w = ReadReg(MAADR2)
    v(2) = w.byte0
    v(3) = w.byte1

    w = ReadReg(MAADR3)
    v(4) = w.byte0
    v(5) = w.byte1
end sub

//
//---------------------------------------------------------------------------
// MACInit()
// Output:
//      0 if mac present, -1 otherwise
// Overview:
//  MACInit sets up the registers in the ENCX24J600 so that normal operation can begin
//---------------------------------------------------------------------------
//
public function MACInit(eEthOpenFlags as byte=ETH_OPEN_DEFAULT) as shortint
    dim retries as byte

    dim
        phymode as word,    // PHCON1 setting
        phyana  as word     // PHANA setting

   // Set up the SPI module on the PIC for communications with the J600
    enc.Initialize()

    // clear rev to indicate that we haven't id'd the chip
    ENCRevID = 0

    // perform a reliable reset
    // v2.0: check for valid result after reset, and retry if invalid
    retries = 0
    while (true)
        result = SendSystemReset()
        if (result = 0) then
            result = -1
            ENCRevID = MACGetRevID()
            // check devID bits for ENC624J600 family
            if ((ENCRevID and $E0) = %00100000) then
                result = 0
            endif
        endif

        if (result = 0) then
            break
        else
            inc(retries)
            result = retries            // return number of retries
            if (retries > 10) then     // fail....
                exit
            endif
        endif
        delayus(100)
    end while

    // initialize RX tracking variables and other control state flags
    wNextPacketPointer = RXSTART
    wCurrentPacketPointer = $0000
    ENC100Flags.bWasDiscarded = true
    ENC100Flags.PoweredDown = false
    ENC100Flags.CryptoEnabled = false

    // set up TX/RX/UDA buffer addresses
    WriteReg(ETXST, TXSTART)
    WriteReg(ERXST, RXSTART)
    WriteReg(ERXTAIL, ENC100_RAM_SIZE-2)
    WriteReg(EUDAST, ENC100_RAM_SIZE)
    WriteReg(EUDAND, ENC100_RAM_SIZE+1)

    // use ENCx24J600 preprogrammed MAC address
    MACGetAddress(AppConfig.MyMACAddr.v)

    // configure LEDA to display LINK status, LEDB to display TX/RX activity
    WriteReg(EIDLED, $2600)

    // enable INT output for receive PKTIE and LINKIE
    // added v2.3 to enable addtl status bits to count errors
    WriteReg(EIE, EIE_INTIE or EIE_LINKIE or EIE_PKTIE) ' or EIE_TXABTIE or EIE_RXABTIE)

    //
    // normally, we set PHY Auto-negotiation to support 10BaseT Half duplex, 10BaseT Full duplex,
    // 100BaseTX Half Duplex, 100BaseTX Full Duplex
    // v2.0 - remove advertising symmetric PAUSE capability (PHANA_ADPAUS0), since
    // we don't currently set up the chip to support PAUSE
    // note: we set the PHCON1 bits in all modes... they will be ignored if autoneg is enabled
    //
    if (eEthOpenFlags.bits(ETH_OPEN_AUTO_BIT) = 1) then
        phymode = PHCON1_ANEN or PHCON1_RENEG       // enable autonegotiation, and restart neg
    else
        phymode = $0000                             // manual
    endif
    phyana  = PHANA_ADIEEE0     // default: ADPAUS[1:0]=00 (no PAUSE), ADIEEE[4:0]=00001 (802.3)
    if (eEthOpenFlags.bits(ETH_OPEN_100_BIT) = 1) then          // 100Mbit modes
        phymode.bits(PHCON1_SPD100_BIT) = 1
        if (eEthOpenFlags.bits(ETH_OPEN_FDUPLEX_BIT) = 1) then
            phyana.bits(PHANA_AD100FD_BIT) = 1
            phymode.bits(PHCON1_PFULDPX_BIT) = 1
        endif
        if (eEthOpenFlags.bits(ETH_OPEN_HDUPLEX_BIT) = 1) then
            phyana.bits(PHANA_AD100_BIT) = 1
        endif
    endif
    if (eEthOpenFlags.bits(ETH_OPEN_10_BIT) = 1) then           // 10Mbit modes
        phymode.bits(PHCON1_SPD100_BIT) = 0
        if (eEthOpenFlags.bits(ETH_OPEN_FDUPLEX_BIT) = 1) then
            phyana.bits(PHANA_AD10FD_BIT) = 1
            phymode.bits(PHCON1_PFULDPX_BIT) = 1
        endif
        if (eEthOpenFlags.bits(ETH_OPEN_HDUPLEX_BIT) = 1) then
            phyana.bits(PHANA_AD10_BIT) = 1
        endif
    endif
    // loopback mode support
    if (eEthOpenFlags.bits(ETH_OPEN_MAC_LOOPBACK_BIT) = 1) then
        BFSReg(MACON1, MACON1_LOOPBK)
        'phymode.bits(PHCON1_PSLEEP_BIT) = 1        // shut down phy for mac test
    elseif (eEthOpenFlags.bits(ETH_OPEN_PHY_LOOPBACK_BIT) = 1) then
        phymode = phymode or PHCON1_PLOOPBK
    endif

    // set PHY autonegotiation and link modes
    WritePHYReg(PHANA, phyana)
    WritePHYReg(PHCON1, phymode)

    // if not using autonegotiation, set mac duplex to match phy
    if (eEthOpenFlags.bits(ETH_OPEN_AUTO_BIT) = 0) then
        delayus(250)             // give the phy a chance to update
        MACUpdateDuplexStatus()
    endif

    // enable RX packet reception
    BFSReg(ECON1, ECON1_RXEN)
end function

//
//---------------------------------------------------------------------------
// module initialize
//---------------------------------------------------------------------------
//
// resolve "not initialized"
ENCRevID = 0
ENC100Flags.b = 0
wCurrentPacketPointer = 0
wNextPacketPointer = 0
vCurrentBank = 0

end module
