{
*****************************************************************************
*  Name    : IP.BAS                                                         *
*  Date    : 30/08/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module IP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "MAC.bas"

Public Const
   IP_PROT_ICMP  =  (1),
   IP_PROT_TCP   =  (6),
   IP_PROT_UDP   =  (17)
   
// IP packet header definition
Public Structure IP_HEADER
   VersionIHL As Byte
   TypeOfService As Byte
   TotalLength As Word
   Identification As Word
   FragmentInfo As Word
   TimeToLive As Byte
   Protocol As Byte
   HeaderChecksum As Word
   SourceAddress As IP_ADDR
   DestAddress As IP_ADDR
End Structure

(*
 *********************************************************************
 * Function:        BOOL IPIsTxReady(BOOL HighPriority)
 * PreCondition:    None
 * Input:           None
 * Output:          TRUE if transmit buffer is empty
 *                  FALSE if transmit buffer is not empty
 * Side Effects:    None
 * Note:            None
 ********************************************************************/
*)
Public Dim IPIsTxReady As MAC.MACIsTxReady
(*
 *********************************************************************
 * Macro:           IPSetTxBuffer(a, b)
 * PreCondition:    None
 * Input:           a       - Buffer identifier
 *                  b       - Offset
 * Output:          Next Read/Write access to transmit buffer 'a'
 *                  set to offset 'b'
 * Side Effects:    None
 * Note:            None
 ********************************************************************/
*)
Public Function IPSetTxBuffer(b As Word) As Word
   result = MAC.MACSetWritePtr(b + BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER))
End Function   
(*
 *********************************************************************
 * Macro:           IPPutArray(a, b)
 * PreCondition:    IPIsTxReady() == TRUE
 * Input:           a       - Data buffer
 *                  b       - Buffer length
 * Output:          None
 * Side Effects:    None
 * Note:            Data is copied To IP data area.
 ********************************************************************/
*)
Public Dim IPPutArray As MAC.MACPutArray
(*
 *********************************************************************
 * Macro:           IPDiscard()
 * PreCondition:    MACGetHeader() == TRUE
 * Input:           None
 * Output:          Current packet is discarded And buffer is
 *                  freed-up
 * Side Effects:    None
 * Note:            None
 ********************************************************************/
*)
Public Dim IPDiscard As MAC.MACDiscardRx
(*
 *********************************************************************
 * Macro:           IPGetArray(a, b)
 * PreCondition:    MACGetHeader() == TRUE
 * Input:           a       - Data buffer
 *                  b       - Buffer length
 * Output:          None
 * Side Effects:    None
 * Note:            Data is copied from IP data To given buffer
 ********************************************************************/
*)
Public Dim IPGetArray As MAC.MACGetArray

// This is left shifted by 4.  Actual value is 0x04.
Const
   IPv4                     = ($40),
   IP_VERSION               = IPv4

// IHL (Internet Header Length) is # of DWORDs in a header.
// Since, we do not support options, our IP header length will be
// minimum i.e. 20 bytes : IHL = 20 / 4 = 5.
Const
   IP_IHL                   = ($05),
   IP_SERVICE_NW_CTRL       = ($07),
   IP_SERVICE_IN_CTRL       = ($06),
   IP_SERVICE_ECP           = ($05),
   IP_SERVICE_OVR           = ($04),
   IP_SERVICE_FLASH         = ($03),
   IP_SERVICE_IMM           = ($02),
   IP_SERVICE_PRIOR         = ($01),
   IP_SERVICE_ROUTINE       = ($00),

   IP_SERVICE_N_DELAY       = ($00),
   IP_SERCICE_L_DELAY       = ($08),
   IP_SERVICE_N_THRPT       = ($00),
   IP_SERVICE_H_THRPT       = ($10),
   IP_SERVICE_N_RELIB       = ($00),
   IP_SERVICE_H_RELIB       = ($20),

   IP_SERVICE               = (IP_SERVICE_ROUTINE Or IP_SERVICE_N_DELAY),
   MY_IP_TTL                = (100)  // Time-To-Live in hops

//module variables...
Dim
   _Identifier As Word,
   IPHeaderLen As Byte

Sub SwapIPHeader(ByRef h As IP_HEADER)
   h.TotalLength      = NETUtils.swaps(h.TotalLength)
   h.Identification   = NETUtils.swaps(h.Identification)
   h.HeaderChecksum   = NETUtils.swaps(h.HeaderChecksum)
End Sub
(*
 *********************************************************************
 * Function:        BOOL IPGetHeader( IP_ADDR    *localIP,
 *                                    NODE_INFO  *remote,
 *                                    Byte        *Protocol,
 *                                    Word        *len)
 * PreCondition:    MACGetHeader() == TRUE
 * Input:           localIP     - Local node IP Address As received
 *                                in current IP header.
 *                                If this information is Not required
 *                                caller may pass NULL value.
 *                  remote      - Remote node info
 *                  Protocol    - Current packet Protocol
 *                  len         - Current packet data length
 * Output:          TRUE, If valid packet was received
 *                  FALSE otherwise
 * Side Effects:    None
 * Note:            Only one IP message can be received.
 *                  Caller may Not transmit And receive a message
 *                  at the same time.
 ********************************************************************/
*)
Public Function IPGetHeader(ByRef localIP As IP_ADDR, ByRef remote As NODE_INFO, ByRef Protocol As Byte, ByRef len As Word) As Boolean
   Dim CalcChecksum As WORD_VAL
   Dim header As IP_HEADER

   // Read IP header.
   MACGetArray(@header, SizeOf(header))

   // Make sure that this is an IPv4 packet.
   If  (header.VersionIHL And $f0) <> IP_VERSION Then
      result = false
      Exit
   EndIf

   // Throw this packet away if it is a fragment.  
   // We don't have enough RAM for IP fragment reconstruction.
   If (header.FragmentInfo And $FF1F) <> 0 Then
      result = false
      Exit
   EndIf
		
   IPHeaderLen = (header.VersionIHL And $0f) << 2

   // Validate the IP header.  If it is correct, the checksum 
   // will come out to 0x0000 (because the header contains a 
   // precomputed checksum).  A corrupt header will have a 
   // nonzero checksum.
   CalcChecksum.Val = MACCalcRxChecksum(0, IPHeaderLen)

   // Seek to the end of the IP header
   MACSetReadPtrInRx(IPHeaderLen)

   If CalcChecksum.Val <> 0 Then
      // Bad packet. The function caller will be notified by means of the FALSE 
      // return value and it should discard the packet.
      result = false
      Exit
   EndIf

   // Network to host conversion.
   SwapIPHeader(header)

   // If caller is intrested, return destination IP address
   // as seen in this IP header.
   If @localIP <> 0 Then
      localIP.Val = header.DestAddress.Val
   EndIf  

   remote.IPAddr.Val  = header.SourceAddress.Val
   Protocol           = header.Protocol
   len 				  = header.TotalLength - IPHeaderLen

   result = true
End Function
(*
 *********************************************************************
 * Function: Word IPPutHeader(NODE_INFO *remote,
 *           				  Byte Protocol,
 *                			  Word len)
 * PreCondition:    IPIsTxReady() == TRUE
 * Input:           *remote     - Destination node address
 *                  Protocol    - Current packet Protocol
 *                  len         - Current packet data length
 * Output:          (Word)0
 * Side Effects:    None
 * Note:            Only one IP message can be transmitted at any
 *                  time.
 ********************************************************************/
*)
Public Function IPPutHeader(ByRef remote As NODE_INFO, Protocol As Byte, len As Word) As Word
   Dim header As IP_HEADER
   IPHeaderLen = SizeOf(IP_HEADER)

   Inc(_Identifier)
   header.VersionIHL       = IP_VERSION Or IP_IHL
   header.TypeOfService    = IP_SERVICE
   header.TotalLength      = SizeOf(header) + len
   header.Identification   = _Identifier
   header.FragmentInfo     = 0
   header.TimeToLive       = MY_IP_TTL
   header.Protocol         = Protocol
   header.HeaderChecksum   = 0
   header.SourceAddress    = AppConfig.MyIPAddr

   header.DestAddress.Val  = remote.IPAddr.Val
   SwapIPHeader(header)
   header.HeaderChecksum   = NETUtils.CalcIPChecksum(@header, SizeOf(header))

   MACPutHeader(remote.MACAddr, MAC_IP, (SizeOf(header) + len))
   MACPutArray(@header, SizeOf(header))
   result = $0000
End Function
(*
 *********************************************************************
 * Function:        IPSetRxBuffer(Word Offset)
 * PreCondition:    IPHeaderLen must have been intialized by 
 *					IPGetHeader() Or IPPutHeader()
 * Input:           Offset from beginning of IP data field
 * Output:          Next Read/Write Access To receive buffer is
 *                  set To Offset 
 * Side Effects:    None
 * Note:            None
 ********************************************************************/
*)
Public Sub IPSetRxBuffer(Offset As Word) 
   MACSetReadPtrInRx(Offset + IPHeaderLen)
End Sub

// initialise module...
_Identifier = 0




