{
*****************************************************************************
*  Name    : NBNS.BAS                                                       *
*  Date    : 03/09/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module NBNS
 
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "UDP.bas"

Const
   NBNS_PORT    = (137)

Structure NBNS_HEADER
   TransactionID As WORD_VAL
   Flags As WORD_VAL
   Questions As WORD_VAL
   Answers As WORD_VAL
   AuthoritativeRecords As WORD_VAL
   AdditionalRecords As WORD_VAL
End Structure

// statics...
Const
   NBNS_HOME = 0,
   NBNS_OPEN_SOCKET = 1,
   NBNS_LISTEN = 2
Dim 
   smNBNS As Byte,
   MySocket As UDP_SOCKET
(*
 *********************************************************************
 * Function:        void NBNSPutName(BYTE *String)
 ********************************************************************/
*) 
Sub NBNSPutName(ByRef Str() As Byte)
   Dim i, j As Byte

   UDPPut(32)	// NetBIOS names are always 32 bytes long (16 decoded bytes)
   i = 0
   FSR2 = @Str
   While i < 16
      j = POSTINC2
	  UDPPut((j >> 4) + Byte("A"))
	  UDPPut((j And $0F) + Byte("A"))
      Inc(i)
   Wend
   UDPPut($00)
End Sub
(*
 *********************************************************************
 * Function:        void NBNSGetName(BYTE *String)
 ********************************************************************/
*) 
Sub NBNSGetName(ByRef Str() As Byte)
   
   Dim i, j, k As Byte

   FSR2 = @Str
   UDPGet(i)
   
   If i > 32 Then
      INDF0 = 0
      Exit
   EndIf

   While i > 0
      Dec(i)
	  UDPGet(j)
	  j = j - Byte("A")
	  k = j << 4
	  Dec(i)
	  UDPGet(j)
	  j = j - Byte("A")

	  POSTINC2 = k Or j
   Wend
End Sub

(*
 *********************************************************************
 * Function:        void NBNSTask(void)
 * PreCondition:    None
 * Input:           None
 * Output:          Sends responses to NetBIOS name requests
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Sub NBNSTask()
   Dim i As Byte
   Dim _Type, Class As WORD_VAL
   Dim NBNSHeader As NBNS_HEADER
   Dim NameString(16) As Byte

   Select smNBNS
      Case NBNS_HOME:
	     Inc(smNBNS)
 
      Case NBNS_OPEN_SOCKET:
	     MySocket = UDPOpen(NBNS_PORT, 0, NBNS_PORT)
		 If MySocket = INVALID_UDP_SOCKET Then
            Exit
         EndIf
		 Inc(smNBNS)

      Case NBNS_LISTEN:
         If UDPIsGetReady(MySocket) = 0 Then
            Exit
         EndIf
         
         // Respond only to name requests sent to us from nodes on the same subnet
		 // This prevents us from sending out the wrong IP address information if 
		 // we haven't gotten a DHCP lease yet.
         If (remoteNode.IPAddr.Val And AppConfig.MyMask.Val) <> (AppConfig.MyIPAddr.Val And AppConfig.MyMask.Val) Then
            UDPDiscard()
            Exit
         EndIf			


		 // Retrieve the NBNS header and de-big-endian it
		 UDPGet(NBNSHeader.TransactionID.v(1))
		 UDPGet(NBNSHeader.TransactionID.v(0))
		 UDPGet(NBNSHeader.Flags.v(1))
		 UDPGet(NBNSHeader.Flags.v(0))
		 UDPGet(NBNSHeader.Questions.v(1))
		 UDPGet(NBNSHeader.Questions.v(0))
		 UDPGet(NBNSHeader.Answers.v(1))
		 UDPGet(NBNSHeader.Answers.v(0))
		 UDPGet(NBNSHeader.AuthoritativeRecords.v(1))
		 UDPGet(NBNSHeader.AuthoritativeRecords.v(0))
		 UDPGet(NBNSHeader.AdditionalRecords.v(1))
		 UDPGet(NBNSHeader.AdditionalRecords.v(0))

		 // Remove all questions
		 While NBNSHeader.Questions.Val <> 0
		    NBNSHeader.Questions.Val = NBNSHeader.Questions.Val - 1
			NBNSGetName(NameString)
			UDPGet(i)				// <??> Trailing character on string
			UDPGet(_Type.v(1))		// Question type
			UDPGet(_Type.v(0))
			UDPGet(Class.v(1))	// Question class
			UDPGet(Class.v(0))
				           
			If (_Type.Val = $0020) And (Class.Val = $0001) And strequal(NameString,AppConfig.NetBIOSName) Then
  			   While UDPIsPutReady(MySocket) = 0
               Wend
			   NBNSHeader.Flags.Val = $8400

               UDPPut(NBNSHeader.TransactionID.v(1))
			   UDPPut(NBNSHeader.TransactionID.v(0))
			   UDPPut(NBNSHeader.Flags.v(1))
			   UDPPut(NBNSHeader.Flags.v(0))
			   UDPPut($00)	// 0x0000 Questions
			   UDPPut($00)
			   UDPPut($00)	// 0x0001 Answers
			   UDPPut($01)
			   UDPPut($00)	// 0x0000 Athoritative records
			   UDPPut($00)
			   UDPPut($00)	// 0x0000 Additional records
			   UDPPut($00)

               NBNSPutName(AppConfig.NetBIOSName)
               UDPPut($00)	// 0x0020 Type: NetBIOS
			   UDPPut($20)
               UDPPut($00)	// 0x0001 Class: Internet
			   UDPPut($01)
			   UDPPut($00)	// 0x00000000 Time To Live
			   UDPPut($00)
			   UDPPut($00)
			   UDPPut($00)

			   UDPPut($00)	// 0x0006 Data length
			   UDPPut($06)	
			   UDPPut($60)	// 0x6000 Flags: H-node, Unique
			   UDPPut($00)
			   UDPPut(AppConfig.MyIPAddr.v(0))	// Put out IP address
               UDPPut(AppConfig.MyIPAddr.v(1))
			   UDPPut(AppConfig.MyIPAddr.v(2))
			   UDPPut(AppConfig.MyIPAddr.v(3))
			   
               // Change the destination address to the unicast address of the last received packet
		       UDPSocketInfo(MySocket).remoteNode = remoteNode

			   UDPFlush()
            EndIf
         Wend   
         UDPDiscard()
   End Select
End Sub

// initialise...
#if defined(STACK_USE_NBNS)
smNBNS = NBNS_HOME
#endif

