{
*****************************************************************************
*  Name    : NETConfig.BAS                                                  *
*  Author  : David John Barker                                              *
*  Notice  : Copyright (c) 2007 Mecanique                                   *
*          : All Rights Reserved                                            *
*  Date    : 28/08/2007                                                     *
*  Version : 1.0                                                            *
*  Notes   :                                                                *
*          :                                                                *
*****************************************************************************
}
Module NETConfig

#option NET_TCP = false
#option NET_UDP = false
#option NET_ICMP = false
#option NET_DNS = false
#option NET_DHCP = false
#option NET_SMTP = false
#option NET_NBNS = false
#option NET_SNTP = false

(*
   this may seem a little convoluted, but the Microchip stack uses defines
   to direct compilation, but if new layers are added to Swordfish it would
   be better to use options. To preserve Microchip naming conventions throughout
   the TCPIP stack, use options to switch on defines...
*)

// TCP...
#if NET_TCP         
#define STACK_USE_TCP
#endif

// UDP...
#if NET_UDP
#define STACK_USE_UDP 
#endif

// Ping (ICMP)
#if NET_ICMP
#define STACK_USE_ICMP 
#endif

// Domain Name Service (DNS)
#if NET_DNS
#define STACK_USE_DNS	
#endif

// Dynamic Host Configuration Protocol (DHCP)
#if NET_DHCP
#define STACK_USE_DHCP_CLIENT           
#endif

// Simple Mail Transfer Protocol (SMTP) 
#if NET_SMTP
#define STACK_USE_SMTP_CLIENT 
#endif

// Net BIOS Name Service (NBNS)
#if NET_NBNS          
#define STACK_USE_NBNS					
#endif

// Simple Network Time Protocol (SNTP)
#if NET_SNTP
#define STACK_USE_SNTP                  
#endif

' ***************** #define STACK_USE_HTTP_SERVER
' ***************** #define STACK_USE_MPFS
' ***************** #define STACK_USE_DHCP_SERVER
' ***************** #define STACK_USE_FTP_SERVER
//#define STACK_USE_SNMP_SERVER
//#define STACK_USE_TFTP_CLIENT
' ***************** #define STACK_USE_TELNET_SERVER			// Telnet server
' ***************** #define STACK_USE_ANNOUNCE				// Microchip Embedded Ethernet Device Discoverer server/client

//#define STACK_USE_REBOOT_SERVER			// Module for resetting this PIC remotely.  Primarily useful for a Bootloader.

(*
 * If html pages are stored in internal program memory,
 * comment MPFS_USE_EEPROM and include an MPFS image (.c file) 
 * in the project.  If html pages are stored in external EEPROM 
 * memory, uncomment MPFS_USE_EEPROM
*)
#define MPFS_USE_EEPROM

(*
 * Number of bytes to be reserved before MPFS storage is to start.
 *
 * These bytes host application configurations such as IP Address,
 * MAC Address, and any other required variables.
 *
 * After making any change to this variable, MPFS.exe must be
 * executed with correct block size.
 * See MPFS.exe help message by executing MPFS /?
*)
#define _MPFS_RESERVE_BLOCK              = (64)
Public Const MPFS_RESERVE_BLOCK = _MPFS_RESERVE_BLOCK

(*
 * Maximum number of simultaneously open MPFS2 files.
*)
#define _MAX_MPFS_HANDLES				= (7)
Public Const MAX_MPFS_HANDLES = _MAX_MPFS_HANDLES

// When MPFS is stored in internal program memory (instead of external EEPROM), FTP is not supported.
'#if Not defined(MPFS_USE_EEPROM)
'	#undefine STACK_USE_FTP_SERVER
'#endif

#if defined(STACK_USE_FTP_SERVER) Or defined(STACK_USE_HTTP_SERVER)
	#undefine STACK_USE_MPFS
    #define STACK_USE_MPFS
#endif

// Comment following line if SNMP TRAP support is needed
//#define SNMP_TRAP_DISABLED

// Make sure that STACK_USE_DNS is defined if a service 
// depends on it
#if defined(STACK_USE_SMTP_CLIENT) Or defined(STACK_USE_SNTP)
   #if Not defined(STACK_USE_DNS)
      #define STACK_USE_DNS
   #endif
#endif      

// Make sure that STACK_CLIENT_MODE is defined if a service 
// depends on it
#if defined(STACK_USE_FTP_SERVER) Or _
	defined(STACK_USE_SNMP_SERVER) Or _
	defined(STACK_USE_DNS) Or _
	defined(STACK_USE_GENERIC_TCP_CLIENT_EXAMPLE) Or _
	defined(STACK_USE_TFTP_CLIENT) Or _
	defined(STACK_USE_SMTP_CLIENT) Or _
	defined(STACK_USE_ICMP) 
	#if Not defined(STACK_CLIENT_MODE)
	    #define STACK_CLIENT_MODE
	#endif
#endif

// Make sure that STACK_USE_TCP is defined if a service depends on 
// it
#if defined(STACK_USE_HTTP_SERVER) Or _
	defined(STACK_USE_FTP_SERVER) Or _
	defined(STACK_USE_TELNET_SERVER) Or _
	defined(STACK_USE_SMTP_CLIENT) 
    #if Not defined(STACK_USE_TCP)
        #define STACK_USE_TCP
    #endif
#endif

// Make sure that STACK_USE_UDP is defined if a service depends 
// on it
#if defined(STACK_USE_DHCP_CLIENT) Or _
	defined(STACK_USE_DNS) Or _
	defined(STACK_USE_NBNS) Or _
	defined(STACK_USE_SNMP_SERVER) Or _
	defined(STACK_USE_TFTP_CLIENT) Or _
	defined(STACK_USE_ANNOUNCE) Or _
	defined(STACK_USE_SNTP)
    #if Not defined(STACK_USE_UDP)
        #define STACK_USE_UDP
    #endif
#endif

(*
 * TCP configurations
 * To minmize page update, match number of sockets And
 * HTTP connections With different page sources in a
 * page.
 * For example, If page contains reference To 3 more pages,
 * browser may try To open 4 simultaneous HTTP connections,
 * And To minimize browser delay, set HTTP connections To
 * 4, MAX_TCP_SOCKETS To 4.
 * If you are using other applications, you should
 * keep at least one socket available For them.
*)

// Maximum TCP sockets to be defined.
// Note that each TCP socket consumes 27 bytes of RAM (sizeof(TCB_STUB)), 
// and a goodly chunk of Ethernet buffer RAM for each socket
#define _MAX_TCP_SOCKETS            = (8)  
#define _TCP_TX_FIFO_SIZE	        = (200)
#define _TCP_RX_FIFO_SIZE	        = (200)
#if defined(STACK_USE_TCP)
   #ifndef RESERVED_TCP_MEMORY
      #define RESERVED_TCP_MEMORY	    = _MAX_TCP_SOCKETS * ((_TCP_TX_FIFO_SIZE + 1) + (_TCP_RX_FIFO_SIZE + 1) + 36) //sizeof(TCB)))
   #endif
#endif

Public Const
   MAX_TCP_SOCKETS                  = _MAX_TCP_SOCKETS,
   TCP_TX_FIFO_SIZE	                = _TCP_TX_FIFO_SIZE,
   TCP_RX_FIFO_SIZE	                = _TCP_RX_FIFO_SIZE

// Maximum avaialble UDP Sockets
#define _MAX_UDP_SOCKETS            = (6)
Public Const MAX_UDP_SOCKETS = _MAX_UDP_SOCKETS

// 
// HTTP2 Server options
//

// Maximum numbers of simultaneous HTTP connections allowed.
// Each connection consumes 10 bytes and a TCP socket
#define _MAX_HTTP_CONNECTIONS	   = (3)
Public Const MAX_HTTP_CONNECTIONS = _MAX_HTTP_CONNECTIONS

// Indicate what file to serve when no specific one is requested
#define _HTTP_DEFAULT_FILE		    = "index.htm"
#define _HTTPS_DEFAULT_FILE		    = "index.htm"
#define _HTTP_DEFAULT_LEN		    = 10 //for buffer overrun protection
										 //set to longest length of above two strings
Public Const
   HTTP_DEFAULT_FILE = _HTTP_DEFAULT_FILE,
   HTTPS_DEFAULT_FILE = _HTTPS_DEFAULT_FILE,
   HTTP_DEFAULT_LEN = _HTTP_DEFAULT_LEN

// Configure MPFS over HTTP updating
// Comment this line to disable updating via HTTP
#define _HTTP_MPFS_UPLOAD	        = "mpfsupload"
Public Const HTTP_MPFS_UPLOAD = _HTTP_MPFS_UPLOAD
 
//#define HTTP_MPFS_UPLOAD_REQUIRES_AUTH	//Uncomment to require auth on MPFS uploads
										// Certain firewall and router combinations cause
										// the MPFS2 Utility to fail when uploading.  If
										// this happens, comment out this definition.

// Define which HTTP modules to use
#define HTTP_USE_POST					// Enable POST support (~0.5k ROM)
#define HTTP_USE_COOKIES				// Enable cookie support (~0.9k ROM)
#define HTTP_USE_AUTHENTICATION			// Enable basic authentication support (~0.9k ROM)

//#define HTTP_NO_AUTH_WITHOUT_SSL		// Uncomment to require SSL before requesting a password
#define _HTTP_SSL_ONLY_CHAR = $ff	    // Files beginning with this character will only be served over HTTPS
										// Set to NUL to require for all files
										// Set to 0xff to require for no files
Public Const HTTP_SSL_ONLY_CHAR = _HTTP_SSL_ONLY_CHAR


										


