{
*****************************************************************************
*  Name    : SMTP.BAS                                                       *
*  Date    : 10/09/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module SMTP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "ARP.bas"
Include "TCP.bas"
Include "DNS.bas"
Include "Tick.bas"

Public Const
   SMTP_SUCCESS	        = ($0000),
   SMTP_RESOLVE_ERROR	= ($8000),
   SMTP_CONNECT_ERROR	= ($8001)

Structure TPointer
   szRAM As Word
   szROM As TROMAddress Union
End Structure
   
Structure TROMPointers
   Server As Boolean
   Username As Boolean
   Password As Boolean
   SendTo As Boolean
   CC As Boolean
   BCC As Boolean
   From As Boolean
   Subject As Boolean
   OtherHeaders As Boolean
   Body As Boolean
End Structure

Structure SMTP_POINTERS
   Server As TPointer 
   Username As TPointer
   Password As TPointer
   SendTo As TPointer
   CC As TPointer
   BCC As TPointer
   From As TPointer
   Subject As TPointer
   OtherHeaders As TPointer
   Body As TPointer
   ROMPointers As TROMPointers
End Structure

Public Dim
   SMTPClient As SMTP_POINTERS

Const
   SMTP_PORT = 25

Dim
   SMTPServer As NODE_INFO,
   MySocket As TCP_SOCKET

Const
   CR_PERIOD_SEEK_CR = 0,
   CR_PERIOD_SEEK_LF = 1,
   CR_PERIOD_SEEK_PERIOD = 2,
   CR_PERIOD_NEED_INSERTION = 3
   
Structure TCRPeriod
   Pos As Word
   State As Byte Union
End Structure
Dim CRPeriod As TCRPeriod

Const
   TRANSPORT_HOME = 0,
   TRANSPORT_BEGIN = 1,
   TRANSPORT_NAME_RESOLVE = 2,
   TRANSPORT_ARP_START_RESOLVE = 3,
   TRANSPORT_ARP_IN_PROGRESS = 4,
   TRANSPORT_ARP_RESOLVED = 5,
   TRANSPORT_SOCKET_OBTAINED = 6,
   TRANSPORT_CLOSE = 7
Dim TransportState As Byte   

Public Const
   SMTP_HOME = 0,
   SMTP_HELO = 1,
   SMTP_HELO_ACK = 2,	
   SMTP_AUTH_LOGIN = 3,
   SMTP_AUTH_LOGIN_ACK = 4,
   SMTP_AUTH_USERNAME = 5,
   SMTP_AUTH_USERNAME_ACK = 6,
   SMTP_AUTH_PASSWORD = 7,
   SMTP_AUTH_PASSWORD_ACK = 8,
   SMTP_MAILFROM = 9,
   SMTP_MAILFROM_ACK = 10,
   SMTP_RCPTTO_INIT = 11,
   SMTP_RCPTTO = 12,
   SMTP_RCPTTO_ACK = 13,
   SMTP_RCPTTO_ISDONE = 14,
   SMTP_RCPTTOCC_INIT = 15,
   SMTP_RCPTTOCC = 16,
   SMTP_RCPTTOCC_ACK = 17,
   SMTP_RCPTTOCC_ISDONE = 18,
   SMTP_RCPTTOBCC_INIT = 19,
   SMTP_RCPTTOBCC = 20,
   SMTP_RCPTTOBCC_ACK = 21,
   SMTP_RCPTTOBCC_ISDONE = 22,
   SMTP_DATA = 23,
   SMTP_DATA_ACK = 24,
   SMTP_DATA_HEADER = 25,
   SMTP_DATA_BODY_INIT= 26,
   SMTP_DATA_BODY = 27,
   SMTP_DATA_BODY_ACK = 28,
   SMTP_QUIT_INIT = 29,
   SMTP_QUIT = 30
Public Dim SMTPState As Byte

Const
   PUTHEADERS_FROM_INIT = 0,
   PUTHEADERS_FROM = 1,
   PUTHEADERS_TO_INIT = 2,
   PUTHEADERS_TO = 3,
   PUTHEADERS_CC_INIT = 4,
   PUTHEADERS_CC = 5,
   PUTHEADERS_SUBJECT_INIT = 6,
   PUTHEADERS_SUBJECT = 7,
   PUTHEADERS_OTHER_INIT = 8,
   PUTHEADERS_OTHER = 9,
   PUTHEADERS_DONE = 10
Dim PutHeadersState As Byte

Const
   RX_BYTE_0 = 0,
   RX_BYTE_1 = 1,
   RX_BYTE_2 = 2,
   RX_BYTE_3 = 3,
   RX_SEEK_CR = 4,
   RX_SEEK_LF = 5
Dim RXParserState As Byte

Structure TSMTPBits
   RXSkipResponse As Boolean
   SMTPInUse As Boolean
   SentSuccessfully As Boolean
   ReadyToStart As Boolean
   ReadyToFinish As Boolean
   ConnectedOnce As Boolean
End Structure

Structure TSMTPFlags
   Val As Byte
   Bits As TSMTPBits Union
End Structure
Dim SMTPFlags As TSMTPFlags
Dim ResponseCode As Word

Const
   LF = #10,
   CR = #13,
   CRLF = CR + LF
   
(*
 *********************************************************************
 * Function:        BOOL SMTPBeginUsage(void)
 * PreCondition:    Stack is initialized()
 * Input:           None
 * Output:          TRUE: If no SMTP operations are in progress and this application
 *                  has successfully taken ownership of the SMTP module
 *					FALSE: If the SMTP module is currently being used by some other 
 *                  application.  Call SMTPBeginUsage() some time later (after returning 
 *                  to the main() program loop).
 * Side Effects:    None
 * Overview:        Call DNSBeginUsage() and make sure it returns TRUE before calling 
 *                  any DNS APIs.  Call DNSEndUsage() when this application no longer needs 
 *                  the DNS module so that other applications may make use of it.
 * Note:            None
 ********************************************************************/
*)
Public Function SMTPBeginUsage() As Boolean
   If SMTPFlags.Bits.SMTPInUse Then
      result = false
      Exit
   EndIf

   SMTPFlags.Val = $00
   SMTPFlags.Bits.SMTPInUse = true
   TransportState = TRANSPORT_BEGIN
   RXParserState = RX_BYTE_0
   SMTPState = SMTP_HOME
   memset(@SMTPClient, $00, SizeOf(SMTPClient))
   result = true
End Function
(*
 *********************************************************************
 * Function:        WORD SMTPEndUsage(void)
 * PreCondition:    SMTPBeginUsage() returned TRUE on a previous call.
 * Input:           None
 * Output:          SMTP_SUCCESS (0x0000): on success
 *					SMTP_RESOLVE_ERROR (0x8000): if unable to resolve SMTP server
 *					SMTP_CONNECT_ERROR (0x8001): if unable to connect to SMTP server 
 *                  or connection lost
 *					1-199 and 300-999: last SMTP server response code
 * Side Effects:    None
 * Overview:        Releases ownership of the SMTP module caused by a 
 *					previous call to SMTPBeginUsage()
 * Note:            None
 ********************************************************************/
*)
Public Function SMTPEndUsage() As Word
   If Not SMTPFlags.Bits.SMTPInUse Then
      result = 0
	  Exit
   EndIf

   // Release the DNS module, if in use
   If TransportState = TRANSPORT_NAME_RESOLVE Then
      DNSEndUsage()
   EndIf   
	
   // Release the TCP socket, if in use
   If MySocket <> INVALID_SOCKET Then
      TCPDisconnect(MySocket)
      MySocket = INVALID_SOCKET
   EndIf

   // Release the SMTP module
   SMTPFlags.Bits.SMTPInUse = false
   TransportState = TRANSPORT_HOME

   If SMTPFlags.Bits.SentSuccessfully Then
      result = 0
   Else
      result = ResponseCode
   EndIf
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Public Inline Sub SMTPSendMail()
   SMTPFlags.Bits.ReadyToStart = true
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Public Function SMTPIsBusy() As Boolean
   result = TransportState <> TRANSPORT_HOME
End Function   
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function SMTPIsPutReady() As Word
   If SMTPState <> SMTP_DATA_BODY Then
      result = 0
      Exit
   EndIf
   result = TCPIsPutReady(MySocket)
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function SMTPPut(c As Byte) As Boolean
   If CRPeriod.State = CR_PERIOD_NEED_INSERTION Then
      If TCPPut(MySocket, ".") Then
         CRPeriod.State = CR_PERIOD_SEEK_CR
      Else
	     result = false
         Exit
      EndIf
   EndIf

   Select CRPeriod.State
      Case CR_PERIOD_SEEK_CR:
	     If c = 13 Then
            Inc(CRPeriod.State)
         EndIf
      Case CR_PERIOD_SEEK_LF:
		 If c = 10 Then
            Inc(CRPeriod.State)
         ElseIf c <> 13 Then
            Dec(CRPeriod.State)
         EndIf
      Case CR_PERIOD_SEEK_PERIOD:
	     If c = Byte(".") Then
		    Inc(CRPeriod.State)	// CR_PERIOD_NEED_INSERTION
         ElseIf c = 13 Then
            Dec(CRPeriod.State)
		 Else
		    CRPeriod.State = CR_PERIOD_SEEK_CR
		 EndIf
   End Select
   If Not TCPPut(MySocket, c) Then
      result = false
   Else
      result = true
   EndIf
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function SMTPPutArray(Data As Word, Len As Word) As Word
   
   result = 0

   // Must use SMTPPut() instead of TCPPutArray because of transparancy replacements of "\r\n." with "\r\n.."
   While Len <> 0
      FSR0 = Data
      Inc(Data)
      If SMTPPut(INDF0) Then
         Inc(result)
      Else
         Dec(Data)
         Break
      EndIf
      Dec(Len)
   Wend
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function SMTPPutROMArray(Data As TROMAddress, Len As Word) As Word
   Dim Val As Byte
   result = 0
   
   
   // Must use SMTPPut() instead of TCPPutArray because of transparancy replacements of "\r\n." with "\r\n.."
   While Len <> 0
      Val = ReadROM(Data)
      Inc(Data)
      If SMTPPut(Val) Then
         Inc(result)
      Else
         Dec(Data)
         Break
      EndIf
      Dec(Len)
   Wend
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function SMTPPutString(Data As Word) As Word
   result = 0

   // Must use SMTPPut() instead of TCPPutArray because of transparancy replacements of "\r\n." with "\r\n.."
   FSR0 = Data
   While INDF0 <> 0
      Inc(Data)
      If SMTPPut(INDF0) Then
	     Inc(result)
      Else
         Dec(Data)
         Break
      EndIf
      FSR0 = Data
   Wend
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function SMTPPutROMString(Data As TROMAddress) As Word
   Dim Val As Byte
   result = 0

   // Must use SMTPPut() instead of TCPPutArray because of transparancy replacements of "\r\n." with "\r\n.."
   Val = ReadROM(Data)
   While Val <> 0
      Inc(Data)
      If SMTPPut(INDF0) Then
	     Inc(result)
      Else
         Dec(Data)
         Break
      EndIf
      Val = ReadROM(Data)
   Wend
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Sub SMTPFlush()
   TCPFlush(MySocket)
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Sub SMTPPutDone()
   SMTPFlags.Bits.ReadyToFinish = true
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Structure TBits
   FoundOpenBracket As Boolean
   FoundAt As Boolean
End Structure
Structure TParseStates
   Val As Byte
   Bits As TBits Union
End Structure    
Function FindEmailAddress(Str As Word, ByRef wLen As Word) As Word
   Dim lpStart As Word
   Dim c As Byte
   Dim ParseStates As TParseStates

   lpStart = Str
   wLen = $0000
   ParseStates.Val = $00

   FSR0 = Str
   c = INDF0
   While c <> 0
      Inc(Str)
      If c = Byte("<") Then
         ParseStates.Bits.FoundOpenBracket = true
         lpStart = Str
		 wLen = Word(-1)
      ElseIf c = Byte("@") Then
         ParseStates.Bits.FoundAt = true
      EndIf   

      If Not ParseStates.Bits.FoundOpenBracket And Not ParseStates.Bits.FoundAt And (c = Byte(" ") Or c = Byte(",")) Then
         lpStart = Str
         FSR0 = Str
         c = INDF0
         Continue
      ElseIf c = Byte(",") Then
         Break
      EndIf   

      If ParseStates.Bits.FoundOpenBracket And ParseStates.Bits.FoundAt Then
         If c = Byte(">") Then
            Break
         EndIf
      EndIf
      
	  // Advance to next character
	  wLen = wLen + 1
      FSR0 = Str
      c = INDF0
   Wend

   If Not ParseStates.Bits.FoundAt Then
      wLen = 0
   EndIf   
   result = lpStart
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Function FindROMEmailAddress(Str As TROMAddress, ByRef wLen As Word) As TROMAddress
   Dim lpStart As Word
   Dim c As Byte
   Dim ParseStates As TParseStates

   lpStart = Str
   wLen = $0000
   ParseStates.Val = $00

   c = ReadROM(Str)
   While c <> 0
      Inc(Str)
      If c = Byte("<") Then
         ParseStates.Bits.FoundOpenBracket = true
         lpStart = Str
		 wLen = Word(-1)
      ElseIf c = Byte("@") Then
         ParseStates.Bits.FoundAt = true
      EndIf   

      If Not ParseStates.Bits.FoundOpenBracket And Not ParseStates.Bits.FoundAt And (c = Byte(" ") Or c = Byte(",")) Then
         lpStart = Str
         c = ReadROM(Str)
         Continue
      ElseIf c = Byte(",") Then
         Break
      EndIf   

      If ParseStates.Bits.FoundOpenBracket And ParseStates.Bits.FoundAt Then
         If c = Byte(">") Then
            Break
         EndIf
      EndIf
      
	  // Advance to next character
	  wLen = wLen + 1
      c = ReadROM(Str)
   Wend

   If Not ParseStates.Bits.FoundAt Then
      wLen = 0
   EndIf   
   result = lpStart
End Function
(*
 *********************************************************************
 * Function:        void SMTPTask(void)
 * PreCondition:    Stack is initialized()
 * Input:           None
 * Output:          None
 * Side Effects:    Uses ARP module
 * Overview:        None
 * Note:            This is a work is in progress
 ********************************************************************/
*)
// statics
Dim
   Timer As TICK,
   RXBuffer(4) As Byte,
   ROMStrPtr, ROMStrPtr2 As Word,
   RAMStrPtr As Word,
   wAddressLength As Word
Public Sub SMTPTask()
   Dim i As Byte
   Dim w As Word
   Dim vBase64Buffer(4) As Byte

   If TransportState = TRANSPORT_HOME Then
      // SMTPBeginUsage() is the only function which will kick 
	  // the state machine into the next state
	  GoTo TransportBreak
   EndIf

   If TransportState = TRANSPORT_BEGIN Then
      // Wait for the user to program all the pointers and then 
	  // call SMTPSendMail()
	  If Not SMTPFlags.Bits.ReadyToStart Then
	     GoTo TransportBreak
	  EndIf

      // Obtain ownership of the DNS resolution module
	  If Not DNSBeginUsage Then
         GoTo TransportBreak
      EndIf

      // Obtain the IP address associated with the SMTP mail server
	  If SMTPClient.Server.szRAM <> 0 Or SMTPClient.Server.szROM <> 0 Then
         If SMTPClient.ROMPointers.Server Then
		    DNSResolveROM(SMTPClient.Server.szROM, DNS_TYPE_A)
		 Else
            DNSResolve(SMTPClient.Server.szRAM, DNS_TYPE_A)
         EndIf
      Else
	     // If we don't have a mail server, try to send the mail 
		 // directly to the destination SMTP server
		 If SMTPClient.SendTo.szRAM <> 0 And Not SMTPClient.ROMPointers.SendTo Then
		    SMTPClient.Server.szRAM = strchr(SMTPClient.SendTo.szRAM, "@")
			SMTPClient.ROMPointers.Server = false
		 ElseIf SMTPClient.SendTo.szROM <> 0 And SMTPClient.ROMPointers.SendTo Then
		    SMTPClient.Server.szROM = strchrpgm(SMTPClient.SendTo.szROM, "@")
		    SMTPClient.ROMPointers.Server = true
		 EndIf

         If Not (SMTPClient.Server.szRAM <> 0 Or SMTPClient.Server.szROM <> 0) Then
            If SMTPClient.CC.szRAM <> 0 And Not SMTPClient.ROMPointers.CC Then
               SMTPClient.Server.szRAM = strchr(SMTPClient.CC.szRAM, "@")
               SMTPClient.ROMPointers.Server = false
            ElseIf SMTPClient.CC.szROM <> 0 And SMTPClient.ROMPointers.CC Then
               SMTPClient.Server.szROM = strchrpgm(SMTPClient.CC.szROM, "@")
			   SMTPClient.ROMPointers.Server = true
		    EndIf
         EndIf
     
     	 If Not (SMTPClient.Server.szRAM <> 0 Or SMTPClient.Server.szROM <> 0) Then
		    If SMTPClient.BCC.szRAM <> 0 And Not SMTPClient.ROMPointers.BCC Then
               SMTPClient.Server.szRAM = strchr(SMTPClient.BCC.szRAM, "@")
               SMTPClient.ROMPointers.Server = false
		    ElseIf SMTPClient.BCC.szROM <> 0 And SMTPClient.ROMPointers.BCC Then
               SMTPClient.Server.szROM = strchrpgm(SMTPClient.BCC.szROM, "@")
			   SMTPClient.ROMPointers.Server = true
		    EndIf
         EndIf

		 // See if we found a hostname anywhere which we could resolve
		 If Not (SMTPClient.Server.szRAM <> 0 Or SMTPClient.Server.szROM <> 0) Then
		    DNSEndUsage()
		    TransportState = TRANSPORT_HOME
		    GoTo TransportBreak
		 EndIf

		 // Skip over the @ sign and resolve the host name
		 If SMTPClient.ROMPointers.Server Then
		    Inc(SMTPClient.Server.szROM)
		    DNSResolveROM(SMTPClient.Server.szROM, DNS_TYPE_MX)
		 Else   
		    Inc(SMTPClient.Server.szRAM)
		    DNSResolve(SMTPClient.Server.szRAM, DNS_TYPE_MX)
		 EndIf
     EndIf
			
     Timer = TickGet
     Inc(TransportState)
     GoTo TransportBreak
   EndIf

   If TransportState = TRANSPORT_NAME_RESOLVE Then
      If Not DNSIsResolved(SMTPServer.IPAddr) Then
	  
         // Timeout after 6 seconds of unsuccessful DNS resolution
		 If TickGet() - Timer > 6 * TICK_SECOND Then
		    ResponseCode = SMTP_RESOLVE_ERROR
			TransportState = TRANSPORT_HOME
			DNSEndUsage()		
	     EndIf
         GoTo TransportBreak
      EndIf
      
      // Release the DNS module, we no longer need it
	 If Not DNSEndUsage Then
	    // An invalid IP address was returned from the DNS 
		// server.  Quit and fail permanantly if host is not valid.
		ResponseCode = SMTP_RESOLVE_ERROR
		TransportState = TRANSPORT_HOME
		GoTo TransportBreak
     EndIf
	 Inc(TransportState)
	 // No need to break here      
   EndIf
   
   If TransportState = TRANSPORT_ARP_START_RESOLVE Then
      // Obtain the MAC address associated with the server's IP address (either 
      // direct MAC address on same subnet, or the MAC address of the Gateway machine)
      ARPResolve(SMTPServer.IPAddr)
	  Timer = TickGet
      Inc(TransportState)
      GoTo TransportBreak
   EndIf

   If TransportState = TRANSPORT_ARP_IN_PROGRESS Then
   
      // Check to see if the MAC address has been resovlved
	  If Not ARPIsResolved(SMTPServer.IPAddr, SMTPServer.MACAddr) Then
	     // Time out after 2 seconds of attempting to perform ARP
		 If TickGet() - Timer > 2 * TICK_SECOND Then
		    ResponseCode = SMTP_RESOLVE_ERROR
			TransportState = TRANSPORT_HOME
	     EndIf
         GoTo TransportBreak
      EndIf
	  
      Inc(TransportState)
      Timer = TickGet
      // No break; fall into SM_ARP_RESOLVED
   EndIf
   
   If TransportState = TRANSPORT_ARP_RESOLVED Then
   
      // Connect a TCP socket to the remote SMTP server
	  MySocket = TCPConnect(SMTPServer, SMTP_PORT)
			
      // Abort operation if no TCP sockets are available
	  // If this ever happens, incrementing MAX_TCP_SOCKETS in 
	  // TCPIPConfig. may help (at the expense of more global memory 
	  // resources).
	  If MySocket = INVALID_SOCKET Then
         GoTo TransportBreak
      EndIf
	  Inc(TransportState)
	  Timer = TickGet
      // No break; fall into SM_SOCKET_OBTAINED   
   EndIf
 
		If TransportState = TRANSPORT_SOCKET_OBTAINED Then
			If Not TCPIsConnected(MySocket) Then

				// Don't stick around in the wrong state if the
				// server was connected, but then disconnected us.
				// Also time out if we can't establish the connection 
				// to the SMTP server
				If SMTPFlags.Bits.ConnectedOnce Or (TickGet - Timer) > (8 * TICK_SECOND) Then
				   ResponseCode = SMTP_CONNECT_ERROR
				   TransportState = TRANSPORT_CLOSE	   
				EndIf
				GoTo TransportBreak
			EndIf
			SMTPFlags.Bits.ConnectedOnce = true

			// See if the server sent us anything
			While TCPIsGetReady(MySocket) <> 0
				TCPGet(MySocket, i)			
				Select RXParserState
					Case RX_BYTE_0, RX_BYTE_1, RX_BYTE_2
						RXBuffer(RXParserState) = i
						Inc(RXParserState)
	
					Case RX_BYTE_3:
						Select Char(i)
							Case " ":
								SMTPFlags.Bits.RXSkipResponse = false
								Inc(RXParserState)
							Case "-":
								SMTPFlags.Bits.RXSkipResponse = true
								Inc(RXParserState)
							Case #13
								RXParserState = RX_SEEK_LF
						End Select
	
					Case RX_SEEK_CR:
						If i = 13 Then
							Inc(RXParserState)
						EndIf
	
					Case RX_SEEK_LF:
						

      
                        // If we received the whole command
						If i = 10 Then
							RXParserState = RX_BYTE_0

							If Not SMTPFlags.Bits.RXSkipResponse Then
								// The server sent us a response code
								// Null terminate the ASCII reponse code so we can convert it to an integer
								RXBuffer(3) = 0
								ResponseCode = atoi(@RXBuffer)

								// Handle the response
								Select SMTPState
									Case SMTP_HELO_ACK:
										If ResponseCode >= 200 And ResponseCode <= 299 Then
                                            If SMTPClient.Username.szRAM <> 0 Or SMTPClient.Username.szROM <> 0 Then
												SMTPState = SMTP_AUTH_LOGIN
											Else
												SMTPState = SMTP_MAILFROM
											EndIf	
										Else
											SMTPState = SMTP_QUIT_INIT
										EndIf

									Case SMTP_AUTH_LOGIN_ACK, SMTP_AUTH_USERNAME_ACK
										If ResponseCode = 334 Then
											Inc(SMTPState)
										Else
											SMTPState = SMTP_QUIT_INIT
										EndIf	

									Case SMTP_AUTH_PASSWORD_ACK:
										If ResponseCode = 235 Then
											Inc(SMTPState)
										Else
											SMTPState = SMTP_QUIT_INIT
										EndIf	

									Case SMTP_HOME, SMTP_MAILFROM_ACK, SMTP_RCPTTO_ACK, SMTP_RCPTTOCC_ACK, SMTP_RCPTTOBCC_ACK
										If ResponseCode >= 200 And ResponseCode <= 299 Then
											Inc(SMTPState)
										Else
											SMTPState = SMTP_QUIT_INIT
										EndIf
							
									Case SMTP_DATA_ACK:
										If ResponseCode = 354 Then
											Inc(SMTPState)
										Else
											SMTPState = SMTP_QUIT_INIT
										EndIf
							
									Case SMTP_DATA_BODY_ACK:
										If ResponseCode >= 200 And ResponseCode <= 299 Then
											SMTPFlags.Bits.SentSuccessfully = true
										EndIf	
										SMTPState = SMTP_QUIT_INIT
								End Select
							EndIf
						ElseIf i <> 13 Then
							Dec(RXParserState)
						EndIf
				End Select
			Wend

			// Generate new data in the TX buffer, as needed, if possible
			If TCPIsPutReady(MySocket) < 64 Then
				GoTo TransportBreak
            EndIf

			'select SMTPState
				If SMTPState = SMTP_HELO Then                       
					TCPPutROMString(MySocket, @("HELO SWORDFISH" + #13 + #10))
					TCPFlush(MySocket)
					Inc(SMTPState)
					GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_AUTH_LOGIN Then
					// Note: This state is only entered from SMTP_HELO_ACK if the application 
					// has specified a Username to use (either SMTPClient.Username.szROM or 
					// SMTPClient.Username.szRAM is non-NULL)
					TCPPutROMString(MySocket, @("AUTH LOGIN" + #13 + #10))
					TCPFlush(MySocket)
					Inc(SMTPState)
					GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_AUTH_USERNAME Then
					// Base 64 encode and transmit the username.
					If SMTPClient.ROMPointers.Username Then
						ROMStrPtr = SMTPClient.Username.szROM
						w = strlenpgm(ROMStrPtr)
					Else
						RAMStrPtr = SMTPClient.Username.szRAM
						w = strlen(RAMStrPtr)
					EndIf

					While w <> 0
						i = 0
						While (i < w) And (i < SizeOf(vBase64Buffer) * 3 / 4)
							If SMTPClient.ROMPointers.Username Then
							   vBase64Buffer(i) = ReadROM(ROMStrPtr)
                               Inc(ROMStrPtr)
							Else
								vBase64Buffer(i) = ReadRAM(RAMStrPtr)
                                Inc(RAMStrPtr)
                            EndIf
                            Inc(i)
						Wend
						w = w - i
						Base64Encode(@vBase64Buffer, i, @vBase64Buffer, SizeOf(vBase64Buffer))
						TCPPutArray(MySocket, @vBase64Buffer, SizeOf(vBase64Buffer))
					Wend
					TCPPutROMString(MySocket, @(#13 + #10))
					TCPFlush(MySocket)
					Inc(SMTPState)
					GoTo SMTPBreak
                 EndIf

				If SMTPState = SMTP_AUTH_PASSWORD Then
					// Base 64 encode and transmit the password
					If SMTPClient.ROMPointers.Password Then
						ROMStrPtr = SMTPClient.Password.szROM
						w = strlenpgm(ROMStrPtr)
					Else
						RAMStrPtr = SMTPClient.Password.szRAM
						w = strlen(RAMStrPtr)
					EndIf

					While w <> 0
						i = 0
						While (i < w) And (i < SizeOf(vBase64Buffer) * 3 / 4)
							If SMTPClient.ROMPointers.Password Then
								vBase64Buffer(i) = ReadROM(ROMStrPtr)
                                Inc(ROMStrPtr)
							Else
								vBase64Buffer(i) = ReadRAM(RAMStrPtr)
                                Inc(RAMStrPtr)
                            EndIf
                            Inc(i)
					    Wend
						w = w - i
						Base64Encode(@vBase64Buffer, i, @vBase64Buffer, SizeOf(vBase64Buffer))
						TCPPutArray(MySocket, @vBase64Buffer, SizeOf(vBase64Buffer))
					Wend
					TCPPutROMString(MySocket, @(#13 + #10))
					TCPFlush(MySocket)
					Inc(SMTPState)
                    GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_MAILFROM Then
					// Send MAIL FROM header.  Note that this is for the SMTP server validation, 
					// not what actually will be displayed in the recipients mail client as a 
					// return address.
					TCPPutROMString(MySocket, @"MAIL FROM:<") 
					If SMTPClient.ROMPointers.From Then
						ROMStrPtr = FindROMEmailAddress(SMTPClient.From.szROM, wAddressLength)
						TCPPutROMArray(MySocket, ROMStrPtr, wAddressLength)
					Else
						RAMStrPtr = FindEmailAddress(SMTPClient.From.szRAM, wAddressLength)
						TCPPutArray(MySocket, RAMStrPtr, wAddressLength)
					EndIf
					TCPPutROMString(MySocket, @(">" + #13 + #10))
					TCPFlush(MySocket)
					Inc(SMTPState)
					GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_RCPTTO_INIT Then
				
					// See if there are any (To) recipients to process
					If SMTPClient.SendTo.szRAM <> 0 And Not SMTPClient.ROMPointers.SendTo Then
						RAMStrPtr = FindEmailAddress(SMTPClient.SendTo.szRAM, wAddressLength)
						If wAddressLength <> 0 Then
							SMTPState = SMTP_RCPTTO
							GoTo SMTPBreak
						EndIf
					EndIf

					If SMTPClient.SendTo.szROM <> 0 And SMTPClient.ROMPointers.SendTo Then
						ROMStrPtr = FindROMEmailAddress(SMTPClient.SendTo.szROM, wAddressLength)
						If wAddressLength <> 0 Then
							SMTPState = SMTP_RCPTTO
							GoTo SMTPBreak
						EndIf
					EndIf
					
					SMTPState = SMTP_RCPTTOCC_INIT
					GoTo SMTPBreak
				EndIf

				If SMTPState = SMTP_RCPTTO Or SMTPState = SMTP_RCPTTOCC Or SMTPState = SMTP_RCPTTOBCC Then
				
					TCPPutROMString(MySocket, @"RCPT TO:<")  
					
					If (SMTPClient.ROMPointers.SendTo And (SMTPState = SMTP_RCPTTO)) Or 
					   (SMTPClient.ROMPointers.CC  And (SMTPState = SMTP_RCPTTOCC)) Or 
					   (SMTPClient.ROMPointers.BCC And (SMTPState = SMTP_RCPTTOBCC)) Then					
						TCPPutROMArray(MySocket, ROMStrPtr, wAddressLength)
					Else
						TCPPutArray(MySocket, RAMStrPtr, wAddressLength)
					EndIf
                    	
					TCPPutROMString(MySocket, @(">" + #13 + #10))
					TCPFlush(MySocket)
					Inc(SMTPState)
					GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_RCPTTO_ISDONE Then
					// See if we have any more (To) recipients to process
					// If we do, we must roll back a couple of states
					If SMTPClient.ROMPointers.SendTo Then
						ROMStrPtr = FindROMEmailAddress(ROMStrPtr + wAddressLength, wAddressLength)
					Else
						RAMStrPtr = FindEmailAddress(RAMStrPtr + wAddressLength, wAddressLength)
					EndIf	
	
					If wAddressLength <> 0 Then
						SMTPState = SMTP_RCPTTO
						GoTo SMTPBreak
					EndIf

					// All done with To field
					Inc(SMTPState)
					//No break
                EndIf
                
				If SMTPState = SMTP_RCPTTOCC_INIT Then
					// See if there are any Carbon Copy (CC) recipients to process
					If SMTPClient.CC.szRAM <> 0 And Not SMTPClient.ROMPointers.CC Then
						RAMStrPtr = FindEmailAddress(SMTPClient.CC.szRAM, wAddressLength)
						If wAddressLength <> 0 Then
							SMTPState = SMTP_RCPTTOCC
							GoTo SMTPBreak
						EndIf
					EndIf
					If SMTPClient.CC.szROM <> 0 And SMTPClient.ROMPointers.CC Then
						ROMStrPtr = FindROMEmailAddress(SMTPClient.CC.szROM, wAddressLength)
						If wAddressLength <> 0 Then
							SMTPState = SMTP_RCPTTOCC
							GoTo SMTPBreak
						EndIf
					EndIf
					SMTPState = SMTP_RCPTTOBCC_INIT
					GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_RCPTTOCC_ISDONE Then
					// See if we have any more Carbon Copy (CC) recipients to process
					// If we do, we must roll back a couple of states
					If SMTPClient.ROMPointers.CC Then
						ROMStrPtr = FindROMEmailAddress(ROMStrPtr + wAddressLength, wAddressLength)
					Else
						RAMStrPtr = FindEmailAddress(RAMStrPtr + wAddressLength, wAddressLength)
					EndIf	

					If wAddressLength <> 0 Then
						SMTPState = SMTP_RCPTTOCC
						GoTo SMTPBreak
					EndIf

					// All done with CC field
					Inc(SMTPState)
					//No break
                EndIf
                
				If SMTPState = SMTP_RCPTTOBCC_INIT Then
					// See if there are any Blind Carbon Copy (BCC) recipients to process
					If SMTPClient.BCC.szRAM <> 0 And Not SMTPClient.ROMPointers.BCC Then
						RAMStrPtr = FindEmailAddress(SMTPClient.BCC.szRAM, wAddressLength)
						If wAddressLength <> 0 Then
							SMTPState = SMTP_RCPTTOBCC
							GoTo SMTPBreak
						EndIf
					EndIf

					If SMTPClient.BCC.szROM <> 0 And SMTPClient.ROMPointers.BCC Then
						ROMStrPtr = FindROMEmailAddress(SMTPClient.BCC.szROM, wAddressLength)
						If wAddressLength <> 0 Then
							SMTPState = SMTP_RCPTTOBCC
							GoTo SMTPBreak
						EndIf
                    EndIf

					// All done with BCC field
					SMTPState = SMTP_DATA
					GoTo SMTPBreak
                EndIf
                
				If SMTPState = SMTP_RCPTTOBCC_ISDONE Then
					// See if we have any more Blind Carbon Copy (CC) recipients to process
					// If we do, we must roll back a couple of states
					If SMTPClient.ROMPointers.BCC Then
						ROMStrPtr = FindROMEmailAddress(ROMStrPtr + wAddressLength, wAddressLength)
					Else
						RAMStrPtr = FindEmailAddress(RAMStrPtr + wAddressLength, wAddressLength)
					EndIf	

					If wAddressLength <> 0 Then
						SMTPState = SMTP_RCPTTOBCC
						GoTo SMTPBreak
					EndIf

					// All done with BCC field
					Inc(SMTPState)
					//No break
                EndIf
                
				If SMTPState = SMTP_DATA Then
					TCPPutROMString(MySocket, @("DATA" + #13 + #10))
					Inc(SMTPState)
					PutHeadersState = PUTHEADERS_FROM_INIT
					TCPFlush(MySocket)
					GoTo SMTPBreak
                EndIf

				If SMTPState = SMTP_DATA_HEADER Then
					While PutHeadersState <> PUTHEADERS_DONE And (TCPIsPutReady(MySocket) > 64)
						Select PutHeadersState
							Case PUTHEADERS_FROM_INIT:
								If SMTPClient.From.szRAM <> 0 Or SMTPClient.From.szROM <> 0 Then
									PutHeadersState = PUTHEADERS_FROM
									TCPPutROMString(MySocket, @"From: ")
								Else
									PutHeadersState = PUTHEADERS_TO_INIT
								EndIf
								
							Case PUTHEADERS_FROM:
								If SMTPClient.ROMPointers.From Then
									SMTPClient.From.szROM = TCPPutROMString(MySocket, SMTPClient.From.szROM)
									If ReadROM(SMTPClient.From.szROM) = 0 Then
										PutHeadersState = PUTHEADERS_TO_INIT
									EndIf	
								Else
									SMTPClient.From.szRAM = TCPPutString(MySocket, SMTPClient.From.szRAM)
									If ReadRAM(SMTPClient.From.szRAM) = 0 Then
										PutHeadersState = PUTHEADERS_TO_INIT
									EndIf
                                EndIf

							Case PUTHEADERS_TO_INIT:
								If SMTPClient.SendTo.szRAM <> 0 Or SMTPClient.SendTo.szROM <> 0 Then
									PutHeadersState = PUTHEADERS_TO
									TCPPutROMString(MySocket, @(#13 + #10 + "To: "))
								Else
									PutHeadersState = PUTHEADERS_CC_INIT
								EndIf
								
							Case PUTHEADERS_TO:
								If SMTPClient.ROMPointers.SendTo Then
									SMTPClient.SendTo.szROM = TCPPutROMString(MySocket, SMTPClient.SendTo.szROM)
									If ReadROM(SMTPClient.SendTo.szROM) = 0 Then
										PutHeadersState = PUTHEADERS_CC_INIT
									EndIf	
								Else
									SMTPClient.SendTo.szRAM = TCPPutString(MySocket, SMTPClient.SendTo.szRAM)
									If ReadRAM(SMTPClient.SendTo.szRAM) = 0 Then
										PutHeadersState = PUTHEADERS_CC_INIT
									EndIf	
								EndIf

							Case PUTHEADERS_CC_INIT:
								If SMTPClient.CC.szRAM <> 0 Or SMTPClient.CC.szROM <> 0 Then
									PutHeadersState = PUTHEADERS_CC
									TCPPutROMString(MySocket, @(#13 + #10 + "CC: "))
								Else
									PutHeadersState = PUTHEADERS_SUBJECT_INIT
								EndIf
								
							Case PUTHEADERS_CC:
								If SMTPClient.ROMPointers.CC Then
									SMTPClient.CC.szROM = TCPPutROMString(MySocket, SMTPClient.CC.szROM)
									If ReadROM(SMTPClient.CC.szROM) = 0 Then
										PutHeadersState = PUTHEADERS_SUBJECT_INIT
									EndIf	
								Else
									SMTPClient.CC.szRAM = TCPPutString(MySocket, SMTPClient.CC.szRAM)
									If ReadRAM(SMTPClient.CC.szRAM) = 0 Then
										PutHeadersState = PUTHEADERS_SUBJECT_INIT
									EndIf
                                EndIf

							Case PUTHEADERS_SUBJECT_INIT:
								If SMTPClient.Subject.szRAM <> 0 Or SMTPClient.Subject.szROM <> 0 Then
									PutHeadersState = PUTHEADERS_SUBJECT
									TCPPutROMString(MySocket, @(#13 + #10 + "Subject: "))
								Else
									PutHeadersState = PUTHEADERS_OTHER_INIT
								EndIf
								
							Case PUTHEADERS_SUBJECT:
								If SMTPClient.ROMPointers.Subject Then
									SMTPClient.Subject.szROM = TCPPutROMString(MySocket, SMTPClient.Subject.szROM)
									If ReadROM(SMTPClient.Subject.szROM) = 0 Then
										PutHeadersState = PUTHEADERS_OTHER_INIT
									EndIf	
								Else
									SMTPClient.Subject.szRAM = TCPPutString(MySocket, SMTPClient.Subject.szRAM)
									If ReadRAM(SMTPClient.Subject.szRAM) = 0 Then
										PutHeadersState = PUTHEADERS_OTHER_INIT
									EndIf
                                EndIf

							Case PUTHEADERS_OTHER_INIT:
								TCPPutROMArray(MySocket, @(#13 + #10), 2)
								If SMTPClient.OtherHeaders.szRAM <> 0 Or SMTPClient.OtherHeaders.szROM <> 0 Then
									PutHeadersState = PUTHEADERS_OTHER
								Else
								    TCPPutROMArray(MySocket, @(#13 + #10), 2)
									PutHeadersState = PUTHEADERS_DONE
									Inc(SMTPState)
                                EndIf
								
							Case PUTHEADERS_OTHER:
								If SMTPClient.ROMPointers.OtherHeaders Then
									SMTPClient.OtherHeaders.szROM = TCPPutROMString(MySocket, SMTPClient.OtherHeaders.szROM)
									If ReadROM(SMTPClient.OtherHeaders.szROM) = 0 Then
										TCPPutROMArray(MySocket, @(#13 + #10), 2)
										PutHeadersState = PUTHEADERS_DONE
										Inc(SMTPState)
                                    EndIf
								Else
									SMTPClient.OtherHeaders.szRAM = TCPPutString(MySocket, SMTPClient.OtherHeaders.szRAM)
									If ReadRAM(SMTPClient.OtherHeaders.szRAM) = 0 Then
										TCPPutROMArray(MySocket, @(#13 + #10), 2)
										PutHeadersState = PUTHEADERS_DONE
										Inc(SMTPState)
									EndIf
                                EndIf
						End Select
					Wend
					TCPFlush(MySocket)
					GoTo SMTPBreak
		        EndIf
		        
				If SMTPState = SMTP_DATA_BODY_INIT Then
					Inc(SMTPState)
					RAMStrPtr = SMTPClient.Body.szRAM
					ROMStrPtr2 = @(#13 + #10 + "." + #13 + #10)
					CRPeriod.Pos = strstrrampgm(RAMStrPtr, @(#13 + #10 + "."))
					// No break here
		        EndIf
		        
				If SMTPState = SMTP_DATA_BODY Then
					If SMTPClient.Body.szRAM <> 0 Or SMTPClient.Body.szROM <> 0 Then
						If ReadROM(ROMStrPtr2) <> 0 Then
							
                            // Put the application data, doing the transparancy replacement of "\r\n." with "\r\n.."
							While CRPeriod.Pos <> 0 
								Inc(CRPeriod.Pos,3)
								RAMStrPtr = RAMStrPtr + TCPPutArray(MySocket, RAMStrPtr, CRPeriod.Pos - RAMStrPtr)
								If RAMStrPtr = CRPeriod.Pos Then
									If Not TCPPut(MySocket, ".") Then
										Dec(CRPeriod.Pos,3)
										Break
                                    EndIf
								Else
									Dec(CRPeriod.Pos,3)
									Break
								EndIf
								CRPeriod.Pos = strstrrampgm(RAMStrPtr, @(#13 + #10 + "."))
							Wend
							
							// If we get down here, either all replacements have been made or there is no remaining space in the TCP output buffer
							RAMStrPtr = TCPPutString(MySocket, RAMStrPtr)
							ROMStrPtr2 = TCPPutROMString(MySocket, ROMStrPtr2)
							TCPFlush(MySocket)
						EndIf
					Else
						If SMTPFlags.Bits.ReadyToFinish Then
							If ReadROM(ROMStrPtr2) <> 0 Then
								ROMStrPtr2 = TCPPutROMString(MySocket, ROMStrPtr2)
								TCPFlush(MySocket)
							EndIf
						EndIf
					EndIf

					If ReadROM(ROMStrPtr2) = 0 Then
						Inc(SMTPState)
                    EndIf
                    GoTo SMTPBreak
		        EndIf
		        
				If SMTPState = SMTP_QUIT_INIT Then
					Inc(SMTPState)
					ROMStrPtr = @("QUIT" + #13 + #10)
					// No break here
                EndIf
                
				If SMTPState = SMTP_QUIT Then
					If ReadROM(ROMStrPtr) <> 0 Then
						ROMStrPtr = TCPPutROMString(MySocket, ROMStrPtr)
						TCPFlush(MySocket)
					EndIf

					If ReadROM(ROMStrPtr) = 0 Then
						TransportState = TRANSPORT_CLOSE
					EndIf
                    GoTo SMTPBreak
                EndIf
			SMTPBreak:  
			GoTo TransportBreak
         EndIf      
     

   If TransportState = TRANSPORT_CLOSE Then

      // Close the socket so it can be used by other modules
      TCPDisconnect(MySocket)
      MySocket = INVALID_SOCKET
      
      // Go back to doing nothing
	  TransportState = TRANSPORT_HOME
   EndIf

TransportBreak:

End Sub

// initialise
#if defined(STACK_USE_SMTP_CLIENT)
Clear(SMTPClient)
RXParserState = RX_BYTE_0
SMTPState = SMTP_HOME
MySocket = INVALID_SOCKET
TransportState = TRANSPORT_HOME
SMTPFlags.Val = 0
#endif
