{
*****************************************************************************
*  Name    : SNTP.BAS                                                       *
*  Date    : 18/09/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module SNTP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "UDP.bas"
Include "ARP.bas"
Include "DNS.bas"
Include "Tick.bas"

Const
   NTP_SERVER_PORT	    = (123),
   NTP_QUERY_INTERVAL   = (10 * 60 * TICK_SECOND),	  // Resynchronize date/time every 10 minutes
   NTP_EPOCH 			= (86400 * (365 * 70 + 17)),  // 0:0:0 1/1/1970
   NTP_REPLY_TIMEOUT	= (6 * TICK_SECOND) 		  // Wait 6 seconds for timeout

// These are normally available network time servers
// the actual IP returned from the pool will vary every
// minute so as to spread the load around stratum 1 timeservers
// for best accuracy and network overhead you should locate the 
// pool server closest to your geography, but it will still work
// if you use the global pool.ntp.org address or choose the wrong 
// one or ship your embedded device to another geography.
//const NTP_SERVER = "pool.ntp.org"
Const NTP_SERVER = "europe.pool.ntp.org"
//const NTP_SERVER = "asia.pool.ntp.org"
//const NTP_SERVER = "oceania.pool.ntp.org"
//const NTP_SERVER = "north-america.pool.ntp.org"
//const NTP_SERVER = "south-america.pool.ntp.org"
//const NTP_SERVER = "africa.pool.ntp.org"

Structure NTPFlags 
   _mode As Byte                 // 3    0..2
   _versionNumber As Byte Union  // 3    3..5
   _leapIndicator As Byte Union  // 2    6..7
End Structure

Structure NTP_PACKET
   Flags As NTPFlags
   stratum As Byte
   poll As Byte
   precision As Byte
   root_delay As DWORD
   root_dispersion As DWORD
   ref_identifier As DWORD
   ref_ts_secs As DWORD
   ref_ts_fraq As DWORD
   orig_ts_secs As DWORD
   orig_ts_fraq As DWORD
   recv_ts_secs As DWORD
   recv_ts_fraq As DWORD
   tx_ts_secs As DWORD
   tx_ts_fraq As DWORD
End Structure

// module statics
Dim
   dwSNTPSeconds As DWORD,
   dwLastUpdateTick As DWORD

// subroutine statics...
Dim
   Server As NODE_INFO,
   dwTimer As DWORD,
   MySocket As UDP_SOCKET,
   SNTPState As Byte
Const
   SM_HOME = 0,
   SM_NAME_RESOLVE = 1,
   SM_ARP_START_RESOLVE = 2,
   SM_ARP_RESOLVE = 3,
   SM_ARP_START_RESOLVE2 = 4,
   SM_ARP_RESOLVE2 = 5,
   SM_ARP_START_RESOLVE3 = 6,
   SM_ARP_RESOLVE3 = 7,
   SM_ARP_RESOLVE_FAIL = 8,
   SM_UDP_SEND = 9,
   SM_UDP_RECV = 10,
   SM_WAIT = 11

Public Dim SNTPSeconds As dwSNTPSeconds  
(*
 *********************************************************************
 * Function:        void SNTPClient(void)
 * PreCondition:    None
 * Input:           None
 * Output:          Sets the local date and time to match remote NTP 
 *					server value
 * Side Effects:    None
 * Overview:        None
 * Note:            Require one available UDP socket resource.  The 
 *					UDP socket is freed while SNTP module is idle.
 ********************************************************************/
*)
Public Sub SNTPClient()
   Dim pkt As NTP_PACKET
   Dim w As Word

   If SNTPState = SM_HOME Then
      // Obtain ownership of the DNS resolution module
	  If Not DNSBeginUsage Then
         Exit
      EndIf
      
      // Obtain the IP address associated with the server name
	  DNSResolveROM(@NTP_SERVER, DNS_TYPE_A)
      dwTimer = TickGet()
      SNTPState = SM_NAME_RESOLVE
      Exit
   EndIf
      
   If SNTPState = SM_NAME_RESOLVE Then
      // Wait for DNS resolution to complete
	  If Not DNSIsResolved(Server.IPAddr) Then
         If TickGet() - dwTimer > (5 * TICK_SECOND) Then
            DNSEndUsage()
			dwTimer = TickGet
			SNTPState = SM_WAIT
		 EndIf
         Exit
      EndIf
			
      // Obtain DNS resolution result
      If Not DNSEndUsage Then
         // No valid IP address was returned from the DNS 
		 // server.  Quit and fail for a while if host is not valid.
		dwTimer = TickGet
		SNTPState = SM_WAIT
		Exit
	  EndIf
      SNTPState = SM_ARP_START_RESOLVE
      
      // No need to break
   EndIf
   
   If SNTPState = SM_ARP_START_RESOLVE Or SNTPState = SM_ARP_START_RESOLVE2 Or SNTPState = SM_ARP_START_RESOLVE3 Then
      // Obtain the MAC address associated with the server's IP address 
	  ARPResolve(Server.IPAddr)
      dwTimer = TickGet
      Inc(SNTPState)
      Exit
   EndIf

   If SNTPState = SM_ARP_RESOLVE Or SNTPState = SM_ARP_RESOLVE2 Or SNTPState = SM_ARP_RESOLVE3 Then  
      // Wait for the MAC address to finish being obtained
	  If Not ARPIsResolved(Server.IPAddr, Server.MACAddr) Then
         // Time out if too much time is spent in this state
		 If TickGet() - dwTimer > 1 * TICK_SECOND Then
		    // Retransmit ARP request by going to next SM_ARP_START_RESOLVE state or fail by going to SM_ARP_RESOLVE_FAIL state.
			Inc(SNTPState)
		 EndIf
         Exit
      EndIf
	  SNTPState = SM_UDP_SEND
	  Exit
   EndIf

   If SNTPState = SM_ARP_RESOLVE_FAIL Then
      // ARP failed after 3 tries, abort and wait for next time query
	  dwTimer = TickGet
      SNTPState = SM_WAIT
      Exit
   EndIf

   If SNTPState = SM_UDP_SEND Then
      // Open up the sending UDP socket
	  MySocket = UDPOpen(0, Server, NTP_SERVER_PORT)
      If MySocket = INVALID_UDP_SOCKET Then
         Exit
      EndIf

      // Make certain the socket can be written to
	  If UDPIsPutReady(MySocket) = 0 Then
         UDPClose(MySocket)
         Exit
      EndIf

      // Transmit a time request packet
	  memset(@pkt, 0, SizeOf(pkt))
      pkt.Flags._versionNumber = 3	                // NTP Version = 3   0..2
      pkt.Flags._mode = pkt.Flags._mode Or (3 << 3)	// NTP Client  = 3   3..5
      pkt.orig_ts_secs = swapl(NTP_EPOCH)
      UDPPutArray(@pkt, SizeOf(pkt))	
      UDPFlush()	
			
      dwTimer = TickGet()
      SNTPState = SM_UDP_RECV		
      Exit
   EndIf

   If SNTPState = SM_UDP_RECV Then
      // Look for a response time packet
	  If UDPIsGetReady(MySocket) = 0 Then
         If TickGet() - dwTimer > NTP_REPLY_TIMEOUT Then
            
            // Abort the request and wait until the next timeout period
			UDPClose(MySocket)
			dwTimer = TickGet
			SNTPState = SM_WAIT
			Exit
         EndIf
		 Exit
      EndIf
			
      // Get the response time packet
      w = UDPGetArray(@pkt, SizeOf(pkt))
      UDPClose(MySocket)
      dwTimer = TickGet
      SNTPState = SM_WAIT

      // Validate packet size
	  If w <> SizeOf(pkt) Then
         Exit
      EndIf

      // Set out local time to match the returned time
	  dwLastUpdateTick = TickGet()
      dwSNTPSeconds = swapl(pkt.tx_ts_secs) - NTP_EPOCH
 
      // Do rounding.  If the partial seconds is > 0.5 then add 1 to the seconds count.
	  If (pkt.tx_ts_fraq.Byte0 And $80) <> 0 Then
	     Inc(dwSNTPSeconds)
      EndIf
      Exit
   EndIf

   If SNTPState = SM_WAIT Then
      
      // Requery the NTP server after a specified NTP_QUERY_INTERVAL time (ex: 10 minutes) has elapsed.
	  If TickGet - dwTimer > NTP_QUERY_INTERVAL Then
         SNTPState = SM_HOME
      EndIf
   EndIf
End Sub
(*
 *********************************************************************
 * Function:        DWORD SNTPGetUTCSeconds(void)
 * PreCondition:    None
 * Input:           None
 * Output:          DWORD indicating the number of seconds since 
 *					00:00:00 1970
 * Side Effects:    None
 * Overview:        None
 * Note:            Use this time value For Absolute time stamping.
 *					Do Not use this API For time difference 
 *					measurements. Since this value depends on what 
 *					the remote NTP Server periodically provides us 
 *					With, one call To SNTPGetUTCSeconds() could 
 *					return a higher number than a second call To 
 *					SNTPGetUTCSeconds() made at a future time.  Use 
 *					the TickGet*() API For reliable time difference 
 *					calculations.
 ********************************************************************/
*)
Public Function SNTPGetUTCSeconds() As DWORD
   result = dwSNTPSeconds 
   Inc(result, (TickGet - dwLastUpdateTick) / 1000)
End Function

// initialise
#if defined(STACK_USE_SNTP)
SNTPState = SM_HOME
dwSNTPSeconds = 0
dwLastUpdateTick = 0
#endif
   

