{
*****************************************************************************
*  Name    : StackTsk.BAS                                                   *
*  Date    : 28/08/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module StackTsk

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "MAC.bas"
Include "IP.bas"
Include "ARP.bas"

'#if defined(STACK_USE_ICMP)
Include "ICMP.bas"
'#endif
'#if defined(STACK_USE_UDP)
Include "UDP.bas"
'#endif
'#if defined(STACK_USE_TCP)
Include "TCP.bas"
'#endif

'#if defined(STACK_USE_DHCP_CLIENT)
Include "DHCP.bas"
'#endif

// Stack FSM states.
Type SM_STACK = Byte
Const
   SM_STACK_IDLE  = 0,
   SM_STACK_MAC   = 1,
   SM_STACK_IP    = 2,
   SM_STACK_ARP   = 3,
   SM_STACK_TCP   = 4,
   SM_STACK_UDP   = 5
Dim
   smStack As SM_STACK

(*
 *********************************************************************
 * Function:        void StackInit(void)
 * PreCondition:    None
 * Input:           None
 * Output:          Stack and its componets are initialized
 * Side Effects:    None
 * Note:            This function must be called before any of the
 *                  stack or its component routines are used.
 ********************************************************************/
*)
Public Sub StackInit()
   #if defined(STACK_USE_DHCP_CLIENT)
   AppConfig.Flags.bIsDHCPEnabled = true
   AppConfig.Flags.bInConfigMode = true
   #endif
   MACInit()
   ARPInit()
   #if defined(STACK_USE_UDP)
   UDPInit()
   #endif
   #if defined(STACK_USE_TCP)
   TCPInit()
   #endif
End Sub
(*
 *********************************************************************
 * Function:        void StackTask(void)
 * PreCondition:    StackInit() is already called.
 * Input:           None
 * Output:          Stack FSM is executed.
 * Side Effects:    None
 * Note:            This FSM checks for new incoming packets,
 *                  and routes it to appropriate stack components.
 *                  It also performs timed operations.
 *                  This function must be called periodically to
 *                  ensure timely responses.
 ********************************************************************/
*)
Include "usart.bas"
Include "convert.bas"
Public Sub StackTask()
   Dim dataCount As Word
   Dim tempLocalIP As IP_ADDR
   Dim cFrameType As Byte
   Dim cIPFrameType As Byte

   #if defined(STACK_USE_DHCP_CLIENT)
   // Normally, an application would not include  DHCP module
   // if it is not enabled. But in case some one wants to disable
   // DHCP module at run-time, remember to not clear our IP
   // address if link is removed.
   If AppConfig.Flags.bIsDHCPEnabled Then
      If Not MACIsLinked() Then
         AppConfig.MyIPAddr.Val = AppConfig.DefaultIPAddr.Val
         AppConfig.MyMask.Val = AppConfig.DefaultMask.Val
		 DHCPFlags.Bits.bDHCPServerDetected = false
		 AppConfig.Flags.bInConfigMode = false
		 DHCPReset()
      EndIf

      // DHCP must be called all the time even after IP configuration is
	  // discovered.
	  // DHCP has to account lease expiration time and renew the configuration
	  // time.
      DHCPTask()
      If DHCPIsBound Then
         AppConfig.Flags.bInConfigMode = false
      EndIf   
   EndIf
   #endif

   #if defined(STACK_USE_TCP)
   // Perform all TCP time related tasks (retransmit, send acknowledge, close connection, etc)
   TCPTick()
   #endif
   
   // Process as many incomming packets as we can
   While true
      //if using the random module, generate entropy
	  #if defined(STACK_USE_RANDOM)
	  RandomAdd(remoteNode.MACAddr.v(5))
	  #endif   
      
      // We are about to fetch a new packet, make sure that the 
	  // UDP module knows that any old RX data it has laying 
	  // around will now be gone.
	  #if defined(STACK_USE_UDP)
	  UDPDiscard()
	  #endif
	
	  // Fetch a packet (throws old one away, if not thrown away 
	  // yet)
	  If Not MACGetHeader(remoteNode.MACAddr, cFrameType) Then
         Break
      EndIf
      
      // Dispatch the packet to the appropriate handler
      Select cFrameType
         Case MAC_ARP:  
		    ARPProcess()
		    Continue
         
         Case MAC_IP:
            If Not IPGetHeader(tempLocalIP, remoteNode, cIPFrameType, dataCount) Then
               Continue
		    EndIf
		    
            #if defined(STACK_USE_ICMP)
			If cIPFrameType = IP_PROT_ICMP Then	

			   // Process this ICMP packet if it the destination IP address matches our address or one of the broadcast IP addressees
			   If (tempLocalIP.Val = AppConfig.MyIPAddr.Val) Or 
                  (tempLocalIP.Val = $FFFFFFFF) Or 						
                  (tempLocalIP.Val = ((AppConfig.MyIPAddr.Val And AppConfig.MyMask.Val) Or Not AppConfig.MyMask.Val)) Then
                     ICMPProcess(remoteNode, dataCount)			   
               EndIf
               Continue
            EndIf
			#endif
				
			#if defined(STACK_USE_TCP)
			If cIPFrameType = IP_PROT_TCP Then
			   TCPProcess(remoteNode, tempLocalIP, dataCount)
			   Continue
			EndIf
			#endif
				
			#if defined(STACK_USE_UDP)
			If cIPFrameType = IP_PROT_UDP Then

			   // Stop processing packets if we came upon a UDP frame with 
               // application data in it
			   If UDPProcess(remoteNode, tempLocalIP, dataCount) Then
			      Break
               EndIf
			   Continue
			EndIf
			#endif
      End Select
   Wend
End Sub

// initialise
smStack = SM_STACK_IDLE

