{
*****************************************************************************
*  Name    : TCP.BAS                                                        *
*  Date    : 28/08/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module TCP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "MAC.bas"
Include "IP.bas"
Include "Tick.bas"

Public Type TCP_SOCKET = Byte

Public Const
   MAX_TCP_SOCKETS  = NETConfig.MAX_TCP_SOCKETS,
   INVALID_SOCKET   = ($FE),
   UNKNOWN_SOCKET   = ($FF),
   LOOPBACK_IP      = $7f000001	   //127.0.0.1

// TCP States as defined by RFC 793
Type TCP_STATE = Byte
Const
   TCP_LISTEN = 0,
   TCP_SYN_SENT = 1,
   TCP_SYN_RECEIVED = 2,
   TCP_ESTABLISHED = 3,
   TCP_FIN_WAIT_1 = 4,
   TCP_FIN_WAIT_2 = 5,
   TCP_CLOSING = 6,
   TCP_LAST_ACK = 7,
   TCP_CLOSED = 8,
   TCP_LOOPBACK = 9,			//custom addition for loopbacks
   TCP_LOOPBACK_CLOSED = 10		//custom addition for loopbacks    

// TCP Control Block (TCB) Information
// Stubs are stored in local PIC RAM
// Current size is 27 bytes
Structure TSTUBFlags
   bServer As Boolean
   bTimerEnabled As Boolean
   bTimer2Enabled As Boolean
   bDelayedACKTimerEnabled As Boolean
   bOneSegmentReceived As Boolean
   bHalfFullFlush As Boolean
   bTXASAP As Boolean
   bTXFin As Boolean
   bSocketReset As Boolean
   bMemoryExternal As Boolean
   // filler : 6
End Structure

Structure TCB_STUB
   remoteHash As WORD_VAL   // Hash consists of remoteIP, remotePort, localPort for connected sockets.  It is a localPort number only for listening server sockets.
   bufferTxStart As Word	// TCB is located sizeof(TCB) bytes before this address
   bufferRxStart As Word
   bufferEnd As Word
   txHead As Word
   txTail As Word
   rxHead As Word
   rxTail As Word
   eventTime As DWORD		// Packet retransmissions, state changes
   eventTime2 As Word       // Window updates, automatic transmission
   delayedACKTime As Word	// Delayed Acknowledgement timer
   smState As TCP_STATE
   Flags As TSTUBFlags
End Structure

// The rest of the TCB is stored in Ethernet buffer RAM
// Current size is 36 bytes
Structure TTCBFlags
   bFINSent As Boolean
   bSYNSent As Boolean
End Structure
   
Structure TCB
   remote As NODE_INFO           // 10      10
   remotePort As WORD_VAL        // 2       12
   localPort As WORD_VAL         // 2       14

   txUnackedTail As Word         // 2       18
   remoteWindow As Word          // 2       20
   MySEQ As DWORD                // 4       24
   RemoteSEQ As DWORD            // 4       28
   retryCount As Byte            // 1       29
   retryInterval As DWORD        // 4       33
   sHoleSize As SHORT            // 2       35
   wFutureDataSize As Word       // 2       37
   Flags As TTCBFlags            // 1       38
End Structure

Structure SOCKET_INFO
   remote As NODE_INFO
   remotePort As WORD_VAL
End Structure

// Configurable TCP options
#define _TCP_MAX_SEG_SIZE		    = (536)
Const TCP_MAX_SEG_SIZE              = _TCP_MAX_SEG_SIZE

// TCP Timeout and retransmit numbers
Const
   TCP_START_TIMEOUT_VAL As Word          = (TICK_SECOND * 1),
   TCP_DELAYED_ACK_TIMEOUT		          = (TICK_SECOND / 10),
   TCP_FIN_WAIT_2_TIMEOUT                 = (TICK_SECOND * 5),
   TCP_MAX_RETRIES			              = (5),
   TCP_MAX_SYN_RETRIES			          = (2),	// Smaller than all other retries to reduce SYN flood DoS duration
   TCP_AUTO_TRANSMIT_TIMEOUT_VAL As Word  = (TICK_SECOND / 25)

// TCP Flags defined in RFC
Const
   FIN As Byte = ($01),
   SYN As Byte = ($02),
   RST As Byte = ($04),
   PSH As Byte = ($08),
   ACK As Byte = ($10),
   URG As Byte = ($20)

Structure TDataOffset
   _Reserved3  As Byte  // 0..3 : 4
   _Val As Byte Union   // 4..7 : 4
End Structure

Structure TBits
   Val As Byte
   flagFIN As Val.Booleans(0)
   flagSYN As Val.Booleans(1)
   flagRST As Val.Booleans(2)
   flagPSH As Val.Booleans(3)
   flagACK As Val.Booleans(4)
   flagURG As Val.Booleans(5)
   _Reserved2 As Val  // 6..7 : 2
End Structure

Structure TFlags
   Val As Byte
   Bits As TBits Union
End Structure
   
// TCP Header
Structure TCP_HEADER
   SourcePort As Word             // 2    2
   DestPort As Word               // 2    4
   SeqNumber As DWORD             // 4    8
   AckNumber As DWORD             // 4    12
   DataOffset As TDataOffset      // 1    13
   Flags As TFlags                // 1    14
   Window As Word                 // 2    16
   Checksum As Word               // 2    18
   UrgentPointer As Word          // 2    20
End Structure

// TCP Options as defined by RFC
Const
   TCP_OPTIONS_END_OF_LIST      = ($00),
   TCP_OPTIONS_NO_OP            = ($01),
   TCP_OPTIONS_MAX_SEG_SIZE     = ($02)

Structure TCP_OPTIONS
   Kind As Byte
   Length As Byte
   MaxSegSize As WORD_VAL
End Structure

// IP pseudo header as defined by RFC 793
Structure PSEUDO_HEADER
   SourceAddress As IP_ADDR
   DestAddress As IP_ADDR
   Zero As Byte
   Protocol As Byte
   TCPLength As Word
End Structure

Const
   LOCAL_PORT_START_NUMBER      = (1024),
   LOCAL_PORT_END_NUMBER        = (5000)

// Local temp port numbers.
Dim 
   NextPort As Word,                      // client socket
   TCBStubs(MAX_TCP_SOCKETS) As TCB_STUB, // The TCB array is very large!!!
   MyTCB As TCB                           // currect TCB
   
(*
 *********************************************************************
* Function:        void SwapPseudoTCPHeader
********************************************************************/
*)  
Sub SwapPseudoTCPHeader(ByRef h As PSEUDO_HEADER)
   h.TCPLength = swaps(h.TCPLength)
End Sub 
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
// Flushes MyTCB cache and loads up the specified TCB
// Does nothing on cache hit
Dim hCurrentTCB As TCP_SOCKET  // static
Sub SyncTCB(hTCP As TCP_SOCKET)
   Dim PtrSave As Word

   If hCurrentTCB = hTCP Then
      Exit
   EndIf

   If hCurrentTCB <> INVALID_SOCKET Then
      // Save the current TCB
	  PtrSave = MACSetWritePtr(TCBStubs(hCurrentTCB).bufferTxStart - SizeOf(MyTCB))
	  MACPutArray(@MyTCB, SizeOf(MyTCB))
	  MACSetWritePtr(PtrSave)
   EndIf

   // Load up the new TCB
   hCurrentTCB = hTCP
   PtrSave = MACSetReadPtr(TCBStubs(hTCP).bufferTxStart - SizeOf(MyTCB))
   MACGetArray(@MyTCB, SizeOf(MyTCB))
   MACSetReadPtr(PtrSave)
End Sub
(*
*********************************************************************
* Function:        static void CloseSocket(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called
* Input:           hTCP  - Handle of socket To close/reinitialize
* Output:          Given socket information is Reset And any
*                  buffered Bytes held by this socket are discarded.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Sub CloseSocket(hTCP As TCP_SOCKET)
   Dim ps As TCB_STUB
	
   ps = TCBStubs(hTCP)
   SyncTCB(hTCP)

   ps.remoteHash.Val = MyTCB.localPort.Val
   ps.txHead = ps.bufferTxStart
   ps.txTail = ps.bufferTxStart
   ps.rxHead = ps.bufferRxStart
   ps.rxTail = ps.bufferRxStart
   If ps.Flags.bServer Then
      ps.smState = TCP_LISTEN
   Else
      ps.smState = TCP_CLOSED
   EndIf   
   ps.Flags.bTimerEnabled = false
   ps.Flags.bTimer2Enabled = false
   ps.Flags.bDelayedACKTimerEnabled = false
   ps.Flags.bOneSegmentReceived = false
   ps.Flags.bHalfFullFlush = false
   ps.Flags.bTXASAP = false
   ps.Flags.bTXFin = false
   ps.Flags.bSocketReset = true   
   TCBStubs(hTCP) = ps
   
   MyTCB.Flags.bFINSent = false
   MyTCB.Flags.bSYNSent = false
   MyTCB.txUnackedTail = ps.bufferTxStart
   MyTCB.MySEQ = rand()
   MyTCB.sHoleSize = -1
End Sub
(*
*********************************************************************
* Function:        void TCPInit(void)
* PreCondition:    None
* Input:           None
* Output:          TCP is initialized.
* Side Effects:    None
* Overview:        Initialize all socket states
* Note:            This Function is called only once during lifetime
*                  of the application.
********************************************************************/
*)
Public Sub TCPInit()
   Dim hTCP As TCP_SOCKET
   Dim ps As TCB_STUB
	
   // Initialize all sockets.
   For hTCP = 0 To MAX_TCP_SOCKETS - 1
      ps.smState = TCP_CLOSED
	  ps.bufferTxStart	= BASE_TCB_ADDR + SizeOf(TCB) + hTCP * (SizeOf(TCB) + (TCP_TX_FIFO_SIZE + 1) + (TCP_RX_FIFO_SIZE + 1))
	  ps.bufferRxStart	= ps.bufferTxStart + TCP_TX_FIFO_SIZE + 1
	  ps.bufferEnd		= ps.bufferRxStart + TCP_RX_FIFO_SIZE
	  ps.Flags.bServer	= false
      ps.Flags.bMemoryExternal = false
	  TCBStubs(hTCP) = ps
	  CloseSocket(hTCP)
   Next
End Sub
(*
*********************************************************************
* Function:        TCP_SOCKET TCPListen(Word Port)
* PreCondition:    TCPInit() is already called.
* Input:           Port    - A TCP Port To be opened.
* Output:          Given Port is opened And returned on success
*                  INVALID_SOCKET If no more sockets left.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPListen(aPort As Word) As TCP_SOCKET
   Dim hTCP As TCP_SOCKET
   Dim ps As TCB_STUB

   For hTCP = 0 To MAX_TCP_SOCKETS - 1
      ps = TCBStubs(hTCP)

      If ps.smState = TCP_CLOSED Then
         ps.Flags.bServer = true
	     ps.smState = TCP_LISTEN
		 ps.remoteHash.Val = aPort
		 TCBStubs(hTCP) = ps
         SyncTCB(hTCP)
		 MyTCB.localPort.Val = aPort
		 result = hTCP
		 Exit
      EndIf
    Next
   result = INVALID_SOCKET
End Function
(*
*********************************************************************
* Function 		BOOL TCPGetRemoteInfo(TCP_SOCKET hTCP, SOCKET_INFO *RemoteInfo)
* PreCondition:	TCPInit() is already called.
* Input:		hTCP      - Handle of socket To read
* Output:		A new socket is created, connection request is
*				sent And socket handle is returned.
* Side Effects:	None
* Overview:		None
* Note:			None
********************************************************************/
*)
Public Dim RemoteInfo As SOCKET_INFO // static
Public Function TCPGetRemoteInfo(hTCP As TCP_SOCKET) As Word
   SyncTCB(hTCP)
   RemoteInfo.remote = MyTCB.remote
   RemoteInfo.remotePort.Val = MyTCB.remotePort.Val
   result = @RemoteInfo
End Function
(*
 *********************************************************************
* Function:        static void SwapTCPHeader(TCP_HEADER* header)
* PreCondition:    None
* Input:           header      - TCP Header To be swapped.
* Output:          Given header is swapped.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Sub SwapTCPHeader(ByRef header As TCP_HEADER)
   Dim h As TCP_HEADER
   h = header
   h.SourcePort      = swaps(h.SourcePort)
   h.DestPort        = swaps(h.DestPort)
   h.SeqNumber       = swapl(h.SeqNumber)
   h.AckNumber       = swapl(h.AckNumber)
   h.Window          = swaps(h.Window)
   h.Checksum        = swaps(h.Checksum)
   h.UrgentPointer   = swaps(h.UrgentPointer)
   header = h
End Sub
(*
 *********************************************************************
* Function:                  
********************************************************************/
*)
Function GetOffsetVal(Val As Byte) As Byte
   result = Val >> 4
End Function
(*
 *********************************************************************
* Function:                  
********************************************************************/
*)
Sub SetOffsetVal(ByRef CurrentValue As Byte, NewValue As Byte)
   Dim Tmp As Byte
   Tmp = CurrentValue
   CurrentValue = Byte(Tmp And $0F) Or Byte(NewValue << 4)  
End Sub
Sub SendTCP(hTCP As TCP_SOCKET, Flags As Byte, bResetTimer As Boolean)
   Dim ps As TCB_STUB
   Dim wVal As WORD_VAL
   Dim Header As TCP_HEADER
   Dim options As TCP_OPTIONS
   Dim pseudoHeader As PSEUDO_HEADER
   Dim Len As Word
   Dim wEffectiveWindow As Word
	
   ps = TCBStubs(hTCP)

   // Do not send anything on loopbacked ports
   If ps.smState = TCP_LOOPBACK Or ps.smState = TCP_LOOPBACK_CLOSED Then
      Exit
   EndIf

   SyncTCB(hTCP)

   // FINs must be handled specially
   If (Flags And FIN) <> 0 Then
      ps.Flags.bTXFin = true
      TCBStubs(hTCP).Flags = ps.Flags
	  Flags = Flags And Not FIN
   EndIf

   // Status will now be synched, disable automatic future 
   // status transmissions
   ps.Flags.bTimer2Enabled = false
   ps.Flags.bDelayedACKTimerEnabled = false
   ps.Flags.bOneSegmentReceived = false
   ps.Flags.bTXASAP = false
   ps.Flags.bHalfFullFlush = false
   TCBStubs(hTCP).Flags = ps.Flags

   //  Make sure that we can write to the MAC transmit area
   While Not IPIsTxReady 
   Wend

   // Put all socket application data in the TX space
   If (Flags And (SYN Or RST)) <> 0 Then
      // Don't put any data in SYN and RST messages
	  Len = 0
   Else
      // Begin copying any application data over to the TX space
	  If ps.txHead = MyTCB.txUnackedTail Then
         // All caught up on data TX, no real data for this packet
		 Len = 0

		 // If we are to transmit a FIN, make sure we can put one in this packet
		 If ps.Flags.bTXFin Then
		    If MyTCB.remoteWindow <> 0 Then
		       Flags = Flags Or FIN
			EndIf	
         EndIf
      ElseIf ps.txHead > MyTCB.txUnackedTail Then
	     Len = ps.txHead - MyTCB.txUnackedTail
		 wEffectiveWindow = MyTCB.remoteWindow
		 If MyTCB.txUnackedTail >= ps.txTail Then
		    wEffectiveWindow = wEffectiveWindow - (MyTCB.txUnackedTail - ps.txTail)
         Else
		    wEffectiveWindow = wEffectiveWindow - ((ps.bufferRxStart - ps.bufferTxStart) - (ps.txTail - MyTCB.txUnackedTail))
         EndIf
		 If Len > wEffectiveWindow Then
 		    Len = wEffectiveWindow
         EndIf	

		 If Len > TCP_MAX_SEG_SIZE Then
		    Len = TCP_MAX_SEG_SIZE
			ps.Flags.bTXASAP = true
            TCBStubs(hTCP).Flags = ps.Flags
         EndIf

		 // If we are to transmit a FIN, make sure we can put one in this packet
		 If ps.Flags.bTXFin Then
		    If (Len <> wEffectiveWindow) And (Len <> TCP_MAX_SEG_SIZE) Then
			   Flags = Flags Or FIN
			EndIf	
         EndIf

		 MACMemCopyAsync(BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER) + SizeOf(TCP_HEADER), MyTCB.txUnackedTail, Len)
		 MyTCB.txUnackedTail = MyTCB.txUnackedTail + Len
      Else
	     pseudoHeader.TCPLength = ps.bufferRxStart - MyTCB.txUnackedTail
		 Len = pseudoHeader.TCPLength + ps.txHead - ps.bufferTxStart

		 wEffectiveWindow = MyTCB.remoteWindow
		 If MyTCB.txUnackedTail >= ps.txTail Then
		    wEffectiveWindow = wEffectiveWindow - (MyTCB.txUnackedTail - ps.txTail)
		 Else
		    wEffectiveWindow = wEffectiveWindow - ((ps.bufferRxStart - ps.bufferTxStart) - (ps.txTail - MyTCB.txUnackedTail))
		 EndIf	
				
		 If Len > wEffectiveWindow Then
		    Len = wEffectiveWindow
		 EndIf	

		 If Len > TCP_MAX_SEG_SIZE Then
		    Len = TCP_MAX_SEG_SIZE
			ps.Flags.bTXASAP = true
			TCBStubs(hTCP).Flags = ps.Flags
		 EndIf

		 // If we are to transmit a FIN, make sure we can put one in this packet
		 If ps.Flags.bTXFin Then
		    If (Len <> wEffectiveWindow) And (Len <> TCP_MAX_SEG_SIZE) Then
			   Flags = Flags Or FIN
			EndIf	
		 EndIf

		 If pseudoHeader.TCPLength > wEffectiveWindow Then
		    pseudoHeader.TCPLength = wEffectiveWindow
		 EndIf	

		 If pseudoHeader.TCPLength > TCP_MAX_SEG_SIZE Then
		    pseudoHeader.TCPLength = TCP_MAX_SEG_SIZE
		 EndIf	

		 MACMemCopyAsync(BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER) + SizeOf(TCP_HEADER), MyTCB.txUnackedTail, pseudoHeader.TCPLength)
		 pseudoHeader.TCPLength = Len - pseudoHeader.TCPLength
	
		 // Copy any left over chunks of application data over
		 If pseudoHeader.TCPLength <> 0 Then
		    MACMemCopyAsync(BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER) + SizeOf(TCP_HEADER) + (ps.bufferRxStart - MyTCB.txUnackedTail), ps.bufferTxStart, pseudoHeader.TCPLength)
		 EndIf

		 MyTCB.txUnackedTail = MyTCB.txUnackedTail + Len
		 If MyTCB.txUnackedTail >= ps.bufferRxStart Then
		    MyTCB.txUnackedTail = MyTCB.txUnackedTail - (ps.bufferRxStart - ps.bufferTxStart)
		 EndIf	
      EndIf
   EndIf

   // Ensure that all packets with data of some kind are 
   // retransmitted by TCPTick() until acknowledged
   // Pure ACK packets with no data are not ACKed back in TCP
   If Len <> 0 Or (Flags And (SYN Or FIN)) <> 0 Then
      If bResetTimer Then
	     MyTCB.retryCount = 0
		 MyTCB.retryInterval = TCP_START_TIMEOUT_VAL
      EndIf

	  ps.eventTime = TickGet() + MyTCB.retryInterval  
      ps.Flags.bTimerEnabled = true

	  TCBStubs(hTCP).eventTime = ps.eventTime
      TCBStubs(hTCP).Flags = ps.Flags
   EndIf

   header.SourcePort			    = MyTCB.localPort.Val
   header.DestPort				    = MyTCB.remotePort.Val
   header.SeqNumber			        = MyTCB.MySEQ
   header.AckNumber			        = MyTCB.RemoteSEQ
   header.Flags.Bits._Reserved2.6	= 0
   header.Flags.Bits._Reserved2.7	= 0
   header.DataOffset._Reserved3	    = 0
   header.Flags.Val			        = Flags
   header.UrgentPointer             = 0

   // Update our send sequence number and ensure retransmissions 
   // of SYNs and FINs use the right sequence number
   MyTCB.MySEQ = MyTCB.MySEQ + Len
   If (Flags And SYN) <> 0 Then
      If MyTCB.Flags.bSYNSent Then
	     header.SeqNumber = header.SeqNumber - 1
	  Else
	     MyTCB.MySEQ = MyTCB.MySEQ + 1
		MyTCB.Flags.bSYNSent = true
	  EndIf
   EndIf

   If (Flags And FIN) <> 0 Then
      If MyTCB.Flags.bFINSent Then
	     header.SeqNumber = header.SeqNumber - 1
	  Else
	     MyTCB.MySEQ = MyTCB.MySEQ + 1
         MyTCB.Flags.bFINSent = true
	  EndIf
   EndIf

   // Calculate the amount of free space in the RX buffer area of this socket
   If ps.rxHead >= ps.rxTail Then
      header.Window = (ps.bufferEnd - ps.bufferRxStart) - (ps.rxHead - ps.rxTail)
   Else
      header.Window = ps.rxTail - ps.rxHead - 1
   EndIf	

   // Calculate the amount of free space in the MAC RX buffer area and adjust window if needed
   wVal.Val = MACGetFreeRxSize() - 64
   If SHORT(wVal.Val) < 0 Then
      wVal.Val = 0
   EndIf
    	
   // Force the remote node to throttle back if we are running low on general RX buffer space
   If header.Window > wVal.Val Then
      header.Window = wVal.Val
   EndIf	

   SwapTCPHeader(header)
   Len = Len + SizeOf(header)
   SetOffsetVal(header.DataOffset._Val, SizeOf(header) >> 2)

   // Insert the MSS (Maximum Segment Size) TCP option if this is SYN packet
   If (Flags And SYN) <> 0 Then
      Len = Len + SizeOf(options)
	  options.Kind = TCP_OPTIONS_MAX_SEG_SIZE
	  options.Length = $04

	  // Load MSS and swap to big endian
	  options.MaxSegSize.Val = (((TCP_MAX_SEG_SIZE - 4) And $00FF)<<8) Or (((TCP_MAX_SEG_SIZE-4) And $FF00)>>8)
      SetOffsetVal(header.DataOffset._Val, GetOffsetVal(header.DataOffset._Val) + (SizeOf(options) >> 2))
   EndIf

   // Calculate IP pseudoheader checksum.
   pseudoHeader.SourceAddress  = AppConfig.MyIPAddr
   pseudoHeader.DestAddress    = MyTCB.remote.IPAddr
   pseudoHeader.Zero           = $0
   pseudoHeader.Protocol       = IP_PROT_TCP
   pseudoHeader.TCPLength	   = Len
   SwapPseudoTCPHeader(pseudoHeader)
   header.Checksum = Not CalcIPChecksum(@pseudoHeader, SizeOf(pseudoHeader))

   // Write IP header
   MACSetWritePtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER))
   IPPutHeader(MyTCB.remote, IP_PROT_TCP, Len)
   MACPutArray(@header, SizeOf(header))
   If (Flags And SYN) <> 0 Then
      MACPutArray(@options, SizeOf(options))
   EndIf
    
   // Wait for any previous copy operations to complete
   While Not MACIsMemCopyDone()
   Wend

   // Update the TCP checksum
   MACSetReadPtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER))
   wVal.Val = CalcIPBufferChecksum(Len)

   MACSetWritePtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER) + 16)
   MACPutArray(@wVal, SizeOf(Word))

   // Physically start the packet transmission over the network
   MACFlush()
End Sub
(*
*********************************************************************
* Function:        TCP_SOCKET TCPConnect(NODE_INFO* remote,
*                                      Word remotePort)
* PreCondition:    TCPInit() is already called.
* Input:           remote      - remote node address info
*                  remotePort  - remote Port To be connected.
* Output:          A new socket is created, connection request is
*                  sent And socket handle is returned.
* Side Effects:    None
* Overview:        None
* Note:            By default this Function is Not included in
*                  source.  You must define STACK_CLIENT_MODE To
*                  be able To use this Function.
********************************************************************/
*)
Public Function TCPConnect(ByRef remote As NODE_INFO, remotePort As Word) As TCP_SOCKET
   Dim hTCP As TCP_SOCKET
   Dim ps As TCB_STUB

   // Find an available socket
   For hTCP = 0 To MAX_TCP_SOCKETS - 1
      ps = TCBStubs(hTCP)
      If ps.smState <> TCP_CLOSED Then
         Continue
      EndIf
      
      SyncTCB(hTCP)

      // Each new socket that is opened by this node, gets the 
	  // next sequential port number.
	  If NextPort < LOCAL_PORT_START_NUMBER Or NextPort > LOCAL_PORT_END_NUMBER Then
         NextPort = LOCAL_PORT_START_NUMBER
      EndIf   
		
      // Set the non-zero TCB fields
	  MyTCB.localPort.Val = NextPort
      Inc(NextPort)
      MyTCB.remotePort.Val = remotePort
      MyTCB.remote = remote
	  
      // Update the TCB Stub
	  ps.remoteHash.Val = (remote.IPAddr.w(1) + remote.IPAddr.w(0) + remotePort) Xor MyTCB.localPort.Val
      ps.smState = TCP_SYN_SENT
      TCBStubs(hTCP) = ps
		
      // Send TCP SYNchronize (connect) request
	  SendTCP(hTCP, SYN, true)
      result = hTCP
      Exit
   Next

   // If there is no socket available, return error.
   result = INVALID_SOCKET
End Function
(*
*********************************************************************
* Function:        BOOL TCPWasReset(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called.
* Input:           hTCP    - Socket to be checked
* Output:          TRUE: if given socket was reset at one 
*						 point since the last call to 
*						 TCPWasDisconnected().
*                  FALSE: if given socket hasn't been reset since the 
*						  last call to TCPWasReset.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPWasReset(hTCP As TCP_SOCKET) As Boolean
   If TCBStubs(hTCP).Flags.bSocketReset Then
      TCBStubs(hTCP).Flags.bSocketReset = false
      result = true
      Exit
   EndIf
   result = false
End Function
(*
*********************************************************************
* Function:        BOOL TCPIsConnected(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called.
* Input:           hTCP    - Socket To be checked For connection.
* Output:          TRUE    If given socket is connected
*                  FALSE   If given socket is Not connected.
* Side Effects:    None
* Overview:        None
* Note:            A socket is said To be connected only If it is in 
*				   the TCP_ESTABLISHED state
********************************************************************/
*)
Public Function TCPIsConnected(hTCP As TCP_SOCKET) As Boolean
   Dim i As Byte
   
   i = TCBStubs(hTCP).smState
   result = (i = TCP_ESTABLISHED) Or (i = TCP_LOOPBACK)
End Function
(*
*********************************************************************
* Function:        void TCPDisconnect(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called
* Input:           hTCP - Socket To be disconnected.
* Output:          A disconnect request is sent For given socket.  
*				   This Function does nothing If the socket isn't 
*				   currently connected.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Sub TCPDisconnect(hTCP As TCP_SOCKET)
   // Delete all data in the RX FIFO
   // In this stack's API, the application TCP handle is 
   // immediately invalid after calling this function, so there 
   // is no longer any way to receive data from the TCP RX FIFO, 
   // even though the data is still there.  Leaving the data there 
   // could interfere with the remote node sending us a FIN if our
   // RX window is zero
   TCBStubs(hTCP).rxTail = TCBStubs(hTCP).rxHead
	
   Select TCBStubs(hTCP).smState
      // These states are all already closed or don't need explicit disconnecting -- 
      // they will disconnect by themselves after a while
	  //case TCP_CLOSED:
	  //case TCP_LISTEN:
	  //case TCP_CLOSING:
	  //case TCP_TIME_WAIT:
	  //case TCP_LOOPBACK_CLOSED
	  //	return;

	  // This state will close itself after some delay, however, 
	  // this is handled so that the user can call TCPDisconnect() 
	  // twice to immediately close a socket (using an RST) without 
	  // having to get an ACK back from the remote node.  This is 
	  // great for instance when the application determines that 
	  // the remote node has been physically disconnected and 
	  // already knows that no ACK will be returned.  Alternatively, 
	  // if the application needs to immediately reuse the socket 
	  // regardless of what the other node's state is in (half open).
	  Case TCP_FIN_WAIT_1, TCP_FIN_WAIT_2, TCP_LAST_ACK
	     SendTCP(hTCP, RST Or ACK, false)
		 CloseSocket(hTCP)

      Case TCP_SYN_SENT:
	     CloseSocket(hTCP)

      Case TCP_SYN_RECEIVED, TCP_ESTABLISHED
	     // Send the FIN
		 SendTCP(hTCP, FIN Or ACK, true)
		 TCBStubs(hTCP).smState = TCP_FIN_WAIT_1
			
      Case TCP_LOOPBACK:
	     // Indicate to loopback owner that the loop is done
		 // Loopback owner must clean up later
		 TCBStubs(hTCP).smState = TCP_LOOPBACK_CLOSED
   End Select
End Sub
(*
*********************************************************************
* Function:        void TCPFlush(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called.
* Input:           s       - Socket whose data is To be transmitted.
* Output:          None
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Sub TCPFlush(hTCP As TCP_SOCKET)
   SyncTCB(hTCP)
   If TCBStubs(hTCP).txHead <> MyTCB.txUnackedTail Then
      // Send the TCP segment with all unacked bytes
	  SendTCP(hTCP, PSH Or ACK, true)
   EndIf
End Sub
(*
*********************************************************************
* Function:        Word TCPIsPutReady(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called.
* Input:           hTCP: handle of socket To test
* Output:          Number of Bytes that can be immediately placed 
*				   in the transmit buffer.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPIsPutReady(hTCP As TCP_SOCKET) As Word
   Dim ps As TCB_STUB

   ps = TCBStubs(hTCP)

   // Unconnected sockets shouldn't be transmitting anything.
   If Not ((ps.smState = TCP_ESTABLISHED) Or (ps.smState = TCP_LOOPBACK_CLOSED) Or (ps.smState = TCP_LOOPBACK)) Then
      result = 0
	  Exit
   EndIf

   // Calculate the free space in this socket's TX FIFO
   If ps.txHead >= ps.txTail Then
      result = (ps.bufferRxStart - ps.bufferTxStart - 1) - (ps.txHead - ps.txTail)
   Else
      result = ps.txTail - ps.txHead - 1
   EndIf
End Function
(*
*********************************************************************
* Function:        BOOL TCPPut(TCP_SOCKET hTCP, Byte Byte)
* PreCondition:    TCPIsPutReady(s) != 0
* Input:           hTCP: socket handle To use
*                  Byte: a data Byte To send
* Output:          TRUE If given Byte was put in transmit buffer
*                  FALSE If transmit buffer is full.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPPut(hTCP As TCP_SOCKET, _Byte As Byte) As Boolean
   Dim ps As TCB_STUB
   Dim wFreeTXSpace As Word


   wFreeTXSpace = TCPIsPutReady(hTCP)
   If wFreeTXSpace = 0 Then
      result = false
      Exit
   ElseIf wFreeTXSpace = 1 Then // About to run out of space, lets transmit so the remote node might send an ACK back faster
      TCPFlush(hTCP)
   EndIf   	

   ps = TCBStubs(hTCP)

   // Send all current bytes if we are crossing half full
   // This is required to improve performance with the delayed 
   // acknowledgement algorithm
   If Not ps.Flags.bHalfFullFlush And wFreeTXSpace <= ((ps.bufferRxStart - ps.bufferTxStart) >> 1) Then
      TCPFlush(hTCP)	
      ps.Flags.bHalfFullFlush = true
      TCBStubs(hTCP).Flags = ps.Flags
   EndIf

   MACSetWritePtr(ps.txHead)
   MACPut(_Byte)
   Inc(ps.txHead)
   If ps.txHead >= ps.bufferRxStart Then
      ps.txHead = ps.bufferTxStart
   EndIf
   TCBStubs(hTCP).txHead = ps.txHead   

   // Send the last byte as a separate packet (likely will make the remote node send back ACK faster)
   If wFreeTXSpace = 1 Then
      TCPFlush(hTCP)

   // If not already enabled, start a timer so this data will 
   // eventually get sent even if the application doens't call
   // TCPFlush()
   ElseIf Not ps.Flags.bTimer2Enabled Then
      TCBStubs(hTCP).Flags.bTimer2Enabled = true
'      TCBStubs(hTCP).eventTime2 = Word(TickGetDiv256) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL / 256
	  TCBStubs(hTCP).eventTime2 = Word(TickGet) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL
   EndIf
   result = true
End Function
(*
*********************************************************************
* Function:        Word TCPPutArray(TCP_SOCKET hTCP, Byte *data, Word len)
* PreCondition:    None
* Input:           hTCP    - Socket handle To use
*                  data    - Pointer To data To put
*				   len     - Count of Bytes To put
* Output:          Count of Bytes actually placed in the TX buffer
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPPutArray(hTCP As TCP_SOCKET, data As Word, Len As Word) As Word
   Dim wActualLen As Word
   Dim wFreeTXSpace As Word
   Dim wRightLen As Word
   Dim ps As TCB_STUB

   wRightLen = 0
    
   wFreeTXSpace = TCPIsPutReady(hTCP)
   If wFreeTXSpace = 0 Then
      TCPFlush(hTCP)
	  result = 0
      Exit
   EndIf

   wActualLen = wFreeTXSpace
   If wFreeTXSpace > Len Then
      wActualLen = Len
   EndIf   

   // Send all current bytes if we are crossing half full
   // This is required to improve performance with the delayed 
   // acknowledgement algorithm
   ps = TCBStubs(hTCP)
   If Not ps.Flags.bHalfFullFlush And wFreeTXSpace <= ((ps.bufferRxStart - ps.bufferTxStart) >> 1) Then
      TCPFlush(hTCP)	
      TCBStubs(hTCP).Flags.bHalfFullFlush = true
   EndIf
	
   // See if we need a two part put
   If ps.txHead + wActualLen >= ps.bufferRxStart Then
      wRightLen = ps.bufferRxStart - ps.txHead
	  MACSetWritePtr(ps.txHead)
	  MACPutArray(data, wRightLen)
	  data = data + wRightLen
      wActualLen = wActualLen - wRightLen
	  ps.txHead = ps.bufferTxStart
      TCBStubs(hTCP).txHead = ps.txHead
   EndIf

   MACSetWritePtr(ps.txHead)
   MACPutArray(data, wActualLen)
   ps.txHead = ps.txHead + wActualLen
   TCBStubs(hTCP).txHead = ps.txHead
   
   // Send these bytes right now if we are out of TX buffer space
   If wFreeTXSpace <= Len Then
      TCPFlush(hTCP)

   // If not already enabled, start a timer so this data will 
   // eventually get sent even if the application doens't call
   // TCPFlush()
   ElseIf Not ps.Flags.bTimer2Enabled Then
      TCBStubs(hTCP).Flags.bTimer2Enabled = true
'      TCBStubs(hTCP).eventTime2 = Word(TickGetDiv256) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL / 256
	  TCBStubs(hTCP).eventTime2 = Word(TickGet) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL
   EndIf
   result = wActualLen + wRightLen
End Function
(*
 *********************************************************************
* Function:                  
********************************************************************/
*)
Public Function TCPPutROMArray(hTCP As TCP_SOCKET, data As TROMAddress, Len As Word) As Word
   Dim wActualLen As Word
   Dim wFreeTXSpace As Word
   Dim wRightLen As Word
   Dim ps As TCB_STUB

   wRightLen = 0
    
   wFreeTXSpace = TCPIsPutReady(hTCP)
   If wFreeTXSpace = 0 Then
      TCPFlush(hTCP)
	  result = 0
      Exit
   EndIf

   wActualLen = wFreeTXSpace
   If wFreeTXSpace > Len Then
      wActualLen = Len
   EndIf   

   // Send all current bytes if we are crossing half full
   // This is required to improve performance with the delayed 
   // acknowledgement algorithm
   ps = TCBStubs(hTCP)
   If Not ps.Flags.bHalfFullFlush And wFreeTXSpace <= ((ps.bufferRxStart - ps.bufferTxStart) >> 1) Then
      TCPFlush(hTCP)	
      TCBStubs(hTCP).Flags.bHalfFullFlush = true
   EndIf
	
   // See if we need a two part put
   If ps.txHead + wActualLen >= ps.bufferRxStart Then
      wRightLen = ps.bufferRxStart - ps.txHead
	  MACSetWritePtr(ps.txHead)
	  MACPutROMArray(data, wRightLen)
	  data = data + wRightLen
      wActualLen = wActualLen - wRightLen
	  ps.txHead = ps.bufferTxStart
      TCBStubs(hTCP).txHead = ps.txHead
   EndIf

   MACSetWritePtr(ps.txHead)
   MACPutROMArray(data, wActualLen)
   ps.txHead = ps.txHead + wActualLen
   TCBStubs(hTCP).txHead = ps.txHead
   
   // Send these bytes right now if we are out of TX buffer space
   If wFreeTXSpace <= Len Then
      TCPFlush(hTCP)

   // If not already enabled, start a timer so this data will 
   // eventually get sent even if the application doens't call
   // TCPFlush()
   ElseIf Not ps.Flags.bTimer2Enabled Then
      TCBStubs(hTCP).Flags.bTimer2Enabled = true
'      TCBStubs(hTCP).eventTime2 = Word(TickGetDiv256) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL / 256
      TCBStubs(hTCP).eventTime2 = Word(TickGet) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL
   EndIf
   result = wActualLen + wRightLen
End Function
(*
*********************************************************************
* Function:        Word TCPPutString(TCP_SOCKET hTCP, Byte *data)
* PreCondition:    None
* Input:           hTCP    - Socket handle To use
*                  data    - Pointer To null terminated String To put
* Output:          Byte* - *data pointer incremented by the number of 
*							Bytes actually placed in the TX FIFO
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPPutString(hTCP As TCP_SOCKET, data As String) As Word
   result = @data + TCPPutArray(hTCP, @data, strlen(@data))
End Function
Public Function TCPPutString(hTCP As TCP_SOCKET, data As Word) As Word
   result = data + TCPPutArray(hTCP, data, strlen(data))
End Function
Public Function TCPPutROMString(hTCP As TCP_SOCKET, data As TROMAddress) As TROMAddress
   result = data + TCPPutROMArray(hTCP, data, strlenpgm(data))
End Function
(*
*********************************************************************
* Function:		Word TCPGetTxFIFOFull(TCP_SOCKET hTCP)
* PreCondition:	TCPInit() is already called
* Input:		hTCP - TCP socket handle
* Output:		Number of Bytes that are pending in TxFIFO.  
*				0 is returned If the FIFO is completely empty.
* Side Effects:	None
* Overview:		None
* Note:			None
********************************************************************/
*)
Public Function TCPGetTxFIFOFull(hTCP As TCP_SOCKET) As Word
   Dim wDataLen As Word
   Dim wFIFOSize As Word

   // Calculate total usable FIFO size
   wFIFOSize = TCBStubs(hTCP).bufferRxStart - TCBStubs(hTCP).bufferTxStart - 1

   // Find out how many data bytes are free in the TX FIFO
   wDataLen = TCPIsPutReady(hTCP)
   result = wFIFOSize - wDataLen
End Function
(*
*********************************************************************
* Function:        Word TCPIsGetReady(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called.
* Input:           hTCP       - socket To test
* Output:          Number of Bytes that are available in socket 'hTCP' 
*				   For immediate retrieval
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPIsGetReady(hTCP As TCP_SOCKET) As Word
   Dim ps As TCB_STUB
   
   ps = TCBStubs(hTCP)
   
   // If the loopback is closed, reject all input
   If ps.smState = TCP_LOOPBACK_CLOSED Then
      result = 0
      Exit
   EndIf
           
   If ps.rxHead >= ps.rxTail Then
      result = ps.rxHead - ps.rxTail
   Else
      result = (ps.bufferEnd - ps.rxTail + 1) + (ps.rxHead - ps.bufferRxStart)
   EndIf
End Function
(*
*********************************************************************
* Function:        void TCPDiscard(TCP_SOCKET hTCP)
* PreCondition:    TCPInit() is already called.
* Input:           hTCP       - socket handle
* Output:          None
* Side Effects:    None
* Overview:        Removes all pending data from the socket's RX 
*				   FIFO.
* Note:            None
********************************************************************/
*)
Sub TCPDiscard(hTCP As TCP_SOCKET)
   If TCPIsGetReady(hTCP) <> 0 Then
      // Delete all data in the RX buffer
      TCBStubs(hTCP).rxTail = TCBStubs(hTCP).rxHead

  	  // Send a Window update message to the remote node
	  SendTCP(hTCP, ACK, true)
   EndIf
End Sub
(*
*********************************************************************
* Function:        BOOL TCPGet(TCP_SOCKET hTCP, Byte *Byte)
* PreCondition:    TCPInit() is already called
* Input:           hTCP    - socket
*                  Byte    - Pointer To a Byte.
* Output:          TRUE If a Byte was read.
*                  FALSE If Byte was Not read.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPGet(hTCP As TCP_SOCKET, ByRef Val As Byte) As Boolean
   Dim ps As TCB_STUB
   Dim GetReadyCount As Word

   GetReadyCount = TCPIsGetReady(hTCP)
   If GetReadyCount = 0 Then
      result = false
      Exit
   EndIf

   ps = TCBStubs(hTCP)
   MACSetReadPtr(TCBStubs(hTCP).rxTail)
   Val = MACGet()
   Inc(ps.rxTail)
   If ps.rxTail > ps.bufferEnd Then
      ps.rxTail = ps.bufferRxStart
   EndIf   
   TCBStubs(hTCP).rxTail = ps.rxTail
   
   // Send a window update if we've run out of data
   If GetReadyCount = 1 Then
      TCBStubs(hTCP).Flags.bTXASAP = true
      
   // If not already enabled, start a timer so a window 
   // update will get sent to the remote node at some point

   ElseIf Not ps.Flags.bTimer2Enabled Then
      TCBStubs(hTCP).Flags.bTimer2Enabled = true
'      TCBStubs(hTCP).eventTime2 = Word(TickGetDiv256) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL / 256
	  TCBStubs(hTCP).eventTime2 = Word(TickGet) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL
   EndIf
   result = true
End Function
(*
*********************************************************************
* Function:        Word TCPGetArray(TCP_SOCKET hTCP, Byte *buffer,
*                                      Word len)
* PreCondition:    TCPInit() is already called
* Input:           hTCP    - socket handle
*                  buffer  - Buffer To hold received data.
*                  len     - Buffer Length
* Output:          Number of Bytes loaded into buffer.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPGetArray(hTCP As TCP_SOCKET, buffer As Word, Len As Word) As Word
   Dim ActualLen As Word
   Dim RightLen As Word
   Dim ps As TCB_STUB
	
   RightLen = 0
   ps = TCBStubs(hTCP)

   ActualLen = TCPIsGetReady(hTCP)
   If Len > ActualLen Then
      Len = ActualLen
   EndIf   

   // See if we need a two part get
   If ps.rxTail + Len > ps.bufferEnd Then
      RightLen = ps.bufferEnd - ps.rxTail + 1
	  MACSetReadPtr(ps.rxTail)
      MACGetArray(buffer, RightLen)
	  If buffer <> 0 Then
	     buffer = buffer + RightLen
	  EndIf   
      Len = Len - RightLen
      ps.rxTail = ps.bufferRxStart
   EndIf

   MACSetReadPtr(ps.rxTail)
   MACGetArray(buffer, Len)
   ps.rxTail = ps.rxTail + Len
   TCBStubs(hTCP).rxTail = ps.rxTail

   // Send a window update if we've run low on data
   If ActualLen - Len <= Len Then
      TCBStubs(hTCP).Flags.bTXASAP = true

   ElseIf Not ps.Flags.bTimer2Enabled Then
      // If not already enabled, start a timer so a window 
	  // update will get sent to the remote node at some point
	  TCBStubs(hTCP).Flags.bTimer2Enabled = true
'      TCBStubs(hTCP).eventTime2 = Word(TickGetDiv256) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL / 256
	  TCBStubs(hTCP).eventTime2 = Word(TickGet) + TCP_AUTO_TRANSMIT_TIMEOUT_VAL
   EndIf

   result = Len + RightLen
End Function
(*
 *********************************************************************
* Function:		Word TCPGetRxFIFOFree(TCP_SOCKET hTCP)
* PreCondition:	TCPInit() is already called
* Input:		hTCP - TCP socket handle
* Output:		Number of Bytes that can still fit in the RX FIFO.  
*				0 is returned If the FIFO is completely full.
* Side Effects:	None
* Overview:		None
* Note:			None
********************************************************************/
*)
Public Function TCPGetRxFIFOFree(hTCP As TCP_SOCKET) As Word
   Dim wDataLen As Word
   Dim wFIFOSize As Word

   // Calculate total usable FIFO size
   wFIFOSize = TCBStubs(hTCP).bufferEnd - TCBStubs(hTCP).bufferRxStart

   // Find out how many data bytes are actually in the RX FIFO
   wDataLen = TCPIsGetReady(hTCP)
   
   result = wFIFOSize - wDataLen
End Function
(*
/*********************************************************************
* Function:		Word TCPFindArrayEx(TCP_SOCKET hTCP, Byte *cFindArray, Word wLen, Word wStart, Word Wend, BOOL bTextCompare)
* PreCondition:	TCPInit() is already called
* Input:		hTCP       - TCP socket handle
*				cFindArray - Pointer To an array of Bytes To look For
*				wLen       - Length of cFindArray
*				wStart	   - Position within the RX FIFO To begin searching. 0 is the beginning of the FIFO.
*				wSearchLen - Number of Bytes To look at from start position
*				bTextCompare- Search using a Case-insensitive algorithm If TRUE, otherwise, do a binary 1 To 1 search
* Output:		0xFFFF: Array Not found
*				0 To 65534: Location the array was found
* Side Effects:	None
* Overview:		None
* Note:			As an example, If the RX FIFO contains:
*				"PIC MCUs are good." 
*				And cFindArray -> "MCU"
*				And wLen -> 3
*				the value returned by the search Function would be 4
********************************************************************/
*)
Public Function TCPFindArrayEx(hTCP As TCP_SOCKET, cFindArray As Word, wLen As Word, wStart As Word, wSearchLen As Word, bTextCompare As Boolean) As Word
   Dim wDataLen As Word
   Dim wBytesUntilWrap As Word
   Dim wLocation As Word
   Dim wLenStart As Word
   Dim cFindArrayStart As Byte
   Dim i, j, k As Byte
   Dim isFinding As Boolean
   Dim buffer(32) As Byte
   Dim ps As TCB_STUB

   If wLen = 0 Then
      result = 0
      Exit
   EndIf

   // Find out how many bytes are in the RX FIFO and return 
   // immediately if we won't possibly find a match
   wDataLen = TCPIsGetReady(hTCP) - wStart
   If wDataLen < wLen Then
      result = $FFFF
      Exit
   EndIf   

   If wSearchLen <> 0 And wDataLen > wSearchLen Then
      wDataLen = wSearchLen
   EndIf   
		
   ps = TCBStubs(hTCP)
   wLocation = ps.rxTail + wStart
   If wLocation > ps.bufferEnd Then
      wLocation = wLocation - (ps.bufferEnd - ps.bufferRxStart + 1)
   EndIf   
   MACSetReadPtr(wLocation)
   wBytesUntilWrap = ps.bufferEnd - wLocation + 1
   wLocation = wStart
   wLenStart = wLen
   cFindArrayStart = cFindArray
   
   FSR0 = cFindArray
   j = INDF0
   Inc(cFindArray)
   
   isFinding = false
   If bTextCompare Then
      If j >= Byte("a") And j <= Byte("z") Then
	     j = j + (Byte("A") - Byte("a"))
      EndIf
   EndIf

   // Search for the array
   While true

      // Figure out how big of a chunk to read
	  k = SizeOf(buffer)
	  If k > wBytesUntilWrap Then
	     k = wBytesUntilWrap
	  EndIf   
	  If k > wDataLen Then
	     k = wDataLen
	  EndIf   

      // Read a chunk of data into the buffer
	  MACGetArray(@buffer, k)
	  wBytesUntilWrap = wBytesUntilWrap - k

      If wBytesUntilWrap = 0 Then
         MACSetReadPtr(ps.bufferRxStart)
		 wBytesUntilWrap = $FFFF
      EndIf

      // Convert everything to uppercase
	  If bTextCompare Then
	     i = 0
	     While i < k
		    If buffer(i) >= Byte("a") And  buffer(i) <= Byte("z") Then
			   buffer(i) = buffer(i) + (Byte("A") - Byte("a"))
            EndIf

			If j = buffer(i) Then
               Dec(wLen)
               If wLen = 0 Then
                  result =  wLocation - wLenStart + i + 1
                  Exit
               EndIf
               FSR0 = cFindArray
               Inc(cFindArray)
			   j = INDF0
			   isFinding = true
			   If j >= Byte("a") And j <= Byte("z") Then
			      j = j + (Byte("A") - Byte("a"))
			   EndIf   
            Else
			   wLen = wLenStart
			   If isFinding Then
			      cFindArray = cFindArrayStart
                  FSR0 = cFindArray
                  Inc(cFindArray)
			      j = INDF0
			      If j >= Byte("a") And j <= Byte("z") Then
			         j = j + (Byte("A") - Byte("a"))
                  EndIf
                  isFinding = false
               EndIf   
            EndIf   
            Inc(i)
         Wend
      Else	// Compare as is
         i = 0
         While i < k
		    If j = buffer(i) Then
			   Dec(wLen)
               If wLen = 0 Then
			      result = wLocation - wLenStart + i + 1
			      Exit
			   EndIf  
               FSR0 = cFindArray
               Inc(cFindArray)
			   j = INDF0
			   isFinding = true
			Else
			   wLen = wLenStart
			   If isFinding Then
			      cFindArray = cFindArrayStart
                  FSR0 = cFindArray
                  Inc(cFindArray)
			      j = INDF0
			      isFinding = false
			   EndIf   
			EndIf
            Inc(i)
         Wend
      EndIf

      // Check to see if it is impossible to find a match
      wDataLen = wDataLen - k
	  If wDataLen < wLen Then
	     result = $FFFF
         Exit
      EndIf

	  wLocation = wLocation + k
   Wend
End Function
(*
/*********************************************************************
* Function:		Word TCPFindROMArrayEx(TCP_SOCKET hTCP, Byte *cFindArray, Word wLen, Word wStart, Word Wend, BOOL bTextCompare)
* PreCondition:	TCPInit() is already called
* Input:		hTCP       - TCP socket handle
*				cFindArray - Pointer To an array of Bytes To look For
*				wLen       - Length of cFindArray
*				wStart	   - Position within the RX FIFO To begin searching. 0 is the beginning of the FIFO.
*				wSearchLen - Number of Bytes To look at from start position
*				bTextCompare- Search using a Case-insensitive algorithm If TRUE, otherwise, do a binary 1 To 1 search
* Output:		0xFFFF: Array Not found
*				0 To 65534: Location the array was found
* Side Effects:	None
* Overview:		None
* Note:			As an example, If the RX FIFO contains:
*				"PIC MCUs are good." 
*				And cFindArray -> "MCU"
*				And wLen -> 3
*				the value returned by the search Function would be 4
********************************************************************/
*)
Public Function TCPFindROMArrayEx(hTCP As TCP_SOCKET, cFindArray As TROMAddress, wLen As Word, wStart As Word, wSearchLen As Word, bTextCompare As Boolean) As Word
   Dim wDataLen As Word
   Dim wBytesUntilWrap As Word
   Dim wLocation As Word
   Dim wLenStart As Word
   Dim cFindArrayStart As Byte
   Dim i, j, k As Byte
   Dim isFinding As Boolean
   Dim buffer(32) As Byte
   Dim ps As TCB_STUB

   If wLen = 0 Then
      result = 0
      Exit
   EndIf

   // Find out how many bytes are in the RX FIFO and return 
   // immediately if we won't possibly find a match
   wDataLen = TCPIsGetReady(hTCP) - wStart
   If wDataLen < wLen Then
      result = $FFFF
      Exit
   EndIf   

   If wSearchLen <> 0 And wDataLen > wSearchLen Then
      wDataLen = wSearchLen
   EndIf   
		
   ps = TCBStubs(hTCP)
   wLocation = ps.rxTail + wStart
   If wLocation > ps.bufferEnd Then
      wLocation = wLocation - (ps.bufferEnd - ps.bufferRxStart + 1)
   EndIf   
   MACSetReadPtr(wLocation)
   wBytesUntilWrap = ps.bufferEnd - wLocation + 1
   wLocation = wStart
   wLenStart = wLen
   cFindArrayStart = cFindArray
   
   j = ReadROM(cFindArray)
   Inc(cFindArray)
   
   isFinding = false
   If bTextCompare Then
      If j >= Byte("a") And j <= Byte("z") Then
	     j = j + (Byte("A") - Byte("a"))
      EndIf
   EndIf

   // Search for the array
   While true

      // Figure out how big of a chunk to read
	  k = SizeOf(buffer)
	  If k > wBytesUntilWrap Then
	     k = wBytesUntilWrap
	  EndIf   
	  If k > wDataLen Then
	     k = wDataLen
	  EndIf   

      // Read a chunk of data into the buffer
	  MACGetArray(@buffer, k)
	  wBytesUntilWrap = wBytesUntilWrap - k

      If wBytesUntilWrap = 0 Then
         MACSetReadPtr(ps.bufferRxStart)
		 wBytesUntilWrap = $FFFF
      EndIf

      // Convert everything to uppercase
	  If bTextCompare Then
	     i = 0
	     While i < k
		    If buffer(i) >= Byte("a") And  buffer(i) <= Byte("z") Then
			   buffer(i) = buffer(i) + (Byte("A") - Byte("a"))
            EndIf

			If j = buffer(i) Then
               Dec(wLen)
               If wLen = 0 Then
                  result =  wLocation - wLenStart + i + 1
                  Exit
               EndIf
               j = ReadROM(cFindArray)
               Inc(cFindArray)
			   isFinding = true
			   If j >= Byte("a") And j <= Byte("z") Then
			      j = j + (Byte("A") - Byte("a"))
			   EndIf   
            Else
			   wLen = wLenStart
			   If isFinding Then
			      cFindArray = cFindArrayStart
			      j = ReadROM(cFindArray)
                  Inc(cFindArray)
			      If j >= Byte("a") And j <= Byte("z") Then
			         j = j + (Byte("A") - Byte("a"))
                  EndIf
                  isFinding = false
               EndIf   
            EndIf   
            Inc(i)
         Wend
      Else	// Compare as is
         i = 0
         While i < k
		    If j = buffer(i) Then
			   Dec(wLen)
               If wLen = 0 Then
			      result = wLocation - wLenStart + i + 1
			      Exit
			   EndIf  
               j = ReadROM(cFindArray)
               Inc(cFindArray)
			   isFinding = true
			Else
			   wLen = wLenStart
			   If isFinding Then
			      cFindArray = cFindArrayStart
                  j = ReadROM(cFindArray)
                  Inc(cFindArray)
			      isFinding = false
			   EndIf   
			EndIf
            Inc(i)
         Wend
      EndIf

      // Check to see if it is impossible to find a match
      wDataLen = wDataLen - k
	  If wDataLen < wLen Then
	     result = $FFFF
         Exit
      EndIf

	  wLocation = wLocation + k
   Wend
End Function

Public Function TCPFindROMArray(hTCP As TCP_SOCKET, cFindArray As TROMAddress, wLen As Word, wStart As Word, bTextCompare As Boolean) As Word
   result = TCPFindROMArrayEx(hTCP, cFindArray, wLen, wStart, 0, bTextCompare)
End Function

Public Function TCPFindEx(hTCP As TCP_SOCKET, cFind As Byte, wStart As Word, wSearchLen As Word, bTextCompare As Boolean) As Word
   result = TCPFindArrayEx(hTCP, @cFind, SizeOf(cFind), wStart, wSearchLen, bTextCompare)
End Function

Public Function TCPFind(hTCP As TCP_SOCKET, cFind As Byte, wStart As Word, bTextCompare As Boolean) As Word
   result = TCPFindEx(hTCP, cFind, wStart, 0, bTextCompare)
End Function
{
****************************************************************************
* Name    : DoTCPTick (PRIVATE)                                            *
* Purpose :                                                                *
****************************************************************************
}
Sub DoTCPTick(hTCP As TCP_SOCKET, ByRef ps As TCB_STUB)
   Dim bRetransmit As Boolean
   Dim bCloseSocket As Boolean
   Dim vFlags As Byte

   vFlags = $0
   bRetransmit = false
   bCloseSocket = false

   // Transmit ASAP data if the medium is available
   If ps.Flags.bTXASAP Then
      If MACIsTxReady() Then
	     vFlags = ACK
      EndIf	
   EndIf

   // Perform any needed window updates and data transmissions
   If ps.Flags.bTimer2Enabled Then
      // See if the timeout has occured, and we need to send a new window update and pending data
'	  If SHORT(ps.eventTime2 - Word(TickGetDiv256)) <= 0 Then
      If SHORT(ps.eventTime2 - Word(TickGet)) <= 0 Then
	     vFlags = ACK  
      EndIf	
   EndIf

   // Process Delayed ACKnowledgement timer
   If ps.Flags.bDelayedACKTimerEnabled Then
      // See if the timeout has occured and delayed ACK needs to be sent
'	   If SHORT(ps.delayedACKTime - Word(TickGetDiv256)) <= 0 Then
       If SHORT(ps.delayedACKTime - Word(TickGet)) <= 0 Then
	      vFlags = ACK    
       EndIf
   EndIf

   If vFlags <> 0 Then
      SendTCP(hTCP, ACK, true)   
   EndIf

        
   // The TCP_CLOSED, TCP_LISTEN, and sometimes the TCP_ESTABLISHED 
   // state don't need any timeout events
   If Not ps.Flags.bTimerEnabled Then
      Exit
   EndIf

   // If timeout has not occured, do not do anything.
   If LONG(TickGet() - ps.eventTime) < 0 Then
      Exit
   EndIf 

   // Load up extended TCB information
   SyncTCB(hTCP)

   // A timeout has occured.  Respond to this timeout condition
   // depending on what state this socket is in.
   Select ps.smState
      Case TCP_SYN_SENT:
	     // Keep sending SYN until we hear from remote node.
		 // This may be for infinite time, in that case
		 // caller must detect it and do something.
		 vFlags = SYN
		 bRetransmit = true
	
      Case TCP_SYN_RECEIVED:
	     // We must receive ACK before timeout expires.
		 // If not, resend SYN+ACK.
		 // Abort, if maximum attempts counts are reached.
		 If MyTCB.retryCount < TCP_MAX_SYN_RETRIES Then
		   vFlags = SYN Or ACK
		   bRetransmit = true
		 Else
		    If ps.Flags.bServer Then
		       vFlags = RST Or ACK
			   bCloseSocket = true
	        Else
			   vFlags = SYN
			EndIf	
	     EndIf
	
      Case TCP_ESTABLISHED:
	     // Retransmit any unacknowledged data
		 If MyTCB.retryCount < TCP_MAX_RETRIES Then
		    vFlags = PSH Or ACK
			bRetransmit = true
	     Else
		    // No response back for too long, close connection
			// This could happen, for instance, if the communication 
			// medium was lost
			ps.smState = TCP_FIN_WAIT_1
			vFlags = FIN Or ACK
	     EndIf
	
      Case TCP_FIN_WAIT_1:
	      If MyTCB.retryCount < TCP_MAX_RETRIES Then
		     // Send another FIN
			 vFlags = FIN Or ACK
			 bRetransmit = true
	      Else
		     // Close on our own, we can't seem to communicate 
			 // with the remote node anymore
			 vFlags = RST Or ACK
			 bCloseSocket = true
	      EndIf
	
      Case TCP_FIN_WAIT_2:
         // Close on our own, we can't seem to communicate 
		 // with the remote node anymore
		 vFlags = RST Or ACK
		 bCloseSocket = true

      Case TCP_CLOSING:
	    If MyTCB.retryCount < TCP_MAX_RETRIES Then
           // Send another ACK+FIN (the FIN is retransmitted 
           // automatically since it hasn't been acknowledged by 
           // the remote node yet)
		   vFlags = ACK
		   bRetransmit = true
        Else
		   // Close on our own, we can't seem to communicate 
		   // with the remote node anymore
		   vFlags = RST Or ACK
		   bCloseSocket = true
        EndIf
	
      Case TCP_LAST_ACK:
	     // Send some more FINs or close anyway
		 If MyTCB.retryCount < TCP_MAX_RETRIES Then
            vFlags = FIN Or ACK
			bRetransmit = true
	     Else
		    vFlags = RST Or ACK
			bCloseSocket = true
	    EndIf
   End Select

   If vFlags <> 0 Then
      If bRetransmit Then
	     // Set the appropriate retry time
		 MyTCB.retryCount = MyTCB.retryCount + 1
		 MyTCB.retryInterval = MyTCB.retryInterval << 1   

		 // Transmit all unacknowledged data over again
		 // Roll back unacknowledged TX tail pointer to cause retransmit to occur
		 MyTCB.MySEQ = MyTCB.MySEQ - (MyTCB.txUnackedTail - ps.txTail)
		 If MyTCB.txUnackedTail < ps.txTail Then
		    MyTCB.MySEQ = MyTCB.MySEQ - (ps.bufferRxStart - ps.bufferTxStart)
		 EndIf
         MyTCB.txUnackedTail = ps.txTail		
      EndIf
	  SendTCP(hTCP, vFlags, Not bRetransmit)
   EndIf
		
   If bCloseSocket Then
      CloseSocket(hTCP)
   EndIf	
End Sub
(*
*********************************************************************
* Function:        void TCPTick(void)
* PreCondition:    TCPInit() is already called.
* Input:           None
* Output:          Each socket FSM is executed For any timeout
*                  situation.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Sub TCPTick()

   Dim hTCP As TCP_SOCKET

   // Periodically all "not closed" sockets must perform timed operations
   For hTCP = 0 To MAX_TCP_SOCKETS - 1
      DoTCPTick(hTCP,TCBStubs(hTCP))
   Next
End Sub
(*
*********************************************************************
* Function:        static TCP_SOCKET FindMatchingSocket(TCP_HEADER *h,
*                                      NODE_INFO* remote)
* PreCondition:    TCPInit() is already called
* Input:           h           - TCP Header To be matched against.
*                  remote      - Node who sent this header.
* Output:          A socket that matches With given header And remote
*                  node is searched.
*                  If such socket is found, its index is returned
*                  Else INVALID_SOCKET is returned.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Function FindMatchingSocket(ByRef h As TCP_HEADER, ByRef remote As NODE_INFO) As TCP_SOCKET
   Dim ps As TCB_STUB
   Dim hTCP As TCP_SOCKET 
   Dim partialMatch As TCP_SOCKET
   Dim hash As Word

   partialMatch = INVALID_SOCKET
   hash = (remote.IPAddr.w(1) + remote.IPAddr.w(0) + h.SourcePort) Xor h.DestPort

   For hTCP = 0 To MAX_TCP_SOCKETS - 1 
      ps = TCBStubs(hTCP)
	  If ps.smState = TCP_CLOSED Then
         Continue
      ElseIf ps.smState = TCP_LISTEN Then
         If ps.remoteHash.Val = h.DestPort Then
	 	    partialMatch = hTCP
	 	 EndIf   
		 Continue
      ElseIf ps.remoteHash.Val <> hash Then
         Continue
      EndIf

      SyncTCB(hTCP)
	  If h.DestPort = MyTCB.localPort.Val And h.SourcePort = MyTCB.remotePort.Val And remote.IPAddr.Val = MyTCB.remote.IPAddr.Val Then
         result = hTCP
         Exit
      EndIf
   Next

   // We are not listening on this port, nor is a socket using it
   If partialMatch = INVALID_SOCKET Then
      result = INVALID_SOCKET
      Exit
   EndIf      

   // Set up the extended TCB with the info needed to establish a 
   // connection and return this socket to the caller
   ps = TCBStubs(partialMatch)
   SyncTCB(partialMatch)

   ps.remoteHash.Val = hash
  
   MyTCB.remote = remote
   MyTCB.remotePort.Val = h.SourcePort
   MyTCB.localPort.Val = h.DestPort
   
   MyTCB.MySEQ = rand()
   MyTCB.txUnackedTail = ps.bufferTxStart
   TCBStubs(partialMatch) = ps
   
   result = partialMatch
End Function
(*
*********************************************************************
* Function:        static void HandleTCPSeg(TCP_SOCKET hTCP, TCP_HEADER *h, Word len)
* PreCondition:    TCPInit() is already called     And
*                  TCPProcess() is the caller.
* Input:           hTCP        - Socket that owns this segment
*                  h           - TCP Header
*                  len         - Total buffer Length.
* Output:          TCP FSM is executed on given socket With
*                  given TCP segment.
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Sub HandleTCPSeg(ByRef ps As TCB_STUB, hTCP As TCP_SOCKET, ByRef h As TCP_HEADER, Len As Word)
   Dim dwTemp As DWORD
   Dim wTemp As Word
   Dim lMissingBytes As LONG
   Dim wMissingBytes As Word
   Dim wFreeSpace As Word
   Dim localState As TCP_STATE
   Dim localHeaderFlags As Byte
   Dim localAckNumber As DWORD
   Dim localSeqNumber As DWORD
   Dim wSegmentLength As Word
   Dim bSegmentAcceptable As Boolean

   SyncTCB(hTCP)

   // Cache a few variables in local RAM.  
   // PIC18s take a fair amount of code and execution time to 
   // dereference pointers frequently.
   localState = ps.smState
   localHeaderFlags = h.Flags.Val
   localAckNumber = h.AckNumber
   localSeqNumber = h.SeqNumber

   // Handle TCP_LISTEN and TCP_SYN_SENT states
   // Both of these states will return, so code following this 
   // state machine need not check explicitly for these two 
   // states.
   Select localState
      Case TCP_LISTEN:
	     // First: check RST flag
		 If (localHeaderFlags And RST) <> 0 Then
		    CloseSocket(hTCP)	// Unbind remote IP address/port info
			Exit
		 EndIf

		 // Second: check ACK flag, which would be invalid
		 If (localHeaderFlags And ACK) <> 0 Then
		    // Use a believable sequence number and reset the remote node
			MyTCB.MySEQ = localAckNumber
			SendTCP(hTCP, RST, false)
			CloseSocket(hTCP)	// Unbind remote IP address/port info
			Exit
         EndIf

		 // Third: check for SYN flag, which is what we're looking for
		 If (localHeaderFlags And SYN) <> 0 Then
		    // We now have a sequence number for the remote node
			MyTCB.RemoteSEQ = localSeqNumber + 1

			// Set Initial Send Sequence (ISS) number
			// Nothing to do on this step... ISS already set in CloseSocket()
				
			// Respond with SYN + ACK
			SendTCP(hTCP, SYN Or ACK, true)
			ps.smState = TCP_SYN_RECEIVED
         Else
		    CloseSocket(hTCP)	// Unbind remote IP address/port info
         EndIf

		 // Fourth: check for other text and control
		 // Nothing to do since we don't support this
		 Exit

      Case TCP_SYN_SENT:
	     // Second: check the RST bit
		 // This is out of order because this stack has no API for 
		 // notifying the application that the connection seems to 
		 // be failing.  Instead, the application must time out and 
		 // the stack will just keep trying in the mean time.
		 If (localHeaderFlags And RST) <> 0 Then
		    Exit
         EndIf

		 // First: check ACK bit
		 If (localHeaderFlags And ACK) <>0 Then
		    If localAckNumber <> MyTCB.MySEQ Then
               
               // Send a RST packet with SEQ = SEG.ACK, but retain our SEQ 
               // number for arivial of any other SYN+ACK packets
               localSeqNumber = MyTCB.MySEQ	// Save our original SEQ number
               MyTCB.MySEQ = localAckNumber	// Set SEQ = SEG.ACK
               SendTCP(hTCP, RST, true)		// Send the RST
               MyTCB.MySEQ = localSeqNumber	// Restore original SEQ number
			   Exit
            EndIf
         EndIf

		 // Third: check the security and precedence
		 // No such feature in this stack.  We want to accept all connections.

		 // Fourth: check the SYN bit
		 If (localHeaderFlags And SYN) <> 0 Then

		    // We now have an initial sequence number and window size
			MyTCB.RemoteSEQ = localSeqNumber + 1
			MyTCB.remoteWindow = h.Window

			If (localHeaderFlags And ACK) <> 0 Then
			   SendTCP(hTCP, ACK, true)
			   ps.smState = TCP_ESTABLISHED
			   ps.Flags.bTimerEnabled = false
			Else
			   SendTCP(hTCP, SYN Or ACK, true)
			   ps.smState = TCP_SYN_RECEIVED
			EndIf
         EndIf

		 // Fifth: drop the segment if neither SYN or RST is set
		 Exit
   End Select
    
   //
   // First: check the sequence number
   //
   wSegmentLength = Len
   If (localHeaderFlags And FIN) <> 0 Then
      Inc(wSegmentLength)
   EndIf
    	
   If (localHeaderFlags And SYN) <> 0 Then
      Inc(wSegmentLength)
   EndIf

   // Calculate the RX FIFO space
   If ps.rxHead >= ps.rxTail Then
      wFreeSpace = (ps.bufferEnd - ps.bufferRxStart) - (ps.rxHead - ps.rxTail)
   Else
      wFreeSpace = ps.rxTail - ps.rxHead - 1
   EndIf	

   // Calculate the number of bytes ahead of our head pointer this segment skips
   lMissingBytes = localSeqNumber - MyTCB.RemoteSEQ
   wMissingBytes = lMissingBytes
	
   // Run TCP acceptability tests to verify that this packet has a valid sequence number
   bSegmentAcceptable = FALSE
   If wSegmentLength <> 0 Then
      // Check to see if we have free space, and if so, if any of the data falls within the freespace
	  If wFreeSpace <> 0 Then
	     // RCV.NXT =< SEG.SEQ < RCV.NXT+RCV.WND
		 If (lMissingBytes >= 0) And (wFreeSpace > lMissingBytes) Then
		    bSegmentAcceptable = TRUE
		 Else
		    // RCV.NXT =< SEG.SEQ+SEG.LEN-1 < RCV.NXT+RCV.WND
			If (lMissingBytes + LONG(wSegmentLength) > 0) And (lMissingBytes <= LONG(wFreeSpace - wSegmentLength)) Then
			   bSegmentAcceptable = TRUE
			EndIf	
		 EndIf
			
		 If lMissingBytes < LONG(wFreeSpace) And (SHORT(wMissingBytes) + SHORT(wSegmentLength) > 0) Then
		    bSegmentAcceptable = TRUE
		 EndIf	
      EndIf
	  // Segments with data are not acceptable if we have no free space
   Else
      // Zero length packets are acceptable if they fall within our free space window
	  // SEG.SEQ = RCV.NXT
	  If lMissingBytes = 0 Then
	     bSegmentAcceptable = TRUE
	  Else
	     // RCV.NXT =< SEG.SEQ < RCV.NXT+RCV.WND
		 If (lMissingBytes >= 0) And (wFreeSpace > lMissingBytes) Then
		    bSegmentAcceptable = TRUE
	     EndIf	
      EndIf
   EndIf
	
   If Not bSegmentAcceptable Then
      // Unacceptable segment, drop it and respond appropriately
	  If (localHeaderFlags And RST) = 0 Then
	     SendTCP(hTCP, ACK, TRUE)
      EndIf	
	  Exit
   EndIf

   //
   // Second: check the RST bit
   //
   //
   // Fourth: check the SYN bit
   //
   // Note, that since the third step is not implemented, we can 
   // combine this second and fourth step into a single operation.
   If (localHeaderFlags And (RST Or SYN)) <> 0 Then
      CloseSocket(hTCP)
	  Exit
   EndIf

   //
   // Third: check the security and precedence
   //
   // Feature not supported.  Let's process this segment.

   //
   // Fifth: check the ACK bit
   //
   If (localHeaderFlags And ACK) = 0 Then
      Exit
   EndIf

   If LocalState = TCP_SYN_RECEIVED Then
      If localAckNumber <> MyTCB.MySEQ Then
         
         // Send a RST packet with SEQ = SEG.ACK, but retain our SEQ 
         // number for arivial of any other correct packets
         localSeqNumber = MyTCB.MySEQ	// Save our original SEQ number
         MyTCB.MySEQ = localAckNumber	// Set SEQ = SEG.ACK
         SendTCP(hTCP, RST, true)		// Send the RST
         MyTCB.MySEQ = localSeqNumber	// Restore original SEQ number
		 Exit
	  EndIf
	  ps.smState = TCP_ESTABLISHED
	  
      // No break
   EndIf
    
   If LocalState = TCP_ESTABLISHED Or LocalState = TCP_FIN_WAIT_1 Or 
      LocalState = TCP_FIN_WAIT_2 Or LocalState = TCP_CLOSING Then
			
      // Drop the packet if it ACKs something we haven't sent
	  If LONG(MyTCB.MySEQ - localAckNumber) < 0 Then
         SendTCP(hTCP, ACK, TRUE)    
	     Exit
      EndIf

	  // Throw away all ACKnowledged TX data:
	  // Calculate what the last acknowledged sequence number was (ignoring any FINs we sent)
	  dwTemp = MyTCB.MySEQ - (Integer(MyTCB.txUnackedTail) - Integer(ps.txTail))
	  If MyTCB.txUnackedTail < ps.txTail Then
	     dwTemp = dwTemp - (ps.bufferRxStart - ps.bufferTxStart)
	  EndIf	
	
	  // Calcluate how many bytes were ACKed with this packet
	  dwTemp = localAckNumber - dwTemp
	  If LONG(dwTemp) > 0 And dwTemp <= ps.bufferRxStart - ps.bufferTxStart Then
	     ps.Flags.bHalfFullFlush = FALSE
	
	     // Bytes ACKed, free up the TX FIFO space
		 ps.txTail = ps.txTail + dwTemp
		 If ps.txTail >= ps.bufferRxStart Then
		    ps.txTail = ps.txTail - (ps.bufferRxStart - ps.bufferTxStart)
		 EndIf
      EndIf

	  // No need to keep our retransmit timer going if we have nothing that needs ACKing anymore
	  If ps.txTail = ps.txHead Then
	     // Make sure there isn't a "FIN byte in our TX FIFO"
		 If Not ps.Flags.bTXFin Then
            ps.Flags.bTimerEnabled = false
         Else
		    // "Throw away" FIN byte from our TX FIFO if it has been ACKed
		    If MyTCB.MySEQ = localAckNumber Then
		       ps.Flags.bTimerEnabled = false
			   ps.Flags.bTXFin = false
			   MyTCB.Flags.bFINSent = false	
            EndIf
		 EndIf
      EndIf

	  // Update the local stored copy of the RemoteWindow
	  // If previously we had a zero window, and now we don't 
	  // immediately send whatever was pending
	  If (MyTCB.remoteWindow = 0) And h.Window <> 0 Then
	     ps.Flags.bTXASAP = true
      EndIf
            	
	  MyTCB.remoteWindow = h.Window

	  // A couple of states must do all of the TCP_ESTABLISHED stuff, but also a little more
	  If localState = TCP_FIN_WAIT_1 Then
	     // Check to see if our FIN has been ACKnowledged
		 If MyTCB.MySEQ = localAckNumber Then
            // Reset our timer for forced closure if the remote node 
			// doesn't send us a FIN in a timely manner.
			ps.eventTime = TickGet() + TCP_FIN_WAIT_2_TIMEOUT
			ps.Flags.bTimerEnabled = true
			ps.smState = TCP_FIN_WAIT_2		
         EndIf	
         
      ElseIf localState = TCP_FIN_WAIT_2 Then
	     // RFC noncompliance:
		 // The remote node should not keep sending us data 
		 // indefinitely after we send a FIN to it.  
		 // However, some bad stacks may still keep sending 
		 // us data indefinitely after ACKing our FIN.  To 
		 // prevent this from locking up our socket, let's 
		 // send a RST right now and close forcefully on 
		 // our side.
		 If (localHeaderFlags And FIN) = 0 Then
		    MyTCB.MySEQ = localAckNumber	// Set SEQ = SEG.ACK
            SendTCP(hTCP, RST Or ACK, FALSE)
			CloseSocket(hTCP)
			Exit
         EndIf
      ElseIf localState = TCP_CLOSING Then
	     // Check to see if our FIN has been ACKnowledged
		 If MyTCB.MySEQ = localAckNumber Then
		    // RFC not recommended: We should be going to 
			// the TCP_TIME_WAIT state right here and 
			// starting a 2MSL timer, but since we have so 
			// few precious sockets, we can't afford to 
			// leave a socket waiting around doing nothing 
			// for a long time.  If the remote node does 
			// not recieve this ACK, it'll have to figure 
			// out on it's own that the connection is now 
			// closed.
			CloseSocket(hTCP)
         EndIf
         Exit
      EndIf
      GoTo DoBreak
   EndIf

   If localState = TCP_LAST_ACK Then
      // Check to see if our FIN has been ACKnowledged
	  If MyTCB.MySEQ = localAckNumber Then
	     CloseSocket(hTCP)
      EndIf
      Exit
   EndIf
    
DoBreak:

   //
   // Sixth: Check the URG bit
   //
   // Urgent packets are not supported in this stack, so we
   // will throw them away instead
   If (localHeaderFlags And URG) <> 0 Then
      Exit
   EndIf

   //
   // Seventh: Process the segment text
   //
   // Throw data away if in a state that doesn't accept data
   If localState = TCP_CLOSING Then
      Exit
   EndIf

   If localState = TCP_LAST_ACK Then
      Exit
   EndIf

   // See if there are bytes we must skip
   If SHORT(wMissingBytes) <= 0 Then
		
      // Position packet read pointer to start of useful data area.
	  IPSetRxBuffer((GetOffsetVal(h.DataOffset._Val) << 2) - wMissingBytes)
	  Len = Len + wMissingBytes
	  
      // Truncate packets that would overflow our TCP RX FIFO
      // and request a retransmit by sending a duplicate ACK
      If len > wFreeSpace Then
         len = wFreeSpace
      EndIf			
         	  
	  MyTCB.RemoteSEQ = MyTCB.RemoteSEQ + Len
	
	  // Copy the application data from the packet into the socket RX FIFO
	  // See if we need a two part copy (spans bufferEnd->bufferRxStart)
	  If ps.rxHead + Len > ps.bufferEnd Then
	     wTemp = ps.bufferEnd - ps.rxHead + 1
		 MACMemCopyAsync(ps.rxHead, Word(-1), wTemp)
		 MACMemCopyAsync(ps.bufferRxStart, Word(-1), Len - wTemp)
		 ps.rxHead = ps.bufferRxStart + (Len - wTemp)
	  Else
		 MACMemCopyAsync(ps.rxHead, Word(-1), Len)
		 ps.rxHead = ps.rxHead + Len
	  EndIf
	
	  // See if we have a hole and other data waiting already in the RX FIFO
	  If MyTCB.sHoleSize <> -1 Then
	     MyTCB.sHoleSize = MyTCB.sHoleSize - Len
		 wTemp = MyTCB.wFutureDataSize + MyTCB.sHoleSize
	
		 // See if we just closed up a hole, and if so, advance head pointer
		 If SHORT(wTemp) < 0 Then
		    MyTCB.sHoleSize = -1
		 ElseIf MyTCB.sHoleSize <= 0 Then
			MyTCB.RemoteSEQ = MyTCB.RemoteSEQ + wTemp
			ps.rxHead = ps.rxHead + wTemp
			If ps.rxHead > ps.bufferEnd Then
			   ps.rxHead = ps.rxHead - (ps.bufferEnd - ps.bufferRxStart + 1)
            EndIf    							
			MyTCB.sHoleSize = -1
		EndIf
      EndIf
	
   // This packet is out of order or we lost a packet, see if we can generate a hole to accomodate it
   ElseIf SHORT(wMissingBytes) > 0 Then

      // Truncate packets that would overflow our TCP RX FIFO
      If (len + wMissingBytes) > wFreeSpace Then
         len = wFreeSpace - wMissingBytes
      EndIf			
            	
      // Position packet read pointer to start of useful data area.
	  IPSetRxBuffer((GetOffsetVal(h.DataOffset._Val) << 2))
	    
	  // See if we need a two part copy (spans bufferEnd->bufferRxStart)
	  If ps.rxHead + wMissingBytes + Len > ps.bufferEnd Then
	     // Calculate number of data bytes to copy before wraparound
		 wTemp = ps.bufferEnd - ps.rxHead + 1 - wMissingBytes
		 If SHORT(wTemp) >= 0 Then
		    MACMemCopyAsync(ps.rxHead + wMissingBytes, Word(-1), wTemp)
			MACMemCopyAsync(ps.bufferRxStart, Word(-1), Len - wTemp)
	     Else
		    MACMemCopyAsync(ps.rxHead + wMissingBytes - (ps.bufferEnd - ps.bufferRxStart + 1), Word(-1), Len)
	     EndIf
      Else
	     MACMemCopyAsync(ps.rxHead + wMissingBytes, Word(-1), Len)
	  EndIf
	
	  // Record the hole is here
	  If MyTCB.sHoleSize = -1 Then
	     MyTCB.sHoleSize = wMissingBytes
		 MyTCB.wFutureDataSize = Len
      Else
	     // We already have a hole, see if we can shrink the hole 
		 // or extend the future data size
		 If wMissingBytes < Word(MyTCB.sHoleSize) Then
            If (wMissingBytes + Len > Word(MyTCB.sHoleSize) + MyTCB.wFutureDataSize) Or (wMissingBytes + Len < Word(MyTCB.sHoleSize)) Then
			   MyTCB.wFutureDataSize = Len
			Else
			   MyTCB.wFutureDataSize = Word(MyTCB.sHoleSize) + MyTCB.wFutureDataSize - wMissingBytes
			EndIf	
			MyTCB.sHoleSize = wMissingBytes
	     ElseIf (wMissingBytes + Len > Word(MyTCB.sHoleSize) + MyTCB.wFutureDataSize) Then
		    // Make sure that there isn't a second hole between 
			// our future data and this TCP segment's future data
			If (wMissingBytes <= Word(MyTCB.sHoleSize) + MyTCB.wFutureDataSize) Then
			   MyTCB.wFutureDataSize = MyTCB.wFutureDataSize + (wMissingBytes + Len - Word(MyTCB.sHoleSize) - MyTCB.wFutureDataSize)
			EndIf	
	     EndIf
      EndIf
	
      // RFC Noncompliance:
	  // Request an immediate fast retransmit from remote node 
	  // by sending three duplicate ACKs.  This might improve 
	  // performance, but then again, it might damage 
	  // performance if the remote node invokes slow restart 
	  // and/or collision avoidance algorithms.
	  //SendTCP(hTCP, ACK, TRUE);
	  //while(!MACIsTxReady());
	  //MACFlush();
	  //while(!MACIsTxReady());
	  //MACFlush();
   EndIf

   // Send back an ACK of the data (+SYN | FIN) we just received, 
   // if any.  To minimize bandwidth waste, we are implementing 
   // the delayed acknowledgement algorithm here, only sending 
   // back an immediate ACK if this is the second segment received.  
   // Otherwise, a 200ms timer will cause the ACK to be transmitted.
   If wSegmentLength <> 0 Then
      // For non-established sockets, let's delete all data in 
	  // the RX buffer immediately after receiving it.  This is 
	  // not really how TCP was intended to operate since a 
	  // socket cannot receive any response after it sends a FIN,
	  // but our TCP application API doesn't readily accomodate
	  // receiving data after calling TCPDisconnect(), which 
	  // invalidates the application TCP handle.  By deleting all 
	  // data, we'll ensure that the RX window is nonzero and 
	  // the remote node will be able to send us a FIN response, 
	  // which needs an RX window of at least 1.
      
      If localState <> TCP_ESTABLISHED Then
         ps.rxTail = ps.rxHead
      EndIf	

	  If ps.Flags.bOneSegmentReceived Then
	     SendTCP(hTCP, ACK, TRUE)
		 SyncTCB(hTCP)
		 // bOneSegmentReceived is cleared in SendTCP(), so no need here
	  Else
	     ps.Flags.bOneSegmentReceived = TRUE	
		
		 // Do not send an ACK immediately back.  Instead, we will 
		 // perform delayed acknowledgements.  To do this, we will 
		 // just start a timer
		 If Not ps.Flags.bDelayedACKTimerEnabled Then
		    ps.Flags.bDelayedACKTimerEnabled = TRUE
		    'ps.delayedACKTime = Word(TickGetDiv256) + Word(TCP_DELAYED_ACK_TIMEOUT >> 8)
			ps.delayedACKTime = TickGet + Word(TICK_SECOND / 5)		
		 EndIf
      EndIf
   EndIf

   // See if all data is acknowledged.  If so, there is no need 
   // to keep the retransmission timer running right now.
   If ps.txHead = ps.txTail Then
      ps.Flags.bTimerEnabled = false
   EndIf

   //
   // Eighth: check the FIN bit
   //
   If (localHeaderFlags And FIN) <> 0 Then
      // Note: Since we don't have a good means of storing "FIN bytes" 
	  // in our TCP RX FIFO, we must ensure that FINs are processed 
	  // in-order.
	  If MyTCB.RemoteSEQ + 1 = localSeqNumber + wSegmentLength Then
	     // FINs are treated as one byte of data for ACK sequencing
		Inc(MyTCB.RemoteSEQ)
        Select localState
		   Case TCP_ESTABLISHED, TCP_SYN_RECEIVED
		      // RFC in exact: Since we don't have an API that 
			  // can allow the user to execute a CLOSE command, 
			  // the stack will automatically perform the close 
			  // operation immediately.  Thus, we are going to 
			  // TCP_LAST_ACK immediately instead of 
			  // TCP_CLOSE_WAIT
			  ps.smState = TCP_LAST_ACK
			  SendTCP(hTCP, FIN Or ACK, TRUE)
			  Exit
	
		   Case TCP_FIN_WAIT_1:
		      If MyTCB.MySEQ = localAckNumber Then
			     // RFC not recommended: We should be going to 
				 // the TCP_TIME_WAIT state right here and 
				 // starting a 2MSL timer, but since we have so 
				 // few precious sockets, we can't afford to 
				 // leave a socket waiting around doing nothing 
				 // for a long time.  If the remote node does 
				 // not recieve this ACK, it'll have to figure 
				 // out on it's own that the connection is now 
				 // closed.
				 SendTCP(hTCP, ACK, false)
				 CloseSocket(hTCP)
				 Exit
			  Else
			     ps.smState = TCP_CLOSING
	          EndIf
	
           Case TCP_FIN_WAIT_2:
		      // RFC not recommended: We should be going to 
			  // the TCP_TIME_WAIT state right here and 
			  // starting a 2MSL timer, but since we have so 
			  // few precious sockets, we can't afford to 
			  // leave a socket waiting around doing nothing 
			  // for a long time.  If the remote node does 
			  // not recieve this ACK, it'll have to figure 
			  // out on it's own that the connection is now 
			  // closed.
			  SendTCP(hTCP, ACK, FALSE)
			  CloseSocket(hTCP)
			  Exit
         End Select

	     // Acknowledge receipt of FIN
	     SendTCP(hTCP, ACK, TRUE)
      EndIf
   EndIf
End Sub
(*
*********************************************************************
* Function:        BOOL TCPProcess(NODE_INFO* remote,
*                                  IP_ADDR *localIP,
*                                  Word len)
* PreCondition:    TCPInit() is already called     And
*                  TCP segment is ready in MAC buffer
* Input:           remote      - remote node info
*                  len         - Total Length of TCP semgent.
* Output:          TRUE If this Function has completed its task
*                  FALSE otherwise
* Side Effects:    None
* Overview:        None
* Note:            None
********************************************************************/
*)
Public Function TCPProcess(ByRef remote As NODE_INFO, ByRef localIP As IP_ADDR, Len As Word) As Boolean
   Dim TCPHeader As TCP_HEADER
   Dim pseudoHeader As PSEUDO_HEADER
   Dim socket As TCP_SOCKET
   Dim checksum1 As WORD_VAL
   Dim checksum2 As WORD_VAL
   Dim optionsSize As Byte

   // Calculate IP pseudoheader checksum.
   pseudoHeader.SourceAddress      = remote.IPAddr
   pseudoHeader.DestAddress        = localIP
   pseudoHeader.Zero               = $0
   pseudoHeader.Protocol           = IP_PROT_TCP
   pseudoHeader.TCPLength          = Len

   SwapPseudoTCPHeader(pseudoHeader)

   checksum1.Val = Not CalcIPChecksum(@pseudoHeader, SizeOf(pseudoHeader))

   // Now calculate TCP packet checksum in NIC RAM - should match
   // pesudo header checksum
   checksum2.Val = CalcIPBufferChecksum(Len)

   // Compare checksums.
   If checksum1.Val <> checksum2.Val Then
      MACDiscardRx()
      result = false
      Exit
   EndIf

   // Retrieve TCP header.
   IPSetRxBuffer(0)
   MACGetArray(@TCPHeader, SizeOf(TCPHeader))
   SwapTCPHeader(TCPHeader)

   // Skip over options to retrieve data bytes
   optionsSize = (GetOffsetVal(TCPHeader.DataOffset._Val) << 2) - SizeOf(TCPHeader)
   Len = Len - optionsSize - SizeOf(TCPHeader)
   
   // Prevent loopback injection attacks.  Discard packets 
   // arriving externally and addressed from the loopback IP
   If remote.IPAddr.Val = LOOPBACK_IP Then
      MACDiscardRx()
      result = false
      Exit
   EndIf

   // Find matching socket.
   socket = FindMatchingSocket(TCPHeader, remote)
   If socket <> INVALID_SOCKET Then
      HandleTCPSeg(TCBStubs(socket),socket, TCPHeader, Len)
//	else
//		// TODO: RFC 793 specifies that if the socket is closed 
//		// and a segment arrives, we should send back a RST if 
//		// the RST bit in the incoming packet is not set.  The 
//		// code to respond with the RST packet and believable SEQ 
//		// and ACK numbers needs to be implemented.
//		//if not TCPHeader.Flags.bits.flagRST then
//		//	SendTCP(INVALID_SOCKET, RST, true)
//      // endif
   EndIf

   MACDiscardRx()
   result = true   
End Function
(*
*********************************************************************
* Function:        BOOL TCPAdjustFIFOSize(TCP_SOCKET hTCP, Word sRX, Word sTX)
* PreCondition:    TCPInit() is already called.
* Input:           hTCP - TCP socket handle
*				   sRX	- Receive FIFO Size
*				   sTX	- Transmit FIFO Size
* Output:          TRUE: the FIFOs were adjusted
*				   FALSE: If the sRX And sTX were of improper size
* Side Effects:    Zeroes both buffers
* Overview:        Adjusts the RX And TX FIFO sizes For optimal speed
* Note:            sRX + sTX <= TCP_RX_FIFO_SIZE + TCP_TX_FIFO_SIZE
*				   If <, Then TX buffer will get remainder of space.
********************************************************************/
*)
Function TCPAdjustFIFOSize(hTCP As TCP_SOCKET, sRX As Word, sTX As Word, destroyRX As Boolean) As Boolean
   Dim ps As TCB_STUB
   
   // Make sure space is available
   If sRX + sTX > TCP_TX_FIFO_SIZE + TCP_RX_FIFO_SIZE Then
      result = false
      Exit
   EndIf

   // Load up info on this socket
   ps = TCBStubs(hTCP)
   SyncTCB(hTCP)

   // Move the RX buffer pointer - it's the one that divides the two
   ps.bufferRxStart = ps.bufferTxStart + sTX + 1

   // Empty the TX buffer
   MyTCB.txUnackedTail = ps.bufferTxStart
   ps.txTail = ps.bufferTxStart
   ps.txHead = ps.bufferTxStart
	
   // Empty the RX buffer
   If destroyRX Or ps.rxTail < ps.bufferRxStart Or ps.rxHead < ps.bufferRxStart Then
      ps.rxTail = ps.bufferRxStart
	  ps.rxHead = ps.bufferRxStart
   EndIf
   TCBStubs(hTCP) = ps

   // Send a window update to notify remote node of change
   If ps.smState = TCP_ESTABLISHED Then
      SendTCP(hTCP, ACK, TRUE)
   EndIf		
   result = true
End Function
(*
*********************************************************************
 * Function:        TCP_SOCKET TCPOpenLoopback(Word DestPort)
 * PreCondition:    TCPInit() is already called
 * Input:           DestPort - the desired listening Port
 * Output:          A socket that matches With given header And remote
 *                  node is searched.
 *                  If such socket is found, its index is returned
 *                  Else INVALID_SOCKET is returned.
 * Side Effects:    None
 * Overview:        None
 * Note:            Only For incoming connections now, could be
 *					expanded To support outgoing connections.
********************************************************************/
*)
Function TCPOpenLoopback(DestPort As Word) As TCP_SOCKET
   Dim ps As TCB_STUB
   Dim hTCP As TCP_SOCKET

   //find a socket listening on the right port
   For hTCP = 0 To MAX_TCP_SOCKETS - 1 
      ps = TCBStubs(hTCP)
      If ps.smState = TCP_LISTEN And ps.remoteHash.Val = DestPort Then

         //lock the socket
		 ps.smState = TCP_LOOPBACK
			
		 //reset all the buffer pointers
		 ps.txHead = ps.bufferTxStart
		 ps.txTail = ps.bufferTxStart
		 ps.rxHead = ps.bufferRxStart
		 ps.rxTail = ps.bufferRxStart
         TCBStubs(hTCP) = ps
        
		 //return the socket handle
		 result = hTCP
		 Exit
      EndIf
   Next  

   //if no socket was found, then return a flag
   result = INVALID_SOCKET
End Function
(*
 *********************************************************************
 * Function:        BOOL TCPCloseLoopback(TCP_SOCKET hTCP)
 * PreCondition:    TCPInit() is already called
 * Input:           hTCP - the TCP loopback handle To close
 * Output:          TRUE If the loopback was closed
 *					FALSE If the socket was Not a loopback
 * Side Effects:    None
 * Overview:        None
 * Note:            Only For incoming connections now, could be
 *					expanded To support outgoing connections.
********************************************************************/
*)
Function TCPCloseLoopback(hTCP As TCP_SOCKET) As Boolean

   If TCBStubs(hTCP).smState <> TCP_LOOPBACK And TCBStubs(hTCP).smState <> TCP_LOOPBACK_CLOSED Then
      result = false
      Exit
   EndIf
   
   CloseSocket(hTCP)
   result = true
End Function
(*
 *********************************************************************
 * Function:        BOOL TCPIsLoopback(TCP_SOCKET hTCP)
 * PreCondition:    TCPInit() is already called
 * Input:           hTCP - the TCP loopback handle To close
 * Output:          TRUE If the socket is a loopback
 *					FALSE If the socket is Not a loopback
 * Side Effects:    None
 * Overview:        None
 * Note:            None
********************************************************************/
*)
Public Function TCPIsLoopback(hTCP As TCP_SOCKET) As Boolean
   If TCBStubs(hTCP).smState <> TCP_LOOPBACK And TCBStubs(hTCP).smState <> TCP_LOOPBACK_CLOSED Then
      result = false
   Else
      result = true
   EndIf
End Function
(*
 *********************************************************************
 * Function:        Word TCPInject(TCP_SOCKET hTCP, Byte *data, 
 *									Word len)
 * PreCondition:    TCPInit() is already called.
 * Input:           hTCP    - Socket handle To use
 *                  data    - Pointer To data To inject
 *				   len     - Number of Bytes in data
 * Output:          Count of Bytes actually placed in the RX buffer
 * Side Effects:    None
 * Overview:        Injects data into a socket's RX FIFO
 * Note:            None
********************************************************************/
*)
Function TCPInject(hTCP As TCP_SOCKET, buffer As Word, Len As Word) As Word
   Dim wTemp As Word
   Dim ps As TCB_STUB

   ps = TCBStubs(hTCP)
	
   //if the connection is finished, pretend to accept bytes
   If ps.smState = TCP_LOOPBACK_CLOSED Then
      result = Len
      Exit
   EndIf
	
   //make sure there's enough space
   wTemp = TCPGetRxFIFOFree(hTCP)
   If wTemp < Len Then
      Len = wTemp
   EndIf   
	
   // Copy the data from the buffer into the RX FIFO
   // See if we need a two part copy (spans bufferEnd->bufferRxStart)
   If ps.rxHead + Len > ps.bufferEnd Then
      wTemp = ps.bufferEnd - ps.rxHead + 1
      MACSetWritePtr(ps.rxHead)
	  MACPutArray(buffer, wTemp)
      buffer = buffer + wTemp
      MACSetWritePtr(ps.bufferRxStart)
      MACPutArray(buffer, Len - wTemp)
	  TCBStubs(hTCP).rxHead = ps.bufferRxStart + (Len - wTemp)
   Else
      MACSetWritePtr(ps.rxHead)
      MACPutArray(buffer, Len)
      TCBStubs(hTCP).rxHead = ps.rxHead + Len
   EndIf
   result = Len
End Function
(*
 *********************************************************************
 * Function:        Word TCPSteal(TCP_SOCKET hTCP, Byte *data, 
 *									Word len)
 * PreCondition:    TCPInit() is already called.
 * Input:           hTCP    - Socket handle To use
 *                  data    - Pointer To a buffer To fill
 *				   len     - Number of Bytes To read
 * Output:          Count of Bytes actually placed in *data
 * Side Effects:    None
 * Overview:        Removes data from a socket's TX FIFO and places
 *					it in *data 
 * Note:            None
 ********************************************************************/
*)
Function TCPSteal(hTCP As TCP_SOCKET, buffer As Word, Len As Word) As Word
   Dim ActualLen As Word
   Dim RightLen As Word
   Dim ps As TCB_STUB
   
   RightLen = 0
   ps = TCBStubs(hTCP)

   ActualLen = TCPGetTxFIFOFull(hTCP)
   If Len > ActualLen Then
      Len = ActualLen
   EndIf   

   // See if we need a two part get	
   If ps.txTail + Len >= ps.bufferRxStart Then
      RightLen = ps.bufferRxStart - ps.txTail
      MACSetReadPtr(ps.txTail)
      MACGetArray(buffer, RightLen)
      If buffer <> 0 Then
         buffer = buffer + RightLen
      EndIf   
      Len = Len - RightLen
      ps.txTail = ps.bufferTxStart
	  TCBStubs(hTCP).txTail = ps.txTail
   EndIf
   MACSetReadPtr(ps.txTail)
   MACGetArray(buffer, Len)
   TCBStubs(hTCP).txTail = ps.txTail + Len
   
   result = Len + RightLen
End Function

// originally defined as macros
Public Function TCPSetFullDuplex(hTCP As TCP_SOCKET) As Boolean
   result = TCPAdjustFIFOSize(hTCP, TCP_RX_FIFO_SIZE, TCP_TX_FIFO_SIZE, true)
End Function
Public Function TCPSetHalfDuplexRead(hTCP As TCP_SOCKET) As Boolean
   result = TCPAdjustFIFOSize(hTCP, TCP_RX_FIFO_SIZE + TCP_TX_FIFO_SIZE - 5, 5, false)
End Function
Public Function TCPSetHalfDuplexWrite(hTCP As TCP_SOCKET) As Boolean
   result = TCPAdjustFIFOSize(hTCP, 5, TCP_RX_FIFO_SIZE + TCP_TX_FIFO_SIZE - 5, true)
End Function   


// initialise
#if defined(STACK_USE_TCP)
hCurrentTCB = INVALID_SOCKET
NextPort = 0
#endif

