{
*****************************************************************************
*  Name    : Tick.bas                                                       *
*  Date    : 9/26/2015                                                      *
*  Version : 1.1 - Based around Microchip TCP/IP stack version 4.11         *
*  Notes   : v1.1 - add #option NET_TICK_PRIORITY                           *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module NETTick

Public Type TICK = LongWord 
//
// This value is used by TCP To implement timeout actions.
// If SNMP Module is in use, this value should be 100 As required
// by SNMP protocol unless main application is providing separate
// Tick which is 10mS.
//
#option TICKS_PER_SECOND   = 1000                  // 1000=1ms, 100=10ms

// tick interrupt (TMR0)
#option NET_TICK_PRIORITY = ipHigh       
const TICK_PRIORITY = NET_TICK_PRIORITY

Public Const
   TICK_SECOND = TICKS_PER_SECOND

// Clock frequency value
// This value is used to calculate Tick Counter value
Const
   CLOCK_FREQ        = _clock * 1000000,
   INSTR_FREQ        = CLOCK_FREQ / 4,  
   TICK_LOAD As Word = ((-(INSTR_FREQ + 2 + TICKS_PER_SECOND / 2) / 4 / TICKS_PER_SECOND) + 4)
   
// local variables...
Dim
   TickCount As TICK,    
   TMR0IF As INTCON.bits(2),
   TMR0IE As INTCON.bits(5),
   TMR0IP As INTCON2.bits(2),
   TMR0ON As T0CON.bits(7)

//
// *********************************************************************
// * Function:        TICK TickGetDiff(TICK a, TICK b)
// ********************************************************************/
//   
Public Function TickGetDiff(a As TICK, b As TICK) As TICK
   result = a - b
End Function

//
// *********************************************************************
// * Function:                                          
// *********************************************************************
// 
Interrupt OnTick(TICK_PRIORITY)
   Dim Val As Word
   If (TMR0IF = 1) Then
      Val.Byte0 = TMR0L     // reading TMR0L latches TMR0H
      Val.Byte1 = TMR0H
      Val = Val + TICK_LOAD
      TMR0H = Val.Byte1
      TMR0L = Val.Byte0     // writing TMR0L updates TMR0H
      Inc(TickCount)
      TMR0IF = 0
   EndIf
End Interrupt

//
// *********************************************************************
// * Function:        void TickInit(void)
// * PreCondition:    None
// * Input:           None
// * Output:          Tick manager is initialized.
// * Side Effects:    None
// * Overview:        Initializes Timer0 as a tick counter.
// * Note:            None
// ********************************************************************/
//
Public Sub TickInit()
   T0CON = $01          // 16-bit, inst cycles, prescaler 1:4

   // buffered write to TMR0...
   TMR0H = Byte(TICK_LOAD >> 8)
   TMR0L = Byte(TICK_LOAD)     // writing TMR0L updates TMR0H
   
 #if (NET_TICK_PRIORITY = ipLow)
   TMR0IP = 0
 #else
   TMR0IP = 1
 #endif
   TMR0IF = 0
   TMR0ON = 1
   TMR0IE = 1
   Enable(OnTick)
End Sub

//
// *********************************************************************
// * Function:        TICK TickGet(void)
// * PreCondition:    None
// * Input:           None
// * Output:          Current tick value is given
// *                  1 tick represents approximately 10ms
// * Side Effects:    None
// * Overview:        None
// * Note:            None
// ********************************************************************/
//
Public Function TickGet() As TICK
   TMR0IE = 0       // disable tick intr
   result = TickCount
   TMR0IE = 1       // reenable tick
End Function

// initialisation
TickCount = 0

end module
