//
// eth_phy.bas
// ethernet mac and phy open modes
//
module eth_phy

//
// eEthOpenFlags
//
public const
  // Link capabilities flags:
    // Use auto negotiation. set the following flags to specify your choices
    ETH_OPEN_AUTO_BIT           = 0,
    // Use full duplex or full duplex negotiation capability needed
    ETH_OPEN_FDUPLEX_BIT        = 1,
    // Use half duplex or half duplex negotiation capability needed
    ETH_OPEN_HDUPLEX_BIT        = 2,
    // Use 100MBps or 100MBps negotiation capability needed
    ETH_OPEN_100_BIT            = 3,
    // Use 10MBps or 10MBps negotiation capability needed
    ETH_OPEN_10_BIT             = 4,
  // MAC flags:
    // Allow huge packets RX/TX
    ETH_OPEN_HUGE_PKTS_BIT      = 5,
    // Loopbacked at the MAC level
    ETH_OPEN_MAC_LOOPBACK_BIT   = 6,
    // When PHY is loopback-ed, negotiation will be disabled!
    ETH_OPEN_PHY_LOOPBACK_BIT   = 7,

    ETH_OPEN_AUTO           = (1<<ETH_OPEN_AUTO_BIT),
    ETH_OPEN_FDUPLEX        = (1<<ETH_OPEN_FDUPLEX_BIT),
    ETH_OPEN_HDUPLEX        = (1<<ETH_OPEN_HDUPLEX_BIT),
    ETH_OPEN_100            = (1<<ETH_OPEN_100_BIT),
    ETH_OPEN_10             = (1<<ETH_OPEN_10_BIT),
    ETH_OPEN_HUGE_PKTS      = (1<<ETH_OPEN_HUGE_PKTS_BIT),
    ETH_OPEN_MAC_LOOPBACK   = (1<<ETH_OPEN_MAC_LOOPBACK_BIT),
    ETH_OPEN_PHY_LOOPBACK   = (1<<ETH_OPEN_PHY_LOOPBACK_BIT)

// All capabilities default
public const
    ETH_OPEN_DEFAULT = (ETH_OPEN_AUTO + ETH_OPEN_FDUPLEX + ETH_OPEN_HDUPLEX + ETH_OPEN_100 + ETH_OPEN_10)

//
// eEthMacPauseType (Ethernet MAC Pause Types)
//
public const
    // No PAUSE capabilities
    ETH_MAC_PAUSE_TYPE_NONE     = $00,
    // Supports symmetric PAUSE
    ETH_MAC_PAUSE_TYPE_PAUSE    = $01,
    // Supports ASM_DIR
    ETH_MAC_PAUSE_TYPE_ASM_DIR  = $02,
    // The previous two values converted to TX/RX capabilities:
    // Enable MAC TX pause support
    ETH_MAC_PAUSE_TYPE_EN_TX    = $04,
    // Enable MAC RX pause support
    ETH_MAC_PAUSE_TYPE_EN_RX    = $08,
  // All types of pause
    ETH_MAC_PAUSE_ALL  = (ETH_MAC_PAUSE_TYPE_PAUSE + ETH_MAC_PAUSE_TYPE_ASM_DIR +  ETH_MAC_PAUSE_TYPE_EN_TX + ETH_MAC_PAUSE_TYPE_EN_RX),
  // All pause capabilities our MAC supports
    ETH_MAC_PAUSE_CPBL_MASK = ETH_MAC_PAUSE_TYPE_NONE   //ETH_MAC_PAUSE_ALL

end module
