{
*****************************************************************************
*  Name    : MAC.BAS                                                        *
*  Date    : 9/18/2015                                                      *
*  Version : 1.3 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   : ver 1.3 - add support for ENCx24J600 10/100 ethernet           *
*          : ver 1.2 - add support for 18FxxJ60 w/internal ethernet         *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module MAC

//
// This MAC module is the interface to all other stack modules - none of the
// stack modules will interface to a hardware layer (ie ENC28J60) directly
//

//
// select which MAC interface to use from one of the following options:
//  #option ETH_MAC = ETHXXJ60      ' J60 internal ethernet
//  #option ETH_MAC = ENC28J60      ' default (unless internal)
//  #option ETH_MAC = ENCx24J600
// by default, if internal is available (device file contains '_ethernet = 1') then it's used
// if you wish to use the ENC28J60/ENCx24J600 over the internal then set '#option ETH_MAC'
// as appropriate before including the network files
//
#if (_ethernet = 1)         // use internal ethernet
  #option ETH_MAC = ETHXXJ60
#else                       // default MAC selection (if not internal)
  #option ETH_MAC = ENC28J60
#endif

//
// MAC RAM definitions
//
// The MAC_TX_BUFFER_COUNT must be equal to MAX_TCP_SOCKETS + 1
// (1 reserved for high priority messages), or else calls to
// TCPPut may fail when multiple TCP sockets have data pending
// in the output buffer that hasn't been ACKed.  Changing this
// value is recommended only if the ramifications of doing so
// are properly understood. 
#define _MAC_TX_BUFFER_SIZE     = (1500)
 
// This value will change if TCP is enabled...
#ifndef RESERVED_TCP_MEMORY
   #define RESERVED_TCP_MEMORY = 0
#endif

#ifndef STACK_USE_HTTP_SERVER
   #define RESERVED_HTTP_MEMORY = 0
#endif

#ifndef STACK_USE_SSL_SERVER
   #define RESERVED_SSL_MEMORY = 0
#endif

// ENC28J60 or PIC18F97J60 family internal Ethernet controller
#if (ETH_MAC = ETHXXJ60) or (ETH_MAC = ENC28J60)
  #define _RAMSIZE         = (8*1024)     
  #define _TXSTART         = (_RAMSIZE - (1+1518+7) - RESERVED_TCP_MEMORY - RESERVED_HTTP_MEMORY - RESERVED_SSL_MEMORY)
  #define _RXSTART         = (0)                      // Should be an even memory address; must be 0 for errata
  #define _RXSTOP          = ((_TXSTART-2) or $0001)    // Odd for errata workaround
  #define _RXSIZE          = (_RXSTOP-_RXSTART+1)
  #define _BASE_TX_ADDR    = (_TXSTART + 1)
  #define _BASE_TCB_ADDR   = (_BASE_TX_ADDR + (1514+7))
  #define _BASE_HTTPB_ADDR = (_BASE_TCB_ADDR + RESERVED_TCP_MEMORY)
  #define _BASE_SSLB_ADDR  = (_BASE_HTTPB_ADDR + RESERVED_HTTP_MEMORY)
// ENC424J600 10/100 Ethernet controller
#elseif (ETH_MAC = ENCx24J600)
  #define _RAMSIZE         = (24*1024)      
  #define _TXSTART         = (0)
  #define _RXSTART         = (_TXSTART + 1518  + RESERVED_TCP_MEMORY + RESERVED_HTTP_MEMORY + RESERVED_SSL_MEMORY)
  #define _RXSTOP          = (_RAMSIZE-1)
  #define _RXSIZE          = (_RXSTOP-_RXSTART+1)
  #define _BASE_TX_ADDR    = (_TXSTART)
  #define _BASE_TCB_ADDR   = (_BASE_TX_ADDR + 1518)
  #define _BASE_HTTPB_ADDR = (_BASE_TCB_ADDR + RESERVED_TCP_MEMORY)
  #define _BASE_SSLB_ADDR  = (_BASE_HTTPB_ADDR + RESERVED_HTTP_MEMORY)
#endif

#if (_RXSIZE < 1400) or (_RXSIZE > _RAMSIZE)
   #warning "Ethernet RX buffer is tiny.  Reduce MAX_TCP_SOCKETS, TCP_TX_FIFO_SIZE, Or TCP_RX_FIFO_SIZE"
#endif

// MAC selection
#if (ETH_MAC = ETHXXJ60)
// 18FxxJ60 with internal ethernet
Include "ETH97J60.bas"
Public Type
   ETHER_HEADER = MAC_ETHER_HEADER   
Public Const
   MAC_IP           = ETH97J60.MAC_IP,
   MAC_ARP          = ETH97J60.MAC_ARP,
   MAC_UNKNOWN      = ETH97J60.MAC_UNKNOWN,
   MAC_TX_BUFFER_SIZE = ETH97J60.MAC_TX_BUFFER_SIZE,
   RAMSIZE          = ETH97J60.RAMSIZE,
   TXSTART          = ETH97J60.TXSTART,
   RXSTART          = ETH97J60.RXSTART,
   RXSTOP           = ETH97J60.RXSTOP,
   RXSIZE           = ETH97J60.RXSIZE,
   BASE_TX_ADDR     = ETH97J60.BASE_TX_ADDR,
   BASE_TCB_ADDR    = ETH97J60.BASE_TCB_ADDR,
   BASE_HTTPB_ADDR  = ETH97J60.BASE_HTTPB_ADDR
Public Dim
   MACInit As ETH97J60.MACInit,
   MACIsLinked As ETH97J60.MACIsLinked,
   MACGetHeader As ETH97J60.MACGetHeader,
   MACSetReadPtrInRx As ETH97J60.MACSetReadPtrInRx,
   MACSetWritePtr As ETH97J60.MACSetWritePtr,
   MACSetReadPtr As ETH97J60.MACSetReadPtr,
   MACGet As ETH97J60.MACGet,
   MACGetArray As ETH97J60.MACGetArray,
   MACDiscardRx As ETH97J60.MACDiscardRx,
   MACGetFreeRxSize As ETH97J60.MACGetFreeRxSize,
   MACMemCopyAsync As ETH97J60.MACMemCopyAsync,
   MACIsMemCopyDone As ETH97J60.MACIsMemCopyDone,
   MACPutHeader As ETH97J60.MACPutHeader,
   MACIsTxReady As ETH97J60.MACIsTxReady,
   MACPut As ETH97J60.MACPut,
   MACPutArray As ETH97J60.MACPutArray,
   MACPutROMArray As ETH97J60.MACPutROMArray,
   MACFlush As ETH97J60.MACFlush,
   MACCalcRxChecksum As ETH97J60.MACCalcRxChecksum,
   CalcIPBufferChecksum As ETH97J60.CalcIPBufferChecksum

// not internal... must be an external MAC
#elseif (ETH_MAC = ENC28J60)
// import ENC28J60
Include "ENC28J60.bas"
Public Type
   ETHER_HEADER = MAC_ETHER_HEADER   
Public Const
   MAC_IP           = ENC28J60.MAC_IP,
   MAC_ARP          = ENC28J60.MAC_ARP,
   MAC_UNKNOWN      = ENC28J60.MAC_UNKNOWN,
   MAC_TX_BUFFER_SIZE = ENC28J60.MAC_TX_BUFFER_SIZE,
   RAMSIZE          = ENC28J60.RAMSIZE,
   TXSTART          = ENC28J60.TXSTART,
   RXSTART          = ENC28J60.RXSTART,
   RXSTOP           = ENC28J60.RXSTOP,
   RXSIZE           = ENC28J60.RXSIZE,
   BASE_TX_ADDR     = ENC28J60.BASE_TX_ADDR,
   BASE_TCB_ADDR    = ENC28J60.BASE_TCB_ADDR,
   BASE_HTTPB_ADDR  = ENC28J60.BASE_HTTPB_ADDR
Public Dim
   MACInit As ENC28J60.MACInit,
   MACIsLinked As ENC28J60.MACIsLinked,
   MACGetHeader As ENC28J60.MACGetHeader,
   MACSetReadPtrInRx As ENC28J60.MACSetReadPtrInRx,
   MACSetWritePtr As ENC28J60.MACSetWritePtr,
   MACSetReadPtr As ENC28J60.MACSetReadPtr,
   MACGet As ENC28J60.MACGet,
   MACGetArray As ENC28J60.MACGetArray,
   MACDiscardRx As ENC28J60.MACDiscardRx,
   MACGetFreeRxSize As ENC28J60.MACGetFreeRxSize,
   MACMemCopyAsync As ENC28J60.MACMemCopyAsync,
   MACIsMemCopyDone As ENC28J60.MACIsMemCopyDone,
   MACPutHeader As ENC28J60.MACPutHeader,
   MACIsTxReady As ENC28J60.MACIsTxReady,
   MACPut As ENC28J60.MACPut,
   MACPutArray As ENC28J60.MACPutArray,
   MACPutROMArray As ENC28J60.MACPutROMArray,
   MACFlush As ENC28J60.MACFlush,
   MACCalcRxChecksum As ENC28J60.MACCalcRxChecksum,
   CalcIPBufferChecksum As ENC28J60.CalcIPBufferChecksum

#elseif (ETH_MAC = ENCx24J600)
// import ENCx24J600
Include "ENCx24J600.bas"
Public Type
   ETHER_HEADER = MAC_ETHER_HEADER   
Public Const
   MAC_IP           = ENCx24J600.MAC_IP,
   MAC_ARP          = ENCx24J600.MAC_ARP,
   MAC_UNKNOWN      = ENCx24J600.MAC_UNKNOWN,
   MAC_TX_BUFFER_SIZE = ENCx24J600.MAC_TX_BUFFER_SIZE,
   RAMSIZE          = ENCx24J600.RAMSIZE,
   TXSTART          = ENCx24J600.TXSTART,
   RXSTART          = ENCx24J600.RXSTART,
   RXSTOP           = ENCx24J600.RXSTOP,
   RXSIZE           = ENCx24J600.RXSIZE,
   BASE_TX_ADDR     = ENCx24J600.BASE_TX_ADDR,
   BASE_TCB_ADDR    = ENCx24J600.BASE_TCB_ADDR,
   BASE_HTTPB_ADDR  = ENCx24J600.BASE_HTTPB_ADDR
Public Dim
   MACInit As ENCx24J600.MACInit,
   MACIsLinked As ENCx24J600.MACIsLinked,
   MACGetHeader As ENCx24J600.MACGetHeader,
   MACSetReadPtrInRx As ENCx24J600.MACSetReadPtrInRx,
   MACSetWritePtr As ENCx24J600.MACSetWritePtr,
   MACSetReadPtr As ENCx24J600.MACSetReadPtr,
   MACGet As ENCx24J600.MACGet,
   MACGetArray As ENCx24J600.MACGetArray,
   MACDiscardRx As ENCx24J600.MACDiscardRx,
   MACGetFreeRxSize As ENCx24J600.MACGetFreeRxSize,
   MACMemCopyAsync As ENCx24J600.MACMemCopyAsync,
   MACIsMemCopyDone As ENCx24J600.MACIsMemCopyDone,
   MACPutHeader As ENCx24J600.MACPutHeader,
   MACIsTxReady As ENCx24J600.MACIsTxReady,
   MACPut As ENCx24J600.MACPut,
   MACPutArray As ENCx24J600.MACPutArray,
   MACPutROMArray As ENCx24J600.MACPutROMArray,
   MACFlush As ENCx24J600.MACFlush,
   MACCalcRxChecksum As ENCx24J600.MACCalcRxChecksum,
   CalcIPBufferChecksum As ENCx24J600.CalcIPBufferChecksum

// other implementations go here...
#else
   #error "invalid Ethernet MAC selection"
   {
      It is very important that with any new harware implementation, the types,
      constants, functions and subroutines shown above in the ENC28J60 import
      section are included
   }
#endif

end module
