program EthernetTest

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
'{
// LJCV Electronics eIP-10 18F2620 with ENC28J60 (http://www.ljcv.net/eip10/)
device = 18F2620
clock = 40
#option USER_INIT = "EIP10_2620.bas"
#option ETH_MAC = ENC28J60
'}

{
// Microchip explorer18 (18F8722) with pictail plus (ETH_MAC = ENCx24J600)
device = 18F8722
clock = 32
#option USER_INIT = "EXP18_8722.bas"
#option ETH_MAC = ENCx24J600
}

{
// Microchip explorer18 with 18F87K22 PIM and pictail plus (ETH_MAC = ENCx24J600)
device = 18F87K22
clock = 64
#option USER_INIT = "EXP18_87K22.bas"
#option ETH_MAC = ENCx24J600
}

{
// Olimex PIC-WEB board 18F67J60 (https://www.olimex.com/Products/PIC/Development/PIC-WEB/)
device = 18F67J60
// select clock - either 25MHz or 41.6667MHz
'clock = 25      // default
clock = 41.6667    // 5xPLL=enabled, PLL Prescaler=div3, PLL Postscaler=disabled
#option USER_INIT = "PIC_WEB_67J60.bas"
'#option ETH_MAC = ETHXXJ60
}

//
//-----------------------------------------------------------------------------
// hardware setup
//-----------------------------------------------------------------------------
//
// include the board support USER_INIT file specified above
include USER_INIT

// USART support
// note: recommend using the modified USART.BAS module v1.6
#option USART_BRGH = true
#option USART_BRG16 = true
Include "usart.bas"

// for the ENC28J60 or ETHXXJ60, the default duplex setting is HALF_DUPLEX
// FULL_DUPLEX may or may not work depending on your network hardware setup 
// (since they don't support auto-negotiation). The ENCx24J600 will negotiate
// 10/100 speed and duplex settings
'#option ETH_DUPLEX = FULL_DUPLEX

// important configurations...
#option NET_ICMP = true   // enable ping 
#option NET_DHCP = true   // enable Dynamic Host Configuration Protocol (DHCP)
Include "NETApp.bas"      // always first!

// program helpers...
Include "NETAppDisplay.bas"

// program start...
usart.SetBaudrate(br115200)

#if NET_DHCP
usart.Write("Waiting for DHCP...",13,10)
#endif

// loop forever...
While true
    // This task performs normal stack task including checking for incoming 
    // packet, type of packet and calling appropriate stack entity to process it.
    NETApp.Task

    // If DHCP is enabled, display configuration when a DHCP event has occurred...
    #if NET_DHCP
    If NETApp.DHCPEvent() Then
        DisplayConfig()
    EndIf

    // read USART value - if equal to "d" then display config info...
    If usart.DataAvailable Then
        If usart.ReadByte = Byte("d") Then
            DisplayConfig()
        EndIf
    EndIf
    #endif
Wend

end program

