{
****************************************************************
*  Name    : 18F97J60                                          *
*  Author  : David John Barker                                 *
*  Notice  : Copyright (c) 2015 Mecanique                      *
*          : All Rights Reserved                               *
*  Date    : 2/26/2015                                         *
****************************************************************
}

module SystemTypes

// system header...
#const _core = $0012                   // processor core
#const _ram_banks = $0F                // 15 RAM bank(s) used
#variable _maxaccess = $60             // access ram is 96 bytes
#variable _maxram = $0E80              // 3712 bytes of user RAM
#variable _maxrom = $01FFF8            // 127 KB of program ROM
#const _eeprom = $0000                 // 0 bytes of EEPROM
#const _eeprom_start = $F00000         // EEPROM start address
#const _ports = $09                    // 9 available ports
#const _no_porti = $01                 // no PORTI
#const _ccp = $02                      // 2 CCP module(s) available
#const _eccp = $03                     // 3 ECCP module(s) available
#const _mssp = $02                     // 2 MSSP module(s) available
#const _usart = $02                    // 2 USART(s) available
#const _adc = $10                      // 16 ADC channels available
#const _adres = $0A                    // 10 bit ADC resolution
#const _comparator = $02               // 2 comparator(s) available
#const _psp = $01                      // onboard Parallel Slave Port (PSP) available
#const _can = $00                      // onboard CAN is NOT supported
#const _usb = $00                      // USB is NOT supported
#const _ethernet = $01                 // onboard Ethernet available
#const _flash_write = $01              // FLASH has write capability
#const _wdt_type = $00                 // WDT watchdog

// special function registers...
public system
   MAADR5 as byte absolute $0E80,
   MAADR6 as byte absolute $0E81,
   MAADR3 as byte absolute $0E82,
   MAADR4 as byte absolute $0E83,
   MAADR1 as byte absolute $0E84,
   MAADR2 as byte absolute $0E85,
   MISTAT as byte absolute $0E8A,
   EFLOCON as byte absolute $0E97,
   EPAUS as byte absolute $0E98,
   EPAUSL as byte absolute $0E98,
   EPAUSH as byte absolute $0E99,
   MACON1 as byte absolute $0EA0,
   MACON3 as byte absolute $0EA2,
   MACON4 as byte absolute $0EA3,
   MABBIPG as byte absolute $0EA4,
   MAIPG as byte absolute $0EA6,
   MAIPGL as byte absolute $0EA6,
   MAIPGH as byte absolute $0EA7,
   MAMXFL as byte absolute $0EAA,
   MAMXFLL as byte absolute $0EAA,
   MAMXFLH as byte absolute $0EAB,
   MICMD as byte absolute $0EB2,
   MIREGADR as byte absolute $0EB4,
   MIWR as byte absolute $0EB6,
   MIWRL as byte absolute $0EB6,
   MIWRH as byte absolute $0EB7,
   MIRD as byte absolute $0EB8,
   MIRDL as byte absolute $0EB8,
   MIRDH as byte absolute $0EB9,
   EHT0 as byte absolute $0EC0,
   EHT1 as byte absolute $0EC1,
   EHT2 as byte absolute $0EC2,
   EHT3 as byte absolute $0EC3,
   EHT4 as byte absolute $0EC4,
   EHT5 as byte absolute $0EC5,
   EHT6 as byte absolute $0EC6,
   EHT7 as byte absolute $0EC7,
   EPMM0 as byte absolute $0EC8,
   EPMM1 as byte absolute $0EC9,
   EPMM2 as byte absolute $0ECA,
   EPMM3 as byte absolute $0ECB,
   EPMM4 as byte absolute $0ECC,
   EPMM5 as byte absolute $0ECD,
   EPMM6 as byte absolute $0ECE,
   EPMM7 as byte absolute $0ECF,
   EPMCS as byte absolute $0ED0,
   EPMCSL as byte absolute $0ED0,
   EPMCSH as byte absolute $0ED1,
   EPMO as byte absolute $0ED4,
   EPMOL as byte absolute $0ED4,
   EPMOH as byte absolute $0ED5,
   ERXFCON as byte absolute $0ED8,
   EPKTCNT as byte absolute $0ED9,
   EWRPT as byte absolute $0EE2,
   EWRPTL as byte absolute $0EE2,
   EWRPTH as byte absolute $0EE3,
   ETXST as byte absolute $0EE4,
   ETXSTL as byte absolute $0EE4,
   ETXSTH as byte absolute $0EE5,
   ETXND as byte absolute $0EE6,
   ETXNDL as byte absolute $0EE6,
   ETXNDH as byte absolute $0EE7,
   ERXST as byte absolute $0EE8,
   ERXSTL as byte absolute $0EE8,
   ERXSTH as byte absolute $0EE9,
   ERXND as byte absolute $0EEA,
   ERXNDL as byte absolute $0EEA,
   ERXNDH as byte absolute $0EEB,
   ERXRDPT as byte absolute $0EEC,
   ERXRDPTL as byte absolute $0EEC,
   ERXRDPTH as byte absolute $0EED,
   ERXWRPT as byte absolute $0EEE,
   ERXWRPTL as byte absolute $0EEE,
   ERXWRPTH as byte absolute $0EEF,
   EDMAST as byte absolute $0EF0,
   EDMASTL as byte absolute $0EF0,
   EDMASTH as byte absolute $0EF1,
   EDMAND as byte absolute $0EF2,
   EDMANDL as byte absolute $0EF2,
   EDMANDH as byte absolute $0EF3,
   EDMADST as byte absolute $0EF4,
   EDMADSTL as byte absolute $0EF4,
   EDMADSTH as byte absolute $0EF5,
   EDMACS as byte absolute $0EF6,
   EDMACSL as byte absolute $0EF6,
   EDMACSH as byte absolute $0EF7,
   EIE as byte absolute $0EFB,
   ESTAT as byte absolute $0EFD,
   ECON2 as byte absolute $0EFE,
   EIR as byte absolute $0F60,
   // EDATA as byte absolute $0F61,
   SSP2CON2 as byte absolute $0F62,
   SSP2CON1 as byte absolute $0F63,
   SSP2STAT as byte absolute $0F64,
   SSP2ADD as byte absolute $0F65,
   SSP2BUF as byte absolute $0F66,
   ECCP2DEL as byte absolute $0F67,
   ECCP2AS as byte absolute $0F68,
   ECCP3DEL as byte absolute $0F69,
   ECCP3AS as byte absolute $0F6A,
   RCSTA2 as byte absolute $0F6B,
   TXSTA2 as byte absolute $0F6C,
   TXREG2 as byte absolute $0F6D,
   RCREG2 as byte absolute $0F6E,
   SPBRG2 as byte absolute $0F6F,
   CCP5CON as byte absolute $0F70,
   CCPR5 as byte absolute $0F71,
   CCPR5L as byte absolute $0F71,
   CCPR5H as byte absolute $0F72,
   CCP4CON as byte absolute $0F73,
   CCPR4 as byte absolute $0F74,
   CCPR4L as byte absolute $0F74,
   CCPR4H as byte absolute $0F75,
   T4CON as byte absolute $0F76,
   PR4 as byte absolute $0F77,
   TMR4 as byte absolute $0F78,
   ECCP1DEL as byte absolute $0F79,
   ERDPT as byte absolute $0F7A,
   ERDPTL as byte absolute $0F7A,
   ERDPTH as byte absolute $0F7B,
   BAUDCON2 as byte absolute $0F7C,
   BAUDCTL2 as byte absolute $0F7C,
   SPBRGH2 as byte absolute $0F7D,
   BAUDCON as byte absolute $0F7E,
   BAUDCON1 as byte absolute $0F7E,
   BAUDCTL as byte absolute $0F7E,
   BAUDCTL1 as byte absolute $0F7E,
   SPBRGH as byte absolute $0F7F,
   SPBRGH1 as byte absolute $0F7F,
   LATA as byte absolute $0F89,
   LATB as byte absolute $0F8A,
   LATC as byte absolute $0F8B,
   LATD as byte absolute $0F8C,
   LATE as byte absolute $0F8D,
   LATF as byte absolute $0F8E,
   LATG as byte absolute $0F8F,
   LATH as byte absolute $0F90,
   LATJ as byte absolute $0F91,
   DDRA as byte absolute $0F92,
   TRISA as byte absolute $0F92,
   DDRB as byte absolute $0F93,
   TRISB as byte absolute $0F93,
   DDRC as byte absolute $0F94,
   TRISC as byte absolute $0F94,
   DDRD as byte absolute $0F95,
   TRISD as byte absolute $0F95,
   DDRE as byte absolute $0F96,
   TRISE as byte absolute $0F96,
   DDRF as byte absolute $0F97,
   TRISF as byte absolute $0F97,
   DDRG as byte absolute $0F98,
   TRISG as byte absolute $0F98,
   DDRH as byte absolute $0F99,
   TRISH as byte absolute $0F99,
   DDRJ as byte absolute $0F9A,
   TRISJ as byte absolute $0F9A,
   OSCTUNE as byte absolute $0F9B,
   MEMCON as byte absolute $0F9C,
   PIE1 as byte absolute $0F9D,
   PIR1 as byte absolute $0F9E,
   IPR1 as byte absolute $0F9F,
   PIE2 as byte absolute $0FA0,
   PIR2 as byte absolute $0FA1,
   IPR2 as byte absolute $0FA2,
   PIE3 as byte absolute $0FA3,
   PIR3 as byte absolute $0FA4,
   IPR3 as byte absolute $0FA5,
   EECON1 as byte absolute $0FA6,
   EECON2 as byte absolute $0FA7,
   RCSTA as byte absolute $0FAB,
   RCSTA1 as byte absolute $0FAB,
   TXSTA as byte absolute $0FAC,
   TXSTA1 as byte absolute $0FAC,
   TXREG as byte absolute $0FAD,
   TXREG1 as byte absolute $0FAD,
   RCREG as byte absolute $0FAE,
   RCREG1 as byte absolute $0FAE,
   SPBRG as byte absolute $0FAF,
   SPBRG1 as byte absolute $0FAF,
   PSPCON as byte absolute $0FB0,
   T3CON as byte absolute $0FB1,
   TMR3 as byte absolute $0FB2,
   TMR3L as byte absolute $0FB2,
   TMR3H as byte absolute $0FB3,
   CMCON as byte absolute $0FB4,
   CVRCON as byte absolute $0FB5,
   ECCP1AS as byte absolute $0FB6,
   CCP3CON as byte absolute $0FB7,
   ECCP3CON as byte absolute $0FB7,
   CCPR3 as byte absolute $0FB8,
   CCPR3L as byte absolute $0FB8,
   CCPR3H as byte absolute $0FB9,
   CCP2CON as byte absolute $0FBA,
   ECCP2CON as byte absolute $0FBA,
   CCPR2 as byte absolute $0FBB,
   CCPR2L as byte absolute $0FBB,
   CCPR2H as byte absolute $0FBC,
   CCP1CON as byte absolute $0FBD,
   ECCP1CON as byte absolute $0FBD,
   CCPR1 as byte absolute $0FBE,
   CCPR1L as byte absolute $0FBE,
   CCPR1H as byte absolute $0FBF,
   ADCON2 as byte absolute $0FC0,
   ADCON1 as byte absolute $0FC1,
   ADCON0 as byte absolute $0FC2,
   ADRES as byte absolute $0FC3,
   ADRESL as byte absolute $0FC3,
   ADRESH as byte absolute $0FC4,
   SSP1CON2 as byte absolute $0FC5,
   SSPCON2 as byte absolute $0FC5,
   SSP1CON1 as byte absolute $0FC6,
   SSPCON1 as byte absolute $0FC6,
   SSP1STAT as byte absolute $0FC7,
   SSPSTAT as byte absolute $0FC7,
   SSP1ADD as byte absolute $0FC8,
   SSPADD as byte absolute $0FC8,
   SSP1BUF as byte absolute $0FC9,
   SSPBUF as byte absolute $0FC9,
   T2CON as byte absolute $0FCA,
   PR2 as byte absolute $0FCB,
   TMR2 as byte absolute $0FCC,
   T1CON as byte absolute $0FCD,
   TMR1 as byte absolute $0FCE,
   TMR1L as byte absolute $0FCE,
   TMR1H as byte absolute $0FCF,
   RCON as byte absolute $0FD0,
   WDTCON as byte absolute $0FD1,
   ECON1 as byte absolute $0FD2,
   OSCCON as byte absolute $0FD3,
   T0CON as byte absolute $0FD5,
   TMR0 as byte absolute $0FD6,
   TMR0L as byte absolute $0FD6,
   TMR0H as byte absolute $0FD7,
   STATUS as byte absolute $0FD8,
   FSR2L as byte absolute $0FD9,
   FSR2H as byte absolute $0FDA,
   PLUSW2 as byte absolute $0FDB,
   PREINC2 as byte absolute $0FDC,
   POSTDEC2 as byte absolute $0FDD,
   POSTINC2 as byte absolute $0FDE,
   INDF2 as byte absolute $0FDF,
   BSR as byte absolute $0FE0,
   FSR1L as byte absolute $0FE1,
   FSR1H as byte absolute $0FE2,
   PLUSW1 as byte absolute $0FE3,
   PREINC1 as byte absolute $0FE4,
   POSTDEC1 as byte absolute $0FE5,
   POSTINC1 as byte absolute $0FE6,
   INDF1 as byte absolute $0FE7,
   WREG as byte absolute $0FE8,
   FSR0L as byte absolute $0FE9,
   FSR0H as byte absolute $0FEA,
   PLUSW0 as byte absolute $0FEB,
   PREINC0 as byte absolute $0FEC,
   POSTDEC0 as byte absolute $0FED,
   POSTINC0 as byte absolute $0FEE,
   INDF0 as byte absolute $0FEF,
   INTCON3 as byte absolute $0FF0,
   INTCON2 as byte absolute $0FF1,
   INTCON as byte absolute $0FF2,
   PRODL as byte absolute $0FF3,
   PRODH as byte absolute $0FF4,
   TABLAT as byte absolute $0FF5,
   TBLPTR as byte absolute $0FF6,
   TBLPTRL as byte absolute $0FF6,
   TBLPTRH as byte absolute $0FF7,
   TBLPTRU as byte absolute $0FF8,
   PC as byte absolute $0FF9,
   PCL as byte absolute $0FF9,
   PCLATH as byte absolute $0FFA,
   PCLATU as byte absolute $0FFB,
   STKPTR as byte absolute $0FFC,
   TOS as byte absolute $0FFD,
   TOSL as byte absolute $0FFD,
   TOSH as byte absolute $0FFE,
   TOSU as byte absolute $0FFF

// system ports...
public system port
   PORTA as byte absolute $0F80,
   PORTB as byte absolute $0F81,
   PORTC as byte absolute $0F82,
   PORTD as byte absolute $0F83,
   PORTE as byte absolute $0F84,
   PORTF as byte absolute $0F85,
   PORTG as byte absolute $0F86,
   PORTH as byte absolute $0F87,
   PORTJ as byte absolute $0F88

// alias...
public dim
   EDATA as byte absolute $0F61,
   PROD as PRODL.AsWord,
   FSR0 as FSR0L.AsWord,
   FSR1 as FSR1L.AsWord,
   FSR2 as FSR2L.AsWord,
   TABLEPTR as TBLPTRL.AsWord

// configuration fuses...
public config
   WDT(WDT) = [OFF, ON],
   STVR(STVR) = [OFF, ON],
   XINST(XINST) = [OFF, ON],
   DEBUG(DEBUG) = [ON, OFF],
   CP0(CP0) = [ON, OFF],
   FOSC(FOSC) = [HS, HSPLL, EC, ECPLL],
   FOSC2(FOSC2) = [OFF, ON],
   FCMEN(FCMEN) = [OFF, ON],
   IESO(IESO) = [OFF, ON],
   WDTPS(WDTPS) = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768],
   EASHFT(EASHFT) = [OFF, ON],
   MODE(MODE) = [XM20, XM16, XM12, MM],
   BW(BW) = [8, 16],
   WAIT(WAIT) = [ON, OFF],
   CCP2MX(CCP2MX) = [OFF, ON],
   ECCPMX(ECCPMX) = [OFF, ON],
   ETHLED(ETHLED) = [OFF, ON]

// default fuses...
config
   WDT = OFF,
   STVR = ON,
   XINST = OFF,
   DEBUG = OFF,
   FOSC = HS,
   FOSC2 = ON,
   FCMEN = OFF,
   IESO = OFF,
   WDTPS = 128
