//
//-----------------------------------------------------------------------------
// microchip Explorer18 board 18F8722 with PICTailPlus ENCx24J600 10/100
//-----------------------------------------------------------------------------
//
module user_init

#define EXP18_8722

// call userInit at startup (true/false)
#option USER_AUTO_INIT = true

//
//-----------------------------------------------------------------------------
// config settings
//-----------------------------------------------------------------------------
//

// config
#if (_clock = 10)       // 10MHz xtal
  config OSC = HS
#elseif (_clock = 40)   // 10MHz xtal + PLL
  config OSC = HSPLL
#elseif (_clock = 8)    // intosc
  config  OSC = INTIO67      // Internal oscillator block, port function on RA6 and RA7
#elseif (_clock = 32)   // intosc + pll
  config  OSC = INTIO67      // Internal oscillator block, port function on RA6 and RA7
#elseif (_clock = 64) and (_device = 18F87K22)
  config  OSC = INTIO67      // Internal oscillator block, port function on RA6 and RA7
#else
  #error "unsupported _clock"
#endif

config
    FCMEN = OFF,        // Fail-Safe Clock Monitor disabled
    IESO = OFF,         // Two-Speed Start-up disabled
    PWRT = ON,          // PWRT enabled
    BOREN = SBORDIS,    // Brown-out Reset enabled in hardware only (SBOREN is disabled)
    BORV = 3,           // Minimum setting
    WDT = OFF,          // WDT disabled (control is placed on the SWDTEN bit)
    WDTPS = 32768,      // 1:32768
    MODE = MC,          // Microcontroller mode
    ADDRBW = ADDR20BIT, // 20-bit Address Bus
    DATABW = DATA16BIT, // 16-bit External Bus mode
    WAIT = OFF,         // Wait selections are unavailable for table reads and table writes
    CCP2MX = PORTC,     // ECCP2 input/output is multiplexed with RC1
    ECCPMX = PORTE,     // ECCP1/3 (P1B/P1C/P3B/P3C) are multiplexed onto RE6, RE5, RE4 and RE3 respectively
    LPT1OSC = OFF,      // Timer1 configured for higher power operation
    MCLRE = ON,         // MCLR pin enabled; RG5 input pin disabled
    STVREN = ON,        // Stack full/underflow will cause Reset
    LVP = OFF,          // Single-Supply ICSP disabled
    BBSIZ = BB2K,       // 1K word (2 Kbytes) Boot Block size
    XINST = OFF,        // Instruction set extension and Indexed Addressing mode disabled (Legacy mode)
    DEBUG = OFF,        // Background debugger disabled, RB6 and RB7 configured as general purpose I/O pins
    CP0 = OFF,          // Block 0 (000800, 001000 or 002000-003FFFh) not code-protected
    CP1 = OFF,          // Block 1 (004000-007FFFh) not code-protected
    CP2 = OFF,          // Block 2 (008000-00BFFFh) not code-protected
    CP3 = OFF,          // Block 3 (00C000-00FFFFh) not code-protected
    CP4 = OFF,          // Block 4 (010000-013FFFh) not code-protected
    CP5 = OFF,          // Block 5 (014000-017FFFh) not code-protected
    CP6 = OFF,          // Block 6 (01BFFF-018000h) not code-protected
    CP7 = OFF,          // Block 7 (01C000-01FFFFh) not code-protected
    CPB = OFF,          // Boot Block (000000-0007FFh) not code-protected
    CPD = OFF,          // Data EEPROM not code-protected
    WRT0 = OFF,         // Block 0 (000800, 001000 or 002000-003FFFh) not write-protected
    WRT1 = OFF,         // Block 1 (004000-007FFFh) not write-protected
    WRT2 = OFF,         // Block 2 (008000-00BFFFh) not write-protected
    WRT3 = OFF,         // Block 3 (00C000-00FFFFh) not write-protected
    WRT4 = OFF,         // Block 4 (010000-013FFFh) not write-protected
    WRT5 = OFF,         // Block 5 (014000-017FFFh) not write-protected
    WRT6 = OFF,         // Block 6 (01BFFF-018000h) not write-protected
    WRT7 = OFF,         // Block 7 (01C000-01FFFFh) not write-protected
    WRTC = OFF,         // Configuration registers (300000-3000FFh) not write-protected
    WRTB = OFF,         // Boot Block (000000-007FFF, 000FFF or 001FFFh) not write-protected
    WRTD = OFF,         // Data EEPROM not write-protected
    EBTR0 = OFF,        // Block 0 (000800, 001000 or 002000-003FFFh) not protected from table reads executed in other blocks
    EBTR1 = OFF,        // Block 1 (004000-007FFFh) not protected from table reads executed in other blocks
    EBTR2 = OFF,        // Block 2 (008000-00BFFFh) not protected from table reads executed in other blocks
    EBTR3 = OFF,        // Block 3 (00C000-00FFFFh) not protected from table reads executed in other blocks
    EBTR4 = OFF,        // Block 4 (010000-013FFFh) not protected from table reads executed in other blocks
    EBTR5 = OFF,        // Block 5 (014000-017FFFh) not protected from table reads executed in other blocks
    EBTR6 = OFF,        // Block 6 (018000-01BFFFh) not protected from table reads executed in other blocks
    EBTR7 = OFF,        // Block 7 (01C000-01FFFFh) not protected from table reads executed in other blocks
    EBTRB = OFF         // Boot Block (000000-007FFF, 000FFF or 001FFFh) not protected from table reads executed in other blocks

// USART support
// note: be sure to use the modified USART.BAS module that supports non-contiguous 
// SPBRG/SPBRGH registers otherwise the baudrate will not be correct
#option USART_BRGH = true
#option USART_BRG16 = true

//
//-----------------------------------------------------------------------------
// Explorer18 hardware
//-----------------------------------------------------------------------------
//
// LEDs
// LED D1 (active high, RD7:RD0, enabled via jumper JP1)
public dim LED1         as PORTD.0

// Switches
// S1 pushbutton sw w/ext pullup
public dim INPUT_SW     as PORTB.0

// Analog
// trimpot connected to RA0/AN0
public dim TRIM_POT     as PORTA.0
// MCP9701A temp sensor connected to RA1/AN1
public dim TEMP_SENSOR  as PORTA.1

// Serial SPI EEPROM (enabled/disabled by jumper JP2)
// 25LC256 SPI
{
#define SPI_EEPROM_MSSP = MSSP1
#define SPI_EEPROM_CS = PORTA.3     // disconnect via JP2
#define SPI_EEPROM_SCK = PORTC.3    // SCK1
#define SPI_EEPROM_SDI = PORTC.4    // SDI1
#define SPI_EEPROM_SDO = PORTC.5    // SDO1
}

// PICTailPlus fast ethernet board (via PICTail daughter board header)
#option ENC_MSSP = 1        // MSSP1
#option ENC_CS   = PORTB.3  // CSn  (J3-21 /CS)
#option ENC_INT  = PORTB.2  // INTn to mcu RB2/INT2 (J3-23)
{
#option ENC_SDO = PORTC.5   // SDO1 (J3-8 SI)
#option ENC_SDI = PORTC.4   // SDI1 (J3-10 SO)
#option ENC_SCK = PORTC.3   // SCK1 (J3-12 SCK)
}

// addtl PICTailPlus IO pins
#option ENC_SHDN = PORTB.5  // SHDNn low-active shutdown from mcu (J3-17)
#option ENC_MDIX = PORTB.4  // MDIX ethernet crossover logic from mcu (J3-19)

dim ENC_SHDN_IO as ENC_SHDN.ENC_SHDN@
dim ENC_MDIX_IO as ENC_MDIX.ENC_MDIX@

//
//-----------------------------------------------------------------------------
// userInitClock() - setup 18F8722 clock
//-----------------------------------------------------------------------------
//
public sub userInitClock()
    const PLLEN = 6     // OSCTUNE PLL bit
    
  #if (_clock = 10) or (_clock = 40)         // HS or HSPLL
    // nothing to do...
  #elseif (_clock = 32)     // INTOSC 8MHZ + 4x PLL
    OSCCON = $70            // INTOSC 8MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 1     // 4x pllen on
  #elseif (_clock = 8)
    OSCCON = $70            // INTOSC 8MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 0     // pllen off
  #else
    #error "unsupported _clock selection"
  #endif
    // allow time for the PLL to lock (if it's enabled)
    if (OSCTUNE.bits(PLLEN) = 1) then
        delayms(2)
    endif
end sub

//
//-----------------------------------------------------------------------------
// userInit() - setup clock, IO pins and other req'd regs
//-----------------------------------------------------------------------------
//
public sub userInit()
    userInitClock()
    
    // init the addtl PICTailPlus IO pins
    high(ENC_SHDN_IO)
    low(ENC_MDIX_IO)

    // init led & switch
    low(LED1)
    input(INPUT_SW)
end sub

//
//-----------------------------------------------------------------------------
// userClockCheck() - toggle led at 1ms rate to verify clock setting
//-----------------------------------------------------------------------------
//
public sub userClockCheck()
    while(true)
        LED1 = 1
        delayus(500)
        LED1 = 0
        delayus(500)
    end while
end sub

//
//-----------------------------------------------------------------------------
// module startup
//-----------------------------------------------------------------------------
//
#if (USER_AUTO_INIT)
  userInit()
#else
  #warning "userInit() NOT called at startup"
#endif

end module
