//
//-----------------------------------------------------------------------------
// microchip Explorer18 board with 18F87K22 PIM and PICTailPlus ENCx24J600 10/100
//-----------------------------------------------------------------------------
//
module user_init

#define EXP18_87K22
// optional Plug-in Module 
#define PIM_18F87K22

// call userInit at startup (true/false)
#option USER_AUTO_INIT = true

//
//-----------------------------------------------------------------------------
// config settings
//-----------------------------------------------------------------------------
//
// config
config PLLCFG = OFF       // Disabled
#if (_clock = 10)       // 10MHz xtal
  config FOSC = HS1
#elseif (_clock = 40)   // 10MHz xtal + PLL
  config FOSC = HS1
  config PLLCFG = ON
#elseif (_clock = 8)    // intosc
  config  FOSC = INTIO2      // Internal oscillator block, port function on RA6 and RA7
#elseif (_clock = 32)   // intosc + pll
  config  FOSC = INTIO2      // Internal oscillator block, port function on RA6 and RA7
#elseif (_clock = 64)
  config  FOSC = INTIO2     // Internal oscillator block, port function on RA6 and RA7
#else
  #error "unsupported _clock"
#endif

config
    RETEN = ON,         // Disabled - Controlled by SRETEN bit
    INTOSCSEL = HIGH,   // LF-INTOSC in High-power mode during Sleep
    SOSCSEL = HIGH,     // High Power SOSC circuit selected
    XINST = OFF,        // Disabled
    FCMEN = OFF,        // Disabled
    IESO = OFF,         // Disabled
    PWRTEN = OFF,       // Disabled
    BOREN = SBORDIS,    // Enabled in hardware, SBOREN disabled
    BORV = 3,           // 1.8V
    BORPWR = ZPBORMV,   // ZPBORMV instead of BORMV is selected
    WDTEN = SWDTDIS,    // WDT controlled by SWDTEN bit setting
    WDTPS = 1048576,    // 1:1048576
    RTCOSC = SOSCREF,   // RTCC uses SOSC
    EASHFT = ON,        // Address Shifting enabled
    ABW = MM,           // 8-bit address bus
    BW = 16,            // 16-bit external bus mode
    WAIT = OFF,         // Disabled
    CCP2MX = PORTC,     // RC1
    ECCPMX = PORTE,     // Enhanced CCP1/3 [P1B/P1C/P3B/P3C] muxed with RE6/RE5/RE4/RE3
    MSSPMSK = MSK7,     // 7 Bit address masking mode
    MCLRE = ON,         // MCLR Enabled, RG5 Disabled
    STVREN = ON,        // Enabled
    BBSIZ = BB2K,       // 2K word Boot Block size
    DEBUG = OFF,        // Disabled
    CP0 = OFF,          // Disabled
    CP1 = OFF,          // Disabled
    CP2 = OFF,          // Disabled
    CP3 = OFF,          // Disabled
    CP4 = OFF,          // Disabled
    CP5 = OFF,          // Disabled
    CP6 = OFF,          // Disabled
    CP7 = OFF,          // Disabled
    CPB = OFF,          // Disabled
    CPD = OFF,          // Disabled
    WRT0 = OFF,         // Disabled
    WRT1 = OFF,         // Disabled
    WRT2 = OFF,         // Disabled
    WRT3 = OFF,         // Disabled
    WRT4 = OFF,         // Disabled
    WRT5 = OFF,         // Disabled
    WRT6 = OFF,         // Disabled
    WRT7 = OFF,         // Disabled
    WRTC = OFF,         // Disabled
    WRTB = OFF,         // Disabled
    WRTD = OFF,         // Disabled
    EBRT0 = OFF,        // Disabled
    EBRT1 = OFF,        // Disabled
    EBRT2 = OFF,        // Disabled
    EBRT3 = OFF,        // Disabled
    EBRT4 = OFF,        // Disabled
    EBRT5 = OFF,        // Disabled
    EBRT6 = OFF,        // Disabled
    EBRT7 = OFF,        // Disabled
    EBRTB = OFF         // Disabled

// USART support
// note: be sure to use the modified USART.BAS module that supports non-contiguous 
// SPBRG/SPBRGH registers otherwise the baudrate will not be correct
#option USART_BRGH = true
#option USART_BRG16 = true

//
//-----------------------------------------------------------------------------
// Explorer18 hardware
//-----------------------------------------------------------------------------
//
// LEDs
// LED D1 (active high, RD7:RD0, enabled via jumper JP1)
public dim LED1         as PORTD.0

// Switches
// S1 pushbutton sw w/ext pullup
public dim INPUT_SW     as PORTB.0

// Analog
// trimpot connected to RA0/AN0
public dim TRIM_POT     as PORTA.0
// MCP9701A temp sensor connected to RA1/AN1
public dim TEMP_SENSOR  as PORTA.1

// Serial SPI EEPROM (enabled/disabled by jumper JP2)
// 25LC256 SPI
{
#define SPI_EEPROM_MSSP = MSSP1
#define SPI_EEPROM_CS = PORTA.3     // disconnect via JP2
#define SPI_EEPROM_SCK = PORTC.3    // SCK1
#define SPI_EEPROM_SDI = PORTC.4    // SDI1
#define SPI_EEPROM_SDO = PORTC.5    // SDO1
}

// PICTailPlus fast ethernet board (via PICTail daughter board header)
#option ENC_MSSP = 1        // MSSP1
#option ENC_CS   = PORTB.3  // CSn  (J3-21 /CS)
#option ENC_INT  = PORTB.2  // INTn to mcu RB2/INT2 (J3-23)
{
#option ENC_SDO = PORTC.5   // SDO1 (J3-8 SI)
#option ENC_SDI = PORTC.4   // SDI1 (J3-10 SO)
#option ENC_SCK = PORTC.3   // SCK1 (J3-12 SCK)
}

// addtl PICTailPlus IO pins
#option ENC_SHDN = PORTB.5  // SHDNn low-active shutdown from mcu (J3-17)
#option ENC_MDIX = PORTB.4  // MDIX ethernet crossover logic from mcu (J3-19)

dim ENC_SHDN_IO as ENC_SHDN.ENC_SHDN@
dim ENC_MDIX_IO as ENC_MDIX.ENC_MDIX@

//
//-----------------------------------------------------------------------------
// userInitClock() - setup 18F87K22 clock
//-----------------------------------------------------------------------------
//
public sub userInitClock()
    const PLLEN = 6     // OSCTUNE PLL bit
    
  #if (_clock = 10) or (_clock = 40)         // HS or HSPLL
    // nothing to do...
  #elseif (_clock = 64)     // INTOSC 16MHZ + 4x PLL
    OSCCON = $70            // INTOSC 16MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 1     // 4x pllen on
  #elseif (_clock = 32)     // INTOSC 8MHZ + 4x PLL
    OSCCON = $60            // INTOSC 8MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 1     // 4x pllen on
  #elseif (_clock = 8)
    OSCCON = $60            // INTOSC 8MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 0     // pllen off
  #else
    #error "unsupported _clock selection"
  #endif
    // allow time for the PLL to lock (if it's enabled)
    if (OSCTUNE.bits(PLLEN) = 1) then
        delayms(2)
    endif
end sub

//
//-----------------------------------------------------------------------------
// userInit() - setup clock, IO pins and other req'd regs
//-----------------------------------------------------------------------------
//
public sub userInit()
    userInitClock()
    
    // init the addtl PICTailPlus IO pins
    high(ENC_SHDN_IO)
    low(ENC_MDIX_IO)

    // init led & switch
    low(LED1)
    input(INPUT_SW)
end sub

//
//-----------------------------------------------------------------------------
// userClockCheck() - toggle led at 1ms rate to verify clock setting
//-----------------------------------------------------------------------------
//
public sub userClockCheck()
    while(true)
        LED1 = 1
        delayus(500)
        LED1 = 0
        delayus(500)
    end while
end sub

//
//-----------------------------------------------------------------------------
// module startup
//-----------------------------------------------------------------------------
//
#if (USER_AUTO_INIT)
  userInit()
#else
  #warning "userInit() NOT called at startup"
#endif

end module
