//
//-----------------------------------------------------------------------------
// Olimex PIC-WEB board 18F67J60
//-----------------------------------------------------------------------------
//
module user_init

#define PIC_WEB_67J60

// call userInit at startup (true/false)
#option USER_AUTO_INIT = true

//
//-----------------------------------------------------------------------------
// config settings
//-----------------------------------------------------------------------------
//
// config
#if (_clock = 25)
  config FOSC = HS
#else   // assume 41.6667MHz
  config FOSC = HSPLL
#endif
config
   FOSC2 = ON,      // primary osc
   ETHLED = ON      // RA0/RA1 function as ethernet LEDA/LEDB

// USART support
// note: be sure to use the modified USART.BAS v1.6 module that supports non-contiguous 
// SPBRG/SPBRGH registers otherwise the baudrate may not be correct
// for the J60 the best setting is BRGH=1 and BRG16=1 to support all baudrates
#option USART_BRGH = true
#option USART_BRG16 = true

//
//-----------------------------------------------------------------------------
// Olimex PIC-WEB 18F67J60 hardware
//-----------------------------------------------------------------------------
//
// LEDs
// LED D1 (red, low-active)
public dim LED1         as PORTB.4

// Switches
// BUT pushbutton sw w/ext pullup
public dim INPUT_SW     as PORTB.0

// Analog
// trimpot connected to RF3/AN8
public dim TRIM_POT     as PORTF.3
// temp termistor connected to RF4/AN9
public dim TEMP_SENSOR  as PORTF.4

// Serial Flash
// PIC-Web uses ATMEL flash AT45DB011
{
#define SPIFLASH_CS_TRIS        (TRISFbits.TRISF7)
#define SPIFLASH_CS_IO          (LATFbits.LATF7)
#define SPIFLASH_SCK_TRIS       (TRISCbits.TRISC3)
#define SPIFLASH_SDI_TRIS       (TRISCbits.TRISC4)
#define SPIFLASH_SDI_IO         (PORTCbits.RC4)
#define SPIFLASH_SDO_TRIS       (TRISCbits.TRISC5)
#define SPIFLASH_SPI_IF         (PIR1bits.SSPIF)
#define SPIFLASH_SSPBUF         (SSP1BUF)
#define SPIFLASH_SPICON1        (SSP1CON1)
#define SPIFLASH_SPICON1bits    (SSP1CON1bits)
#define SPIFLASH_SPICON2        (SSP1CON2)
#define SPIFLASH_SPISTAT        (SSP1STAT)
#define SPIFLASH_SPISTATbits    (SSP1STATbits)
}

//
//-----------------------------------------------------------------------------
// userInitClock() - setup 18FxxJ60 clock
//-----------------------------------------------------------------------------
//
public sub userInitClock()
    const PLLEN = 6         // pllen bit in the osctune reg

  #if (_clock = 25)
    OSCCON  = %00000000      // SCS[1:0] = primary osc
    OSCTUNE = %00000000      // default
  #else   // assume 41.66MHz
    OSCCON  = %00000000      // SCS[1:0] = primary osc
    OSCTUNE = %01000000      // 5xPLL=enabled, PLL Prescaler=div3, PLL Postscaler=disabled
  #endif

    // allow time for the PLL to lock (if it's enabled)
    if (OSCTUNE.bits(PLLEN) = 1) then
        delayms(2)
    endif
end sub

//
//-----------------------------------------------------------------------------
// userInit() - setup clock, IO pins and other req'd regs
//-----------------------------------------------------------------------------
//
public sub userInit()
    userInitClock()
    
    // init led & switch
    low(LED1)
    input(INPUT_SW)
end sub

//
//-----------------------------------------------------------------------------
// userClockCheck() - toggle led at 1ms rate to verify clock setting
//-----------------------------------------------------------------------------
//
public sub userClockCheck()
    while(true)
        LED1 = 1
        delayus(500)
        LED1 = 0
        delayus(500)
    end while
end sub

//
//-----------------------------------------------------------------------------
// module startup
//-----------------------------------------------------------------------------
//
#if (USER_AUTO_INIT)
  userInit()
#else
  #warning "userInit() NOT called at startup"
#endif

end module
